#ifndef AGECLASSHANDLER_H_
#define AGECLASSHANDLER_H_

#include <string>
using std::string;
#include <vector>
using std::vector;

#ifdef USE_MPI
#include "MPIUtil.h"
#endif

class TreeContainer;
class Tree;

#include <xercesc/util/XercesVersion.hpp>

namespace XERCES_CPP_NAMESPACE {
	class DOMElement;
}

class AgeClassHandler {
public:
	AgeClassHandler();
	~AgeClassHandler();

#ifdef USE_MPI
	inline void broadcast(int const rootId, MPI_Comm const comm) {
		MPIUtil::broadcast(ageClassNames, rootId, comm);
		MPIUtil::broadcast(ageClassBounds, rootId, comm);
	}
#endif
	/**
	 * Adds an age class to the handler.
	 * Returns the id of the age class.
	 */
	unsigned int addAgeClass(string const & name, int minimumAge);
	/**
	 * Determines the age class of a Tree.
	 * The age classes must have been added in ascending order, otherwise determineAgeClasses() won't work.
	 * If the tree is younger than the youngest age class, 0 will be returned;
	 * i.e. the tree is considered to belong to the youngest age class.
	 */
	unsigned int determineAgeClass(Tree const & curTree) const;
	/**
	 * Returns the number of age classes.
	 */
	unsigned int getNumberOfAgeClasses() const {
		return ageClassNames.size();
	}
	/**
	 * Returns the name of the age class identified by @ageClassId.
	 */
	string const & getAgeClassName(unsigned int ageClassId) const {
		return ageClassNames[ageClassId];
	}
	unsigned int getAgeClassId(string const & name) const;
	/**
	 * Loads XML data for the AgeClassHandler.
	 *
	 * Precondition:	@ageClasses must not be null.
	 */
	void loadData(XERCES_CPP_NAMESPACE::DOMElement const * ageClasses);
private:
	/**
	 * Loads XML data for one age class.
	 *
	 * Precondition:	@ageClass must not be null.
	 */
	void loadAgeClass(XERCES_CPP_NAMESPACE::DOMElement const * ageClass);

	vector<string> ageClassNames;
	vector<int> ageClassBounds;
};

#endif /* AGECLASSHANDLER_H_ */
