#include "AssortmentVolumeModel.h"
#include "Tree.h"
#include "Species.h"
#include "SpeciesContainer.h"
#include "Compartment.h"
#include "AssortmentContainer.h"
#include "VolumeModel.h"
#include "StumpModel.h"
#include "DiameterModel.h"

#include <algorithm>
using std::fill;

AssortmentVolumeModel::AssortmentVolumeModel(Compartment const & comp)
: compartment(&comp)
{
}

void AssortmentVolumeModel::calculateForTree(Tree const & tree, vector<double> & result) {
	size_t const numAssortments = AssortmentContainer::size();
	result.resize(numAssortments);

	double const d20h = DiameterModel::estimateDiameterAt20PercentOfHeight(tree);
	double h0 = StumpModel::calculateStumpHeightOf(tree);
	double d0 = DiameterModel::calculateDiameterAtHeight(tree, h0, d20h);

	// Loop over all assortments in order
	for (size_t index = 0; index < numAssortments; ++index) {
		Assortment const & assortment = AssortmentContainer::get(index);
		Assortment::Constraints const & constraints = assortment.getConstraints(tree.getSpeciesId());
		if (constraints.minDiameter > 0 && constraints.logLength > 0) {
			// The assortment has constraints for the current species
			while (d0 >= constraints.minDiameter) {
				double const h1 = h0 + constraints.logLength;
				if (h1 < tree.getMark(TreeMarks::HEIGHT)) {
					double const d1 = DiameterModel::calculateDiameterAtHeight(tree, h1, d20h);
					if (d1 >= constraints.minDiameter) {
						// We are able to make a log whose diameter at both ends
						// is at least as wide as the requirement.
						// Add the log's volume to the appropriate assortment bin.
						result[index] += VolumeModel::calculateStemVolumeOf(&tree, h0, h1);
						// The next log will start where the previous ended.
						h0 = h1;
						d0 = d1;
						continue;
					}
				}
				// We cannot make any more logs with these constraints.
				// Switch to the next assortment type.
				break;
			}
		} else {
			// The current assortment doesn't have any constraints. Add the rest
			// of the stem volume to the assortment bin.
			result[index] += VolumeModel::calculateStemVolumeOf(&tree, h0, tree.getMark(TreeMarks::HEIGHT));
			break;
		}
	}
}

void AssortmentVolumeModel::calculateForCompartment(vector<double> & result, int toInclude) {
	size_t const numAssortments = AssortmentContainer::size();
	result.resize(numAssortments);
	fill(result.begin(), result.end(), 0.0);
	for(Tree * curTree = compartment->getFirstTree();
			curTree != NULL;
			curTree = curTree->getNextTreeInCompartment()) {
		if (toInclude == ALL_TREES || curTree->isMarked()) {
			calculateForTree(*curTree, result);
		}
	}
}

void AssortmentVolumeModel::calculateForCompartment(vector< vector<double> > & result, int toInclude) {
	size_t const numAssortments = AssortmentContainer::size();
	size_t const numSpecies = SpeciesContainer::size();
	result.resize(numSpecies);
	for (size_t speciesId = 0; speciesId < numSpecies; ++speciesId) {
		result[speciesId].resize(numAssortments);
		fill(result[speciesId].begin(), result[speciesId].end(), 0.0);
	}
	for(Tree * curTree = compartment->getFirstTree();
			curTree != NULL;
			curTree = curTree->getNextTreeInCompartment()) {
		if (toInclude == ALL_TREES || curTree->isMarked()) {
			unsigned int speciesId = curTree->getSpeciesId();
			calculateForTree(*curTree, result[speciesId]);
		}
	}
}
