#ifdef CALCULATION_INSPECTION

#include "CalculationInspector.h"
#include "Tree.h"
#include "CommonData.h"
#include "Vector.h"
#include "TreeMarks.h"
#include "Model.h"
#include "SecondaryModel.h"
#include "RegenerationModelData.h"
#include "RegenerationModel.h"
#include "Log.h"
#include <sstream>
#include <set>
#include <vector>
#include <map>
#include <cstdio>
#include <cstdarg>
#include <algorithm>

#ifdef USE_MPI
#include "MPIUtil.h"
#endif

#include "XercesString.h"
#include "XercesUtilities.h"
#include <xercesc/util/XMLString.hpp>
#include <xercesc/dom/DOMElement.hpp>
#include <xercesc/dom/DOMNodeList.hpp>

XERCES_CPP_NAMESPACE_USE
using namespace std;

#define MSG_SIZE 100000

CalculationInspector* CalculationInspector::singleton = NULL;


CalculationInspector& CalculationInspector::Instance() {
	if (singleton == NULL) {
		singleton = new CalculationInspector;
	}
	return *singleton;
}

void CalculationInspector::Destroy() {
	delete singleton;
	singleton = NULL;
}

CalculationInspector::CalculationInspector() :
	enableInspection(false),
	inspectInitialStand(false), inspectGrowth(false), inspectMortality(false),
	inspectReproduction(false), inspectRegeneration(false),
	inspectCompetitionDistance(false),
	inspectWeibullIteration(false), inspectWeibullProbability(false),
	firstIteration(0), lastIteration(0),
	msgBuf(new char[MSG_SIZE]) {
}

CalculationInspector::~CalculationInspector() {
	delete [] msgBuf;
	msgBuf = NULL;
}

#ifdef USE_MPI
void CalculationInspector::broadcast(int const rootId, MPI_Comm const comm) {
	MPIUtil::broadcast(enableInspection, rootId, comm);
	MPIUtil::broadcast(inspectInitialStand, rootId, comm);
	MPIUtil::broadcast(inspectGrowth, rootId, comm);
	MPIUtil::broadcast(inspectMortality, rootId, comm);
	MPIUtil::broadcast(inspectReproduction, rootId, comm);
	MPIUtil::broadcast(inspectRegeneration, rootId, comm);
	MPIUtil::broadcast(inspectCompetitionDistance, rootId, comm);
	MPIUtil::broadcast(inspectWeibullIteration, rootId, comm);
	MPIUtil::broadcast(inspectWeibullProbability, rootId, comm);
	MPIUtil::broadcast(firstIteration, rootId, comm);
	MPIUtil::broadcast(lastIteration, rootId, comm);
}
#endif

void CalculationInspector::inspectModel(string const & modelName,
		Tree const & tree, int treeIndex, int markId,
		CommonData const & data,
		Vector const & modelVariables, Vector const & secondaryVariables,
		double const result) {
	ostr << "Calculating ";
	if (markId >= 0)
		ostr << "'" << TreeMarks::getMarkName(markId) << "' in ";
	ostr << "model '" << modelName << "' for tree";
	if (treeIndex >= 0)
		ostr << " [" << treeIndex << "]";
	ostr << " at (" << tree.getX() << ", " << tree.getY() << ")";
	flushOutput();
	ostr << "y = ";
	data.writeModel(ostr);
	flushOutput();

	set<unsigned int> factorIds;
	set<unsigned int> secondaryFactorIds;
	vector<ModelTerm> modelTerms = data.getModelTerms();
	for (vector<ModelTerm>::const_iterator i = modelTerms.begin(); i != modelTerms.end(); ++i) {
		vector<ModelFactor> modelFactors = i->getFactors();
		for (vector<ModelFactor>::const_iterator j = modelFactors.begin(); j != modelFactors.end(); ++j) {
			const unsigned int id = j->getVariableId();
			if (id != Model::INTERCEPT)
				factorIds.insert(id);
		}
		vector<SecondaryModelTerm> secondaryTerms = i->getCoefficient().getTerms();
		for (vector<SecondaryModelTerm>::const_iterator j = secondaryTerms.begin(); j != secondaryTerms.end(); ++j) {
			vector<SecondaryModelFactor> secondaryFactors = j->getFactors();
			for (vector<SecondaryModelFactor>::const_iterator k = secondaryFactors.begin(); k != secondaryFactors.end(); ++k) {
				const unsigned int id = k->getVariableId();
				if (id != SecondaryModel::INTERCEPT)
					secondaryFactorIds.insert(id);
			}
		}
	}

	if (factorIds.size() > 0) {
		ostr << "Variables: ";
		for (set<unsigned int>::const_iterator i = factorIds.begin(); i != factorIds.end();) {
			ostr << Model::getModelVariableName(*i) << " = " << modelVariables[*i];
			if (++i != factorIds.end())
				ostr << ", ";
		}
		flushOutput();
	}
	if (secondaryFactorIds.size() > 0) {
		ostr << "Secondary: ";
		for (set<unsigned int>::const_iterator i = secondaryFactorIds.begin(); i != secondaryFactorIds.end();) {
			ostr << SecondaryModel::getSecondaryVariableName(*i) << " = " << secondaryVariables[*i];
			if (++i != secondaryFactorIds.end())
				ostr << ", ";
		}
		flushOutput();
	}

	ostr << "y = " << result;
	flushOutput();
}

void CalculationInspector::inspectRegenerationModel(
		RegenerationModelData const & data, Vector const & variables, double const result) {
	ostr << "Calculating number of new trees per hectare in model 'Regeneration'";
	flushOutput();
	ostr << "y = ";
	data.writeModel(ostr);
	flushOutput();

	ostr << "Variables: ";
	bool firstTerm = true;
	const Vector& coefficients = data.getCoefficients();
	for (unsigned int i = 0; i < RegenerationModel::NUMBER_OF_VARIABLES; i++) {
		if (coefficients[i] != 0.0 && i != RegenerationModel::INTERCEPT) {
			if (firstTerm) {
				firstTerm = false;
			} else {
				ostr << ", ";
			}
			ostr << RegenerationModel::getVariableName(i) << " = " << variables[i];
		}
	}
	flushOutput();

	ostr << "y = " << result;
	flushOutput();
}

void CalculationInspector::print(string const & fmt, ...)
{
	va_list ap;
	va_start(ap, fmt);
	vsnprintf (msgBuf, MSG_SIZE, fmt.c_str(), ap);
	va_end(ap);

	ostr << msgBuf;
	flushOutput();
}

void CalculationInspector::loadData(XERCES_CPP_NAMESPACE::DOMElement const * data) {

	// eUnrecognized denotes an unrecognized tag.
	enum Tags {eUnrecognized, eInspectModel, eInspectThinning, eEnablePause, eIterationInterval, eOutput};

	map<string, int> tagMap;
	tagMap["inspectModel"] = eInspectModel;
	tagMap["inspectThinning"] = eInspectThinning;
	tagMap["iterationInterval"] = eIterationInterval;
	tagMap["output"] = eOutput;

	enableInspection = true;

	for (DOMElement * element = XercesUtilities::getFirstChildDOMElement(data); element != 0; element
			= XercesUtilities::getNextSiblingDOMElement(element)) {
		XercesString tagName = XMLString::replicate(element->getTagName());
		switch (tagMap[tagName.toString()]) {
		case (eInspectModel): {
			string model = XercesUtilities::getFirstChildTextContentAsString(element);
			enableInspectionForModel(model);
			break;
		}
		case (eInspectThinning): {
			string thinning = XercesUtilities::getFirstChildTextContentAsString(element);
			enableInspectionForThinning(thinning);
			break;
		}
		case (eIterationInterval): {
			int first = XercesUtilities::getAttributeValueAsInt(element, "first");
			int last = XercesUtilities::getAttributeValueAsInt(element, "last");
			if (first <= 0 || last <= 0 || first > last) {
				ostringstream ss;
				ss << "Invalid iteration interval given: " << first << " - " << last << ". ";
				ss << "Attributes 'first' and 'last' must be positive and 'first' <= 'last'";
				Log::log(-1, ss.str().c_str());
				throw "Invalid iteration interval given";
			}
			firstIteration = static_cast<unsigned int>(first);
			lastIteration = static_cast<unsigned int>(last);
			break;
		}
		case (eOutput): {
			// Not used
			break;
		}
		default: { // Handles eUnrecognized.
			throw "Unrecognized XML-element: "+tagName.toString();
		}
		}
	}

	Log::log(2, "Calculation inspection is enabled");
	if (firstIteration > 0 && lastIteration > 0) {
		Log::log(3, "Inspecting iterations %u through %u", firstIteration, lastIteration);
	}
}

void CalculationInspector::enableInspectionForModel(const string & model) {
	string name(model.length(), ' ');
	transform(model.begin(), model.end(), name.begin(), ::tolower);
	if (name == "initial stand") {
		inspectInitialStand = true;
	} else if (name == "growth") {
		inspectGrowth = true;
	} else if (name == "mortality") {
		inspectMortality = true;
	} else if (name == "reproduction") {
		inspectReproduction = true;
	} else if (name == "regeneration") {
		inspectRegeneration = true;
	} else if (name == "competition distance" || name == "longest branch") {
		inspectCompetitionDistance = true;
	} else if (name == "all") {
		inspectInitialStand = true;
		inspectGrowth = true;
		inspectMortality = true;
		inspectReproduction = true;
		inspectRegeneration = true;
		inspectCompetitionDistance = true;
	} else {
		throw "Invalid value for inspectModel: " + model;
	}
}

void CalculationInspector::enableInspectionForThinning(const string & thinning) {
	string name(thinning.length(), ' ');
	transform(thinning.begin(), thinning.end(), name.begin(), ::tolower);
	if (name == "weibull iteration") {
		inspectWeibullIteration = true;
	} else if (name == "weibull thinning probability") {
		inspectWeibullProbability = true;
	} else {
		throw "Invalid value for inspectThinning: " + thinning;
	}
}

void CalculationInspector::flushOutput() {
	Log::log(-1, ostr.str().c_str());
	ostr.str("");
	ostr.clear();
}

#endif
