#ifdef CALCULATION_INSPECTION

#ifndef CALCULATIONINSPECTOR_H_
#define CALCULATIONINSPECTOR_H_

#ifdef USE_MPI
#include "MPIUtil.h"
#endif

#include <string>
#include <sstream>
using std::string;
using std::ostringstream;

#include <xercesc/util/XercesVersion.hpp>

namespace XERCES_CPP_NAMESPACE
{
	class DOMElement;
}

class Tree;
class CommonData;
class Vector;
class RegenerationModelData;

/**
 * Singleton class that can be used to print information about calculations.
 * It can either be used in interactive mode, where text is printed to standard
 * output, or, it can be used to print to the log file.
 */
class CalculationInspector {
public:
	/**
	 * Get singleton instance
	 */
	static CalculationInspector& Instance();
	/**
	 * Destroy singleton instance
	 */
	static void Destroy();

	CalculationInspector();
	~CalculationInspector();

#ifdef USE_MPI
	void broadcast(int const rootId, MPI_Comm const comm);
#endif

	/**
	 * Print the formula, the variables, the result and general information
	 * about a calculation.
	 * @param tree The target tree
	 * @param treeIndex Array index of the tree. Use -1 if not applicable
	 * @param markId The ID of the mark. Use -1 if not applicable
	 * @param data Model data
	 * @param modelVariables Model variables
	 * @param secondaryVariables Secondary-level model variables
	 * @param result The result of the calculation
	 */
	void inspectModel(string const & modelName,
			Tree const & tree, int treeIndex, int markId,
			CommonData const & data,
			Vector const & modelVariables, Vector const & secondaryVariables,
			double const result);

	/**
	 * Print information about a regeneration model calculation
	 */
	void inspectRegenerationModel(RegenerationModelData const & data,
			Vector const & variables, double const result);

	/**
	 * Print formatted text
	 */
	void print(string const & fmt, ...);

	/**
	 * Set the current iteration
	 */
	inline void setCurrentIteration(unsigned int iteration) {
		currentIteration = iteration;
	}

	inline bool isInspectionEnabled() const {
		return enableInspection;
	}
	inline bool shouldInspect() const {
		return enableInspection && iterationWithinInterval(currentIteration);
	}
	inline bool shouldInspectInitialStand() const {
		return enableInspection && inspectInitialStand;
	}
	inline bool shouldInspectGrowth() const {
		return enableInspection && inspectGrowth && iterationWithinInterval(currentIteration);
	}
	inline bool shouldInspectMortality() const {
		return enableInspection && inspectMortality && iterationWithinInterval(currentIteration);
	}
	inline bool shouldInspectReproduction() const {
		return enableInspection && inspectReproduction && iterationWithinInterval(currentIteration);
	}
	inline bool shouldInspectRegeneration() const {
		return enableInspection && inspectRegeneration && iterationWithinInterval(currentIteration);
	}
	inline bool shouldInspectCompetitionDistance() const {
		return enableInspection && inspectCompetitionDistance && iterationWithinInterval(currentIteration);
	}
	inline bool shouldInspectWeibullIteration() const {
		return enableInspection && inspectWeibullIteration && iterationWithinInterval(currentIteration);
	}
	inline bool shouldInspectWeibullProbability() const {
		return enableInspection && inspectWeibullProbability && iterationWithinInterval(currentIteration);
	}

	void loadData(XERCES_CPP_NAMESPACE::DOMElement const * data);
private:
	inline bool iterationWithinInterval(unsigned int iteration) const {
		return (firstIteration == 0 && lastIteration == 0) ||
				(iteration >= firstIteration && iteration <= lastIteration);
	}

	void enableInspectionForModel(const string & model);
	void enableInspectionForThinning(const string & thinning);

	void flushOutput();

	CalculationInspector(CalculationInspector const &);
	CalculationInspector& operator=(CalculationInspector const &);

	static CalculationInspector* singleton;

	bool enableInspection;
	bool inspectInitialStand;
	bool inspectGrowth;
	bool inspectMortality;
	bool inspectReproduction;
	bool inspectRegeneration;
	bool inspectCompetitionDistance;
	bool inspectWeibullIteration;
	bool inspectWeibullProbability;

	unsigned int firstIteration;
	unsigned int lastIteration;
	unsigned int currentIteration;

	ostringstream ostr;
	char* msgBuf;
};

#endif /* CALCULATIONINSPECTOR_H_ */

#endif
