#include "CompartmentStatistics.h"
#include "Tree.h"
#include "SimulationData.h"
#include "CompartmentData.h"
#include "ForestType.h"
#include "ForestTypeContainer.h"
#include "SizeClassHandler.h"
#include "EvaluatorFunctions.h"
#include "AssortmentContainer.h"
#include "Species.h"
#include "SpeciesContainer.h"
#include "roundToInt.h"
#ifdef USE_MPI
#include "MPIUtil.h"
#endif
#include "Log.h"

#include <iomanip>
using std::setw;
using std::fixed;
#include <iostream>
#include <algorithm>
using std::fill;
#include <cmath>
using std::max;
using std::sort;

CompartmentStatistics::CompartmentStatistics(SimulationData const & simData, CompartmentData const & compData)
: simulationData(simData)
, compartmentData(compData)
, replication(0)
{
	const unsigned char numberOfSizeClasses = simulationData.getSizeClassHandler().getNumberOfSizeClasses();
	// Setup space for tree statistics.
	treeStatistics.resize(simulationData.getNumberOfIterations()+1);
	for(unsigned int i = 0; i < simulationData.getNumberOfIterations()+1; ++i) {
		treeStatistics[i].resize(numberOfSizeClasses);
		for(unsigned char j = 0; j < numberOfSizeClasses; ++j) {
			treeStatistics[i][j].resize(NUMBER_OF_STATISTICS);
		}
	}
	resizeStatistics(basalAreaGrowthStatistics);
	resizeStatistics(basalAreaFinalStatistics);
	resizeStatistics(volumeGrowthStatistics);
	resizeStatistics(volumeFinalStatistics);
	resizeStatistics(volumeDeadStatistics);
	resizeStatistics(volumeMarkedStatistics);
	resizeStatistics(volumeNewStatistics);
	assortmentVolumeStatistics.resize(simulationData.getNumberOfIterations() + 1);
	removalAssortmentVolumeStatistics.resize(simulationData.getNumberOfIterations() + 1);
}

CompartmentStatistics::~CompartmentStatistics()
{
}

void CompartmentStatistics::addDeadTreeStatistics(unsigned int iteration, Tree const & curTree)
{
	SizeClassHandler const & sizeClassHandler = simulationData.getSizeClassHandler();
	treeStatistics[iteration][sizeClassHandler.determineSizeClass(curTree)][DEAD]++;
}

void CompartmentStatistics::addNewTreeStatistics(unsigned int iteration, Tree const & curTree)
{
	SizeClassHandler const & sizeClassHandler = simulationData.getSizeClassHandler();
	treeStatistics[iteration][sizeClassHandler.determineSizeClass(curTree)][NEW]++;
}
void CompartmentStatistics::addGrowthTreeStatistics(unsigned int iteration, Tree const & curTree)
{
	SizeClassHandler const & sizeClassHandler = simulationData.getSizeClassHandler();
	treeStatistics[iteration][sizeClassHandler.determineSizeClass(curTree)][GROWTH]++;
}

void CompartmentStatistics::addRemovalTreeStatistics(unsigned int iteration, Tree const & curTree)
{
	SizeClassHandler const & sizeClassHandler = simulationData.getSizeClassHandler();
	treeStatistics[iteration][sizeClassHandler.determineSizeClass(curTree)][REMOVAL]++;
}


void CompartmentStatistics::addFinalTreeStatistics(unsigned int iteration, Tree const & curTree)
{
	SizeClassHandler const & sizeClassHandler = simulationData.getSizeClassHandler();
	treeStatistics[iteration][sizeClassHandler.determineSizeClass(curTree)][FINAL]++;
}

void CompartmentStatistics::addAssortmentVolumes(unsigned int iteration, vector< vector<double> > const & volumes) {
	unsigned int numReplications = simulationData.getNumberOfReplications();
	assortmentVolumeStatistics[iteration].resize(volumes.size());
	for (size_t speciesId = 0, n = volumes.size(); speciesId < n; ++speciesId) {
		assortmentVolumeStatistics[iteration][speciesId].resize(volumes[speciesId].size());
		for (size_t assortId = 0, m = volumes[speciesId].size(); assortId < m; ++assortId) {
			assortmentVolumeStatistics[iteration][speciesId][assortId].resize(numReplications);
			assortmentVolumeStatistics[iteration][speciesId][assortId][replication] += volumes[speciesId][assortId];
		}
	}
}

void CompartmentStatistics::addRemovalAssortmentVolumes(unsigned int iteration, vector< vector<double> > const & volumes) {
	unsigned int numReplications = simulationData.getNumberOfReplications();
	removalAssortmentVolumeStatistics[iteration].resize(volumes.size());
	for (size_t speciesId = 0, n = volumes.size(); speciesId < n; ++speciesId) {
		removalAssortmentVolumeStatistics[iteration][speciesId].resize(volumes[speciesId].size());
		for (size_t assortId = 0, m = volumes[speciesId].size(); assortId < m; ++assortId) {
			removalAssortmentVolumeStatistics[iteration][speciesId][assortId].resize(numReplications);
			removalAssortmentVolumeStatistics[iteration][speciesId][assortId][replication] += volumes[speciesId][assortId];
		}
	}
}

#ifdef USE_MPI
void CompartmentStatistics::combineStatistics() {
	unsigned int numIterations = simulationData.getNumberOfIterations();
	unsigned int maxAge = 0;
	if(!treeAgeStatistics.empty()) {
		maxAge = treeAgeStatistics.rbegin()->first;
	}
	MPIUtil::max(maxAge,MPIUtil::getActiveCommunicator());

	vector<uint64_t> agesBuffer(maxAge+1,0);
	for(map<unsigned int,uint64_t>::const_iterator iter = treeAgeStatistics.begin(); iter!=treeAgeStatistics.end(); iter++) {
		agesBuffer[iter->first] = iter->second;
	}

	MPIUtil::sum(agesBuffer,MPIUtil::getActiveCommunicator());

	for(unsigned int age=0; age<=maxAge; age++) {
		if(agesBuffer[age] != 0) {
			treeAgeStatistics[age] = agesBuffer[age];
		}
	}
	agesBuffer.clear();

	for (unsigned int i = 0; i < numIterations + 1; i++)
		MPIUtil::sum(basalAreaFinalStatistics[i],MPIUtil::getActiveCommunicator());
	for (unsigned int i = 0; i < numIterations + 1; i++)
		MPIUtil::sum(volumeFinalStatistics[i],MPIUtil::getActiveCommunicator());
	for (unsigned int i = 0; i < numIterations + 1; i++)
		MPIUtil::sum(volumeDeadStatistics[i],MPIUtil::getActiveCommunicator());
	for (unsigned int i = 0; i < numIterations + 1; i++)
		MPIUtil::sum(volumeMarkedStatistics[i],MPIUtil::getActiveCommunicator());
	for (unsigned int i = 0; i < numIterations + 1; i++)
		MPIUtil::sum(volumeNewStatistics[i],MPIUtil::getActiveCommunicator());

	SizeClassHandler const & sizeClassHandler = simulationData.getSizeClassHandler();
	size_t const size = (simulationData.getNumberOfIterations()+1)*sizeClassHandler.getNumberOfSizeClasses()*NUMBER_OF_STATISTICS;
	vector<uint64_t> buffer;
	buffer.resize(size);

	size_t ptr = 0;
	for(unsigned int i = 0; i < simulationData.getNumberOfIterations()+1; ++i) {
		for(unsigned char j = 0; j < sizeClassHandler.getNumberOfSizeClasses(); ++j) {
			for(unsigned char k = 0; k < NUMBER_OF_STATISTICS; ++k) {
				buffer[ptr++] = treeStatistics[i][j][k];
			}
		}
	}

	MPIUtil::sum(buffer,MPIUtil::getActiveCommunicator());

	ptr = 0;
	for(unsigned int i = 0; i < simulationData.getNumberOfIterations()+1; ++i) {
		for(unsigned char j = 0; j < sizeClassHandler.getNumberOfSizeClasses(); ++j) {
			for(unsigned char k = 0; k < NUMBER_OF_STATISTICS; ++k) {
				treeStatistics[i][j][k] = buffer[ptr++];
			}
		}
	}

	for (size_t iteration = 0, n = assortmentVolumeStatistics.size(); iteration < n; ++iteration) {
		for (size_t speciesId = 0, m = assortmentVolumeStatistics[iteration].size(); speciesId < m; ++speciesId) {
			for (size_t assortId = 0, p = assortmentVolumeStatistics[iteration][speciesId].size(); assortId < p; ++assortId) {
				MPIUtil::sum(assortmentVolumeStatistics[iteration][speciesId][assortId], MPIUtil::getActiveCommunicator());
			}
		}
	}
	for (size_t iteration = 0, n = removalAssortmentVolumeStatistics.size(); iteration < n; ++iteration) {
		for (size_t speciesId = 0, m = removalAssortmentVolumeStatistics[iteration].size(); speciesId < m; ++speciesId) {
			for (size_t assortId = 0, p = removalAssortmentVolumeStatistics[iteration][speciesId].size(); assortId < p; ++assortId) {
				MPIUtil::sum(removalAssortmentVolumeStatistics[iteration][speciesId][assortId], MPIUtil::getActiveCommunicator());
			}
		}
	}
}
#endif


void CompartmentStatistics::printSummary(ostream & out) const
{
	unsigned int const numReplications = simulationData.getNumberOfReplications();
	unsigned int const numIterations = simulationData.getNumberOfIterations() + 1;
	double const fullAreaInHectares = compartmentData.getCompartmentPolygon()->getFullArea() / 10000.0;
	int p0025 = roundToInt(0.025 * numReplications) - 1;
	if (p0025 < 0) {
		p0025 = 0;
	}
	int p0975 = roundToInt(0.975 * numReplications) - 1;

	out.precision(2);
	out << "Compartment area:        " << fixed << fullAreaInHectares << " hectares\n";
	out << "Age of initial stand:    " << fixed << compartmentData.getInitialStandAge() << " years\n";

	out << *ForestTypeContainer::get(compartmentData.getForestTypeId()) << "\n";

	out << "\n      STATISTICS SUMMARY\n";
	out << "---------------------------------\n";
	out << "Frequencies of ages of the final trees, averages over " << numReplications << " replications:\n";
	out << setw(5) << "Age" << setw(17) << "Number of trees";
	out << "\n";
	for(map<unsigned int,uint64_t>::const_iterator iter = treeAgeStatistics.begin(); iter!=treeAgeStatistics.end(); iter++) {
		out << setw(4) << iter->first << setw(17) << fixed << iter->second / numReplications;
		out << "\n";
	}

	vector< vector<double> > sums(7);
	for (size_t i = 0; i < 7; i++) {
		sums[i].resize(numReplications, 0.0);
	}

	out.precision(2);
	out << "\n      Iteration statistics\n";
	out << "---------------------------------\n";
	out << "Values averaged over all replications:\n";
	out << "Iteration" << setw(19) << "Basal area growth" << setw(18) << "Final basal area";
	out << setw(15) << "Volume growth";
	out << setw(18) << "Dead tree volume" << setw(17) << "New tree volume";
	out << setw(11) << "Removal" << setw(15) << "Final volume";
	out << "\n";
	out << setw(24) << "(m^2/ha)" << setw(18) << "(m^2/ha)";
	out << setw(18) << "(m^3/ha)" << setw(15) << "(m^3/ha)" << setw(18) << "(m^3/ha)" << setw(14) << "(m^3/ha)" << setw(14) << "(m^3/ha)";
	out << "\n";
	for(unsigned int i = 0; i < numIterations; ++i) {
		double const basalAreaGrowth = computeAvg(basalAreaGrowthStatistics[i]) / fullAreaInHectares;
		double const basalAreaFinal = computeAvg(basalAreaFinalStatistics[i]) / fullAreaInHectares;
		double const volumeGrowth = computeAvg(volumeGrowthStatistics[i]) / fullAreaInHectares;
		double const deadVolume = computeAvg(volumeDeadStatistics[i]) / fullAreaInHectares;
		double const newVolume = computeAvg(volumeNewStatistics[i]) / fullAreaInHectares;
		double const markedVolume = computeAvg(volumeMarkedStatistics[i]) / fullAreaInHectares;
		double const finalVolume = computeAvg(volumeFinalStatistics[i]) / fullAreaInHectares;
		accumulateSums(basalAreaGrowthStatistics[i], sums[0]);
		accumulateSums(volumeGrowthStatistics[i], sums[2]);
		accumulateSums(volumeDeadStatistics[i], sums[3]);
		accumulateSums(volumeNewStatistics[i], sums[4]);
		accumulateSums(volumeMarkedStatistics[i], sums[5]);
		out << setw(5) << fixed << i;
		out << setw(17) << fixed << basalAreaGrowth;
		out << setw(18) << fixed << basalAreaFinal;
		out << setw(18) << fixed << volumeGrowth;
		out << setw(15) << fixed << deadVolume;
		out << setw(18) << fixed << newVolume;
		out << setw(14) << fixed << markedVolume;
		out << setw(14) << fixed << finalVolume;
		out << "\n";
	}
	accumulateSums(basalAreaFinalStatistics[numIterations - 1], sums[1]);
	accumulateSums(volumeFinalStatistics[numIterations - 1], sums[6]);
	out << setw(5) << "Total";
	out << setw(17) << fixed << computeAvg(sums[0]) / fullAreaInHectares;
	out << setw(18) << " ";
	out << setw(18) << fixed << computeAvg(sums[2]) / fullAreaInHectares;
	out << setw(15) << fixed << computeAvg(sums[3]) / fullAreaInHectares;
	out << setw(18) << fixed << computeAvg(sums[4]) / fullAreaInHectares;
	out << setw(14) << fixed << computeAvg(sums[5]) / fullAreaInHectares;
	out << "\n";

	for (size_t i = 0; i < 7; i++) {
		sort(sums[i].begin(), sums[i].end());
	}
	out << "-----\n";
	out << setw(5) << "Lower";
	out << setw(17) << fixed << sums[0][p0025] / fullAreaInHectares;
	out << setw(18) << fixed << sums[1][p0025] / fullAreaInHectares; 
	out << setw(18) << fixed << sums[2][p0025] / fullAreaInHectares;
	out << setw(15) << fixed << sums[3][p0025] / fullAreaInHectares;
	out << setw(18) << fixed << sums[4][p0025] / fullAreaInHectares;
	out << setw(14) << fixed << sums[5][p0025] / fullAreaInHectares;
	out << setw(14) << fixed << sums[6][p0025] / fullAreaInHectares;
	out << "\n";
	out << setw(5) << "Upper";
	out << setw(17) << fixed << sums[0][p0975] / fullAreaInHectares;
	out << setw(18) << fixed << sums[1][p0975] / fullAreaInHectares; 
	out << setw(18) << fixed << sums[2][p0975] / fullAreaInHectares;
	out << setw(15) << fixed << sums[3][p0975] / fullAreaInHectares;
	out << setw(18) << fixed << sums[4][p0975] / fullAreaInHectares;
	out << setw(14) << fixed << sums[5][p0975] / fullAreaInHectares;
	out << setw(14) << fixed << sums[6][p0975] / fullAreaInHectares;
	out << "\n";

	out << "\n      Stem statistics\n";
	out << "---------------------------------\n";
	out << "Total number of stems per iteration. Values averaged over all replications:\n";
	out << setw(10) << "Iteration" << setw(25) << "Growth" << setw(15) << "Mortality" << setw(15) << "Reproduction" << setw(15) << "Removal" << setw(16) << "Final\n";
	for(unsigned int i = 0; i < numIterations; ++i) {
		out << setw(6) << i << "\n";
		vector<uint64_t> treeNumberSums(NUMBER_OF_STATISTICS);
		SizeClassHandler const & sizeClassHandler = simulationData.getSizeClassHandler();
		for(unsigned char j = 0; j < sizeClassHandler.getNumberOfSizeClasses(); ++j) {
			out << setw(20) << sizeClassHandler.getSizeClassName(j);
			for(unsigned int k = 0; k < NUMBER_OF_STATISTICS; ++k) {
				out << setw(15) << treeStatistics[i][j][k] / numReplications;
				treeNumberSums[k] += treeStatistics[i][j][k];
			}
			out << "\n";
		}
		out << setw(20) << "Total";
		for(unsigned int j = 0; j < NUMBER_OF_STATISTICS; ++j) {
			out << setw(15) << treeNumberSums[j] / numReplications;
		}
		out << "\n";
	}
	out << "\n";

	if (AssortmentContainer::size() > 0) {
		printAssortmentSummary(out);
		printRemovalAssortmentSummary(out);
	}
}

void CompartmentStatistics::printAssortmentSummary(ostream & out) const {
	unsigned int const numIterations = simulationData.getNumberOfIterations() + 1;
	unsigned int const numReplications = simulationData.getNumberOfReplications();
	unsigned int const numSpecies = SpeciesContainer::size();
	unsigned int const numAssortments = AssortmentContainer::size();
	int p0025 = roundToInt(0.025 * numReplications) - 1;
	if (p0025 < 0) {
		p0025 = 0;
	}
	int p0975 = roundToInt(0.975 * numReplications) - 1;

	vector< vector< vector<double> > > statistics(2);
	for (unsigned int i = 0; i < 2; i++) {
		statistics[i].resize(numSpecies);
		for (unsigned int speciesId = 0; speciesId < numSpecies; ++speciesId) {
			statistics[i][speciesId].resize(numAssortments);
		}
	}
	vector<unsigned int> speciesNameLengths(numSpecies);

	out.precision(2);
	out << "\n      Assortment volumes\n";
	out << "---------------------------------\n";
	out << "Values averaged over all replications:\n";
	out << setw(10) << "Iteration" << setw(16) << "Assortment" << setw(20) << "Volume (m^3/ha)" << "\n";
	out << setw(26) << "";
	for (unsigned int speciesId = 0; speciesId < numSpecies; ++speciesId) {
		string speciesName = SpeciesContainer::get(speciesId)->getName();
		unsigned int len = speciesName.length();
		speciesNameLengths[speciesId] = max(len, 10U);
		out << setw(speciesNameLengths[speciesId] + 2) << speciesName;
	}
	out << setw(12) << "All" << setw(12) << "Percentage" << "\n";
	for (unsigned int iteration = 0; iteration < numIterations; ++iteration) {
		for (unsigned int speciesId = 0; speciesId < numSpecies; ++speciesId) {
			for (unsigned int assortId = 0; assortId < numAssortments; ++assortId) {
				statistics[0][speciesId][assortId] = computeAvg(assortmentVolumeStatistics[iteration][speciesId][assortId]);
			}
		}
		out << setw(6) << iteration << "\n";
		printAssortmentIterationStatistics(out, statistics[0], speciesNameLengths);
	}
	out << "\n";

	vector< vector< vector<double> > > finalVolumes = assortmentVolumeStatistics[numIterations - 1];
	for (unsigned int speciesId = 0; speciesId < numSpecies; ++speciesId) {
		for (unsigned int assortId = 0; assortId < numAssortments; ++assortId) {
			sort(finalVolumes[speciesId][assortId].begin(), finalVolumes[speciesId][assortId].end());
			statistics[0][speciesId][assortId] = finalVolumes[speciesId][assortId][p0025];
			statistics[1][speciesId][assortId] = finalVolumes[speciesId][assortId][p0975];
		}
	}
	out << setw(6) << "Lower limit (3rd percentile)" << "\n";
	printAssortmentIterationStatistics(out, statistics[0], speciesNameLengths, false);
	out << "\n";
	out << setw(6) << "Upper limit (98th percentile)" << "\n";
	printAssortmentIterationStatistics(out, statistics[1], speciesNameLengths, false);
	out << "\n";
}

void CompartmentStatistics::printRemovalAssortmentSummary(ostream & out) const {
	unsigned int const numIterations = simulationData.getNumberOfIterations() + 1;
	unsigned int const numReplications = simulationData.getNumberOfReplications();
	unsigned int const numSpecies = SpeciesContainer::size();
	unsigned int const numAssortments = AssortmentContainer::size();
	int p0025 = roundToInt(0.025 * numReplications) - 1;
	if (p0025 < 0) {
		p0025 = 0;
	}
	int p0975 = roundToInt(0.975 * numReplications) - 1;

	vector< vector< vector<double> > > totals(numSpecies);
	for (unsigned int speciesId = 0; speciesId < numSpecies; ++speciesId) {
		totals[speciesId].resize(numAssortments);
		for (unsigned int assortId = 0; assortId < numAssortments; ++assortId) {
			totals[speciesId][assortId].resize(numReplications, 0.0);
		}
	}
	vector< vector< vector<double> > > statistics(3);
	for (unsigned int i = 0; i < 3; i++) {
		statistics[i].resize(numSpecies);
		for (unsigned int speciesId = 0; speciesId < numSpecies; ++speciesId) {
			statistics[i][speciesId].resize(numAssortments);
		}
	}
	vector<unsigned int> speciesNameLengths(numSpecies);

	out.precision(2);
	out << "\n      Removal assortment volumes\n";
	out << "---------------------------------\n";
	out << "Values averaged over all replications:\n";
	out << setw(10) << "Iteration" << setw(16) << "Assortment" << setw(20) << "Volume (m^3/ha)" << "\n";
	out << setw(26) << "";
	for (unsigned int speciesId = 0; speciesId < numSpecies; ++speciesId) {
		string speciesName = SpeciesContainer::get(speciesId)->getName();
		unsigned int len = speciesName.length();
		speciesNameLengths[speciesId] = max(len, 10U);
		out << setw(speciesNameLengths[speciesId] + 2) << speciesName;
	}
	out << setw(12) << "All" << setw(12) << "Percentage" << "\n";
	for (unsigned int iteration = 1; iteration < numIterations; ++iteration) {
		for (unsigned int speciesId = 0; speciesId < numSpecies; ++speciesId) {
			for (unsigned int assortId = 0; assortId < numAssortments; ++assortId) {
				statistics[0][speciesId][assortId] = computeAvg(removalAssortmentVolumeStatistics[iteration][speciesId][assortId]);
				accumulateSums(removalAssortmentVolumeStatistics[iteration][speciesId][assortId], totals[speciesId][assortId]);
			}
		}
		out << setw(6) << iteration << "\n";
		printAssortmentIterationStatistics(out, statistics[0], speciesNameLengths);
	}

	for (unsigned int speciesId = 0; speciesId < numSpecies; ++speciesId) {
		for (unsigned int assortId = 0; assortId < numAssortments; ++assortId) {
			sort(totals[speciesId][assortId].begin(), totals[speciesId][assortId].end());
			statistics[0][speciesId][assortId] = computeAvg(totals[speciesId][assortId]);
			statistics[1][speciesId][assortId] = totals[speciesId][assortId][p0025];
			statistics[2][speciesId][assortId] = totals[speciesId][assortId][p0975];
		}
	}
	out << setw(6) << "Total" << "\n";
	printAssortmentIterationStatistics(out, statistics[0], speciesNameLengths);
	out << "\n";
	out << setw(6) << "Lower limit (3rd percentile)" << "\n";
	printAssortmentIterationStatistics(out, statistics[1], speciesNameLengths, false);
	out << "\n";
	out << setw(6) << "Upper limit (98th percentile)" << "\n";
	printAssortmentIterationStatistics(out, statistics[2], speciesNameLengths, false);
	out << "\n";
}

void CompartmentStatistics::printAssortmentIterationStatistics(
		ostream & out, vector< vector<double> > const & statistics,
		vector<unsigned int> const & speciesNameLengths,
		bool sumUpTotals) const {
	unsigned int const numSpecies = SpeciesContainer::size();
	unsigned int const numAssortments = AssortmentContainer::size();
	double const fullAreaInHectares = compartmentData.getCompartmentPolygon()->getFullArea() / 10000.0;
	vector<double> speciesTotals(numSpecies, 0.0);

	double totalVolumeRemoved = 0.0;
	for (unsigned int speciesId = 0; speciesId < numSpecies; ++speciesId) {
		totalVolumeRemoved += accumulate(statistics[speciesId].begin(), statistics[speciesId].end(), 0.0);
	}

	for (unsigned int assortId = 0; assortId < numAssortments; ++assortId) {
		Assortment const & assortment = AssortmentContainer::get(assortId);
		double assortmentTotal = 0.0;
		out << setw(26) << assortment.getName();
		for (unsigned int speciesId = 0; speciesId < numSpecies; ++speciesId) {
			double const volume = statistics[speciesId][assortId];
			assortmentTotal += volume;
			speciesTotals[speciesId] += volume;
			out << setw(speciesNameLengths[speciesId] + 2) << (volume / fullAreaInHectares);
		}
		if (sumUpTotals) {
			out << setw(12) << (assortmentTotal / fullAreaInHectares);
			out << setw(12) << EvaluatorFunctions::safeDivision(assortmentTotal * 100.0, totalVolumeRemoved) << "\n";
		} else {
			out << "\n";
		}
	}
	if (sumUpTotals) {
		out << setw(26) << "Total";
		double volumeTotal = 0.0;
		for (unsigned int speciesId = 0; speciesId < numSpecies; ++speciesId) {
			double const volume = speciesTotals[speciesId];
			volumeTotal += volume;
			out << setw(speciesNameLengths[speciesId] + 2) << (volume / fullAreaInHectares);
		}
		out << setw(12) << (volumeTotal / fullAreaInHectares);
		out << setw(12) << EvaluatorFunctions::safeDivision(volumeTotal * 100.0, totalVolumeRemoved) << "\n";
	} else {
		out << "\n";
	}
}

void CompartmentStatistics::resizeStatistics(vector< vector<double> > & statistics) const {
	unsigned int numIterations = simulationData.getNumberOfIterations();
	unsigned int numReplications = simulationData.getNumberOfReplications();
	statistics.resize(numIterations + 1);
	for (unsigned int i = 0; i < numIterations + 1; i++) {
		statistics[i].resize(numReplications, 0.0);
	}
}

