#ifndef COMPARTMENTSTATISTICS_H_
#define COMPARTMENTSTATISTICS_H_

#ifdef USE_MPI
#include <mpi.h>
#endif
#include <stdint.h>
#include <ostream>
using std::ostream;
#include <vector>
using std::vector;
#include <map>
using std::map;
#include <numeric>
#include <algorithm>
#include <functional>
#include <cassert>

class SimulationData;
class CompartmentData;
class Tree;

/**
 * Collects statistics of the number of trees of different types and size classes.
 */
class CompartmentStatistics
{
public:
	enum TreeStatistics { GROWTH = 0, DEAD, NEW, REMOVAL, FINAL, NUMBER_OF_STATISTICS };

	CompartmentStatistics(SimulationData const & simData, CompartmentData const & compData);
	~CompartmentStatistics();

	/**
	 * Set the current replication
	 */
	inline void setReplication(unsigned int rep) {
		replication = rep;
	}

	/**
	 * Add one tree to the dead tree statistics.
	 */
	void addDeadTreeStatistics(unsigned int iteration, Tree const & curTree);
	/**
	 * Add one tree to the new tree statistics.
	 */
	void addNewTreeStatistics(unsigned int iteration, Tree const & curTree);
	/**
	 * Add one tree to the growth tree statistics.
	 */
	void addGrowthTreeStatistics(unsigned int iteration, Tree const & curTree);
	/**
	 * Add one tree to the removal tree statistics
	 */
	void addRemovalTreeStatistics(unsigned int iteration, Tree const & curTree);
	/**
	 * Adds one tree to the iteration final tree statistics.
	 */
	void addFinalTreeStatistics(unsigned int iteration, Tree const & curTree);
	/**
	 * Stores the basal area growth for the given iteration
	 */
	inline void addBasalAreaGrowth(unsigned int iteration, double area) {
		basalAreaGrowthStatistics[iteration][replication] += area;
	}
	/**
	 * Stores the final basal area for the given iteration
	 */
	inline void addBasalAreaFinal(unsigned int iteration, double area) {
		basalAreaFinalStatistics[iteration][replication] += area;
	}
	/**
	 * Stores the volume growth for the given iteration
	 */
	inline void addVolumeGrowth(unsigned int iteration, double volume) {
		volumeGrowthStatistics[iteration][replication] += volume;
	}
	/**
	 * Stores the total volume for the given iteration, after changes applied.
	 */
	inline void addVolumeFinal(unsigned int iteration, double volume) {
		volumeFinalStatistics[iteration][replication] += volume;
	}
	/**
	 * Stores the volume of dead trees for the given iteration.
	 */
	inline void addVolumeDead(unsigned int iteration, double volume) {
		volumeDeadStatistics[iteration][replication] += volume;
	}
	/**
	 * Stores the volume of marked trees for the given iteration.
	 */
	inline void addVolumeMarked(unsigned int iteration, double volume) {
		volumeMarkedStatistics[iteration][replication] += volume;
	}
	/**
	 * Stores the volume of new trees for the given iteration.
	 */
	inline void addVolumeNew(unsigned int iteration, double volume) {
		volumeNewStatistics[iteration][replication] += volume;
	}
	/**
	 * Adds to the frequency of trees of the given age.
	 */
	inline void addAgeFrequencies(unsigned int age, unsigned int number) {
		treeAgeStatistics[age] += number;
	}
	/**
	 * Stores the volume of available assortments at the given iteration.
	 */
	void addAssortmentVolumes(unsigned int iteration, vector< vector<double> > const & volumes);
	/**
	 * Stores the volume of removal assortments at the given iteration.
	 */
	void addRemovalAssortmentVolumes(unsigned int iteration, vector< vector<double> > const & volumes);

#ifdef USE_MPI
	/*
	 * Combines cohort and tree statistics from all processors
	 */
	void combineStatistics();
#endif
	/**
	 * Prints a summary of the statistics to the specified ostream.
	 */
	void printSummary(ostream & out) const;
private:
	void printAssortmentSummary(ostream & out) const;
	void printRemovalAssortmentSummary(ostream & out) const;
	
	void printAssortmentIterationStatistics(ostream & out,
			vector< vector<double> > const & statistics,
			vector<unsigned int> const & speciesNameLengths,
			bool sumUpTotals = true) const;
	
	void resizeStatistics(vector< vector<double> > & statistics) const;

	static inline double computeAvg(vector<double> const & statistics) {
		return std::accumulate(statistics.begin(), statistics.end(), 0.0) / (double)(statistics.size());
	}

	static inline void accumulateSums(vector<double> const & statistics, vector<double> & sums) {
		//std::transform(statistics.begin(), statistics.end(), sums.begin(), sums.begin(), std::plus<double>());
		for (size_t i = 0, n = statistics.size(); i < n; i++)
			sums[i] += statistics[i];
	}

	SimulationData const &			simulationData;
	CompartmentData const &			compartmentData;

	// tree age statistics. mapping frequency to age
	map<unsigned int, uint64_t>	treeAgeStatistics;

	// treestatistics are stored for each iteration (plus initialstand), each sizeclass and the number of statistics we have.
	// numberOfIterations+1 * numberOfSizeClasses * numberOfStatistics.
	vector< vector< vector<uint64_t> > >	treeStatistics;

	// basal area and volume statistics are stored for each iteration and each replication.
	vector< vector<double> >	basalAreaGrowthStatistics;
	vector< vector<double> >	basalAreaFinalStatistics;
	vector< vector<double> >	volumeGrowthStatistics;
	vector< vector<double> >	volumeFinalStatistics;
	vector< vector<double> >	volumeDeadStatistics;
	vector< vector<double> >	volumeMarkedStatistics;
	vector< vector<double> >	volumeNewStatistics;
	// First index is iteration number. Second index is species id. Third index is assortment id.
	// Fourth index is replication number.
	vector< vector < vector< vector<double> > > > assortmentVolumeStatistics;
	vector< vector < vector< vector<double> > > > removalAssortmentVolumeStatistics;

	unsigned int replication;

	friend class StandStatisticsSummary;
};

#endif /*COMPARTMENTSTATISTICS_H_*/
