#include "Forest.h"
#include "Polygon.h"
#include "SimulationData.h"
#include "CompartmentStatistics.h"
#include "RandomNumber.h"
#include "StandardNormalRandomNumberGenerator.h"
#include "Model.h"
#include "Timer.h"
#include "Compartment.h"
#include "parameters.h"
#include "PaddedField.h"
#include "InitialCompartmentGenerator.h"
#ifdef INCLUDE_ENVIRONMENTAL_EFFECTS
#include "LocalCorrelation.h"
#include "Matrix.h"
#endif
#include "CompartmentDataContainer.h"
#include "CompartmentData.h"
#include "ReproductionData.h"
#include "Tree.h"
#include "GrowthData.h"
#include "MortalityData.h"
#include "Log.h"
#include "LimitFinder.h"
#include "UniformRandomNumberGenerator.h"
#include "ManagementTemplate.h"
#include "TemplateContainer.h"
#include "EvaluatorFunctions.h"
#include "VolumeModel.h"
#include "SecondaryModel.h"
#include "TreePersistence.h"
#include "ForestInitializer.h"
#include "AssortmentVolumeModel.h"
#include "MarkTransformer.h"
#include "ForestType.h"
#include "ForestTypeContainer.h"
#include "SpeciesContainer.h"
#include "BasalAreaModel.h"
#include "TreeCharacteristicsVerifier.h"

#ifdef USE_MPI
#ifdef INCLUDE_ENVIRONMENTAL_EFFECTS
#include "TreeSorter.h"
#endif
#endif

#include <stdint.h>
#include <algorithm>
using std::max;
using std::distance;

#ifdef USE_MPI
#include "MPIUtil.h"
#include "TreeSender.h"
#ifdef COMPARISONSEEDING
#include "CompartmentCommunicator.h"
#endif
#endif

#ifdef CALCULATION_INSPECTION
#include "CalculationInspector.h"
#endif

#include <iostream>
using std::cout;

Forest::Forest(CompartmentDataContainer const & dataCont, SimulationData const & simData, int rep) :
	dataContainer(dataCont), simulationData(simData), replication(rep) {
	createInitialForest();
	for (TreeContainer::const_comp_iterator compIter = allTrees.compBegin(); compIter != allTrees.compEnd(); ++compIter) {
		Compartment * const comp = *compIter;
		CompartmentStatistics & stats = comp->getCompartmentData()->getCompartmentStatistics();
		stats.setReplication(rep);
	}
}

Forest::~Forest() {
}

#ifdef USE_MPI
void Forest::transferForeignTrees() {
	Timer treeTransfer;
	vector<Tree*> acceptedReceivedTrees = TreeSender::transferForeignTrees(allTrees);
	allTrees.addForeignTrees(acceptedReceivedTrees);
	vector<Tree*>().swap(acceptedReceivedTrees); // get rid of pointers
	Log::log(3, "Transfer of trees between processes took %.2fs.", treeTransfer.getElapsed());

#ifdef COMPARISONSEEDING
	Timer compartmentTransfer;
	otherCompartmentTrees.clearForeignTrees();

	for(TreeContainer::const_comp_iterator compIter=allTrees.compBegin(); compIter!=allTrees.compEnd(); ++compIter) {
		Compartment * const compartment = *compIter;
		if(compartment->getPolygon()->isActive()) {
			vector<Tree*> receivedTrees;
			CompartmentCommunicator::transferCompartmentTrees(allTrees, compartment, receivedTrees);
			otherCompartmentTrees.addForeignTrees(receivedTrees);
		}
	}
	Log::log(3, "Transfer of compartment trees between processes took %.2fs.", compartmentTransfer.getElapsed());
#endif // comparisonseeding
}

void Forest::clearForeignTrees() {
	allTrees.clearForeignTrees();
#ifdef COMPARISONSEEDING
	otherCompartmentTrees.clearForeignTrees(); // clear the trees needed for comparisonseeding too.
#endif // comparisonseeding
}
#endif // use_mpi

void Forest::createInitialForest() {
	LimitFinder & limitFinder = LimitFinder::Instance();
	limitFinder.reset();
	// Setting up all initial compartments according to settings in the CompartmentDatas.
	Log::log(2, "Setting up initial forest");

	// Creating compartment objects.
	for (CompartmentDataContainer::const_iterator compDataIter=dataContainer.begin(); compDataIter != dataContainer.end(); ++compDataIter) {
		allTrees.addCompartment(new Compartment(**compDataIter));
	}

	// Get the filename from the input file. If it is empty, nothing will be loaded.
	string const loadFileName = simulationData.getInitialStandFilename();
	ForestInitializer::load(loadFileName, allTrees);

	// Generate the missing data.
	ForestInitializer::generate(simulationData.getCompartmentGenerator(),replication,allTrees);

	TreeCharacteristicsVerifier::checkMarks(allTrees);

	calculateCompartmentCharacteristics();
	// Update the limitfinder to get the correct largest diameter and competition radius for the next iteration.
	limitFinder.update(allTrees);
	limitFinder.setLimitingMode(limitedByCompRad);
	Log::log(2, "Competition radius for next iteration: %f (m)", limitFinder.getLargestLimitingValue());
#ifdef USE_MPI
	allTrees.clearForeignTrees();
#endif
}

void Forest::simulate() {
	Log::log(2, "Initial forest size: %d", allTrees.size());

	unsigned int iteration = 0;
	calculateAssortmentStatistics(iteration);
	addFinalTreeStatistics(iteration);

#ifdef USE_MPI
	ForestSplitter & forestSplitter = ForestSplitter::Instance();
#endif

	for (; iteration < simulationData.getNumberOfIterations(); iteration++) {
#ifdef USE_MPI
		if (iteration > 0 &&
				(simulationData.hasIterationLoadBalancing() &&
				 forestSplitter.needRebalancing(allTrees, simulationData.getIterationLoadBalancingDetectionPct())) ||
				(forestSplitter.isActiveAreaTooSmall(allTrees))) {
			Timer rebalancing;
			Log::log(2, "Rebalancing load for iteration %d", iteration + 1);
			forestSplitter.split(&allTrees);
			TreeSender::rebalanceTreesFast(allTrees);
			for (TreeContainer::const_comp_iterator compIter = allTrees.compBegin(), compIterEnd = allTrees.compEnd();
					compIter != compIterEnd; ++compIter) {
				(*compIter)->communicateManagementHistory();
			}
			Log::log(2, "Rebalancing took %.2fs. %d trees in forest.", rebalancing.getElapsed(), allTrees.size());
		}
#endif
		Log::log(2, "Iteration %d", iteration + 1);
		simulateIteration(iteration + 1);
	}

	// Add age statistics for this replication
	storeAgeStatistics();

	// Write results to disk.
	if (simulationData.isSaveFinalTrees()) {
		Log::log(2, "Writing trees to disk.");
		PaddedField replicationNumber(lengthOfNumberString, '0');
		replicationNumber.setValue(replication);
		string const filename = simulationData.getOutputDirectory(true) +
				simulationData.getSimulationName() + replicationFileName +
				replicationNumber.str() + fileNameSuffix;
		TreePersistence::write(allTrees,filename);
	} else {
		Log::log(2, "NOT Writing trees to disk.");
	}
}

void Forest::simulateIteration(unsigned int iteration) {
	Timer t;
	Log::log(3, "Iteration started");
#ifdef CALCULATION_INSPECTION
	CalculationInspector& inspector = CalculationInspector::Instance();
	if (inspector.isInspectionEnabled()) {
		inspector.setCurrentIteration(iteration);
	}
#endif
#ifdef INCLUDE_ENVIRONMENTAL_EFFECTS
	if (simulationData.hasGrowthEnvironmentalEffects() ||
			simulationData.hasMortalityEnvironmentalEffects()) {
		// Random numbers are needed for environmental effects.
		// Generate them here so they can be sent to neighbouring processes. 
		bool includeSets[Tree::NUMBER_OF_RND_NUM_SETS];
		includeSets[0] = simulationData.hasGrowthEnvironmentalEffects();
		includeSets[1] = simulationData.hasMortalityEnvironmentalEffects();
		allTrees.assignRandomNumbers(includeSets);
#ifdef USE_MPI
		// Sort the trees so that correlated random numbers can be calculated in
		// the correct order.
		Log::log(3, "Sorting trees for local Cholesky decomposition.");
		Timer sortingTime;
		TreeSorter::sortForLocalCholesky(allTrees);
		Log::log(3, "Tree sorting took %.2fs.", sortingTime.getElapsed());
#endif
	}
#endif
#ifdef USE_MPI
	transferForeignTrees();
#endif

	if (simulationData.hasForestGrowth()) {
		Timer growth;
		Log::log(3, "Calculating forest growth...");
		calculateGrowth(iteration);
		TreeCharacteristicsVerifier::checkGrowth(allTrees);
		Log::log(3, "Growth phase took %.2fs.", growth.getElapsed());
	}

	unsigned int numberOfDeadTrees = 0;
	if (simulationData.hasForestMortality()) {
		Timer mortality;
		Log::log(3, "Calculating forest mortality...");
		numberOfDeadTrees = calculateMortality(iteration);
		Log::log(3, "Number of dead trees: %d", numberOfDeadTrees);
		Log::log(3, "Mortality phase took %.2fs.", mortality.getElapsed());
	}

	Timer reproduction;
	Log::log(3, "Calculating forest reproduction...");
	calculateReproduction(iteration);
	Log::log(3, "Reproduction phase took %.2fs.", reproduction.getElapsed());

	// The growth, mortality and reproduction phases are complete, commit all changes.
	Timer commitChanges;
	Log::log(3, "Applying changes from growth, mortality and reproduction phases");

	// Update marks and age of preexisting trees
	growAndAgeTrees();

	calculateGrowthStatistics(iteration);

	addGrowthTreeStatistics(iteration);

	// Remove dead trees
	allTrees.eraseDeadAndMarked();

	// Add new trees
	addNewTrees(iteration);

	Log::log(3, "Applying changes took %.2fs.", commitChanges.getElapsed());

	Timer characteristicsTimer;
	Log::log(3, "Calculating compartment characteristics...");
	calculateCompartmentCharacteristics();
	Log::log(3, "Characteristics calculations took %.2fs.", characteristicsTimer.getElapsed());

	Timer managementTimer;
	Log::log(3, "Calculating forest management");
	calculateManagement(iteration);
	Log::log(3, "Management phase took %.2fs.", managementTimer.getElapsed());

#ifdef USE_MPI
	allTrees.clearForeignTrees(); // Empty the list of trees which came from other processes for this iteration
	otherCompartmentTrees.clearForeignTrees(); // clear the trees needed for comparisonseeding too.
#endif
	allTrees.clearRectangles(); // Remove the rectangles pointers to trees as they are no longer needed

	Timer statisticsTimer;
	Log::log(3, "Calculating iteration statistics...");

	addRemovalStatistics(iteration);
	calculateRemovalAssortmentStatistics(iteration);

	// Remove marked trees
	allTrees.eraseDeadAndMarked();

	// Calculate and add statistics for the final stand
	calculateAssortmentStatistics(iteration);

	addFinalTreeStatistics(iteration);

	calculateCompartmentCharacteristics();

	Log::log(3, "Statistics calculations took %.2fs.", statisticsTimer.getElapsed());

	// Reset and update the limitfinder to get the correct largest diameter and competition radius for the next iteration.
	LimitFinder & limitFinder = LimitFinder::Instance();
	limitFinder.reset();
	limitFinder.update(allTrees);
	limitFinder.setLimitingMode(limitedByCompRad);
	Log::log(2, "Competition radius for next iteration: %f (m)", limitFinder.getLargestLimitingValue());

	if (simulationData.isSaveIterationTrees()) {
		// Write results to disk.
		Log::log(2, "Writing trees to disk");
		PaddedField replicationNumber(lengthOfNumberString, '0');
		replicationNumber.setValue(replication);
		PaddedField iterationNumber(lengthOfNumberString, '0');
		iterationNumber.setValue(iteration);
		string filename = simulationData.getOutputDirectory(true)
				+ simulationData.getSimulationName() + replicationFileName + replicationNumber.str()
				+ "_iteration_" + iterationNumber.str() + fileNameSuffix;
		TreePersistence::write(allTrees,filename);
	}

	unsigned int const numTrees = allTrees.size();
	unsigned int const sizeTree = sizeof(Tree);
	double const memTrees = ((double) (numTrees * sizeTree) / 1024.0) / 1024.0;
	Log::log(2, "Iteration %d completed in %.2fs. %d trees in forest.", iteration, t.getElapsed(), numTrees);
	Log::log(2, "Memory required for Tree objects: %d * %d bytes = %.2f megabytes.", numTrees, sizeTree, memTrees);
}

void Forest::calculateGrowthStatistics(unsigned int iteration) {
	size_t const numComp = allTrees.compSize();
	vector<double> basalAreas(numComp, 0.0);
	vector<double> volumes(numComp, 0.0);

	for(TreeContainer::const_comp_iterator compIter=allTrees.compBegin(); compIter!=allTrees.compEnd(); ++compIter) {
		Compartment * const comp = *compIter;
		size_t const index = distance(allTrees.compBegin(),compIter);
		basalAreas[index] = EvaluatorFunctions::calculateBasalAreaSum(comp);
		volumes[index] = VolumeModel::calculateStemVolumeOf(comp);
	}
#ifdef USE_MPI
	MPIUtil::sum(basalAreas, MPIUtil::getActiveCommunicator());
	MPIUtil::sum(volumes, MPIUtil::getActiveCommunicator());
#endif
	for(TreeContainer::const_comp_iterator compIter=allTrees.compBegin(); compIter!=allTrees.compEnd(); ++compIter) {
		Compartment * const comp = *compIter;
		size_t const index = distance(allTrees.compBegin(),compIter);
		CompartmentStatistics & stats = comp->getCompartmentData()->getCompartmentStatistics();
		// These values are already combined over MPI, do not combine in CompartmentStatistics.
		double const basalAreaGrowth = basalAreas[index] - comp->getTotalBasalArea();
		stats.addBasalAreaGrowth(iteration, basalAreaGrowth);
		double const volumeGrowth = volumes[index] - comp->getTotalVolume();
		stats.addVolumeGrowth(iteration, volumeGrowth);
	}
}

void Forest::calculateAssortmentStatistics(unsigned int iteration) {
	vector< vector<double> > volumes;
	for (TreeContainer::const_comp_iterator compIter = allTrees.compBegin(); compIter != allTrees.compEnd(); ++compIter) {
		Compartment * const comp = *compIter;
		AssortmentVolumeModel volumeModel(*comp);
		volumeModel.calculateForCompartment(volumes, AssortmentVolumeModel::ALL_TREES);
		CompartmentStatistics & stats = comp->getCompartmentData()->getCompartmentStatistics();
		stats.addAssortmentVolumes(iteration, volumes);
	}
}

void Forest::calculateRemovalAssortmentStatistics(unsigned int iteration) {
	vector< vector<double> > volumes;
	for (TreeContainer::const_comp_iterator compIter = allTrees.compBegin(); compIter != allTrees.compEnd(); ++compIter) {
		Compartment * const comp = *compIter;
		AssortmentVolumeModel volumeModel(*comp);
		volumeModel.calculateForCompartment(volumes, AssortmentVolumeModel::MARKED_TREES);
		CompartmentStatistics & stats = comp->getCompartmentData()->getCompartmentStatistics();
		stats.addRemovalAssortmentVolumes(iteration, volumes);
	}
}

void Forest::calculateCompartmentCharacteristics() {
	size_t const numComp = allTrees.compSize();
	vector<double> diameterSums(numComp);
	vector<double> basalAreaSums(numComp);
	vector<double> volumeSums(numComp);
	vector<uint64_t> sizes(numComp);

	for(TreeContainer::const_comp_iterator compIter=allTrees.compBegin(); compIter!=allTrees.compEnd(); ++compIter) {
		Compartment const * const comp = *compIter;
		size_t const index = distance(allTrees.compBegin(),compIter);
		diameterSums[index] = EvaluatorFunctions::calculateMarkSum(comp, TreeMarks::DIAMETER);
		basalAreaSums[index] = EvaluatorFunctions::calculateBasalAreaSum(comp);
		volumeSums[index] = VolumeModel::calculateStemVolumeOf(comp);
		sizes[index] = comp->size();
	}
#ifdef USE_MPI
	/*
	 * I think compartment communication doesn't make sense here.
	 * Pros: Using the compartmentcommunicator means smaller groups of processes participating
	 * Cons: Processes go compartment by compartment (waiting for each other).
	 * If we use compartmentcommunicators here we have to repeat this after rebalancing.
	 *
	 * It is actually useful to update the values in processes which don't have part of a
	 * compartment as they may have a part of it after a rebalancing. This way, they don't
	 * have to redo this calculation&communication.
	 */
	MPIUtil::sum(diameterSums, MPIUtil::getActiveCommunicator());
	MPIUtil::sum(basalAreaSums, MPIUtil::getActiveCommunicator());
	MPIUtil::sum(sizes, MPIUtil::getActiveCommunicator());
	MPIUtil::sum(volumeSums, MPIUtil::getActiveCommunicator());
#endif
	for(TreeContainer::const_comp_iterator compIter=allTrees.compBegin(); compIter!=allTrees.compEnd(); ++compIter) {
		Compartment * const comp = *compIter;
		size_t const index = distance(allTrees.compBegin(),compIter);
		comp->setTotalBasalArea(basalAreaSums[index]);
		comp->setTotalSize(sizes[index]);
		comp->setTotalDiameter(diameterSums[index]);
		comp->setTotalVolume(volumeSums[index]);
	}
}

void Forest::calculateGrowth(unsigned int iteration) {
	const unsigned int numberOfTrees = allTrees.size();

	LimitFinder & limitFinder = LimitFinder::Instance();
	limitFinder.setLimitingMode(limitedByCompRad);
	double const competitionRadius = limitFinder.getLargestLimitingValue();

#ifdef CALCULATION_INSPECTION
	CalculationInspector& inspector = CalculationInspector::Instance();
#endif

#ifdef INCLUDE_ENVIRONMENTAL_EFFECTS
	vector<Vector> environmentalEffects(TreeMarks::NUMBER_OF_MARKS);
	for (unsigned int index=0; index<TreeMarks::NUMBER_OF_MARKS; index++) {
		Log::log(4, "Will allocate %dx%d = %d bytes for environmental effects vector", numberOfTrees, sizeof(double),
				numberOfTrees*sizeof(double));
		environmentalEffects[index] = Vector(numberOfTrees);
		GrowthData const & growthData = simulationData.getGrowthData(index);
		// Growthenvironmentaleffects can be enabled/disabled.
		if (growthData.hasEnvironmentalEffects()) {
			allTrees.setLimitingValue(competitionRadius * simulationData.getCorrelationDistanceFactor());

			LocalCorrelation correlation;
			correlation.setCompetitionRadius(competitionRadius);
			correlation.setDistanceFactor(simulationData.getCorrelationDistanceFactor());
			correlation.setCoefficients(growthData.getEnvironmentalEffectCoefficients());
			correlation.setRandomNumberSetIndex(0);
			correlation.setMarkIndex(index);
			for (unsigned int curTreeIndex = 0; curTreeIndex < numberOfTrees; ++curTreeIndex) {
				Tree const * const curTree = allTrees[curTreeIndex];
				double const x = correlation.calculateCorrelatedRandomNumber(curTree, allTrees.getTreesCloseTo(curTree));
				environmentalEffects[index][curTreeIndex] = x;
			}
			unsigned int numCalculations = correlation.getNumberOfCalculations();
			double avgMatrixSize = correlation.getAverageMatrixSize();
			unsigned int maxMatrixSize = correlation.getMaximumMatrixSize();
			Log::log(4, "Performed %u local covariance calculations with average matrix size %3.2fx%3.2f and maximum matrix size %ux%u",
					numCalculations, avgMatrixSize, avgMatrixSize, maxMatrixSize, maxMatrixSize);
		}
	}
#endif
	vector<bool> markHasGrowth(TreeMarks::NUMBER_OF_MARKS);
	for (unsigned int index = 0; index < TreeMarks::NUMBER_OF_MARKS; index++) {
		GrowthData const & growthData = simulationData.getGrowthData(index);
		markHasGrowth[index] = growthData.hasTreeGrowth();
	}
	// Calculate competition indexes.
	// We only have to go through the trees nearby, not the whole forest.
	// CompetitionRadius limiting factor here.
	allTrees.setLimitingValue(competitionRadius);
	// Preallocate temp-vectors here to avoid reallocation
	Vector secondaryVariables(SecondaryModel::numberOfVariables());
	Vector modelVariables(Model::NUMBER_OF_MODEL_VARIABLES);
	for (unsigned int curTreeIndex = 0; curTreeIndex < numberOfTrees; curTreeIndex++) {
		Tree * const curTree = allTrees[curTreeIndex];

		SecondaryModel::calculateSecondaryVariables(
				curTree, curTree->getParentCompartment()->getCompartmentData(),
				simulationData.getSizeClassHandler(),
				simulationData.getAgeClassHandler(),
				simulationData.getHeightClassHandler(),
				secondaryVariables);

		for (unsigned int index = 0; index < TreeMarks::NUMBER_OF_MARKS; index++) {
			if (!markHasGrowth[index]) {
				continue;
			}
			modelVariables.clear();
			modelVariables[Model::TAU] = curTree->getInitialHeterogeneity();
#ifdef COMPARISONSEEDING
			UniformRandomNumberGenerator::setSeeds(curTree,103);
#endif
			modelVariables[Model::XI] = 0.0;
			GrowthData const & growthData = simulationData.getGrowthData(index);
#ifdef INCLUDE_ENVIRONMENTAL_EFFECTS
			if (growthData.hasEnvironmentalEffects())
				modelVariables[Model::XI] = environmentalEffects[index][curTreeIndex];
#endif
			// Calculate competition indexes for this tree.
			Model::calculateModelVariables(curTree, allTrees.getTreesCloseTo(curTree), competitionRadius,
					modelVariables);
			const double y = Model::calculateMark(modelVariables, secondaryVariables, growthData.getModelTerms());
			const double growth = exp(y);
#ifdef CALCULATION_INSPECTION
			if (inspector.shouldInspectGrowth()) {
				string const markName = TreeMarks::getMarkName(index);
				inspector.inspectModel("Growth", *curTree, curTreeIndex, index, growthData, modelVariables, secondaryVariables, y);
				inspector.print("Setting %s growth to: exp(y) = %f", markName.c_str(), growth);
			}
#endif
			curTree->setGrowth(index, growth);
		}
	}
}

unsigned int Forest::calculateMortality(unsigned int iteration) {
	unsigned int numberOfDeadTrees = 0;
	unsigned int const numberOfTrees = allTrees.size();

#ifdef CALCULATION_INSPECTION
	CalculationInspector& inspector = CalculationInspector::Instance();
#endif

	for (unsigned int index = 0; index < TreeMarks::NUMBER_OF_MARKS; index++) {
		const string markName = TreeMarks::getMarkName(index);
		if (simulationData.getMortalityData(index).hasRegularTreeMortality()) {
			double const threshold =
					simulationData.getMortalityData(index).getRegularMortalityThresholdValue();
			for (unsigned int curTreeIndex = 0; curTreeIndex < numberOfTrees; curTreeIndex++) {
				Tree * const curTree = allTrees[curTreeIndex];
				double const growth = curTree->getGrowth(index);
#ifdef CALCULATION_INSPECTION
				if (inspector.shouldInspectMortality())
					inspector.print("Regular %s mortality of tree [%u] at (%f, %f): %f < %f?",
							markName.c_str(), curTreeIndex, curTree->getX(), curTree->getY(), growth, threshold);
#endif
				if (growth < threshold) {
					// Mark the tree as dead.
					curTree->die();
					numberOfDeadTrees++;
#ifdef CALCULATION_INSPECTION
					if (inspector.shouldInspectMortality())
						inspector.print("-> Tree dies.");
#endif
				}
			}
		}
	}

	LimitFinder & limitFinder = LimitFinder::Instance();
	limitFinder.setLimitingMode(limitedByCompRad);
	double const competitionRadius = limitFinder.getLargestLimitingValue();

#ifdef INCLUDE_ENVIRONMENTAL_EFFECTS
	vector<Vector> environmentalEffects(TreeMarks::NUMBER_OF_MARKS);
	for (unsigned int index = 0; index < TreeMarks::NUMBER_OF_MARKS; index++) {
		Log::log(4, "Will allocate %dx%d = %d bytes for environmental effects vector", numberOfTrees, sizeof(double),
				numberOfTrees*sizeof(double));
		environmentalEffects[index] = Vector(numberOfTrees);
		MortalityData const & mortalityData = simulationData.getMortalityData(index);
		// Mortality environmentaleffects can be enabled/disabled.
		if (mortalityData.hasEnvironmentalEffects()) {
			allTrees.setLimitingValue(competitionRadius * simulationData.getCorrelationDistanceFactor());

			LocalCorrelation correlation;
			correlation.setCompetitionRadius(competitionRadius);
			correlation.setDistanceFactor(simulationData.getCorrelationDistanceFactor());
			correlation.setCoefficients(mortalityData.getEnvironmentalEffectCoefficients());
			correlation.setRandomNumberSetIndex(1);
			correlation.setMarkIndex(index);
			for (unsigned int curTreeIndex = 0; curTreeIndex < numberOfTrees; ++curTreeIndex) {
				Tree const * const curTree = allTrees[curTreeIndex];
				double const x = correlation.calculateCorrelatedRandomNumber(curTree, allTrees.getTreesCloseTo(curTree));
				environmentalEffects[index][curTreeIndex] = x;
			}
			unsigned int numCalculations = correlation.getNumberOfCalculations();
			double avgMatrixSize = correlation.getAverageMatrixSize();
			unsigned int maxMatrixSize = correlation.getMaximumMatrixSize();
			Log::log(2, "Performed %u local covariance calculations with average matrix size %3.2fx%3.2f and maximum matrix size %ux%u",
					numCalculations, avgMatrixSize, avgMatrixSize, maxMatrixSize, maxMatrixSize);
		}
	}
#endif

	// Calculate tree mortality using rejection sampling.
	vector<bool> markHasMortality(TreeMarks::NUMBER_OF_MARKS);
	for (unsigned int index = 0; index < TreeMarks::NUMBER_OF_MARKS; index++) {
		MortalityData const & mortalityData = simulationData.getMortalityData(index);
		markHasMortality[index] = mortalityData.hasIrregularTreeMortality();
	}
	// Calculate competition indexes.
	// We only have to go through the trees nearby, not the whole forest.
	// CompetitionRadius limiting factor here.
	allTrees.setLimitingValue(competitionRadius);

	Vector modelVariables(Model::NUMBER_OF_MODEL_VARIABLES);
	Vector secondaryVariables(SecondaryModel::numberOfVariables());

	for (unsigned int curTreeIndex = 0; curTreeIndex < numberOfTrees; curTreeIndex++) {
		Tree * const curTree = allTrees[curTreeIndex];
		SecondaryModel::calculateSecondaryVariables(
				curTree, curTree->getParentCompartment()->getCompartmentData(),
				simulationData.getSizeClassHandler(),
				simulationData.getAgeClassHandler(),
				simulationData.getHeightClassHandler(),
				secondaryVariables);

		for (unsigned int index = 0; index < TreeMarks::NUMBER_OF_MARKS; index++) {
			if (!markHasMortality[index] || curTree->isDead()) {
				continue;
			}
			modelVariables.clear(); // Epsilon is always zero in mortality phase
			modelVariables[Model::TAU] = curTree->getInitialHeterogeneity();
#ifdef COMPARISONSEEDING
			UniformRandomNumberGenerator::setSeeds(curTree,13);
#endif
			double const u = RandomNumber::get();

			modelVariables[Model::XI] = 0.0;
			MortalityData const & mortalityData = simulationData.getMortalityData(index);
#ifdef INCLUDE_ENVIRONMENTAL_EFFECTS
			if (mortalityData.hasEnvironmentalEffects())
				modelVariables[Model::XI] = environmentalEffects[index][curTreeIndex];
#endif
			Model::calculateModelVariables(curTree, allTrees.getTreesCloseTo(curTree), competitionRadius,
					modelVariables);
			double const y = Model::calculateMark(modelVariables, secondaryVariables, mortalityData.getModelTerms());

			double const sample = 1.0 / (1.0 + exp(-y));

#ifdef CALCULATION_INSPECTION
			if (inspector.shouldInspectMortality()) {
				inspector.inspectModel("Mortality", *curTree, curTreeIndex, index, mortalityData, modelVariables, secondaryVariables, y);
				inspector.print("sample = 1 / (1 + exp(-y)) = %f", sample);
				inspector.print("%f <= %f ?", u, sample);
			}
#endif

			if (u <= sample) {
#ifdef CALCULATION_INSPECTION
				if (inspector.shouldInspectMortality())
					inspector.print("-> Tree dies.");
#endif
				curTree->die(); // Accept sample.
				numberOfDeadTrees++;
			}
		}
	}

	return numberOfDeadTrees;
}

void Forest::calculateManagement(unsigned int iteration) {
	for(TreeContainer::const_comp_iterator compIter=allTrees.compBegin(); compIter!=allTrees.compEnd(); ++compIter) {
		Compartment * const compartment = *compIter;
		CompartmentData const * const compartmentData = compartment->getCompartmentData();
		vector<unsigned int> const & templateIds = compartmentData->getManagementTemplateIds();
		for (size_t i = 0; i < templateIds.size(); i++) {
			ManagementTemplate const * const managementTemplate = TemplateContainer::get(templateIds[i]);
			managementTemplate->apply(iteration, simulationData.getIterationLength(), compartment, &allTrees);
			// Mark removed trees as dead so that subsequential management methods
			// may disregard them.
			for (Tree * tree = compartment->getFirstTree(); tree != NULL; tree = tree->getNextTreeInCompartment()) {
				if (tree->isMarked())
					tree->die();
			}
		}
	}
}

void Forest::calculateReproduction(unsigned int iteration) {
	/* Setting the limiting value for the rectanglecontainer already here to avoid
	 * changing it repeatedly in the loop for all compartments!
	 * The first part is limited by competition radius and the second by diameters!
	 */
	LimitFinder & limitFinder = LimitFinder::Instance();
	limitFinder.setLimitingMode(limitedByDiameterAndCompRad);
	double const limitingValue = limitFinder.getLargestLimitingValue();
	allTrees.setLimitingValue(limitingValue);

	TreeContainer newTrees(false);

	/* Calculate all tree positions first
	 */
	for(TreeContainer::const_comp_iterator compIter = allTrees.compBegin(); compIter != allTrees.compEnd(); ++compIter) {
		Compartment * const compartment = *compIter;
		if (compartment->getPolygon()->isActive()) {
#ifdef COMPARISONSEEDING
			CompartmentData const * const compartmentData = compartment->getCompartmentData();
			size_t const compartmentId = compartmentData->getIdNumber();
			UniformRandomNumberGenerator::setSeed(2145.0*(compartmentId+1)+iteration+replication*1001);
#endif
			compartment->calculateNewTreePositions(iteration, simulationData, allTrees, otherCompartmentTrees);
			list<Tree*> compartmentNewTrees = compartment->getNewTrees();
			newTrees.addAll(compartmentNewTrees);
		}
	}

#ifdef INCLUDE_ENVIRONMENTAL_EFFECTS
	if (simulationData.hasReproductionEnvironmentalEffects()) {
		// Reproduction phase uses environmental effects to generate new tree marks.
		// Since the correlated random numbers are only calculated between new trees,
		// we only need to assign random numbers to new trees.
		bool includeSets[Tree::NUMBER_OF_RND_NUM_SETS] = {true, false};
		newTrees.assignRandomNumbers(includeSets);
#ifdef USE_MPI
		// Sort the trees so that correlated random numbers can be calculated in
		// the correct order.
		Log::log(3, "Sorting new trees for local Cholesky decomposition.");
		Timer sortingTime;
		TreeSorter::sortForLocalCholesky(newTrees);
		Log::log(3, "Tree sorting took %.2fs.", sortingTime.getElapsed());
		// We need to communicate new trees between processes to distribute random numbers
		Timer treeTransfer;
		vector<Tree*> acceptedReceivedTrees = TreeSender::transferForeignTrees(newTrees);
		newTrees.addForeignTrees(acceptedReceivedTrees);
		vector<Tree*>().swap(acceptedReceivedTrees); // get rid of pointers
		Log::log(3, "Transfer of new trees between processes took %.2fs.", treeTransfer.getElapsed());
#endif
	}
#endif

	/* Calculate all tree diameters after the positions
	 * Done for all compartments in parallel
	 */
	calculateNewTreeMarks(newTrees, iteration);

	TreeCharacteristicsVerifier::checkMarks(newTrees);
}

void Forest::calculateNewTreeMarks(TreeContainer & newTrees, unsigned int iteration) {
	Log::log(4, "Calculating new tree marks for all compartments");
	// We only have to go through the trees nearby, not the whole forest.
	// CompetitionRadius limiting factor here. To avoid checking for updated diameters
	// in the reproduction for each compartment, the limitfinder is updated earlier!
	LimitFinder & limitFinder = LimitFinder::Instance();
	limitFinder.setLimitingMode(limitedByCompRad);
	double const competitionRadius = limitFinder.getLargestLimitingValue();

	size_t const numberOfNewTrees = newTrees.size();
	Vector modelVariables(Model::NUMBER_OF_MODEL_VARIABLES);
	Vector secondaryVariables(SecondaryModel::numberOfVariables());

#ifdef CALCULATION_INSPECTION
	CalculationInspector& inspector = CalculationInspector::Instance();
#endif

	// Optionally calculate environmental effects.
	// Set tree diameters for all new trees in all compartments.

#ifdef INCLUDE_ENVIRONMENTAL_EFFECTS
	vector<Vector> environmentalEffects(TreeMarks::NUMBER_OF_MARKS);
	for (unsigned int index = 0; index < TreeMarks::NUMBER_OF_MARKS; index++) {
		Log::log(4, "Will allocate %dx%d = %d bytes for environmental effects vector",
				numberOfNewTrees, sizeof(double), numberOfNewTrees*sizeof(double));
		environmentalEffects[index] = Vector(numberOfNewTrees);
		ReproductionData const & reproductionData = simulationData.getReproductionData(index);
		if (reproductionData.hasEnvironmentalEffects()) {
			newTrees.setLimitingValue(competitionRadius * simulationData.getCorrelationDistanceFactor());

			LocalCorrelation correlation;
			correlation.setCompetitionRadius(competitionRadius);
			correlation.setDistanceFactor(simulationData.getCorrelationDistanceFactor());
			correlation.setCoefficients(reproductionData.getEnvironmentalEffectCoefficients());
			correlation.setRandomNumberSetIndex(0);
			correlation.setMarkIndex(index);
			for (unsigned int curTreeIndex = 0; curTreeIndex < numberOfNewTrees; ++curTreeIndex) {
				Tree const * const curTree = newTrees[curTreeIndex];
				double const x = correlation.calculateCorrelatedRandomNumber(curTree, newTrees.getTreesCloseTo(curTree));
				environmentalEffects[index][curTreeIndex] = x;
			}
			unsigned int numCalculations = correlation.getNumberOfCalculations();
			double avgMatrixSize = correlation.getAverageMatrixSize();
			unsigned int maxMatrixSize = correlation.getMaximumMatrixSize();
			Log::log(4, "Performed %u local covariance calculations with average matrix size %3.2fx%3.2f and maximum matrix size %ux%u",
					numCalculations, avgMatrixSize, avgMatrixSize, maxMatrixSize, maxMatrixSize);
		}
	}
#endif

	// Calculate tree diameters and update new tree statistics.
	for (unsigned int index = 0; index < TreeMarks::NUMBER_OF_MARKS; index++) {
		for (size_t curTreeIndex = 0; curTreeIndex < numberOfNewTrees; curTreeIndex++) {
			Tree * const curTree = newTrees[curTreeIndex];
			CompartmentData * const compartmentData = curTree->getParentCompartment()->getCompartmentData();

			SecondaryModel::calculateSecondaryVariables(
					curTree,
					compartmentData,
					simulationData.getSizeClassHandler(),
					simulationData.getAgeClassHandler(),
					simulationData.getHeightClassHandler(),
					secondaryVariables);

			modelVariables.clear();
#ifdef COMPARISONSEEDING
			UniformRandomNumberGenerator::setSeeds(curTree,83);
#endif
			modelVariables[Model::XI] = 0.0;
			ReproductionData const & reproductionData = simulationData.getReproductionData(index);
#ifdef INCLUDE_ENVIRONMENTAL_EFFECTS
			if (reproductionData.hasEnvironmentalEffects())
				modelVariables[Model::XI] = environmentalEffects[index][curTreeIndex];
#endif
			modelVariables[Model::TAU] = (index == 0) ? StandardNormalRandomNumberGenerator::get()
					: (double) curTree->getInitialHeterogeneity();
			if (index == 0) {
				curTree->setInitialHeterogeneity(modelVariables[Model::TAU]);
			}

			Model::calculateModelVariables(curTree, allTrees.getTreesCloseTo(curTree), competitionRadius,
					modelVariables);
			const double initialMark = Model::calculateMark(modelVariables, secondaryVariables,
					reproductionData.getModelTerms());
#ifdef CALCULATION_INSPECTION
			if (inspector.shouldInspectReproduction()) {
				inspector.inspectModel("Reproduction", *curTree, curTreeIndex, index, reproductionData, modelVariables, secondaryVariables, initialMark);
			}
#endif
			curTree->setMark(index, initialMark);
		}

		// Transform tree marks
		const string markName = TreeMarks::getMarkName(index);

		// Calculate an initial CDF for all trees
		MarkTransformer transformer(index);
		transformer.calculateInitialCDF(newTrees);

		// We need to loop over the compartments to provide forest type-specific target distributions
		TreeContainer::const_comp_iterator compIter = allTrees.compBegin(), compEnd = allTrees.compEnd();
		for (; compIter != compEnd; ++compIter) {
			Compartment * const compartment = *compIter;
			CompartmentData * const compData = compartment->getCompartmentData();
			if (compartment->getPolygon()->isActive()) {
				list<Tree*> const & newTreesInCompartment = compartment->getNewTrees();

				unsigned int forestTypeId = compData->getForestTypeId();
				ForestType const * const forestType = ForestTypeContainer::get(forestTypeId);
				MarkTransformation const & transformation = forestType->getReproductionTransformation(index);
				if (transformation.isLoaded()) {
					Log::log(5, "Transforming mark %s of new trees in compartment %d", markName.c_str(), compData->getIdNumber());
					transformer.setTargetTransformation(transformation);

					TreeContainer compartmentTrees(false);
					compartmentTrees.addAll(newTreesInCompartment);
					transformer.transform(compartmentTrees);

#ifdef CALCULATION_INSPECTION
					if (inspector.shouldInspectReproduction()) {
						size_t const numCompTrees = compartmentTrees.size();
						for (size_t curTreeIndex = 0; curTreeIndex < numCompTrees; curTreeIndex++) {
							Tree * const curTree = compartmentTrees[curTreeIndex];
							inspector.print("Transformed mark '%s' of new tree [%d] in compartment %d: %f", markName.c_str(), curTreeIndex, compData->getIdNumber(), curTree->getMark(index));
						}
					}
#endif
				} else {
					Log::log(5, "Skipping transformation of mark %s. No distribution defined.", markName.c_str());
				}

				if (index == TreeMarks::DIAMETER) {
					// Multiply the diameter by the new tree diameter factor
					for (list<Tree*>::const_iterator treeIter = newTreesInCompartment.begin(), endIter = newTreesInCompartment.end();
							treeIter != endIter; ++treeIter) {
						Tree * const curTree = *treeIter;
						Species const * const species = SpeciesContainer::get(curTree->getSpeciesId());
						double const diamFactor = forestType->getNewTreeDiameterFactor(species);
						curTree->setMark(index, curTree->getMark(index) * diamFactor);	
					}
				}
			}
		}
	}

	// Calculate compartment statistics
	TreeContainer::const_comp_iterator compIter = allTrees.compBegin(), compEnd = allTrees.compEnd();
	for (; compIter != compEnd; ++compIter) {
		Compartment * const compartment = *compIter;
		CompartmentData * const compData = compartment->getCompartmentData();
		if (compartment->getPolygon()->isActive()) {
			list<Tree*> const & newTreesInCompartment = compartment->getNewTrees();
			CompartmentStatistics & compartmentStatistics = compData->getCompartmentStatistics();
			for (list<Tree*>::const_iterator treeIter = newTreesInCompartment.begin(), endIter = newTreesInCompartment.end();
					treeIter != endIter; ++treeIter) {
				compartmentStatistics.addNewTreeStatistics(iteration, **treeIter);
			}
		}
	}
}

void Forest::addGrowthTreeStatistics(unsigned int iteration) {
	for(TreeContainer::const_iterator treeIter=allTrees.begin(); treeIter!=allTrees.end(); ++treeIter) {
		Tree const * const curTree = *treeIter;
		Compartment * const curCompartment = curTree->getParentCompartment();
		CompartmentData * const curCompartmentData = curCompartment->getCompartmentData();
		CompartmentStatistics & curCompartmentStatistics = curCompartmentData->getCompartmentStatistics();
		curCompartmentStatistics.addGrowthTreeStatistics(iteration, *curTree);
		if (curTree->isDead()) {
			double const volume = VolumeModel::calculateStemVolumeOf(curTree);
			curCompartmentStatistics.addVolumeDead(iteration, volume); // value combined in CompartmentStatistics
			curCompartmentStatistics.addDeadTreeStatistics(iteration, *curTree);
		}
	}
}

void Forest::addRemovalStatistics(unsigned int iteration) {
	TreeContainer::const_iterator treeIter = allTrees.begin(), endIter = allTrees.end();
	for (; treeIter != endIter; ++treeIter) {
		Tree const * const curTree = *treeIter;
		if (curTree->isMarked()) {
			Compartment const * const compartment = curTree->getParentCompartment();
			CompartmentData * const compartmentData = compartment->getCompartmentData();
			CompartmentStatistics & compartmentStatistics = compartmentData->getCompartmentStatistics();
			compartmentStatistics.addRemovalTreeStatistics(iteration, *curTree);
			compartmentStatistics.addVolumeMarked(iteration, VolumeModel::calculateStemVolumeOf(curTree));
		}
	}
}

void Forest::addFinalTreeStatistics(unsigned int iteration) {
	for(TreeContainer::const_iterator treeIter=allTrees.begin(); treeIter!=allTrees.end(); ++treeIter) {
		Tree const * const curTree = *treeIter;
		Compartment * const curCompartment = curTree->getParentCompartment();
		CompartmentData * const curCompartmentData = curCompartment->getCompartmentData();
		CompartmentStatistics & curCompartmentStatistics = curCompartmentData->getCompartmentStatistics();
		curCompartmentStatistics.addFinalTreeStatistics(iteration, *curTree);

		double const volume = VolumeModel::calculateStemVolumeOf(curTree);
		curCompartmentStatistics.addVolumeFinal(iteration, volume); // value combined in CompartmentStatistics

		double const basalArea = BasalAreaModel::calculate(*curTree);
		curCompartmentStatistics.addBasalAreaFinal(iteration, basalArea); // value combined in CompartmentStatistics
	}
}

void Forest::growAndAgeTrees() {
	unsigned int const iterationLength = simulationData.getIterationLength();
	for(TreeContainer::const_iterator treeIter=allTrees.begin(); treeIter!=allTrees.end(); ++treeIter) {
		Tree * const curTree = *treeIter;
		if (!curTree->isDead()) {
			curTree->grow();
		}
		curTree->addAge(iterationLength);
	}
}

void Forest::addNewTrees(unsigned int iteration) {
	for(TreeContainer::const_comp_iterator compIter=allTrees.compBegin(); compIter!=allTrees.compEnd(); ++compIter) {
		Compartment * const compartment = *compIter;
		CompartmentData * const curCompartmentData = compartment->getCompartmentData();
		CompartmentStatistics & curCompartmentStatistics = curCompartmentData->getCompartmentStatistics();

		double const volume = VolumeModel::calculateStemVolume(compartment->getNewTrees());
		curCompartmentStatistics.addVolumeNew(iteration, volume); // value combined in CompartmentStatistics

		allTrees.addAll(compartment->getNewTrees());
		compartment->addNewTrees();
	}
}

void Forest::storeAgeStatistics() {
	for(TreeContainer::const_comp_iterator compIter=allTrees.compBegin(); compIter!=allTrees.compEnd(); ++compIter) {
		Compartment * const compartment = *compIter;
		CompartmentData * const curCompartmentData = compartment->getCompartmentData();
		CompartmentStatistics & curCompartmentStatistics = curCompartmentData->getCompartmentStatistics();
		for (Tree const * curTree = compartment->getFirstTree(); curTree != NULL; curTree
				= curTree->getNextTreeInCompartment()) {
			curCompartmentStatistics.addAgeFrequencies(curTree->getAge(), 1);
		}
	}
}
