#ifndef HEIGHTCLASSHANDLER_H_
#define HEIGHTCLASSHANDLER_H_

#include <string>
using std::string;
#include <vector>
using std::vector;

#ifdef USE_MPI
#include "MPIUtil.h"
#endif

class TreeContainer;
class Tree;

#include <xercesc/util/XercesVersion.hpp>

namespace XERCES_CPP_NAMESPACE {
	class DOMElement;
}

class HeightClassHandler {
public:
	HeightClassHandler();
	~HeightClassHandler();

#ifdef USE_MPI
	inline void broadcast(int const rootId, MPI_Comm const comm) {
		MPIUtil::broadcast(heightClassNames, rootId, comm);
		MPIUtil::broadcast(heightClassBounds, rootId, comm);
	}
#endif
	/**
	 * Adds a height class to the handler.
	 * Returns the id of the height class.
	 */
	unsigned int addHeightClass(string const & name, double minimumHeight);
	/**
	 * Determines the height class of a Tree.
	 * The height classes must have been added in ascending order, otherwise determineHeightClasses() won't work.
	 * If the tree is shorter than the smallest height class, 0 will be returned;
	 * i.e. the tree is considered to belong to the smallest height class.
	 */
	unsigned int determineHeightClass(Tree const & curTree) const;
	/**
	 * Returns the number of height classes.
	 */
	unsigned int getNumberOfHeightClasses() const {
		return heightClassNames.size();
	}
	/**
	 * Returns the name of the height class identified by @heightClassId.
	 */
	string getHeightClassName(unsigned char heightClassId) const {
		return heightClassNames[heightClassId];
	}
	unsigned char getHeightClassId(string const & name) const;
	/**
	 * Loads XML data for the HeightClassHandler.
	 *
	 * Precondition:	@heightClasses must not be null.
	 */
	void loadData(XERCES_CPP_NAMESPACE::DOMElement const * heightClasses);
private:
	/**
	 * Loads XML data for one height class.
	 *
	 * Precondition:	@heightClass must not be null.
	 */
	void loadHeightClass(XERCES_CPP_NAMESPACE::DOMElement const * heightClass);

	vector<string> heightClassNames;
	vector<double> heightClassBounds;
};

#endif /* HEIGHTCLASSHANDLER_H_ */
