#include "LimitFinder.h"
#include "SimulationData.h"
#include "TreeContainer.h"
#include "Tree.h"
#include "CompetitionDistanceModel.h"

#include <algorithm>
using std::max;
using std::min;
#include <cassert>

#ifdef USE_MPI
#include "MPIUtil.h"
#endif

LimitFinder * LimitFinder::singleton=NULL;

LimitFinder & LimitFinder::Instance(SimulationData const & sim) {
	if (singleton==NULL)
		singleton = new LimitFinder(sim);
	return *singleton;
}
LimitFinder & LimitFinder::Instance() {
	// unsafe method, may return reference to null.
	assert(singleton!=NULL);
	return *singleton;
}

LimitFinder::LimitFinder(SimulationData const & sim) :
	simulationData(sim), largestDiameter(0.0), competitionDistance(0.0), largestNeeded(0.0), currentLimitingMode(limitedByAll) {
}

void LimitFinder::Destroy() {
	delete singleton;
	singleton=NULL;
}

void LimitFinder::reset() {
	largestDiameter = 0.0;
	competitionDistance = 0.0;
}

LimitFinder::~LimitFinder() {

}

void LimitFinder::update(TreeContainer const & trees) {
	size_t const size = trees.size();
	CompetitionDistanceModel competitionDistanceModel(
			simulationData.getSizeClassHandler(),
			simulationData.getAgeClassHandler(),
			simulationData.getHeightClassHandler());
	for (size_t curTreeIndex=0; curTreeIndex<size; curTreeIndex++) {
		keepTrackOfLargest(trees[curTreeIndex], competitionDistanceModel);
	}
#ifdef USE_MPI
	MPIUtil::max(largestDiameter,MPIUtil::getActiveCommunicator());
	MPIUtil::max(competitionDistance,MPIUtil::getActiveCommunicator());
#endif
}

void LimitFinder::keepTrackOfLargest(Tree const * const curTree, CompetitionDistanceModel& competitionDistanceModel) {
	double const diameterThisTree = curTree->getMark(TreeMarks::DIAMETER);
	largestDiameter = max(diameterThisTree, largestDiameter);

	double const competitionDistanceThisTree = competitionDistanceModel.calculate(curTree);
	competitionDistance = max(competitionDistanceThisTree, competitionDistance);
}

void LimitFinder::setLimitingMode(LimitingModes mode) {
	double const fromDiameterInCentimeterToRadiusInMeterCoefficient = 1.0/200.0;
	switch (mode) {
	case (limitedByAll): {
		double const largestRadiusInMeters = largestDiameter*fromDiameterInCentimeterToRadiusInMeterCoefficient;
		double const correlationDistance = competitionDistance * simulationData.getCorrelationDistanceFactor();
		largestNeeded = max(largestRadiusInMeters, correlationDistance);
		break;
	}
	case (limitedByDiameter): {
		double const largestRadiusInMeters = largestDiameter*fromDiameterInCentimeterToRadiusInMeterCoefficient;
		largestNeeded = largestRadiusInMeters;
		break;
	}
	case (limitedByCompRad): {
		largestNeeded = competitionDistance;
		break;
	}
	case (limitedByDiameterAndCompRad): {
		// (Competitionradius or largestDiameter)*2.0 is the minimum value we can use for the division.
		double const largestRadiusInMeters = largestDiameter*fromDiameterInCentimeterToRadiusInMeterCoefficient;
		largestNeeded = max(largestRadiusInMeters, competitionDistance);
		break;
	}
	case (limitedByCorrelationDistance): {
		largestNeeded = competitionDistance * simulationData.getCorrelationDistanceFactor();
		break;
	}
	}
	currentLimitingMode=mode;
}
