#ifdef INCLUDE_ENVIRONMENTAL_EFFECTS

#include "LocalCorrelation.h"
#include "Tree.h"
#include "Log.h"
#include <iostream>
#ifdef USE_MPI
#include "ForestSplitter.h"
#include "TreeSorter.h"
#include <algorithm>
#endif

LocalCorrelation::LocalCorrelation() :
  competitionRadius(0.0)
, distanceFactor(1.0)
, setIndex(0)
, markIndex(0)
, numberOfCalculations(0)
, maximumMatrixSize(0)
, averageMatrixSize(0.0) {
#ifdef USE_MPI
	// Obtain the coordinates of the local process and find the X value that
	// splits the area into two equally sized strips.
	ForestSplitter& forestSplitter = ForestSplitter::Instance();
	activeMinCoord = forestSplitter.getMinCoordinate();
	activeMaxCoord = forestSplitter.getMaxCoordinate();
	splitX = (activeMinCoord.x + activeMaxCoord.x) / 2;
#endif
}

double LocalCorrelation::calculateCorrelatedRandomNumber(
		Tree const * const curTree, vector<Tree*> const & trees) {
#ifdef USE_MPI
	// Determine the current strip (left or right) based on the current tree's position.
	currentStrip = curTree->getX() < splitX ? LEFT_STRIP : RIGHT_STRIP;
#endif
	double const squaredCompetitionRadius = competitionRadius*competitionRadius;

	// Find trees to include in the Cholesky decomposition
	findProcessedNeighbours(curTree, trees);

#ifdef USE_MPI
	TreeSorter::sortProcessedNeighbours(localTrees);
#endif

	unsigned int const numberOfTrees = localTrees.size() + 1;
	double randomNumber = 0.0;
	if (numberOfTrees == 1) {
		// No processed neighbours. Calculating a "lower triangular"
		// matrix is trivial.
		double const lowerTriangular = exp(coefficients[0] / 2.0);
		randomNumber = lowerTriangular * curTree->getRandomNumber(setIndex, markIndex); 
	} else {
		// We have to do a Cholesky decomposition
		localTrees.push_back(curTree); // The target tree is the last element
		buildCovarianceMatrix(squaredCompetitionRadius);
		// Perform Cholesky decomposition
		try {
			covarianceMatrix.choleskyDecompose(Matrix::LOWER_TRIANGULAR);
			// We are only interested in the correlated random number of the target tree.
			// Therefore we multiply the last row of the L matrix by the random numbers.
			Vector const lastRow = covarianceMatrix.getRow(numberOfTrees - 1);
			for (unsigned int i = 0; i < numberOfTrees; ++i) {
				Tree const * const tree = localTrees[i];
				randomNumber += lastRow[i] * tree->getRandomNumber(setIndex, markIndex);
			}
		} catch (char const * msg) {
			// The matrix is not positive definite.
			// The error is already logged to the log file.
			randomNumber = 0.0;
		}
	}

	// Update statistics
	averageMatrixSize = 1.0 / (numberOfCalculations + 1.0) * (numberOfCalculations * averageMatrixSize + numberOfTrees);
	++numberOfCalculations;
	if (numberOfTrees > maximumMatrixSize) {
		maximumMatrixSize = numberOfTrees;
	}

	processedTrees.insert(curTree); // Mark the current tree as processed

	return randomNumber;
}

void LocalCorrelation::findProcessedNeighbours(Tree const * const curTree, vector<Tree*> const & trees) {
	double const squaredInclusionDistance = competitionRadius*competitionRadius*distanceFactor*distanceFactor;
	if (squaredInclusionDistance == 0.0) return;

	localTrees.clear();
	unsigned int const size = trees.size();
	for (unsigned int otherTreeIndex = 0; otherTreeIndex < size; ++otherTreeIndex) {
		Tree const * const otherTree = trees[otherTreeIndex];
		if (curTree != otherTree) {
			double const squaredTreeDistance = curTree->squaredDistanceTo(*otherTree);
			if (squaredTreeDistance <= squaredInclusionDistance && treeIsProcessed(otherTree)) {
				// The other tree is within the local Cholesky distance and has
				// already been processed. Include it.
				localTrees.push_back(otherTree);
			}
		}
	}
}

void LocalCorrelation::buildCovarianceMatrix(double const squaredCompetitionRadius) {
	unsigned int const numberOfTrees = localTrees.size();
	covarianceMatrix.resize(numberOfTrees, numberOfTrees);
	for (unsigned int i = 0; i < numberOfTrees; ++i) {
		covarianceMatrix[i][i] = exp(coefficients[0]); // The diagonal
		for (unsigned int j = 0; j < i; ++j) {
			double const squaredTreeDistance = localTrees[i]->squaredDistanceTo(*localTrees[j]);
			// We still need to check that two trees are within the competition
			// radius from each other.
			if (squaredTreeDistance <= squaredCompetitionRadius) {
				double const treeDistance = sqrt(squaredTreeDistance);
				double const covariance = exp(coefficients[0] + treeDistance * coefficients[1]);
				covarianceMatrix[i][j] = covariance;
			} else {
				covarianceMatrix[i][j] = 0;
			}
		}
	}
}

inline bool LocalCorrelation::treeIsProcessed(const Tree * const tree) const {
	if (processedTrees.find(tree) != processedTrees.end()) {
		// The tree is within the local process' area and has already been
		// processed.
		return true;
	}
#ifdef USE_MPI
	double const x = tree->getX();
	double const y = tree->getY();
	if (x < activeMinCoord.x || x >= activeMaxCoord.x || y < activeMinCoord.y || y >= activeMaxCoord.y) {
		// The tree originates from another process; determine whether it has been
		// processed based on the current strip and the tree's position.
		switch (currentStrip) {
		case LEFT_STRIP:
			// The tree has been processed *only* if it is positioned directly above
			// the left strip of the current process' area.
			return (y >= activeMaxCoord.y) && (x >= activeMinCoord.x) && (x < splitX);
		case RIGHT_STRIP:
			// The tree has *not* been processed if it is positioned directly below
			// the right strip of the current process' area.
			return !((y < activeMinCoord.y) && (x >= splitX) && (x < activeMaxCoord.x));
		}
	}
#endif
	return false;
}

#endif
