#ifdef INCLUDE_ENVIRONMENTAL_EFFECTS
#ifndef LOCALCORRELATION_H_
#define LOCALCORRELATION_H_

#include <vector>
#include <set>
#include <map>
#include <cassert>
#include "Vector.h"
#include "Matrix.h"
#include "Point.h"

using std::vector;
using std::set;
using std::map;

class Tree;

/**
 * LocalCorrelation is used to calculated correlated random numbers for
 * trees. In order to use this class, trees must have been assigned
 * random numbers and been sorted with the
 * TreeSorter::sortForLocalCholesky() method.
 */
class LocalCorrelation {
public:
	LocalCorrelation();

	inline void setCompetitionRadius(double const radius) {
		competitionRadius = radius;
	}
	inline void setDistanceFactor(double const factor) {
		distanceFactor = factor;
	}
	inline void setCoefficients(Vector const & coefficients) {
		assert(coefficients.size() >= 2);
		this->coefficients = coefficients;
	}
	inline void setRandomNumberSetIndex(unsigned int set) {
		setIndex = set;
	}
	inline void setMarkIndex(unsigned int mark) {
		markIndex = mark;
	}

	/**
	 * Calculate a correlated random number for the target tree.
	 * @curTree The target tree
	 * @trees Trees that might be neighbours
	 */
	double calculateCorrelatedRandomNumber(
			Tree const * const curTree, vector<Tree*> const & trees);

	inline unsigned int getNumberOfCalculations() const {
		return numberOfCalculations;
	}

	inline unsigned int getMaximumMatrixSize() const {
		return maximumMatrixSize;
	}

	inline double getAverageMatrixSize() const {
		return averageMatrixSize;
	}

private:
	/**
	 * Finds processed neighbours to curTree which are within the local
	 * Cholesky distance. The result is put in the localTrees vector.
	 */
	void findProcessedNeighbours(Tree const * const curTree, vector<Tree*> const & trees);

	void buildCovarianceMatrix(double const squaredCompetitionRadius);

	bool treeIsProcessed(Tree const * const tree) const;

	double competitionRadius;
	double distanceFactor;
	Vector coefficients;
	unsigned int setIndex;
	unsigned int markIndex;
	set<Tree const *> processedTrees;
	vector<Tree const *> localTrees;
	Matrix covarianceMatrix;
	unsigned int numberOfCalculations;
	unsigned int maximumMatrixSize;
	double averageMatrixSize;
#ifdef USE_MPI
	Point activeMinCoord, activeMaxCoord;
	double splitX;
	enum { LEFT_STRIP, RIGHT_STRIP } currentStrip;
#endif
};

#endif /* LOCALCORRELATION_H_ */
#endif
