#include <stdio.h>
#include <stdarg.h>
#include <sys/time.h>
#include <stdlib.h>
#include <unistd.h>

using namespace std;
#include <iostream>
using std::cout;

#include "Log.h"
#include "Memusage.h"
#ifdef USE_MPI
#include "MPIUtil.h"
#endif

#define MAXLEVEL 10
#define MSG_SIZE 100000

#define FLUSH_AFTER_EACH_LINE 1

const static double toMB = 1.0/1024/1024;

static char* filename = NULL;
static string directory;
static FILE* logFile = NULL;
static char msg[MSG_SIZE];
static double maxmemusage = 0.0;
static double timer[MAXLEVEL];
static int currentLevel=0;
static bool firstCall=true;

unsigned int Log::msgCounts[NUMBER_OF_STATUSES];

void Log::log(int level, const char *fmt, ...) {
	va_list ap;
	va_start(ap, fmt);
	vsnprintf (msg, MSG_SIZE, fmt, ap);
	va_end(ap);
	logMsg(level, msg);
	msgCounts[INFORMATION]++;
}

void Log::logWarning(int level, const char *fmt, ...) {
	va_list ap;
	va_start(ap, fmt);
	vsnprintf (msg, MSG_SIZE, fmt, ap);
	va_end(ap);
	string temp = "{WARNING} ";
	temp += msg;
	logMsg(level, temp.c_str());
	msgCounts[WARNING]++;
}

void Log::logError(int level, const char *fmt, ...) {
	va_list ap;
	va_start(ap, fmt);
	vsnprintf (msg, MSG_SIZE, fmt, ap);
	va_end(ap);
	string temp = "{ERROR} ";
	temp += msg;
	logMsg(level, temp.c_str());
	msgCounts[ERROR]++;
}

void Log::logFailure(int level, const char *fmt, ...) {
	va_list ap;
	va_start(ap, fmt);
	vsnprintf (msg, MSG_SIZE, fmt, ap);
	va_end(ap);
	string temp = "{FAILURE} ";
	temp += msg;
	logMsg(level, temp.c_str());
	msgCounts[FAILURE]++;
}

void Log::finalize() {
	if (filename != NULL) {
		char totalElapsed[10];
		double now = Log::getTimestamp();

		getElapsed(totalElapsed, now-timer[0]);
		double memusage;
		Memusage::getMemusage(&memusage);
		fprintf(logFile, "[%4.1fMB %s]",memusage*toMB,totalElapsed);
		fprintf(logFile, " Maximum memory usage: %.2f\n", maxmemusage*toMB);

		fclose(logFile);

		free(filename);
	}
}

void Log::logMsg(int level, const char *str) {
	if (level >= MAXLEVEL) {
		cout << "Log::log* called with invalid level: "<<level<<" (maxlevel is set to "<<MAXLEVEL<<")\n";
		return;
	}
	if (level == -1)
		level = currentLevel;

	if (firstCall) {
		int i;

#ifdef USE_MPI
		if (MPIUtil::mpiId() == 0)
#endif
			cout << "Initializing log\n";

		for (i=0; i < MAXLEVEL; i++) {
			timer[i]=-1.0;
		}

		filename = (char*) malloc(20);
#ifdef USE_MPI
		int mpiid = MPIUtil::mpiId();
		if (mpiid < 0)
			mpiid = 0;
		sprintf(filename,"%04d-output.txt",mpiid);
#else
		sprintf(filename,"output.txt");
#endif
		string fileWithPath = filename;
		if (directory.length() > 0) {
			fileWithPath = directory+"/"+fileWithPath;
		}

		// truncate or create file
		logFile = fopen(fileWithPath.c_str(),"w");
		if (!logFile) {
			printf("Failed to create output file %s\n",fileWithPath.c_str());
		}

		for (i = 0; i < NUMBER_OF_STATUSES; i++) {
			msgCounts[i] = 0;
		}

		firstCall = false;
	}

	// Clear all timers with higher level than this
	for (int i=level+1; i < MAXLEVEL; i++) {
		timer[i]=-1.0;
	}
	currentLevel = level;

	double now = Log::getTimestamp();
	if (timer[level] < 0) {
		timer[level] = now;
	}

	char levelElapsed[10];
	char totalElapsed[10];

	getElapsed(levelElapsed, now-timer[level]);
	getElapsed(totalElapsed, now-timer[0]);

	double memusage;
	Memusage::getMemusage(&memusage);
	if (memusage > maxmemusage) {
		maxmemusage = memusage;
	}

	char logString[1000];

	sprintf(logString, "[%4.1fMB %s]",memusage*toMB,totalElapsed);
	if (level>0) {
		 sprintf(logString+strlen(logString), "%*c",level*2,' ');
		sprintf(logString+strlen(logString), "[%s]",levelElapsed);
	}
	snprintf(logString+strlen(logString), 900, " %s\n", str);
	fprintf(logFile,logString);

#ifdef LOG_TO_STDOUT
	printf(logString);
#endif

#ifdef FLUSH_AFTER_EACH_LINE
	fflush(logFile);
#endif
}

double Log::getTimestamp() {
	struct timeval tv;
	gettimeofday(&tv, NULL);
	double ret = tv.tv_sec+tv.tv_usec/1000000.0;
	return ret;
}

void Log::getElapsed(char* out, double timeInUnits) {
	char timeUnit = 's';
	if (timeInUnits > 60.0) {
		timeInUnits /=60.0;
		timeUnit = 'm';
	}
	if (timeInUnits > 60.0) {
		timeInUnits /=60.0;
		timeUnit = 'h';
	}
	sprintf(out,"%2.2f%c",timeInUnits,timeUnit);

}

void Log::setDirectory(string logDirectory) {
	directory = logDirectory;

	if (!firstCall) {
		finalize();

		firstCall = true;
	}
}

