#include "MarkTransformer.h"
#include "MarkTransformation.h"
#include "TreeContainer.h"
#include "Tree.h"
#include "Log.h"

#ifdef USE_MPI
#include "MPIUtil.h"
#endif

#include <algorithm>
#include <cmath>
using std::min;
using std::max;
#include <limits>
using std::numeric_limits;
#include <cassert>

MarkTransformer::MarkTransformer(unsigned int const markId) :
	transformation(NULL), markId(markId) {

}

MarkTransformer::~MarkTransformer() {

}

void MarkTransformer::calculateInitialCDF(TreeContainer const & trees) {
	const double tolerance = 0.0000001;
	// Find the minimum and maximum mark values
	double lower = numeric_limits<double>::max();
	double upper = 0.0;
	findMinimumAndMaximumMarks(trees, lower, upper);
#ifdef USE_MPI
	const MPI_Comm communicator = MPIUtil::getActiveCommunicator();
	MPIUtil::min(lower, communicator);
	MPIUtil::max(upper, communicator);
#endif
	if (upper - lower < tolerance) {
		// The range is too narrow. We have to artificially enlargen it so that
		// the cubic spline interpolation won't fail.
		lower -= 0.1;
		upper += 0.1;
	}

	// Number of classes hardcoded to 100. Increasing the value will result in
	// better approximation, but will also increase memory consumption.
	const int numberOfClasses = 100;
	initialCDFClasses.resize(numberOfClasses);
	initialCDFValues.resize(numberOfClasses);

	setupClasses(lower, upper, initialCDFClasses);
	// Offset the lower and upper class bounds to ensure that all trees are
	// included.
	initialCDFClasses.front() -= tolerance;
	initialCDFClasses.back() += tolerance;

	// Calculate the number of trees in each size class
	vector<unsigned int> frequencies(numberOfClasses - 1);
	classifyTrees(trees, initialCDFClasses, frequencies);
	unsigned int totalNumberOfTrees = trees.size();
#ifdef USE_MPI
	MPIUtil::sum(frequencies, communicator);
	MPIUtil::sum(totalNumberOfTrees, communicator);
#endif
	initialCDFValues.front() = 0.0;

	double artificialInc = 0.0;
	int artificialIncIdx = -1;
	for (int i = 0; i < numberOfClasses - 1; i++) {
		if (i <= artificialIncIdx) {
			// Increment artificially
			initialCDFValues[i + 1] = initialCDFValues[i] + artificialInc;
		} else if (frequencies[i] > 0) {
			// Increment as normal
			initialCDFValues[i + 1] = initialCDFValues[i] + static_cast<double>(frequencies[i]) / totalNumberOfTrees;
		} else {
			// The CDF must be monotonically increasing.
			// Find the next class with a frequency greater than 0 and use that value
			// produce artificial points.
			artificialIncIdx = numberOfClasses - 1;
			artificialInc = tolerance;
			for (int j = i + 1; j < numberOfClasses - 1; j++) {
				if (frequencies[j] > 0) {
					artificialIncIdx = j;
					double const temp = static_cast<double>(frequencies[j]) / totalNumberOfTrees;	
					artificialInc = temp / (j - i + 1);
					break;
				}
			}
			initialCDFValues[i + 1] = initialCDFValues[i] + artificialInc;
		}
	}

	initialCDFValues.back() = 1.0;
}

void MarkTransformer::transform(TreeContainer & trees) {
	assert(transformation);
	assert(initialCDFClasses.size() > 0 && initialCDFValues.size() > 0);
	if (trees.size() == 0) {
		return;
	}

	vector<double> const & mappedCDFClasses = transformation->getCDFClasses();
	vector<double> const & mappedCDFValues = transformation->getCDFValues();
	vector<double> initialSpline(initialCDFValues.size());
	vector<double> mappedSpline(mappedCDFValues.size());

	if (mappedCDFClasses.size() != initialCDFClasses.size()) {
		Log::logWarning(-1, "Mapped CDF does not have the same number of classes as the simulated CDF. \
Simulated CDF: %u classes, Mapped CDF: %u classes", initialCDFClasses.size(), mappedCDFClasses.size());
	}

	// Set up the splines
	splineDer(initialCDFClasses, initialCDFValues, initialSpline);
	splineDer(mappedCDFValues, mappedCDFClasses, mappedSpline);

	const unsigned int numberOfTrees = trees.size();
	vector<double> marks(numberOfTrees);
	for (unsigned int i = 0; i < numberOfTrees; ++i) {
		marks[i] = trees[i]->getMark(markId);
	}

	vector<double> uniform(numberOfTrees);

	// Interpolate from marks to uniform using the initial CDF
	splineInt(initialCDFClasses, initialCDFValues, initialSpline, numberOfTrees, marks, uniform);
	// Interpolate from uniform to marks using the mapped CDF
	splineInt(mappedCDFValues, mappedCDFClasses, mappedSpline, numberOfTrees, uniform, marks);

	// Update tree marks
	for (unsigned int i = 0; i < numberOfTrees; ++i) {
		trees[i]->setMark(markId, marks[i]);
	}
}

void MarkTransformer::findMinimumAndMaximumMarks(
		TreeContainer const & trees, double & lower, double & upper) const {
	const size_t size = trees.size();
	for (unsigned int curTreeIndex = 0; curTreeIndex < size; curTreeIndex++) {
		const double mark = trees[curTreeIndex]->getMark(markId);
		lower = min(lower, mark);
		upper = max(upper, mark);
	}
}

void MarkTransformer::setupClasses(double lower, double upper, vector<double> & classBounds) const {
	const unsigned int numberOfClasses = classBounds.size();
	const double increase = (upper - lower) / (numberOfClasses - 1);

	classBounds.front() = lower;
	for (unsigned int i = 1; i < numberOfClasses; ++i) {
		classBounds[i] = classBounds[i-1] + increase;
	}
	classBounds.back() = upper;
}

void MarkTransformer::classifyTrees(
		TreeContainer const & trees, vector<double> const & classBounds, vector<unsigned int> & frequencies) const {
	const unsigned int numberOfTrees = trees.size();
	const unsigned int numberOfClasses = classBounds.size();
	for (unsigned int curTreeIndex = 0; curTreeIndex < numberOfTrees; ++curTreeIndex) {
		const double mark = trees[curTreeIndex]->getMark(markId);
		size_t k0 = 0;
		size_t ki = numberOfClasses - 1;
		while (ki - k0 > 1) {
			size_t k = (ki + k0) / 2;
			if (classBounds[k] > mark) {
				ki = k;
			} else {
				k0 = k;
			}
		}
		frequencies[k0]++;
	}
}

void MarkTransformer::splineDer(
		vector<double> const & xArr, vector<double> const & yArr, vector<double> & y2Arr) const {
	int nc = xArr.size();
	vector<double> temp(nc);

	for (int i = 1; i < nc-1; ++i) {
		double sig = (xArr[i] - xArr[i - 1]) / (xArr[i + 1] - xArr[i - 1]);
		double p = sig * y2Arr[i - 1] + 2.0;
		y2Arr[i] = (sig - 1.0) / p;
		temp[i] = (6.0 * ((yArr[i + 1] - yArr[i]) / (xArr[i + 1] - xArr[i]) - (yArr[i] - yArr[i - 1]) / (xArr[i]
				- xArr[i - 1])) / (xArr[i + 1] - xArr[i - 1]) - sig * temp[i - 1]) / p;
	}
	// Natural upper boundary.
	y2Arr[nc - 1] = 0.0;
	for (int i = nc-2; i >= 0; --i) {
		y2Arr[i] = y2Arr[i] * y2Arr[i + 1] + temp[i];
	}
}

void MarkTransformer::splineInt(
		vector<double> const & xArr, vector<double> const & yArr, vector<double> const & y2Arr,
		unsigned int nt, vector<double> const & inArr, vector<double> & outArr) const {
	const int numberOfClasses = xArr.size();

	for (unsigned int i = 0; i < nt; ++i) {
		// for each tree
		const double zz = inArr[i];
		int k0 = 0;
		int ki = numberOfClasses - 1;
		// find the class for the tree
		while (ki - k0 > 1) {
			int k = (ki + k0) / 2;
			if (xArr[k] > zz) {
				ki = k;
			} else {
				k0 = k;
			}
		}
		double d = xArr[ki] - xArr[k0];
		if (d <= 0.0) {
			Log::logError(-1, "Increment in cumulative frequency is %.2f", d);
			Log::logError(-1, "zz: %.2f, k0: %d, ki: %d, xArr[ki]: %.2f, xArr[k0]: %.2f, d: %d", zz, k0, ki, xArr[ki],
					xArr[k0], d);
			throw "Increment in cumulative frequency is 0.0";
		}
		double upper = (xArr[ki] - zz) / d;
		double lower = (zz - xArr[k0]) / d;
		outArr[i] = upper * yArr[k0] + lower * yArr[ki] + (((upper*upper*upper) - upper) * y2Arr[k0] +
				((lower*lower*lower) - lower) * y2Arr[ki]) * (d*d) / 6.0;
	}
}
