#ifndef MARKTRANSFORMER_H_
#define MARKTRANSFORMER_H_

#include <vector>
using std::vector;

class MarkTransformation;
class TreeContainer;

/**
 * MarkTransformer transforms tree marks to adhere to a target distribution
 * provided by a MarkTransformation object. The method calculateInitialCDF
 * should be called first to build a cumulative distribution of the initial
 * tree marks. Then transform can be called to perform the actual mark
 * transformation.
 */
class MarkTransformer {
public:
	explicit MarkTransformer(unsigned int const markId);
	~MarkTransformer();

	inline void setTargetTransformation(MarkTransformation const & targetTransformation) {
		transformation = &targetTransformation;
	}

	/**
	 * Calculate an initial Cumulative Distribution Function based on the trees
	 * in the TreeContainer. The CDF is then used in the transform() method to
	 * transform the tree marks.
	 */
	void calculateInitialCDF(TreeContainer const & trees);

	/**
	 * Transform tree marks according to the mapped CDF given by the MarkTransformation
	 * object and the initial CDF of the tree marks. The method calculateInitialCDF()
	 * *must* have been called before this method can be used.
	 */
	void transform(TreeContainer & trees);
private:
	void findMinimumAndMaximumMarks(TreeContainer const & trees, double & lower, double & upper) const;
	void setupClasses(double lower, double upper, vector<double> & classBounds) const;
	void classifyTrees(
			TreeContainer const & trees, vector<double> const & classBounds,
			vector<unsigned int> & frequencies) const;

	/**
	 * Setup y2Arr to be used with cubic spline interpolation
	 */
	void splineDer(
			vector<double> const & xArr, vector<double> const & yArr, vector<double> & y2Arr) const;

	/**
	 * Interpolate with cubic splines
	 */
	void splineInt(
			vector<double> const & xArr, vector<double> const & yArr, vector<double> const & y2Arr,
			unsigned int nt, vector<double> const & inArr, vector<double> & outArr) const;

	MarkTransformation const * transformation;
	unsigned int const markId;
	vector<double> initialCDFClasses;
	vector<double> initialCDFValues;
};

#endif /* MARKTRANSFORMER_H_ */
