#include "Matrix.h"
#include <cmath>
#include <cassert>
#include <iostream>
#include <iomanip>
#include "Log.h"

using std::ios;
using std::setw;

Matrix::Matrix()
: array(NULL)
, numberOfColumns(0)
, numberOfRows(0)
, capacity(0)
{
}

Matrix::Matrix(size_t nrOfRows, size_t nrOfCols, double defaultValue)
: array(new double[nrOfRows*nrOfCols])
, numberOfColumns(nrOfCols)
, numberOfRows(nrOfRows)
, capacity(nrOfCols * nrOfRows)
{
//	Log::log(-1,"Created matrix with dimensions %d x %d",nrOfCols,nrOfRows);
	for(size_t i=0; i < numberOfRows*numberOfColumns; ++i) {
		array[i] = defaultValue;
	}
}

Matrix::Matrix(Matrix const & other)
: array(new double[other.numberOfColumns*other.numberOfRows])
, numberOfColumns(other.numberOfColumns)
, numberOfRows(other.numberOfRows)
, capacity(other.numberOfColumns * other.numberOfRows)
{
	copyMatrix(other,*this);
}
Matrix::~Matrix()
{
	delete[] array;
}

void Matrix::resize(size_t nrOfRows, size_t nrOfColumns) {
	numberOfRows = nrOfRows;
	numberOfColumns = nrOfColumns;
	size_t newCapacity = numberOfRows * numberOfColumns;
	if (newCapacity > capacity) {
		delete [] array;
		array = new double[numberOfRows*numberOfColumns];
	}
	capacity = newCapacity;
}

/**
 * Vout = M' * Vin
 */
Vector Matrix::transposedMultiplyWith(Vector const & v) const {
	size_t const vsize = v.size();
	if(getNumberOfRows() != vsize) {
		throw "Error multiplying matrix and vector (dimensions do not match).";
	}

	size_t const numberOfRowsInOpMatrixA = getNumberOfColumns();
	size_t const numberOfColsInOpMatrixA = getNumberOfRows();
	double const scalingFactorAlpha=1.0;
	double const scalingTermBeta=0.0;
	size_t const numberOfColsInMatrixA = getNumberOfColumns();
	size_t const incrementX = 1;
	size_t const incrementY = 1;

	double * matrixA = getDataPointer();
	double * vectorX = v.getDataPointer();

	Vector result(getNumberOfColumns());
	double * vectorY = result.getDataPointer();

	dgemv('N',numberOfRowsInOpMatrixA,numberOfColsInOpMatrixA,scalingFactorAlpha,matrixA,numberOfColsInMatrixA,vectorX,incrementX,scalingTermBeta,vectorY,incrementY);
	return result;
}
/**
 * Vout = M * Vin
 */
Vector Matrix::multiplyWith(Vector const & v) const {
	size_t const vsize = v.size();
	if(getNumberOfColumns() != vsize) {
		throw "Error multiplying matrix and vector (dimensions do not match).";
	}

	size_t const numberOfRowsInOpMatrixA = getNumberOfColumns();
	size_t const numberOfColsInOpMatrixA = getNumberOfRows();
	double const scalingFactorAlpha=1.0;
	double const scalingTermBeta=0.0;
	size_t const numberOfColsInMatrixA = getNumberOfColumns();
	size_t const incrementX = 1;
	size_t const incrementY = 1;

	double * matrixA = getDataPointer();
	double * vectorX = v.getDataPointer();

	Vector result(getNumberOfRows());
	double * vectorY = result.getDataPointer();

	dgemv('T',numberOfRowsInOpMatrixA,numberOfColsInOpMatrixA,scalingFactorAlpha,matrixA,numberOfColsInMatrixA,vectorX,incrementX,scalingTermBeta,vectorY,incrementY);
	return result;
}

void Matrix::choleskyDecompose(OrientationType const type) {
	int size = getNumberOfRows();
	int info = 0;
	// ACML has fortran memory order so 'L' in practice means 'U'...
	if(type==UPPER_TRIANGULAR) {
		// We have an upper triangular input matrix.
		dpotrf('L',size,getDataPointer(),size,&info);
	} else if (type==LOWER_TRIANGULAR) {
		// We have a lower triangular input matrix.
		dpotrf('U',size,getDataPointer(),size,&info);
	}
	if(info!=0) {
		Log::logError(-1,"Info variable in choleskyDecompose is %d, should be zero!",info);
		if(info<0)
			Log::logError(-1,"the %d-th argument had an illegal value!",-info);
		if(info>0)
			Log::logError(-1,"The leading minor of order %d is not positive definite, and the factorization could not be completed.",info);
		throw "Info variable in choleskyDecompose indicates dpotrf failed, see log!";
	}
}

ostream & operator<<(ostream & os, Matrix const & m) {
	os.setf(ios::fixed);
	os.precision(4);
	size_t const mrows = m.getNumberOfRows();
	size_t const mcols = m.getNumberOfColumns();
	for(size_t i = 0; i < mrows; ++i) {
		for(size_t j = 0; j < mcols; ++j) {
			os << setw(12) << m[i][j] << "\t";
		}
		os << "\n";
	}
	return os;
}
