#include "Model.h"
#include "TreeContainer.h"
#include "Compartment.h"
#include "Tree.h"
#include "Vector.h"
#include "Helper.h"
#include "EvaluatorFunctions.h"
#include "ModelFactor.h"
#include "ModelTerm.h"

#include <stdint.h>
#include <cmath>
#include <sstream>
using std::stringstream;

vector<string> Model::modelVariableNames = setupModelVariableNames();

string Model::getModelVariableName(unsigned int id) {
	return modelVariableNames[id];
}

unsigned int Model::getModelVariableId(string const & name) {
	return Helper::getVariableId(name, modelVariableNames);
}

double Model::calculateMark(Vector & modelVariables, Vector & secondaryVariables,
		vector<ModelTerm> const & modelTerms) {
	double result = 0.0;
	for (vector<ModelTerm>::const_iterator termIter = modelTerms.begin(); termIter != modelTerms.end(); ++termIter) {
		result += termIter->evaluateTerm(modelVariables, secondaryVariables);
	}
	return result;
}

void Model::calculateModelVariables(Tree const * const targetTree, vector<Tree*> const & trees,
		float const competitionRadius, Vector & modelVariables) {

	modelVariables[INTERCEPT] = 1.0;

	modelVariables[DIAMETER] = targetTree->getMark(TreeMarks::DIAMETER);

	modelVariables[DIAMETER_LOG] = log(targetTree->getMark(TreeMarks::DIAMETER));

	modelVariables[HEIGHT_DIAMETER_RATIO] = targetTree->getMark(TreeMarks::HEIGHT) / targetTree->getMark(TreeMarks::DIAMETER);

	Compartment const * const compartment = targetTree->getParentCompartment();
	if (compartment != NULL) {
		double const totalVolume =  compartment->getTotalVolume();
		Polygon const * const polygon = compartment->getPolygon();
		double const totalAreaInHectares = polygon != NULL? polygon->getFullArea() * 0.0001: 0.0;
		modelVariables[COMPARTMENT_VOLUME] = EvaluatorFunctions::safeDivision(totalVolume, totalAreaInHectares);
	} else {
		modelVariables[COMPARTMENT_VOLUME] = 0.0;
	}

	unsigned int numberOfCompetitors = 0;
	double const squaredCompetitionRadius = competitionRadius * competitionRadius;
	// Find all competitors and update this tree's competition indexes.
	// Since we only do things in the loop if the competition radius is bigger than 0,
	// we can skip the loop completely in some cases!
	if (competitionRadius > 0.0f) {
		unsigned int const size = trees.size();
		for (unsigned int treeIndex = 0; treeIndex < size; treeIndex++) {
			Tree const * const tree = trees[treeIndex];
			// cout << "Possible competitor: " << *nearbyTree << "\n";
			double const squaredTreeDistance = targetTree->squaredDistanceTo(*tree);
			if (0.0 < squaredTreeDistance && squaredTreeDistance <= squaredCompetitionRadius) {
				// We have a competitor, update competition indexes.
				double const treeDistance = sqrt(squaredTreeDistance);
				++numberOfCompetitors;

				modelVariables[AVG_DIST] += treeDistance;

				if (targetTree->getMark(TreeMarks::DIAMETER) < tree->getMark(TreeMarks::DIAMETER)) {
					modelVariables[INV_DIST] += 1.0 / treeDistance; // Sum of inverse distance to all bigger competitors.
				}

				modelVariables[DIAMETER_SUM] += tree->getMark(TreeMarks::DIAMETER); // Sum of competitors' diameters.

				// Sum of difference between all competitors' marks.
				modelVariables[DIAMETER_DIFF] += targetTree->getMark(TreeMarks::DIAMETER) - tree->getMark(
						TreeMarks::DIAMETER);

				modelVariables[HEIGHT_DIFF] += targetTree->getMark(TreeMarks::HEIGHT) - tree->getMark(
						TreeMarks::HEIGHT);

				double treeRadiusInMetres = (tree->getMark(TreeMarks::DIAMETER)/200.0);
				modelVariables[COMPETITOR_BASAL_AREA] += M_PI*treeRadiusInMetres*treeRadiusInMetres;

				if (targetTree->getMark(TreeMarks::DIAMETER) < tree->getMark(TreeMarks::DIAMETER)) {
					// Sum of quotient between all bigger competitors' diameters and the target tree's diameter.
					modelVariables[DIAMETER_RATIO] += EvaluatorFunctions::safeDivision(tree->getMark(
							TreeMarks::DIAMETER), targetTree->getMark(TreeMarks::DIAMETER));
				}
			}
		}
	}
	if (numberOfCompetitors > 0) {
		double const competitionArea = M_PI * squaredCompetitionRadius;
		double targetTreeRadiusInMetres = targetTree->getMark(TreeMarks::DIAMETER)/200.0;
		double targetTreeBasalArea = M_PI*targetTreeRadiusInMetres*targetTreeRadiusInMetres;

		modelVariables[DENSITY] = EvaluatorFunctions::safeDivision(numberOfCompetitors, competitionArea) * 10000.0; // Competitor density.
		modelVariables[AVG_DIST] /= numberOfCompetitors; // Average distance to competitors.
		modelVariables[RELATIVE_BASAL_AREA] = EvaluatorFunctions::safeDivision(targetTreeBasalArea, modelVariables[COMPETITOR_BASAL_AREA]);
		modelVariables[BA_DIAMETER_RATIO] = EvaluatorFunctions::safeDivision(modelVariables[COMPETITOR_BASAL_AREA], modelVariables[DIAMETER_SUM] / numberOfCompetitors);
		// Convert competitor basal area to square metres per hectare
		modelVariables[COMPETITOR_BASAL_AREA] = EvaluatorFunctions::safeDivision(modelVariables[COMPETITOR_BASAL_AREA], competitionArea) * 10000.0;
	}
}

vector<string> Model::setupModelVariableNames() {
	vector<string> names;
	names.resize(NUMBER_OF_MODEL_VARIABLES);
	names[INTERCEPT] = "Intercept";
	names[DIAMETER] = "Diameter";
	names[DIAMETER_LOG] = "Diameter log";
	names[DENSITY] = "Local density";
	names[AVG_DIST] = "Average distance";
	names[INV_DIST] = "Inverse distance";
	names[DIAMETER_SUM] = "Diameter sum";
	names[DIAMETER_DIFF] = "Diameter difference";
	names[HEIGHT_DIFF] = "Height difference";
	names[HEIGHT_DIAMETER_RATIO] = "Height diameter ratio";
	names[COMPETITOR_BASAL_AREA] = "Basal area of competitors";
	names[RELATIVE_BASAL_AREA] = "Relative basal area";
	names[DIAMETER_RATIO] = "Diameter quotient";
	names[BA_DIAMETER_RATIO] = "Basal area to mean diameter ratio";
	names[COMPARTMENT_VOLUME] = "Compartment volume";
	names[TAU] = "Tau";
	names[XI] = "Xi";
	names[EPSILON] = "Epsilon";
	return names;
}
