#include "SecondaryModel.h"
#include "Tree.h"
#include "Vector.h"
#include "Helper.h"
#include "SpeciesContainer.h"
#include "Species.h"
#include "TemplateContainer.h"
#include "ManagementTemplate.h"
#include "SizeClassHandler.h"
#include "AgeClassHandler.h"
#include "HeightClassHandler.h"
#include "CompartmentData.h"
#include "SecondaryModelTerm.h"

vector<string> SecondaryModel::secondaryVariableNames;

void SecondaryModel::initialize(vector<string> const & indicatorVariables) {
	secondaryVariableNames = setupSecondaryVariableNames();
	for (size_t i = 0, n = indicatorVariables.size(); i < n; i++) {
		secondaryVariableNames.push_back(string("Indicator variable: ") + indicatorVariables[i]);
	}
}

string SecondaryModel::getSecondaryVariableName(unsigned int id) {
	assert(secondaryVariableNames.size() > 0);
	return secondaryVariableNames[id];
}

size_t SecondaryModel::numberOfVariables() {
	return secondaryVariableNames.size();
}

unsigned int SecondaryModel::getSecondaryVariableId(string const & name) {
	assert(secondaryVariableNames.size() > 0);
	return Helper::getVariableId(name, secondaryVariableNames);
}

double SecondaryModel::calculateMark(Vector & secondaryVariables, vector<SecondaryModelTerm> const & terms) {
	double result = 0.0;
	for (vector<SecondaryModelTerm>::const_iterator termIter = terms.begin(); termIter != terms.end(); ++termIter) {
		result += termIter->evaluateTerm(secondaryVariables);
	}
	return result;
}

void SecondaryModel::calculateSecondaryVariables(Tree const * const targetTree,
		CompartmentData const * const compartmentData,
		SizeClassHandler const & sizeClassHandler,
		AgeClassHandler const & ageClassHandler,
		HeightClassHandler const & heightClassHandler,
		Vector & variables) {
	variables.clear();

	variables[INTERCEPT] = 1.0;

	vector<unsigned int> const & templateIds = compartmentData->getManagementTemplateIds();
	if (templateIds.size() > 0) {
		variables[MANAGEMENT] = TemplateContainer::get(templateIds[0])->getMethodSecondaryModelValue();
	}

	unsigned int const speciesId = targetTree->getSpeciesId();
	variables[SPECIES] = SpeciesContainer::get(speciesId)->get(Species::SECONDARY_MODEL_VALUE);

	variables[AGE] = (double) targetTree->getAge();

	variables[AGE_CLASS] = ageClassHandler.determineAgeClass(*targetTree);

	variables[SIZE_CLASS] = sizeClassHandler.determineSizeClass(*targetTree);

	variables[HEIGHT_CLASS] = heightClassHandler.determineHeightClass(*targetTree);

	for (size_t i = 0, n = SpeciesContainer::size(); i < n; i++) {
		variables[INDICATOR_VARIABLE + i] = (speciesId == i? 1.0: 0.0);
	}
}

vector<string> SecondaryModel::setupSecondaryVariableNames() {
	vector<string> names;
	names.resize(INDICATOR_VARIABLE);
	names[INTERCEPT] = "Intercept";
	names[MANAGEMENT] = "Management";
	names[SPECIES] = "Species";
	names[AGE] = "Age";
	names[AGE_CLASS] = "Age class";
	names[SIZE_CLASS] = "Size class";
	names[HEIGHT_CLASS] = "Height class";
	names[EPSILON] = "Epsilon";
	return names;
}
