#ifdef USE_MPI
#include <mpi.h>
#include "MPIUtil.h"
#include "ForestSplitter.h"
#endif
#include "Simulation.h"

#include "SpeciesContainer.h"
#include "CompartmentData.h"
#include "TemplateContainer.h"
#include "ForestTypeContainer.h"
#include "CompartmentStatistics.h"
#include "StandStatisticsSummary.h"
#include "AssortmentContainer.h"
#include "Forest.h"
#include "StandStructureEvaluator.h"
#include "parameters.h"
#include "PaddedField.h"
#include "TreeContainer.h"
#include "File.h"
#include "Timer.h"
#include "UniformRandomNumberGenerator.h"
#include "PoissonRandomNumberGenerator.h"
#include "RandomNumber.h"
#include "Log.h"
#include "CompartmentVerifier.h"
#include "Polygon.h"
#include "LimitFinder.h"
#include "wildcards.h"
#include "SecondaryModel.h"

#include "XercesErrorHandler.h"
#include "XercesString.h"
#include "XercesUtilities.h"
#include <xercesc/dom/DOMException.hpp>
#include <xercesc/dom/DOMElement.hpp>
#include <xercesc/dom/DOMNodeList.hpp>
#include <xercesc/parsers/XercesDOMParser.hpp>
#include <xercesc/util/PlatformUtils.hpp>
#include <xercesc/util/XMLException.hpp>
XERCES_CPP_NAMESPACE_USE

#ifdef CALCULATION_INSPECTION
#include "CalculationInspector.h"
#endif


#include <ios>
using std::ios_base;
#include <iostream>
using std::cout;
using std::cerr;
#include <sstream>
using std::stringstream;
#include <map>
using std::map;

#include <sys/stat.h>
#include <sys/types.h>

Simulation::Simulation()
{
#ifdef USE_MPI
	MPI_Init(0, NULL);
	MPIUtil::findMpiId();
	MPIUtil::findMpiSize(MPI_COMM_WORLD);
#endif
	// Initialize the Limitfinder;
	LimitFinder::Instance(simulationData);
}

Simulation::~Simulation() {
	PoissonRandomNumberGenerator::Destroy();
	TemplateContainer::Destroy();
	SpeciesContainer::Destroy();
	ForestTypeContainer::Destroy();
	AssortmentContainer::Destroy();
#ifdef USE_MPI
	ForestSplitter::Destroy();
	MPIUtil::cleanup();
	Log::log(0, "Number of sums done: %d",MPIUtil::getNumberOfSumsDone());
#endif
#ifdef CALCULATION_INSPECTION
	CalculationInspector::Destroy();
#endif
	LimitFinder::Destroy();
	XMLPlatformUtils::Terminate();
}

void Simulation::run() {
	Log::log(0, "Simulation started.");

#ifdef USE_MPI
	if (MPIUtil::mpiId() == 0)
#endif
		cout << "Simulation started." << endl;
	CompartmentVerifier::verify(dataContainer);

	printModels();

	int lastReplication = simulationData.getNumberOfReplications()-1;
	Log::log(0, "Replications total: %d", simulationData.getNumberOfReplications());

#ifdef USE_MPI
	Log::log(0,"Calculating and setting active area for this process");
	ForestSplitter & forestSplitter = ForestSplitter::Instance(dataContainer);
	forestSplitter.setupProcessCoordinates();
	forestSplitter.split();
	// printNeighbours();
#endif

	StandStructureEvaluator standStructureEvaluator(simulationData, dataContainer);

	// Initialize random number generator.
	Log::log(0,"Seed value for the random number generator: %.2f", UniformRandomNumberGenerator::getSeed());
	UniformRandomNumberGenerator::setSeed(UniformRandomNumberGenerator::getSeed());

	for (int replication = 0; replication <= lastReplication; replication++) {
		Log::log(1, "Starting replication %d", replication);
#ifdef COMPARISONSEEDING
		UniformRandomNumberGenerator::setSeed(214.3*(replication+1));
#endif
		// Create a new forest.
		Forest forest(dataContainer, simulationData, replication);
		forest.simulate();
		standStructureEvaluator.evaluateCompartments(forest.getTreeContainer(),replication);
		Log::log(1, "Replication %d done", replication);
	}
	Log::log(0, "All replications complete");

	createStatisticalOutput(standStructureEvaluator);
	Log::log(0, "Statistics for all compartments written");

#ifdef USE_MPI
	if (MPIUtil::mpiId() == 0)
#endif
		cout << "Simulation complete\n";
	Log::log(0, "Simulation complete");
}

#ifdef USE_MPI
void Simulation::printNeighbours() {
	ForestSplitter & forestSplitter = ForestSplitter::Instance();

	stringstream text;
	for(size_t c=0; c<dataContainer.size(); c++) {
		text << "Compartment " << c << " is active in processes: ";
		vector<int> const compartmentProcesses = forestSplitter.getCompartmentProcesses(c);
		for(size_t p=0; p<compartmentProcesses.size(); p++) {
			text << compartmentProcesses[p] << " ";
		}
		Log::log(1,"%s", text.str().c_str() );
		text.str(string());
	}
}
#endif
void Simulation::printModels() {
	stringstream textstream;
	for(unsigned int index=0; index<TreeMarks::NUMBER_OF_MARKS; index++) {
		Log::log(2,"Mark: %s",TreeMarks::getMarkName(index).c_str());
		textstream << "Growth: ";
		simulationData.getGrowthData(index).writeModel(textstream);
		Log::log(3,"%s",textstream.str().c_str());
		textstream.str(string());
		textstream << "Mortality: ";
		simulationData.getMortalityData(index).writeModel(textstream);
		Log::log(3,"%s",textstream.str().c_str());
		textstream.str(string());
		textstream << "Reproduction: ";
		simulationData.getReproductionData(index).writeModel(textstream);
		Log::log(3,"%s",textstream.str().c_str());
		textstream.str(string());
	}
}

void Simulation::loadInputParameters(string fileName) {
#ifdef USE_MPI
	if(MPIUtil::mpiId()==0) {
#endif
		cout << "Reading parameters from file: " << fileName << endl;
		File::addSearchPath(fileName);

		try {
			XMLPlatformUtils::Initialize();
		}
		catch (XMLException const & xmlException) {
			throw "Error initializing XML parser.";
		}

		XercesDOMParser parser;
		// Use own error handler.
		XercesErrorHandler errorHandler;
		parser.setErrorHandler(&errorHandler);
		// For schema validation. Doesn't seem to work...
		//	parser.setDoNamespaces(true);
		//	parser.setDoSchema(true);

		try {
			parser.parse(fileName.c_str());
			DOMElement * simulation = parser.getDocument()->getDocumentElement();

			if(simulation) {
				string outputDirectory = loadOutputDirectory(parser.getDocument());
				outputDirectory = expandTimestampWildcard(outputDirectory);
				if (outputDirectory.length() > 0) {
					simulationData.setOutputDirectory(outputDirectory);
				}
#ifdef USE_MPI
				// Explicitly broadcast the output directory so that other
				// processes are able to use it for log files.
				MPIUtil::broadcast(outputDirectory, 0);
#endif
				// Find indicator variables
				vector<string> indicatorVariables = loadIndicatorVariables(simulation);
#ifdef USE_MPI
				MPIUtil::broadcast(indicatorVariables, 0);
#endif
				SecondaryModel::initialize(indicatorVariables);

				Log::log(0, "SPATE-HPC - v. %s", VERSION);
				Log::log(0, "Loading input parameters from file '%s'", fileName.c_str());
				loadData(simulation);
			}
		} catch (DOMException const & domException) {
			XercesString msg = XMLString::replicate(domException.getMessage());
			throw msg.toString();
		} catch (XMLException const & xmlException) {
			XercesString msg = XMLString::replicate(xmlException.getMessage());
			cerr << "Error in XML file at line ";
			cerr << xmlException.getSrcFile() << " column " << xmlException.getSrcLine() << "\n";
			throw msg.toString();
		}
#ifdef USE_MPI
	} else {
		// All non-root processes receive the output directory
		string outputDirectory;
		MPIUtil::broadcast(outputDirectory, 0);
		if (outputDirectory.length() > 0) {
			simulationData.setOutputDirectory(outputDirectory);
		}
		// All non-root processes receive indicator variables
		vector<string> indicatorVariables;
		MPIUtil::broadcast(indicatorVariables, 0);
		SecondaryModel::initialize(indicatorVariables);
	}
	broadcastInputParameters();
#endif
	dataContainer.setupIds();
}

void Simulation::createStatisticalOutput(StandStructureEvaluator & standStructureEvaluator) {
#ifdef USE_MPI
	ForestSplitter & forestSplitter = ForestSplitter::Instance();
	Log::log(0,"Combining statistical data");
	for(CompartmentDataContainer::const_iterator compDataIter=dataContainer.begin(); compDataIter!=dataContainer.end(); ++compDataIter) {
		CompartmentData * const compartmentData = *compDataIter;
		CompartmentStatistics & compartmentStatistics = compartmentData->getCompartmentStatistics();
		compartmentStatistics.combineStatistics();
	}
	Log::log(0,"Combine done");
#endif
	for(CompartmentDataContainer::const_iterator compDataIter=dataContainer.begin(); compDataIter!=dataContainer.end(); ++compDataIter) {
		CompartmentData const * const compartmentData = *compDataIter;
		if(compartmentData->getCompartmentPolygon()->isActive()) {
			unsigned int const compartmentId = compartmentData->getIdNumber();
#ifdef USE_MPI
			if(forestSplitter.isFirstProcessInCompartment(compartmentId)) {
#endif
				PaddedField compartmentNumber(lengthOfNumberString, '0');
				compartmentNumber.setValue(compartmentId);
				File out(simulationData.getOutputDirectory(true) +
						simulationData.getSimulationName() + statisticsFileName +
						compartmentNumber.str() + fileNameSuffix, ios_base::out);
				printSummary(compartmentId, out);
				CompartmentStatistics const & compartmentStatistics = compartmentData->getCompartmentStatistics();
				compartmentStatistics.printSummary(out);
				standStructureEvaluator.printSummary(compartmentId, out);
				Log::log(1,"Statistics for compartment id %d written",compartmentId);
#ifdef USE_MPI
			}
#endif
		}
	}

#ifdef USE_MPI
	if (MPIUtil::mpiId() == 0) {
#endif
		StandStatisticsSummary summary(simulationData, dataContainer, standStructureEvaluator);
		File out(simulationData.getOutputDirectory(true) +
				simulationData.getSimulationName() + statisticsTableFileName + fileNameSuffix,
				ios_base::out);
		summary.print(out);
		Log::log(1, "Statistics summary table written");
#ifdef USE_MPI
	}
#endif
}

void Simulation::printSummary(unsigned int compartment, ostream & out) const {
	out << "\n      SIMULATION SUMMARY\n";
	out << "---------------------------------\n";
	out << "Simulation name:         " << simulationData.getSimulationName() << "\n";
	out << "Number of replications:  " << simulationData.getNumberOfReplications() << "\n";
	out << "Compartment Id:          " << compartment << "\n";
	out << "Number of iterations:    " << simulationData.getNumberOfIterations() << "\n";
	out << "Number of years:         " << simulationData.getNumberOfIterations() * simulationData.getIterationLength() << "\n\n";
}

string Simulation::loadOutputDirectory(DOMDocument const * document) {
	XercesString tagName = XMLString::transcode("outputDirectory");
	DOMNodeList * list = document->getElementsByTagName(tagName.xmlCh());
	if (list->getLength() == 0) {
		return "";
	} else if (list->getLength() != 1) {
		throw "Input file contains multiple outputDirectory elements";
	} else {
	    DOMNode *node = list->item(0);
	    if (node->getNodeType() != DOMNode::ELEMENT_NODE) {
	    	throw "Input file contains an invalid outputDirectory element";
	    }
	    string s = XercesUtilities::getFirstChildTextContentAsString((DOMElement*)node);
	    return s;
	}
}

vector<string> Simulation::loadIndicatorVariables(XERCES_CPP_NAMESPACE::DOMElement const * simulation) {
	vector<string> result;
	XercesString tagName;
	DOMDocument * document = simulation->getOwnerDocument();
	tagName = XMLString::transcode("speciesParameters");
	DOMNodeList * list = document->getElementsByTagName(tagName.xmlCh());
	if (list->getLength() == 0) {
		return result;
	}
	if (list->item(0)->getNodeType() != DOMNode::ELEMENT_NODE) {
		throw "Input file contains an invalid speciesParameter element";
	}
	DOMElement * speciesParams = static_cast<DOMElement*>(list->item(0));
	for (DOMElement * species = XercesUtilities::getFirstChildDOMElement(speciesParams); species != 0;
			species = XercesUtilities::getNextSiblingDOMElement(species)) {
		tagName = XMLString::replicate(species->getTagName());
		if (tagName.toString() != "species") {
			continue;
		}
		for (DOMElement * element = XercesUtilities::getFirstChildDOMElement(species); element != 0;
				element = XercesUtilities::getNextSiblingDOMElement(element)) {
			tagName = XMLString::replicate(element->getTagName());
			if (tagName.toString() == "name") {
				result.push_back(XercesUtilities::getFirstChildTextContentAsString(element));
			}
		}
	}
	return result;
}

void Simulation::loadData(DOMElement const * simulation) {
	// eUnrecognized denotes an unrecognized tag.
	enum Tags {eUnrecognized, eGeneralParameters, eCompartmentParameters, eManagementParameters, eSpeciesParameters, eForestTypeParameters, eInspectionParameters, eAssortmentParameters, eNote};

	map<string, int> tagMap;
	tagMap["generalParameters"] = eGeneralParameters;
	tagMap["compartmentParameters"] = eCompartmentParameters;
	tagMap["managementParameters"] = eManagementParameters;
	tagMap["speciesParameters"] = eSpeciesParameters;
	tagMap["forestTypeParameters"] = eForestTypeParameters;
	tagMap["inspectionParameters"] = eInspectionParameters;
	tagMap["assortmentParameters"] = eAssortmentParameters;
	tagMap["note"] = eNote;

	for (DOMElement * element = XercesUtilities::getFirstChildDOMElement(simulation); element != 0; element
			= XercesUtilities::getNextSiblingDOMElement(element)) {
		XercesString tagName = XMLString::replicate(element->getTagName());
		switch (tagMap[tagName.toString()]) {
		case (eGeneralParameters): {
			Log::log(1, "Loading general parameters");
			simulationData.loadData(element);
			break;
		}
		case (eCompartmentParameters): {
			Log::log(1, "Loading compartment parameters");
			dataContainer.loadData(element,simulationData);
			break;
		}
		case (eManagementParameters): {
			Log::log(1, "Loading management parameters");
			TemplateContainer::loadData(element);
			break;
		}
		case (eSpeciesParameters): {
			Log::log(1, "Loading species parameters");
			SpeciesContainer::loadData(element);
			break;
		}
		case (eForestTypeParameters): {
			Log::log(1, "Loading forest type parameters");
			ForestTypeContainer::loadData(element);
			break;
		}
		case (eInspectionParameters): {
#ifdef CALCULATION_INSPECTION
			Log::log(1, "Loading calculation inspection parameters");
			CalculationInspector::Instance().loadData(element);
#else
			Log::log(1, "Calculation inspection not enabled in build. Ignoring parameters.");
#endif
			break;
		}
		case (eAssortmentParameters): {
			Log::log(1, "Loading assortment parameters");
			AssortmentContainer::loadData(element);
			break;
		}
		case (eNote): {
			// Not used by program, only a place to write a description in the input file
			break;
		}
		default: // Handles eUnrecognized.
		{
			throw "Unrecognized XML-element: "+tagName.toString();
		}
		}
	}
}

#ifdef USE_MPI
void Simulation::broadcastInputParameters() {
	int const rootId = 0;
	MPI_Comm const comm = MPI_COMM_WORLD;

	Log::log(0,"MPI broadcast of input parameters from process id %d", rootId);

	simulationData.broadcast(rootId,comm);
	Log::log(0,"Simulation data broadcasted");
	dataContainer.broadcast(rootId,comm,simulationData);
	Log::log(0,"CompartmentData container broadcasted");
	TemplateContainer::broadcast(rootId,comm);
	Log::log(0,"ManagementTemplate container broadcasted");
	SpeciesContainer::broadcast(rootId,comm);
	Log::log(0,"Species container broadcasted");
	ForestTypeContainer::broadcast(rootId,comm);
	Log::log(0,"ForestType container broadcasted");
	AssortmentContainer::broadcast(rootId, comm);
	Log::log(0, "Assortment container broadcasted");
#ifdef CALCULATION_INSPECTION
	CalculationInspector::Instance().broadcast(rootId, comm);
	Log::log(0, "Inspection parameters broadcasted");
#endif

	Log::log(0,"Broadcast of input parameters completed. Number of broadcasts done: %d",MPIUtil::getNumberOfBroadcastsDone());
}
#endif
