#include "SizeClassHandler.h"
#include "Tree.h"
#include "classify.h"
#include "TreeContainer.h"
#include "Log.h"

// Xerces-C++ includes.
#include <xercesc/util/XMLString.hpp>
#include <xercesc/dom/DOMElement.hpp>
#include <xercesc/dom/DOMNodeList.hpp>
XERCES_CPP_NAMESPACE_USE

#include "XercesString.h"
#include "XercesUtilities.h"

#include <iterator>
#include <map>
using std::map;

SizeClassHandler::SizeClassHandler()
{
}

SizeClassHandler::~SizeClassHandler()
{
}

unsigned char SizeClassHandler::addSizeClass(string name, float minimumDiameter)
{
	sizeClassNames.push_back(name);
	sizeClassBounds.push_back(minimumDiameter);
	return getNumberOfSizeClasses() - 1;
}

unsigned char SizeClassHandler::determineSizeClass(Tree const & curTree) const
{
	if(sizeClassBounds.empty()) {
		throw "No size class data available.";
	}
	int temp = classify(sizeClassBounds, curTree.getMark(TreeMarks::DIAMETER));
	unsigned char sizeClass = static_cast<unsigned char>(temp > 0? temp: 0);
	return sizeClass;
}

unsigned char SizeClassHandler::getSizeClassId(string name) const
{
	vector<string>::const_iterator position = find(sizeClassNames.begin(), sizeClassNames.end(), name);
	if(position != sizeClassNames.end()) {
		return distance(sizeClassNames.begin(), position);
	} else {
		throw "Size class not found.";
	}
}

void SizeClassHandler::loadData(DOMElement const * sizeClasses)
{
	// eUnrecognized denotes an unrecognized tag.
	enum Tags { eUnrecognized, eSizeClass };

	map<string, int> tagMap;
	tagMap["sizeClass"] = eSizeClass;

	for(DOMElement * element = XercesUtilities::getFirstChildDOMElement(sizeClasses); element != 0; element = XercesUtilities::getNextSiblingDOMElement(element)) {
		XercesString tagName = XMLString::replicate(element->getTagName());
		switch(tagMap[tagName.toString()])
		{
			case(eSizeClass):
			{
				Log::log(2,"Size class:");
				loadSizeClass(element);
				break;
			}
			default:	// Handles eUnrecognized.
			{
				throw "Unrecognized XML-element: "+tagName.toString();
			}
		}
	}
}

void SizeClassHandler::loadSizeClass(DOMElement const * sizeClass)
{
	string name;
	float minimumDiameter = 0.0;

	// eUnrecognized denotes an unrecognized tag.
	enum Tags { eUnrecognized, eName, eMinimumDiameter };

	map<string, int> tagMap;
	tagMap["name"] = eName;
	tagMap["minimumDiameter"] = eMinimumDiameter;

	for(DOMElement * element = XercesUtilities::getFirstChildDOMElement(sizeClass); element != 0; element = XercesUtilities::getNextSiblingDOMElement(element)) {
		XercesString tagName = XMLString::replicate(element->getTagName());
		switch(tagMap[tagName.toString()])
		{
			case(eName):
			{
				name = XercesUtilities::getFirstChildTextContentAsString(element);
				Log::log(3,"Name: %s", name.c_str());
				break;
			}
			case(eMinimumDiameter):
			{
				minimumDiameter = XercesUtilities::getFirstChildTextContentAsDouble(element);
				Log::log(3,"Minimum diameter: %f", minimumDiameter);
				break;
			}
			default:	// Handles eUnrecognized.
			{
				throw "Unrecognized XML-element: "+tagName.toString();
			}
		}
	}

	addSizeClass(name, minimumDiameter);
}
