#ifndef SIZECLASSHANDLER_H_
#define SIZECLASSHANDLER_H_

#include <string>
using std::string;
#include <vector>
using std::vector;

#ifdef USE_MPI
#include "MPIUtil.h"
#endif

class TreeContainer;
class Tree;

#include <xercesc/util/XercesVersion.hpp>

namespace XERCES_CPP_NAMESPACE {
	class DOMElement;
}

class SizeClassHandler {
public:
	SizeClassHandler();
	~SizeClassHandler();

#ifdef USE_MPI
	inline void broadcast(int const rootId, MPI_Comm const comm) {
		MPIUtil::broadcast(sizeClassNames, rootId, comm);
		MPIUtil::broadcast(sizeClassBounds, rootId, comm);
	}
#endif
	/**
	 * Adds a size class to the handler.
	 * Returns the id of the size class.
	 */
	unsigned char addSizeClass(string name, float minimumDiameter);
	/**
	 * Determines the size class of a Tree.
	 * The size classes must have been added in ascending order, otherwise determineSizeClasses() won't work.
	 * If the tree is smaller than the smallest size class, 0 will be returned;
	 * i.e. the tree is considered to belong to the smallest size class.
	 */
	unsigned char determineSizeClass(Tree const & curTree) const;
	/**
	 * Returns the number of size classes.
	 */
	unsigned char getNumberOfSizeClasses() const {
		return sizeClassNames.size();
	}
	/**
	 * Returns the name of the size class identified by @sizeClassId.
	 */
	string getSizeClassName(unsigned char sizeClassId) const {
		return sizeClassNames[sizeClassId];
	}
	unsigned char getSizeClassId(string name) const;
	/**
	 * Loads XML data for the SizeClassHandler.
	 *
	 * Precondition:	@sizeClasses must not be null.
	 */
	void loadData(XERCES_CPP_NAMESPACE::DOMElement const * sizeClasses);
private:
	/**
	 * Loads XML data for one size class.
	 *
	 * Precondition:	@sizeClass must not be null.
	 */
	void loadSizeClass(XERCES_CPP_NAMESPACE::DOMElement const * sizeClass);

	vector<string> sizeClassNames;
	vector<float> sizeClassBounds;
};

#endif /*SIZECLASSHANDLER_H_*/
