#include "StandStatisticsSummary.h"
#include "SimulationData.h"
#include "CompartmentDataContainer.h"
#include "StandStructureEvaluator.h"
#include "CompartmentData.h"
#include "CompartmentStatsAccumulator.h"
#include "Polygon.h"
#include "ForestType.h"
#include "ForestTypeContainer.h"
#include "SizeClassHandler.h"
#include "SpeciesContainer.h"
#include "Species.h"
#include "AssortmentContainer.h"
#include "Assortment.h"

using std::make_pair;
#include <algorithm>
using std::sort;
#include <numeric>
using std::accumulate;
#include <iomanip>
using std::setw;
using std::fixed;

bool sortPolygonCompartments(
		StandStatisticsSummary::PolygonCompartment const & a,
		StandStatisticsSummary::PolygonCompartment const & b) {
	if (a.polygon->getFilename().compare(b.polygon->getFilename()) < 0) {
		return true;
	} else if (a.polygon->getId() < b.polygon->getId()) {
		return true;
	} else {
		return false;
	}
}


StandStatisticsSummary::StandStatisticsSummary(
		SimulationData const & simData,
		CompartmentDataContainer const & compData,
		StandStructureEvaluator const & evaluator)
: simData(simData)
, compData(compData)
, evaluator(evaluator) {
}

StandStatisticsSummary::~StandStatisticsSummary() {
}

void StandStatisticsSummary::print(ostream & ostr) const {
	printHeader(ostr);

	vector<PolygonCompartment> polyCompVec = generatePolygonCompartmentVector(); 
	for (size_t i = 0, n = polyCompVec.size(); i < n; i++) {
		printPolygon(polyCompVec[i].polygon, polyCompVec[i].comp, ostr);
	}
}

vector<StandStatisticsSummary::PolygonCompartment> StandStatisticsSummary::generatePolygonCompartmentVector() const {
	vector<PolygonCompartment> result;
	CompartmentDataContainer::const_iterator iter, iend;
	for (iter = compData.begin(), iend = compData.end(); iter != iend; ++iter) {
		CompartmentData const * const data = *iter;
		Polygon const * const polygon = data->getCompartmentPolygon();
		vector<Polygon*> const & children = polygon->getChildren();
		if (children.size() == 0) {
			PolygonCompartment temp = {polygon, data};
			result.push_back(temp);
		} else {
			for (size_t i = 0, n = children.size(); i < n; i++) {
				Polygon const * const child = children[i];
				PolygonCompartment temp = {child, data};
				result.push_back(temp);
			}
		}
	}
	sort(result.begin(), result.end(), &sortPolygonCompartments);
	return result;
}

double StandStatisticsSummary::sumAndAverage(vector< vector<double> > const & statistics) const {
	double sum = 0.0;
	for (size_t i = 0, n = statistics.size(); i < n; i++) {
		sum += accumulate(statistics[i].begin(), statistics[i].end(), 0.0) / (double)(statistics[i].size());
	}
	return sum;
}

double StandStatisticsSummary::average(vector<double> const & statistics) const {
	return accumulate(statistics.begin(), statistics.end(), 0.0) / (double)(statistics.size());
}

void StandStatisticsSummary::printHeader(ostream & ostr) const {
	ostr << "Polygon id,Compartment id,";
	ostr << "Polygon area (ha),Compartment area (ha),Forest type,";
	ostr << "Basal area growth (m^2/ha),Final basal area (m^2/ha),";
	ostr << "Volume growth (m^3/ha),Dead tree volume (m^3/ha),";
	ostr << "New tree volume (m^3/ha),Removal (m^3/ha),";
	printAssortmentHeader("Removal", ostr);
	ostr << "Final volume (m^3/ha),";
	printAssortmentHeader("Growing stock", ostr);	
	ostr << "Mean diameter (cm),Mean height (m),Mean volume (l),";
	ostr << "Number of trees in compartment,";
	
	SizeClassHandler const & sizeClassHandler = simData.getSizeClassHandler();
	for (unsigned char i = 0; i < sizeClassHandler.getNumberOfSizeClasses(); ++i) {
		ostr << sizeClassHandler.getSizeClassName(i) << ",";
	}

	for (size_t i = 0, n = SpeciesContainer::size(); i < n; i++) {
		ostr << SpeciesContainer::get(i)->getName() << ",";
	}

	ostr << "Polygon file\n";
}

void StandStatisticsSummary::printAssortmentHeader(string const & prefix, ostream & ostr) const {
	for (size_t i = 0, n = AssortmentContainer::size(); i < n; i++) {
		Assortment const & assortment = AssortmentContainer::get(i);
		for (size_t j = 0, m = SpeciesContainer::size(); j < m; j++) {
			Species const * const species = SpeciesContainer::get(j);
			ostr << prefix << " " << species->getName() << " " << assortment.getName() << " (m^3/ha),";
		}
		ostr << prefix << " " << assortment.getName() << " (m^3/ha),";
	}
	for (size_t j = 0, m = SpeciesContainer::size(); j < m; j++) {
		Species const * const species = SpeciesContainer::get(j);
		ostr << prefix << " " << species->getName() << " (m^3/ha),";
	}
}

void StandStatisticsSummary::printPolygon(Polygon const * polygon, CompartmentData const * comp, ostream & ostr) const {
	unsigned int const numReplications = simData.getNumberOfReplications();
	double const areaInHectares = polygon->getFullArea() / 10000.0;
	double const fullAreaInHectares = comp->getCompartmentPolygon()->getFullArea() / 10000.0;
	ForestType const * const forestType = ForestTypeContainer::get(comp->getForestTypeId());
	CompartmentStatistics const & statistics = comp->getCompartmentStatistics();
	CompartmentStatsAccumulator const & accumulator = evaluator.compartmentStats.find(comp->getIdNumber())->second;
	vector< vector<uint64_t> > const & spStats = accumulator.speciesStatistics;

	double const basalAreaGrowth = sumAndAverage(statistics.basalAreaGrowthStatistics) / fullAreaInHectares; 
	double const volumeGrowth = sumAndAverage(statistics.volumeGrowthStatistics) / fullAreaInHectares;
	double const volumeDead = sumAndAverage(statistics.volumeDeadStatistics) / fullAreaInHectares;
	double const volumeNew = sumAndAverage(statistics.volumeNewStatistics) / fullAreaInHectares;
	double const volumeMarked = sumAndAverage(statistics.volumeMarkedStatistics) / fullAreaInHectares;

	double const basalAreaFinal = average(statistics.basalAreaFinalStatistics.back()) / fullAreaInHectares;
	double const volumeFinal = average(statistics.volumeFinalStatistics.back()) / fullAreaInHectares;

	double const meanDiameter = average(accumulator.meanDiameter);
	double const meanHeight = average(accumulator.meanHeight);
	double const meanVolume = average(accumulator.meanVolume);

	ostr.precision(2);
	ostr << setw(4) << polygon->getId() << ",";
	ostr << setw(4) << comp->getIdNumber() << ",";
	ostr << setw(8) << fixed << areaInHectares << ",";
	ostr << setw(8) << fixed << fullAreaInHectares << ",";
	ostr << forestType->getName() << ",";
	ostr << setw(8) << fixed << basalAreaGrowth << ",";
	ostr << setw(8) << fixed << basalAreaFinal << ",";
	ostr << setw(8) << fixed << volumeGrowth << ",";
	ostr << setw(8) << fixed << volumeDead << ",";
	ostr << setw(8) << fixed << volumeNew << ",";
	ostr << setw(8) << fixed << volumeMarked << ",";
	printRemovalAssortments(polygon, comp, ostr);
	ostr << setw(8) << fixed << volumeFinal << ",";
	printAssortments(polygon, comp, ostr);
	ostr << setw(8) << fixed << meanDiameter << ",";
	ostr << setw(8) << fixed << meanHeight << ",";
	ostr << setw(8) << fixed << meanVolume * 1000.0 << ",";

	uint64_t numberOfTrees = 0;
	SizeClassHandler const & sizeClassHandler = simData.getSizeClassHandler();
	for (size_t i = 0, n = sizeClassHandler.getNumberOfSizeClasses(); i < n; i++) {
		numberOfTrees += statistics.treeStatistics.back()[i][CompartmentStatistics::FINAL];
	}
	numberOfTrees /= numReplications;
	ostr << setw(8) << fixed << numberOfTrees << ",";

	for (size_t i = 0, n = sizeClassHandler.getNumberOfSizeClasses(); i < n; i++) {
		uint64_t const amount = statistics.treeStatistics.back()[i][CompartmentStatistics::FINAL];
		numberOfTrees += amount;
		ostr << setw(8) << amount / numReplications << ",";
	}

	for (size_t i = 0, n = spStats.size(); i < n; i++) {
		uint64_t const sum = accumulate(spStats[i].begin(), spStats[i].end(), 0);
		ostr << setw(8) << sum / numReplications << ",";
	}

	ostr << polygon->getFilename() << "\n";
}

void StandStatisticsSummary::printRemovalAssortments(Polygon const * polygon, CompartmentData const * comp, ostream & ostr) const {
	CompartmentStatistics const & statistics = comp->getCompartmentStatistics();
	unsigned int const numIterations = simData.getNumberOfIterations() + 1;
	unsigned int const numSpecies = SpeciesContainer::size();
	unsigned int const numAssortments = AssortmentContainer::size();
	double const fullAreaInHectares = comp->getCompartmentPolygon()->getFullArea() / 10000.0;

	vector< vector<double> > totals(numSpecies);
	for (unsigned int speciesId = 0; speciesId < numSpecies; ++speciesId) {
		totals[speciesId].resize(numAssortments, 0.0);
	}

	for (unsigned int iteration = 1; iteration < numIterations; ++iteration) {
		for (unsigned int speciesId = 0; speciesId < numSpecies; ++speciesId) {
			for (unsigned int assortId = 0; assortId < numAssortments; ++assortId) {
				totals[speciesId][assortId] += average(statistics.removalAssortmentVolumeStatistics[iteration][speciesId][assortId]);
			}
		}
	}

	ostr.precision(2);
	for (unsigned int assortId = 0; assortId < numAssortments; ++assortId) {
		double sum = 0.0;
		for (unsigned int speciesId = 0; speciesId < numSpecies; ++speciesId) {
			ostr << setw(8) << fixed << totals[speciesId][assortId] / fullAreaInHectares << ",";
			sum += totals[speciesId][assortId];
		}
		ostr << setw(8) << fixed << sum / fullAreaInHectares << ",";
	}
	for (unsigned int speciesId = 0; speciesId < numSpecies; ++speciesId) {
		double const total = accumulate(totals[speciesId].begin(), totals[speciesId].end(), 0.0);
		ostr << setw(8) << fixed << total / fullAreaInHectares << ",";
	}
}

void StandStatisticsSummary::printAssortments(Polygon const * polygon, CompartmentData const * comp, ostream & ostr) const {
	CompartmentStatistics const & statistics = comp->getCompartmentStatistics();
	unsigned int const numIterations = simData.getNumberOfIterations() + 1;
	unsigned int const numSpecies = SpeciesContainer::size();
	unsigned int const numAssortments = AssortmentContainer::size();
	double const fullAreaInHectares = comp->getCompartmentPolygon()->getFullArea() / 10000.0;

	vector< vector< vector<double> > > const & volumes = statistics.assortmentVolumeStatistics[numIterations - 1];

	ostr.precision(2);
	for (unsigned int assortId = 0; assortId < numAssortments; ++assortId) {
		double sum = 0.0;
		for (unsigned int speciesId = 0; speciesId < numSpecies; ++speciesId) {
			double const volume = average(volumes[speciesId][assortId]) / fullAreaInHectares;
			ostr << setw(8) << fixed << volume << ",";
			sum += volume;
		}
		ostr << setw(8) << fixed << sum << ",";
	}
	for (unsigned int speciesId = 0; speciesId < numSpecies; ++speciesId) {
		double total = 0.0;
		for (unsigned int assortId = 0; assortId < numAssortments; ++assortId) {
			total += average(volumes[speciesId][assortId]);
		}
		ostr << setw(8) << fixed << total / fullAreaInHectares << ",";
	}
}

