#include "StandStructureEvaluator.h"
#include "EvaluatorFunctions.h"
#include "Polygon.h"
#include "SimulationData.h"
#include "roundToInt.h"
#ifdef USE_MPI
#include "MPIUtil.h"
#endif
#include "TreeContainer.h"
#include "Compartment.h"
#include "CompartmentData.h"
#include "CompartmentDataContainer.h"
#include "CompartmentStatsAccumulator.h"
#include "SpeciesContainer.h"
#include "Species.h"
#include "VolumeModel.h"
#include <numeric>
#include <cmath>
#include <cassert>
#include <iomanip>
#include <vector>
#include <map>
#include <iterator>
#include <algorithm>
using std::accumulate;
using std::vector;
using std::map;
using std::ios;
using std::ios_base;
using std::fixed;
using std::setw;
using std::distance;
using std::sort;

StandStructureEvaluator::StandStructureEvaluator(SimulationData const & simData, CompartmentDataContainer const & compData) :
	simulationData(simData), compDataContainer(compData), compartmentStats() {
}

StandStructureEvaluator::~StandStructureEvaluator() {
}

void StandStructureEvaluator::evaluateCompartments(TreeContainer & allTrees, int const replication) {
	unsigned int const numReplications = simulationData.getNumberOfReplications();
	size_t const numSpecies = SpeciesContainer::size();
	size_t const numComps = allTrees.compSize();

	vector<double> diameterSums(numComps);
	vector<double> heightSums(numComps);
	vector<double> volumeSums(numComps);
	vector<double> basalAreaSums(numComps);
	vector<uint64_t> sizes(numComps);
	vector< map<size_t, uint64_t> > speciesSums(numSpecies);

	EvaluatorFunctions::calculateSpecies(allTrees,speciesSums);
	for(TreeContainer::const_comp_iterator compIter=allTrees.compBegin(); compIter!=allTrees.compEnd(); ++compIter) {
		Compartment const * const comp = *compIter;
		size_t const index = distance(allTrees.compBegin(),compIter);
		diameterSums[index] = EvaluatorFunctions::calculateMarkSum(comp,TreeMarks::DIAMETER);
		heightSums[index] = EvaluatorFunctions::calculateMarkSum(comp,TreeMarks::HEIGHT);
		volumeSums[index] = VolumeModel::calculateStemVolumeOf(comp);
		basalAreaSums[index] = EvaluatorFunctions::calculateBasalAreaSum(comp);
		sizes[index] = comp->size();
	}
#ifdef USE_MPI
	MPIUtil::sum(diameterSums,MPIUtil::getActiveCommunicator());
	MPIUtil::sum(heightSums,MPIUtil::getActiveCommunicator());
	MPIUtil::sum(volumeSums,MPIUtil::getActiveCommunicator());
	MPIUtil::sum(basalAreaSums,MPIUtil::getActiveCommunicator());
	MPIUtil::sum(sizes,MPIUtil::getActiveCommunicator());
	for(size_t spIndex=0; spIndex<numSpecies; spIndex++) {
		for (CompartmentDataContainer::const_iterator dataIter = compDataContainer.begin(), dataEnd = compDataContainer.end();
				dataIter != dataEnd; ++dataIter) {
			CompartmentData const * const compData = *dataIter;
			size_t const compartmentId = compData->getIdNumber();
			if (speciesSums[spIndex].find(compartmentId) == speciesSums[spIndex].end()) {
				speciesSums[spIndex][compartmentId] = 0;
			}
			MPIUtil::sum(speciesSums[spIndex][compartmentId], MPIUtil::getActiveCommunicator());
		}
	}
#endif
	for(TreeContainer::const_comp_iterator compIter=allTrees.compBegin(); compIter!=allTrees.compEnd(); ++compIter) {
		Compartment const * const comp = *compIter;
		size_t const index = distance(allTrees.compBegin(),compIter);
		Polygon const * const polygon = comp->getPolygon();
		double const fullAreaInHectar = polygon->getFullArea() / 10000.0;
		size_t const compartmentId = comp->getCompartmentData()->getIdNumber();
		if (compartmentStats.find(compartmentId) == compartmentStats.end())
			compartmentStats[compartmentId] = CompartmentStatsAccumulator(numReplications,numSpecies);
		CompartmentStatsAccumulator & curCompStats = compartmentStats.find(compartmentId)->second;
		curCompStats.basalArea[replication] = basalAreaSums[index] / fullAreaInHectar;
		curCompStats.meanDiameter[replication] = EvaluatorFunctions::calculateMean(diameterSums[index],sizes[index]);
		curCompStats.meanHeight[replication] = EvaluatorFunctions::calculateMean(heightSums[index],sizes[index]);
		curCompStats.meanVolume[replication] = EvaluatorFunctions::calculateMean(volumeSums[index],sizes[index]);
		curCompStats.meanVolumePerHectare[replication] = volumeSums[index] / fullAreaInHectar;
		for(size_t spIndex=0; spIndex<numSpecies; spIndex++) {
			curCompStats.speciesStatistics[spIndex][replication] = speciesSums[spIndex][index];
		}
	}
}

void StandStructureEvaluator::printSummary(unsigned int compartmentId, ostream & out) {
	out << "\n          STAND SUMMARY\n";
	out << "---------------------------------\n";
	// Setup array indexes for the 2.5 and 97.5 percentiles.
	int p0025 = roundToInt(0.025 * simulationData.getNumberOfReplications()) - 1;
	if (p0025 < 0) {
		p0025 = 0;
	}
	int p0975 = roundToInt(0.975 * simulationData.getNumberOfReplications()) - 1;
	out << "Approximate 95% Monte Carlo confidence interval:\n";
	out << "Lower = 3rd percentile";
	out << "\t\t(0.025 * " << simulationData.getNumberOfReplications() << " = " << p0025 + 1 << ")\n";
	out << "Upper = 98th percentile";
	out << "\t\t(0.975 * " << simulationData.getNumberOfReplications() << " = " << p0975 + 1 << ")\n";

	double simulatedValuesSum = 0.0;

	assert(compartmentStats.find(compartmentId)!=compartmentStats.end());
	CompartmentStatsAccumulator & curCompStats = compartmentStats.find(compartmentId)->second;

	out << "\nMean tree diameter at breast height (cm):\n";
	writeHeader(out, "");
	sumAndSort(curCompStats.meanDiameter, simulatedValuesSum);
	out << setw(7) << "";
	writeLine(out, 2, curCompStats.meanDiameter[p0025], simulatedValuesSum, curCompStats.meanDiameter[p0975]);

	out << "\nMean tree height (m):\n";
	writeHeader(out, "");
	sumAndSort(curCompStats.meanHeight, simulatedValuesSum);
	out << setw(7) << "";
	writeLine(out, 2, curCompStats.meanHeight[p0025], simulatedValuesSum, curCompStats.meanHeight[p0975]);

	out << "\nMean tree volume (l):\n";
	writeHeader(out, "");
	sumAndSort(curCompStats.meanVolume, simulatedValuesSum);
	out << setw(7) << "";
	writeLine(out, 2, curCompStats.meanVolume[p0025]*1000.0, simulatedValuesSum*1000.0, curCompStats.meanVolume[p0975]*1000.0);

	out << "\nCompartment final volume (m^3 / ha):\n";
	writeHeader(out, "");
	sumAndSort(curCompStats.meanVolumePerHectare, simulatedValuesSum);
	out << setw(7) << "";
	writeLine(out, 2, curCompStats.meanVolumePerHectare[p0025], simulatedValuesSum, curCompStats.meanVolumePerHectare[p0975]);

	out << "\nBasal area (m^2 / ha):\n";
	writeHeader(out, "");
	sumAndSort(curCompStats.basalArea, simulatedValuesSum);
	out << setw(7) << "";
	writeLine(out, 2, curCompStats.basalArea[p0025], simulatedValuesSum, curCompStats.basalArea[p0975]);

	out << "\nFinal species ratio and amount:\n";
	size_t nameLength = std::max(SpeciesContainer::getLongestNameLength() + 2, (size_t)15);
	out << setw(15) << "Percent" << setw(nameLength) << "Species";
	out << setw(15) << "Lower" << setw(15) << "Average" << setw(15) << "Upper" << "\n";
	uint64_t totalTrees = 0;
	vector< vector<uint64_t> > const & spStats = curCompStats.speciesStatistics;
	vector< vector<uint64_t> >::const_iterator spIterFirst, spIter;
	vector< vector<uint64_t> > spSorted(spStats.size());
	vector<bool> showSpecies(spStats.size(), false);
	for(spIter=spStats.begin(); spIter!=spStats.end(); ++spIter) {
		totalTrees += accumulate(spIter->begin(),spIter->end(),0);
		size_t const index = distance(spStats.begin(), spIter);
		spSorted[index] = *spIter;
		sort(spSorted[index].begin(), spSorted[index].end());
	}
	for(spIter=spStats.begin(); spIter!=spStats.end(); ++spIter) {
		uint64_t const sum = accumulate(spIter->begin(),spIter->end(),0);
		if(sum>0) {
			size_t const spIndex = distance(spStats.begin(), spIter);
			showSpecies[spIndex] = true;
			out.precision(1);
			out << setw(15) << fixed << (double)sum/(double)totalTrees * 100.0;
			out << setw(nameLength) << fixed << SpeciesContainer::get(spIndex)->getName();
			out.precision(0);
			double const average = (double)sum/(double)simulationData.getNumberOfReplications();
			out << setw(15) << spSorted[spIndex][p0025];
			out << setw(15) << fixed << average;
			out << setw(15) << spSorted[spIndex][p0975] << "\n";
		}
	}
	out << "\n";
}

void StandStructureEvaluator::sumAndSort(vector<double> & simulatedValues, double & simulatedValuesSum) {
	simulatedValuesSum = accumulate(simulatedValues.begin(), simulatedValues.end(), 0.0);
	sort(simulatedValues.begin(), simulatedValues.end());
}

void StandStructureEvaluator::writeHeader(ostream & out, string firstHeader) const {
	out << setw(8) << firstHeader << setw(15) << "Lower" << setw(15) << "Average" << setw(15) << "Upper\n";
}

void StandStructureEvaluator::writeLine(ostream & out, int precision1, double lowerBound,
		double simulatedValuesSum, double upperBound) const {
	out.precision(precision1);
	out << setw(15) << fixed << lowerBound << setw(15) << fixed << simulatedValuesSum /
	simulationData.getNumberOfReplications() << setw(15) << fixed << upperBound << "\n";
}

