#include "Tree.h"
#include "SpeciesContainer.h"
#include "Species.h"
#include "parameters.h"
#include "ForestInitializer.h"

#include <iomanip>
using std::ios;
using std::setw;
using std::fixed;

Tree::Tree(unsigned int species, unsigned int years, Point pos, double diam)
: position(pos)
, prevTreeInCompartment(NULL)
, nextTreeInCompartment(NULL)
, parentCompartment(NULL)
, initialHeterogeneity(0.0f)
, age(years)
, speciesId(species)
, removalFlags(NORMAL)
{
	previousDbhGrowth[0] = 1.0f;
	previousDbhGrowth[1] = 1.0f;
	for(unsigned int mark=0; mark < TreeMarks::NUMBER_OF_MARKS; mark++) {
		marks[mark] = 0.0f;
		marksGrowth[mark] = 1.0f;
	}
#ifdef INCLUDE_ENVIRONMENTAL_EFFECTS
	for (unsigned int i = 0; i < NUMBER_OF_RND_NUM_SETS * TreeMarks::NUMBER_OF_MARKS; i++) {
		randomNumbers[i] = 0.0f;
	}
#endif
	marks[TreeMarks::DIAMETER] = diam;
}

Tree::~Tree()
{
}

bool Tree::operator<(Tree const & tree) const {
	if(getX() == tree.getX()) {
		if(getY() == tree.getY()) {
			return false;
		}
		if(getY() < tree.getY()) {
			return true;
		} else {
			return false;
		}
	}
	if(getX() < tree.getX()) {
		return true;
	} else {
		return false;
	}
}

float Tree::getThreeYearTotalDbhGrowth(unsigned int iterationLength) const {
	float totalGrowth = marksGrowth[TreeMarks::DIAMETER];
	if(iterationLength<3) totalGrowth *= previousDbhGrowth[1];
	if(iterationLength<2) totalGrowth *= previousDbhGrowth[0];
	return totalGrowth;
}

void Tree::grow() {
	previousDbhGrowth[0] = previousDbhGrowth[1];
	previousDbhGrowth[1] = marksGrowth[TreeMarks::DIAMETER];
	for(unsigned int mark=0; mark < TreeMarks::NUMBER_OF_MARKS; mark++) {
		marks[mark] *= marksGrowth[mark];
		marksGrowth[mark] = 1.0f;
	}
}

ostream & operator<<(ostream & os, Tree const & tree)
{
	os.precision(6);
	os.setf(ios::fixed);
	os << setw(14) << tree.getX() << delimiterChar << setw(14) << tree.getY();
	for(unsigned int index=0; index<TreeMarks::NUMBER_OF_MARKS; index++) {
		os << delimiterChar << setw(12) << tree.getMark(index);
	}
	os << delimiterChar << setw(12) << tree.getInitialHeterogeneity();
	os << delimiterChar << setw(6) << tree.getAge();
	string const & name = SpeciesContainer::get(tree.getSpeciesId())->getName();
	size_t const neededLength = SpeciesContainer::getLongestNameLength();
	os << delimiterChar << setw(neededLength+1) << name;
	os << "\n";
	return os;
}

istream & operator>>(istream & in, Tree & tree)
{
	in.setf(ios::skipws); // Skip white-space.
	string species;
	char delimiter;

	if(ForestInitializer::positionIsLoaded())
		in >> tree.position.x;

	in >> delimiter;
	if(ForestInitializer::positionIsLoaded())
		in >> tree.position.y;

	for(unsigned int index=0; index<TreeMarks::NUMBER_OF_MARKS; index++) {
		in >> delimiter;
		if(ForestInitializer::markIsLoaded(index))
			in >> tree.marks[index];
	}

	in >> delimiter;
	if(ForestInitializer::heterogeneityIsLoaded())
		in >> tree.initialHeterogeneity;

	in >> delimiter;
	if(ForestInitializer::ageIsLoaded())
		in >> tree.age;

	in >> delimiter;
	if(in.eof())
		return in; // break if the line is not complete, signaling that a tree could not be read.
	if(ForestInitializer::speciesIsLoaded()) {
		// Ignore whitespace padding after the delimiter
		while (in.peek() == ' ') in.ignore(1);
		// Read the rest of the line into a temporary buffer with the get() method
		// in order to avoid problems with species whose names contain whitespace.
		size_t const bufsize = SpeciesContainer::getLongestNameLength() + 1;
		char * buf = new char[bufsize];
		try {
			in.get(buf, bufsize);
			size_t const buflen = std::strlen(buf);
			// Remove carriage return if file format is DOS
			if (buflen > 0 && buf[buflen - 1] == '\r') {
				buf[buflen - 1] = 0;
			}
			species = buf;
			delete [] buf;
		} catch (...) {
			delete [] buf;
			throw;
		}
		tree.speciesId = SpeciesContainer::getId(species);
	}
	return in;
}
