#ifndef TREE_H_
#define TREE_H_

#include <iostream>
using std::ostream;
using std::istream;
#include <cassert>
#include <limits>
using std::numeric_limits;

#include "Point.h"
#include "TreeMarks.h"

class TreeContainer;
class Compartment;

class Tree
{
public:
#ifdef INCLUDE_ENVIRONMENTAL_EFFECTS
	enum { NUMBER_OF_RND_NUM_SETS = 2 };
#endif

//FIXME: diameter should be float but the current version of craypat breaks if float is used... nice 10 million euro system
	Tree(unsigned int species, unsigned int years, Point position, double diam = 0.0);
	~Tree();

	// Calculates the distance to another tree.
	inline double distanceTo(Tree const & tree) const { return position.distanceTo(tree.position); }
	inline double squaredDistanceTo(Tree const & tree) const { return position.squaredDistanceTo(tree.position); }

	inline bool operator==(Tree const & tree) const { return *this == tree; }
	inline bool operator!=(Tree const & tree) const { return !(*this == tree); }
	bool operator<(Tree const & tree) const;

	inline Point const & getPosition() const { return position; }
	inline double getX() const { return position.x; }
	inline void setX(double x) { position.x = x; }
	inline double getY() const { return position.y; }
	inline void setY(double y) { position.y = y; }

	inline float getMark(unsigned int mark) const { assert(mark<TreeMarks::NUMBER_OF_MARKS); return marks[mark]; }
	inline void setMark(unsigned int mark, float value) { assert(mark<TreeMarks::NUMBER_OF_MARKS); marks[mark] = value; }

	inline Tree * getNextTreeInCompartment() const { return nextTreeInCompartment; }
	inline Tree * getPrevTreeInCompartment() const { return prevTreeInCompartment; }
	inline Compartment * getParentCompartment() const { return parentCompartment; }

	inline float getGrowth(unsigned int mark) const { assert(mark<TreeMarks::NUMBER_OF_MARKS); return marksGrowth[mark]; }
	inline void setGrowth(unsigned int mark, float value) { assert(mark<TreeMarks::NUMBER_OF_MARKS); marksGrowth[mark] = value; }
	inline float getPreviousDbhGrowth(unsigned int index) const { return previousDbhGrowth[index]; }
	inline void setPreviousDbhGrowth(unsigned int index, float value) {	previousDbhGrowth[index] = value; }
	inline float getInitialHeterogeneity() const { return initialHeterogeneity; }
	inline void setInitialHeterogeneity(float inHet) { initialHeterogeneity=inHet; }
	inline unsigned int getAge() const { return age; }
	inline void setAge(unsigned int years) { age = years; }
	inline void addAge(unsigned int years) { age += years; }
	inline void die() { if(removalFlags==NORMAL) removalFlags=DEAD; else if(removalFlags==MARKED) removalFlags=DEAD_AND_MARKED; }
	inline bool isDead() const { return removalFlags==DEAD || removalFlags==DEAD_AND_MARKED; }
	inline void markForRemoval() { if(removalFlags==NORMAL) removalFlags=MARKED; else if(removalFlags==DEAD) removalFlags=DEAD_AND_MARKED; }
	inline void unmarkRemoval() { if(removalFlags==MARKED) removalFlags=NORMAL; else if(removalFlags==DEAD_AND_MARKED) removalFlags=DEAD; }
	inline bool isMarked() const { return removalFlags==MARKED || removalFlags==DEAD_AND_MARKED; }
	inline unsigned int getSpeciesId() const { return speciesId; }
	inline void setSpeciesId(unsigned int const sp) { speciesId = sp; }

#ifdef INCLUDE_ENVIRONMENTAL_EFFECTS
	inline float getRandomNumber(unsigned int set, unsigned int mark) const {
		return randomNumbers[set * TreeMarks::NUMBER_OF_MARKS + mark];
	}
	inline void setRandomNumber(unsigned int set, unsigned int mark, float value) {
		randomNumbers[set * TreeMarks::NUMBER_OF_MARKS + mark] = value;
	}
#ifdef USE_MPI
	inline float getProcessXPosition() const {
		return processXPosition;
	}
	inline float getProcessYPosition() const {
		return processYPosition;
	}
	inline void setProcessXPosition(float const x) {
		processXPosition = x;
	}
	inline void setProcessYPosition(float const y) {
		processYPosition = y;
	}
#endif
#endif

	float getThreeYearTotalDbhGrowth(unsigned int iterationLength) const;
	void grow();

	friend ostream & operator<<(ostream & os, Tree const & tree);
	friend istream & operator>>(istream & in, Tree & tree);

private:
	enum RemovalState { NORMAL, DEAD, MARKED, DEAD_AND_MARKED };

	Point			position;
	Tree *			prevTreeInCompartment;
	Tree *			nextTreeInCompartment;
	Compartment *	parentCompartment;
	float			marks[TreeMarks::NUMBER_OF_MARKS];
	float			marksGrowth[TreeMarks::NUMBER_OF_MARKS];
	float			previousDbhGrowth[2];
	float			initialHeterogeneity;
#ifdef INCLUDE_ENVIRONMENTAL_EFFECTS
	float			randomNumbers[NUMBER_OF_RND_NUM_SETS * TreeMarks::NUMBER_OF_MARKS];
#ifdef USE_MPI
	float			processXPosition, processYPosition;
#endif
#endif
	unsigned short	age;
	unsigned short	speciesId;
	unsigned char	removalFlags;

	friend class Compartment; // Compartment is the only class allowed to call the following functions
	inline void setNextTreeInCompartment(Tree * t) { nextTreeInCompartment=t; }
	inline void setPrevTreeInCompartment(Tree * t) { prevTreeInCompartment=t; }
	inline void setParentCompartment(Compartment * s) { parentCompartment=s; }
};

ostream & operator<<(ostream & os, Tree const & tree);
istream & operator>>(istream & in, Tree & tree);

#endif /*TREE_H_*/
