/*
 * TreePersistence.cpp
 *
 *  Created on: 2008-nov-04
 *      Author: joschori
 */
#include "TreePersistence.h"
#include "TreeContainer.h"
#include "ForestSplitter.h"
#include "Compartment.h"
#include "Polygon.h"
#include "Tree.h"
#include "File.h"
#include "Log.h"

#include <cassert>
#include <iostream>
using std::ios_base;
#include <list>
using std::list;
#include <sstream>
using std::stringstream;
#include <numeric>
using std::accumulate;

#ifdef USE_MPI
#include "MPIUtil.h"
#endif

void TreePersistence::write(TreeContainer const & trees, string const & filename) {
#ifdef USE_MPI
	stringstream temp;
	temp << Tree(0,0,Point(0.0,0.0));
	// The number of bytes to write, length of a line including newline.
	unsigned int const lineLength = temp.str().size();

	char * oldstylefilename = new char[filename.size()+1];
	strncpy(oldstylefilename,filename.c_str(),filename.size()+1);

	int accessMode = MPI_MODE_WRONLY | MPI_MODE_CREATE;
	MPI_Info info = MPI_INFO_NULL;
	MPI_File fileHandle;

	MPI_File_open(MPIUtil::getActiveCommunicator(),oldstylefilename,accessMode,info,&fileHandle);
#if defined(COMPARISONSEEDING) && defined(INCLUDE_ENVIRONMENTAL_EFFECTS)
	// Write trees to the file in such an order that the result can be verified
	// in a sequential run.
	ForestSplitter const & forestSplitter = ForestSplitter::Instance();
	Point const & activeMinCoord = forestSplitter.getMinCoordinate();
	Point const & activeMaxCoord = forestSplitter.getMaxCoordinate();
	double const splitX = (activeMinCoord.x + activeMaxCoord.x) / 2;
	double const splitY = (activeMinCoord.y + activeMaxCoord.y) / 2;
	
	char * buffer = new char[lineLength];
	MPI_Status status;
	for (int quadrant = 0; quadrant < 4; ++quadrant) {
		Log::log(-1, "Writing trees in quadrant %d", quadrant);
		for (int id = 0; id < MPIUtil::mpiSize(); ++id) {
			Log::log(-1, "Writing trees of process %d", id);
			if (id == MPIUtil::mpiId()) {
				TreeContainer::const_iterator treeIter = trees.begin(), endIter = trees.end();
				for(; treeIter != endIter; ++treeIter) {
					Tree const * const tree = *treeIter;
					double const x = tree->getX();
					double const y = tree->getY();
					if ((quadrant == 0 && x < splitX && y < splitY) ||
					    (quadrant == 1 && x < splitX && y >= splitY) ||
					    (quadrant == 2 && x >= splitX && y < splitY) ||
					    (quadrant == 3 && x >= splitX && y >= splitY)) {
						temp.str(string());
						temp << **treeIter;
						strncpy(buffer, temp.str().c_str(), lineLength);
						MPI_File_write_shared(fileHandle, buffer, lineLength, MPI_CHAR, &status);
					}
				}
				MPI_File_sync(fileHandle);
			}
			MPI_Barrier(MPIUtil::getActiveCommunicator());
		}
	}
#else
	// Find the total needed file size.
	// Depends on the length of each line and the number of lines to be written in all processes.
	size_t const numTrees = trees.size();
	vector<uint64_t> treesInProcesses;
	MPIUtil::allgather(numTrees,treesInProcesses,MPIUtil::getActiveCommunicator());
	uint64_t const totalNumberOfTrees = accumulate(treesInProcesses.begin(),treesInProcesses.end(),(uint64_t)0);

	MPI_Offset const newSize = totalNumberOfTrees * lineLength;
	double const totalSize = (double)(newSize) / 1048576.0;
	double const thisProcessSize = (double)(numTrees * lineLength) / 1048576.0;
	Log::log(-1,"Total file size: %.2lf MB, this process part: %.2lf MB",totalSize,thisProcessSize);
	MPI_File_set_size(fileHandle,newSize);

	uint64_t const numberOfTreesBeforeThis = sumTreesBeforeThisProcess(treesInProcesses);
	MPI_Offset const offset = numberOfTreesBeforeThis * lineLength;
	MPI_File_seek(fileHandle,offset,MPI_SEEK_SET);

	char * buffer = new char[lineLength];
	MPI_Status status;
	for(TreeContainer::const_iterator treeIter = trees.begin(); treeIter!=trees.end(); ++treeIter) {
		temp.str(string());
		temp << **treeIter;
		strncpy(buffer,temp.str().c_str(),lineLength);
		MPI_File_write(fileHandle,buffer,lineLength,MPI_CHAR,&status);
	}
#endif
	// Make sure all processes are ready, then close the file.
	MPI_Barrier(MPIUtil::getActiveCommunicator());
	MPI_File_close(&fileHandle);
	delete[] oldstylefilename;
	delete[] buffer;
#else
	File out(filename, ios_base::out);
	for(TreeContainer::const_iterator treeIter = trees.begin(); treeIter!=trees.end(); ++treeIter) {
		out << **treeIter;
	}
#endif
}

#ifdef USE_MPI
uint64_t TreePersistence::sumTreesBeforeThisProcess(vector<uint64_t> const & treesInProcesses) {
	uint64_t sum=0;
	size_t const mpiId = MPIUtil::mpiId();
	for(size_t index=0; index<mpiId; index++) {
		sum += treesInProcesses[index];
	}
	return sum;
}
#endif

void TreePersistence::read(string const & filename, TreeContainer & trees) {
#ifdef USE_MPI
	ForestSplitter const & forestSplitter = ForestSplitter::Instance();
	Point const & activeMinCoord = forestSplitter.getMinCoordinate();
	Point const & activeMaxCoord = forestSplitter.getMaxCoordinate();
#endif

	File in(filename, ios_base::in);
	Tree tempTree = Tree(0,0,Point(0.0,0.0));
	list<Tree*> acceptedTrees;
	// Go through the whole input file of trees.
	while(true) {
		in >> tempTree;
		// If a tree could not be read due to eof, we are done.
		if(in.eof())
			break;
		Point const & position = tempTree.getPosition();
#ifdef USE_MPI
		if(activeMinCoord <= position && position < activeMaxCoord) {
			// Tree belongs to process, keep it
#endif
			// Find out which compartment the tree belongs to.
			for(TreeContainer::const_comp_iterator compIter = trees.compBegin(); compIter!=trees.compEnd(); ++compIter) {
				Compartment * const comp = *compIter;
				Polygon const * const polygon = comp->getPolygon();
				if(polygon->isActive() && polygon->activeContains(position)) {
					// Tree belongs to this compartment, add it.
					Tree * const newTree = new Tree(tempTree);
					comp->addTree(newTree);
					acceptedTrees.push_back(newTree);
					break;
				}
			}
#ifdef USE_MPI
		}
#endif
	}
	// Add the trees to the treecontainer.
	trees.addAll(acceptedTrees);
}
