#ifndef VECTOR_H_
#define VECTOR_H_

#include <ostream>
using std::ostream;
#include <vector>
using std::vector;
#include <cassert>
#include "acml.h"

class Vector
{
public:
	Vector();
	Vector(Vector const & other); // Copy constructor with deep copy
	Vector(vector<double> const & v); // Implicit conversion.
	Vector(vector<int> const & v); // Implicit conversion from int to double
	Vector(double * row, size_t numberOfElements); // constructor using preallocated data.
	explicit Vector(size_t size, double defaultValue = 0.0);
	~Vector();
	inline void clear() {
		size_t const vsize = size();
		for(size_t i=0; i<vsize; i++) {
			vec[i] = 0.0;
		}
	}
	inline double operator[](size_t index) const { // get-operator.
		return vec[index];
	}
	inline double & operator[](size_t index) { // set-operator.
		return vec[index];
	}
	inline Vector & operator=(Vector const & other) { // Assignment operator with deep copy
		if(ownsData) {
			if(vecsize!=other.vecsize) {
				delete[] vec;
				vecsize=other.vecsize;
				vec = new double[vecsize];
			}
		}
		else {
			assert(other.vecsize==vecsize);
		}
		copyVector(other,*this);
		return *this;
	}
	inline Vector & operator+=(Vector const & v) {
		daxpy(size(), 1.0, v.getDataPointer(), 1, getDataPointer(), 1);
		return *this;
	}
	inline Vector & operator-=(double const val) {
		size_t const vsize = size();
		for(size_t i=0; i < vsize; ++i) {
			vec[i] -= val;
		}
		return *this;
	}
	inline Vector & operator-=(Vector const & v) {
		daxpy(size(), -1.0, v.getDataPointer(), 1, getDataPointer(), 1);
		return *this;
	}
	inline Vector & operator/=(double const divisor) {
		double const factor = 1.0/divisor;
		(*this) *= factor;
		return *this;
	}
	inline Vector & operator*=(double const factor) {
		dscal(size(), factor, getDataPointer(), 1);
		return *this;
	}
	inline static double dotProduct(Vector const & v1, Vector const & v2) {
		return ddot(v1.size(), v1.getDataPointer(), 1, v2.getDataPointer(), 1);
	}
	inline size_t size() const { return vecsize; }
	inline double * getDataPointer() const { return vec; };
	friend Vector operator+(Vector const & left, Vector const & right);
	friend ostream & operator<<(ostream & os, Vector const & v);
private:

	inline static void copyVector(Vector const & from, Vector & to) {
		dcopy(to.size(), from.getDataPointer(), 1, to.getDataPointer(), 1);
	}

	double * vec;
	size_t vecsize;
	bool ownsData;
};

Vector operator+(Vector const & left, Vector const & right);
ostream & operator<<(ostream & os, Vector const & v);

#endif /*VECTOR_H_*/
