#include "VolumeModel.h"
#include "Tree.h"
#include "Compartment.h"
#include "SpeciesContainer.h"
#include "Species.h"
#include "DiameterModel.h"
#include "StumpModel.h"
#include "Log.h"

#include <cmath>
using std::pow;
#include <gsl/gsl_integration.h>
#include <gsl/gsl_errno.h>


struct GSLParameters {
	Tree const * tree;
	double diameterAt20PercentOfHeight;
};


double VolumeModel::calculateStemVolumeOf(Tree const * const targetTree) {
	double const heightInMetres = targetTree->getMark(TreeMarks::HEIGHT);
	double const stumpHeight = StumpModel::calculateStumpHeightOf(*targetTree);
	return calculateStemVolumeOf(targetTree, stumpHeight, heightInMetres);
}

double VolumeModel::calculateStemVolumeOf(Tree const * const targetTree,
		double h0, double h1) {
	if (h0 < 0.0)
		h0 = 0.0;
	if (h0 >= h1)
		return 0;
	double const diameter = targetTree->getMark(TreeMarks::DIAMETER);
	double const heightInMetres = targetTree->getMark(TreeMarks::HEIGHT);
	if (h1 > heightInMetres)
		h1 = heightInMetres;

	/**
	 * This volume formula is only valid for trees with height over 4 meters and a diameter larger than 3cm.
	 *
	 * For smaller trees we use a standard cylinder volume formula weighed by an extra tree coefficient which
	 * typically should be between 0.5 and 0.7.
	 *
	 */
	if (heightInMetres <= 4.0 || diameter <= 3.0) {
		/* TODO Perhaps use an offset to approximate the radius at ground level as the diameter is given at breast height.
		 *  double const dbhoffset = 0.0;
		 *  double const radiusAtGroundInMeters = (diameter+dbhoffset)/200.0;
		 *
		 */
		Species const * species = SpeciesContainer::get(targetTree->getSpeciesId());
		double const treeVolumeCoefficient = species->get(Species::SMALL_VOLUME_COEFFICIENT);
		double const radiusAtGroundInMeters = (diameter) / 200.0;
		double const approxVolumeInCubicMeters = M_PI * radiusAtGroundInMeters * radiusAtGroundInMeters
				* (h1 - h0) * treeVolumeCoefficient;
		return approxVolumeInCubicMeters;
	} else {
		/**
		 * Use the taper curve functions by Laasasenaho to calculate the volume
		 */
		double epsabs = 0.0001;
		double epsrel = 1.0;
		GSLParameters gslParams;
		gslParams.tree = targetTree;
		gslParams.diameterAt20PercentOfHeight = DiameterModel::estimateDiameterAt20PercentOfHeight(*targetTree);
		gsl_function gslFunction;
		gslFunction.function = &gslIntegrand;
		gslFunction.params = &gslParams;
		double volumeInCubicMetres, abserr;
		size_t neval;
		// Integrate over the stem area.
		int status = gsl_integration_qng(&gslFunction, h0, h1, epsabs, epsrel, &volumeInCubicMetres, &abserr, &neval);
		if (status) {
			Log::logError(-1, "Failed to calculate volume of tree at (%f, %f) with diameter %f and height %f. Message was: %s",
					targetTree->getX(), targetTree->getY(), diameter, heightInMetres,
					gsl_strerror(status));
			return 0.0;
		} else {
			return volumeInCubicMetres;
		}
	}
}

double VolumeModel::calculateStemVolumeOf(Compartment const * const compartment) {
	double volumeSum = 0.0;
	for (Tree const * curTree = compartment->getFirstTree(); curTree != NULL; curTree
			= curTree->getNextTreeInCompartment()) {
		volumeSum += calculateStemVolumeOf(curTree);
	}
	return volumeSum;
}

double VolumeModel::calculateBiomassVolumeOf(Tree const * const targetTree) {
	return calculateStemVolumeOf(targetTree) / 0.68;
}

double VolumeModel::calculateBiomassVolumeOf(Compartment const * const compartment) {
	return calculateStemVolumeOf(compartment) / 0.68;
}

double VolumeModel::gslIntegrand(double x, void * params) {
	GSLParameters * gslParams = static_cast<GSLParameters*>(params);
	double const radiusInMetres = 0.005 * DiameterModel::calculateDiameterAtHeight(
			*(gslParams->tree), x, gslParams->diameterAt20PercentOfHeight);
	return M_PI * radiusInMetres * radiusInMetres;
}

