#include "Assortment.h"
#include "Species.h"
#include "SpeciesContainer.h"
#include "XercesString.h"
#include "XercesUtilities.h"
#include <xercesc/util/XMLString.hpp>
#include <xercesc/dom/DOMElement.hpp>
XERCES_CPP_NAMESPACE_USE

#include <limits>
using std::numeric_limits;

Assortment::Assortment() {
	defaultConstraints.minDiameter = numeric_limits<float>::max();
	defaultConstraints.logLength = numeric_limits<float>::max();
}

#ifdef USE_MPI
void Assortment::broadcast(int const rootId, MPI_Comm const comm) {
	MPIUtil::broadcast(name, rootId, comm);
	defaultConstraints.broadcast(rootId, comm);

	unsigned int mapSize = constraints.size();
	MPIUtil::broadcast(mapSize, rootId, comm);
	unsigned int speciesId;
	if (MPIUtil::mpiId() == rootId) {
		for (ConstraintMap::iterator i = constraints.begin(); i != constraints.end(); ++i) {
			speciesId = i->first;
			MPIUtil::broadcast(speciesId, rootId, comm);
			i->second.broadcast(rootId, comm);
		}
	} else {
		for (unsigned int i = 0; i < mapSize; ++i) {
			MPIUtil::broadcast(speciesId, rootId, comm);
			constraints[speciesId].broadcast(rootId, comm);
		}
	}
}
#endif

void Assortment::loadData(XERCES_CPP_NAMESPACE::DOMElement const * assortment) {
	// eUnrecognized denotes an unrecognized tag.
	enum Tags {eUnrecognized, eName, eConstraints};

	map<string, int> tagMap;
	tagMap["name"] = eName;
	tagMap["constraints"] = eConstraints;

	for (DOMElement * element = XercesUtilities::getFirstChildDOMElement(assortment); element != 0; element
			= XercesUtilities::getNextSiblingDOMElement(element)) {
		XercesString tagName = XMLString::replicate(element->getTagName());
		switch (tagMap[tagName.toString()]) {
		case (eName): {
			name = XercesUtilities::getFirstChildTextContentAsString(element);
			break;
		}
		case (eConstraints): {
			string speciesName = XercesUtilities::getAttributeValueAsString(element, "species");
			if (speciesName != "") {
				unsigned int species = SpeciesContainer::getId(speciesName);
				loadConstraints(element, constraints[species]);
			} else {
				loadConstraints(element, defaultConstraints);
			}
			break;
		}
		default: {
			throw "Unrecognized XML-element: "+tagName.toString();
		}
		}
	}
}

void Assortment::loadConstraints(
		XERCES_CPP_NAMESPACE::DOMElement const * data, Constraints & result) {
	// eUnrecognized denotes an unrecognized tag.
	enum Tags {eUnrecognized, eMinDiameter, eLogLength};

	map<string, int> tagMap;
	tagMap["minimumDiameter"] = eMinDiameter;
	tagMap["logLength"] = eLogLength;

	result.minDiameter = 0.0f;
	result.logLength = 0.0f;

	for (DOMElement * element = XercesUtilities::getFirstChildDOMElement(data); element != 0; element
			= XercesUtilities::getNextSiblingDOMElement(element)) {
		XercesString tagName = XMLString::replicate(element->getTagName());
		switch (tagMap[tagName.toString()]) {
		case (eMinDiameter): {
			result.minDiameter = static_cast<float>(XercesUtilities::getFirstChildTextContentAsDouble(element));
			break;
		}
		case (eLogLength): {
			result.logLength = static_cast<float>(XercesUtilities::getFirstChildTextContentAsDouble(element));
			break;
		}
		default: {
			throw "Unrecognized XML-element: "+tagName.toString();
		}
		}
	}
}
