#include "CommonData.h"
#include "Log.h"
#include "SecondaryModel.h"
#include "Model.h"

#include <map>
using std::map;

#include "loadCoefficients.h"
#include "XercesString.h"
#include "XercesUtilities.h"
#include <xercesc/util/XMLString.hpp>
#include <xercesc/dom/DOMElement.hpp>
#include <xercesc/dom/DOMNodeList.hpp>
XERCES_CPP_NAMESPACE_USE

CommonData::CommonData() :
	modelTerms()
#ifdef INCLUDE_ENVIRONMENTAL_EFFECTS
, environmentalEffectCoefficients(ENVIRONMENTAL_EFFECTS)
#endif
{

}

CommonData::~CommonData() {

}

void CommonData::loadData(DOMElement const * data) {
	// eUnrecognized denotes an unrecognized tag.
	enum Tags {eUnrecognized, eModelTerm, eEnvironmentalEffects};

	map<string, int> tagMap;
	tagMap["modelTerm"] = eModelTerm;
	tagMap["environmentalEffects"] = eEnvironmentalEffects;

	for (DOMElement * element = XercesUtilities::getFirstChildDOMElement(data); element != 0; element
			= XercesUtilities::getNextSiblingDOMElement(element)) {
		XercesString tagName = XMLString::replicate(element->getTagName());
		switch (tagMap[tagName.toString()]) {
		case (eModelTerm): {
			loadModelTerm(element);
			break;
		}
		case (eEnvironmentalEffects): {
#ifdef INCLUDE_ENVIRONMENTAL_EFFECTS
			setEnvironmentalEffectCoefficients(loadCoefficients(element));
			Log::log(3,"Environmental effects loaded");
#else
			Log::logWarning(3, "Environmental effects not included in compilation, parameters ignored!");
#endif
			break;
		}
		default: // Handles eUnrecognized.
		{
			loadExtraData(element);
		}
		}
	}
}

void CommonData::loadModelTerm(DOMElement const * modelTerm) {

	ModelTerm term;

	// eUnrecognized denotes an unrecognized tag.
	enum Tags {eUnrecognized, eModelVariable, eTermCoefficient};

	map<string, int> tagMap;
	tagMap["modelVariable"] = eModelVariable;
	tagMap["termCoefficient"] = eTermCoefficient;

	for (DOMElement * element = XercesUtilities::getFirstChildDOMElement(modelTerm); element != 0; element
			= XercesUtilities::getNextSiblingDOMElement(element)) {
		XercesString tagName = XMLString::replicate(element->getTagName());
		switch (tagMap[tagName.toString()]) {
		case (eModelVariable): {
			string const name = XercesUtilities::getAttributeValueAsString(element, "name");
			unsigned int const modelVariableId = Model::getModelVariableId(name);
			double const order = XercesUtilities::getAttributeValueAsDouble(element, "order");
			double const exponent = (order!=0.0) ? order : 1.0; // default exponent 1.0
			ModelFactor const factor = ModelFactor(modelVariableId,exponent);
			term.insert(factor);
			break;
		}
		case (eTermCoefficient): {
			SecondaryModelData data;
			loadTermCoefficient(element,data);
			term.setCoefficient(data);
			break;
		}
		default: // Handles eUnrecognized.
		{
			throw "Unrecognized XML-element: " + tagName.toString();
		}
		}
	}
	insert(term);
}
void CommonData::loadTermCoefficient(DOMElement const * modelCoefficient, SecondaryModelData & params) {

	// eUnrecognized denotes an unrecognized tag.
	enum Tags {eUnrecognized, eSmoothingFunctionTerm};

	map<string, int> tagMap;
	tagMap["smoothingFunctionTerm"] = eSmoothingFunctionTerm;

	for (DOMElement * element = XercesUtilities::getFirstChildDOMElement(modelCoefficient); element != 0; element
			= XercesUtilities::getNextSiblingDOMElement(element)) {
		XercesString tagName = XMLString::replicate(element->getTagName());
		switch (tagMap[tagName.toString()]) {
		case (eSmoothingFunctionTerm): {
			SecondaryModelTerm secondaryTerm(XercesUtilities::getAttributeValueAsDouble(element,"coefficient"));
			loadSmoothingFunctionTerm(element, secondaryTerm);
			params.insert(secondaryTerm);
			break;
		}
		default: // Handles eUnrecognized.
		{
			throw "Unrecognized XML-element: " + tagName.toString();
		}
		}
	}
}
void CommonData::loadSmoothingFunctionTerm(DOMElement const * smoothingFunctionTerm, SecondaryModelTerm & secondaryTerm) {
	// eUnrecognized denotes an unrecognized tag.
	enum Tags {eUnrecognized, eSmoothingFunctionVariable};

	map<string, int> tagMap;
	tagMap["smoothingFunctionVariable"] = eSmoothingFunctionVariable;

	for (DOMElement * element = XercesUtilities::getFirstChildDOMElement(smoothingFunctionTerm); element != 0; element
			= XercesUtilities::getNextSiblingDOMElement(element)) {
		XercesString tagName = XMLString::replicate(element->getTagName());
		switch (tagMap[tagName.toString()]) {
		case (eSmoothingFunctionVariable): {
			string const name = XercesUtilities::getAttributeValueAsString(element, "name");
			unsigned int const secondaryVariableId = SecondaryModel::getSecondaryVariableId(name);
			double const order = XercesUtilities::getAttributeValueAsDouble(element, "order");
			double const exponent = (order!=0.0) ? order : 1.0; // default exponent 1.0
			SecondaryModelFactor const factor(secondaryVariableId,exponent);
			secondaryTerm.insert(factor);
			break;
		}
		default: // Handles eUnrecognized.
		{
			throw "Unrecognized XML-element: " + tagName.toString();
		}
		}
	}

}
