#ifndef COMMONDATA_H_
#define COMMONDATA_H_

#include <cassert>

#include "ModelTerm.h"
#include <vector>
using std::vector;
#include <ostream>
using std::ostream;
using std::endl;

#ifdef USE_MPI
#include "MPIUtil.h"
#endif
#ifdef INCLUDE_ENVIRONMENTAL_EFFECTS
#include "Vector.h"
#endif

#include <xercesc/util/XercesVersion.hpp>

namespace XERCES_CPP_NAMESPACE
{
	class DOMElement;
}

class CommonData {
public:
#ifdef INCLUDE_ENVIRONMENTAL_EFFECTS
	enum NumberOfCoefficients { ENVIRONMENTAL_EFFECTS = 2 };
#endif

	CommonData();
	virtual ~CommonData();

	virtual inline vector<ModelTerm> const & getModelTerms() const {
		return modelTerms;
	}
	virtual inline void writeModel(ostream & out) const {
		for(vector<ModelTerm>::const_iterator termIter=modelTerms.begin(); termIter!=modelTerms.end();) {
			out << *termIter;
			if(++termIter != modelTerms.end())
				out <<" + ";
		}
	}

#ifdef USE_MPI
	virtual inline void broadcast(int const rootId, MPI_Comm const comm) {
		unsigned int numTerms = modelTerms.size();
		MPIUtil::broadcast(numTerms, rootId, comm);
		if(MPIUtil::mpiId()!=rootId)
			modelTerms.resize(numTerms);
		for(unsigned int index=0; index<numTerms; index++) {
			modelTerms[index].broadcast(rootId, comm);
		}
#ifdef INCLUDE_ENVIRONMENTAL_EFFECTS
		MPIUtil::broadcast(environmentalEffectCoefficients, rootId, comm);
#endif
	}
#endif

#ifdef INCLUDE_ENVIRONMENTAL_EFFECTS
	virtual inline double getEnvironmentalEffectCoefficient(unsigned int index) const {
		assert(index < ENVIRONMENTAL_EFFECTS);
		return environmentalEffectCoefficients[index];
	}
	virtual inline void setEnvironmentalEffectCoefficient(unsigned int index, double value) {
		if(index < ENVIRONMENTAL_EFFECTS) {
			environmentalEffectCoefficients[index] = value;
		} else {
			throw "Bad environmental effect index.";
		}
	}
	virtual inline Vector const & getEnvironmentalEffectCoefficients() const {
		return environmentalEffectCoefficients;
	}
	virtual inline void setEnvironmentalEffectCoefficients(Vector const & coefficients) {
		for(unsigned int i = 0; i < coefficients.size(); ++i) {
			setEnvironmentalEffectCoefficient(i, coefficients[i]);
		}
	}
	/**
	 * Returns true if there are environmental effects, otherwise false.
	 */
	virtual inline bool hasEnvironmentalEffects() const {
		return environmentalEffectCoefficients[1] != 0.0;
	}
#endif
	/**
	 * Loads XML data for the CommonData.
	 *
	 * Precondition:	@data must not be null.
	 */
	void loadData(XERCES_CPP_NAMESPACE::DOMElement const * data);

protected:
	vector<ModelTerm> modelTerms;
#ifdef INCLUDE_ENVIRONMENTAL_EFFECTS
	Vector	environmentalEffectCoefficients;
#endif

	/**
	 * Loading values of XML elements within a "modelTerm"-tag.
	 */
	void loadModelTerm(XERCES_CPP_NAMESPACE::DOMElement const * modelTerm);
	/**
	 * Loading values of XML elements within a "termCoefficient"-tag.
	 */
	void loadTermCoefficient(XERCES_CPP_NAMESPACE::DOMElement const * termCoefficient, SecondaryModelData & params);
	/**
	 * Loading values of XML elements within a "termCoefficient"-tag.
	 */
	void loadSmoothingFunctionTerm(XERCES_CPP_NAMESPACE::DOMElement const * smoothingFunctionTerm, SecondaryModelTerm & secondaryTerm);

	virtual inline void insert(ModelTerm const & params) {
		modelTerms.push_back(params);
	}
	virtual inline bool hasParams() const {
		for(vector<ModelTerm>::const_iterator termIter=modelTerms.begin(); termIter!=modelTerms.end(); ++termIter) {
			if(termIter->hasParams())
				return true;
		}
		return false;
	}
	/**
	 * Loads XML data for child classes
	 *
	 * Precondition:	@extraData must not be null.
	 */
	virtual void loadExtraData(XERCES_CPP_NAMESPACE::DOMElement const * extraData) = 0;
};

#endif /*COMMONDATA_H_*/
