#include "CompartmentData.h"
#include "Polygon.h"
#include "SimulationData.h"
#include "Log.h"
#ifdef USE_MPI
#include "MPIUtil.h"
#endif
#include "ForestTypeContainer.h"
#include "TemplateContainer.h"

#include "XercesString.h"
#include "XercesUtilities.h"
#include <xercesc/util/XMLString.hpp>
#include <xercesc/dom/DOMElement.hpp>
XERCES_CPP_NAMESPACE_USE
#include <map>
using std::map;
#include <sstream>
using std::stringstream;

map<string, map<unsigned int, Polygon*> > CompartmentData::polygonData;

CompartmentData::CompartmentData(SimulationData const & sim, unsigned int id) :
	compartmentStatistics(sim, *this), polygon(auto_ptr<Polygon> (new Polygon())), numberOfInitialTreesFactor(1.0),
	ageOfInitialTreesFactor(1.0), idNumber(id),
	forestTypeId(0), initialStandAge(0), forestTypeName("Undefined")
#ifdef USE_MPI
			, compartmentActiveProcesses(MPI_COMM_NULL)
#endif
{
}

CompartmentData::~CompartmentData() {
#ifdef USE_MPI
	freeCompartmentCommunicator();
#endif
}

#ifdef USE_MPI
void CompartmentData::broadcast(int const rootId, MPI_Comm const comm) {
	// no need to broadcast compartmentStatistics as they are initialized correctly by constructor
	// but when the polygon is changed, the the area have to be set in compartmentStatistics.
	polygon->broadcast(rootId, comm);
	MPIUtil::broadcast(numberOfInitialTreesFactor, rootId, comm);
	MPIUtil::broadcast(ageOfInitialTreesFactor, rootId, comm);
	MPIUtil::broadcast(idNumber, rootId, comm);
	MPIUtil::broadcast(forestTypeId, rootId, comm);
	MPIUtil::broadcast(forestTypeName, rootId, comm);
	MPIUtil::broadcast(managementTemplateIds, rootId, comm);
	MPIUtil::broadcast(managementTemplateNames, rootId, comm);
}
#endif

void CompartmentData::loadData(DOMElement const * compartmentData) {

	// eUnrecognized denotes an unrecognized tag.
	enum Tags {
		eUnrecognized, ePolygonFile, eManagementTemplateName, eForestTypeName, eNumberOfInitialTreesFactor,
		eAgeOfInitialTreesFactor
	};

	map<string, int> tagMap;
	tagMap["polygonFile"] = ePolygonFile;
	tagMap["managementTemplateName"] = eManagementTemplateName;
	tagMap["forestTypeName"] = eForestTypeName;
	tagMap["numberOfInitialTreesFactor"] = eNumberOfInitialTreesFactor;
	tagMap["ageOfInitialTreesFactor"] = eAgeOfInitialTreesFactor;

	XercesString name = XMLString::transcode("id");
	if (compartmentData->hasAttribute(name.xmlCh())) {
		int id = XercesUtilities::getAttributeValueAsInt(compartmentData, "id");
		if (id < 0) {
			Log::logError(-1, "Compartment id must not be negative. %d given.", id);
		} else {
			idNumber = (unsigned int)id;
		}
	}

	vector<Polygon *> polygons;

	for (DOMElement * element = XercesUtilities::getFirstChildDOMElement(compartmentData); element != 0; element
			= XercesUtilities::getNextSiblingDOMElement(element)) {
		XercesString tagName = XMLString::replicate(element->getTagName());
		switch (tagMap[tagName.toString()]) {
		case (ePolygonFile): {
			string polygonFilename = XercesUtilities::getFirstChildTextContentAsString(element);
			int id = XercesUtilities::getAttributeValueAsInt(element, "id");

			Polygon * p = getPolygon(polygonFilename, id);
			polygons.push_back(p);
			//Log::log(2, "Compartment %d polygon read from %s, polygon number %d.", idNumber, polygonFilename.c_str(), id);
			break;
		}
		case (eManagementTemplateName): {
			string name = XercesUtilities::getFirstChildTextContentAsString(element);
			managementTemplateNames.push_back(name);
			break;
		}
		case (eForestTypeName): {
			forestTypeName = XercesUtilities::getFirstChildTextContentAsString(element);
			break;
		}
		case (eNumberOfInitialTreesFactor): {
			numberOfInitialTreesFactor = XercesUtilities::getFirstChildTextContentAsDouble(element);
			break;
		}
		case (eAgeOfInitialTreesFactor): {
			ageOfInitialTreesFactor = XercesUtilities::getFirstChildTextContentAsDouble(element);
			break;
		}
		default: // Handles eUnrecognized.
		{
			throw "Unrecognized XML-element: " + tagName.toString();
		}
		}
	}

	if (polygons.size() == 1) {
		setCompartmentPolygon(polygons[0]);
	} else if (polygons.size() > 1) {
		// The compartment has several polygons. Combine them into one.
		setCompartmentPolygon(new Polygon(polygons));
	}
}

void CompartmentData::loadPolygonData(DOMElement const * root) {
	set<string> files = findUsedPolygons(root);
	polygonData.clear();
	for (set<string>::iterator iter = files.begin(); iter != files.end(); iter++) {
		string filename = *iter;
		map<unsigned int, Polygon*> data = Polygon::loadPolygons(filename);
		polygonData.insert(make_pair(filename, data));
	}
}

set<string> CompartmentData::findUsedPolygons(DOMElement const * root) {
	set<string> polygonFiles;
	// polygon filenames are located at simulation/compartment/polygonFile
	for (DOMElement * element = XercesUtilities::getFirstChildDOMElement(root); element != 0; element
			= XercesUtilities::getNextSiblingDOMElement(element)) {
		XercesString tagName = XMLString::replicate(element->getTagName());

		if (tagName.toString() == "compartment") {
			findPolygonFileLoop(polygonFiles, element);
		}
	}
	return polygonFiles;
}

void CompartmentData::findPolygonFileLoop(set<string> & polygonFiles, DOMElement * element) {
	for (DOMElement * compartmentElement = XercesUtilities::getFirstChildDOMElement(element); compartmentElement != 0; compartmentElement
			= XercesUtilities::getNextSiblingDOMElement(compartmentElement)) {
		XercesString compartmentTagName = XMLString::replicate(compartmentElement->getTagName());
		if (compartmentTagName.toString() == "polygonFile") {
			string polygonFileName = XercesUtilities::getFirstChildTextContentAsString(compartmentElement);
			polygonFiles.insert(polygonFileName);
			break;
		}
	}
}

Polygon* CompartmentData::getPolygon(string filename, unsigned int polygonId) {
	// Log::log(-1, "getPolygon(%s, %d)", filename.c_str(), polygonId);
	map<string, map<unsigned int, Polygon*> >::iterator i = polygonData.find(filename);
	if (i != polygonData.end()) {
		map<unsigned int, Polygon*> & polygonsFromFile = i->second;

		map<unsigned int, Polygon*>::iterator iter = polygonsFromFile.find(polygonId);
		if (iter != polygonsFromFile.end()) {
			Polygon * p = iter->second;
			polygonsFromFile.erase(iter);
			return p;
		}

	}

	stringstream ss;
	ss << "Polygon " << filename << ", " << polygonId << " was not found in polygonMap";
	throw ss.str();
}

void CompartmentData::cleanupLoadedPolygons() {
	// Log::log(-1, "cleanupLoadedPolygons");
	while (!polygonData.empty()) {
		map<string, map<unsigned int, Polygon*> >::iterator iter = polygonData.begin();
		map<unsigned int, Polygon*> & filePolygons = iter->second;

		while (!filePolygons.empty()) {
			map<unsigned int, Polygon*>::iterator polyIter = filePolygons.begin();
			// Delete Polygon
			delete polyIter->second;

			filePolygons.erase(polyIter);
		}

		polygonData.erase(iter);
	}
}

void CompartmentData::setUpIds() {
	forestTypeId = ForestTypeContainer::getId(forestTypeName);
	size_t numTemplates = managementTemplateNames.size();
	managementTemplateIds.resize(numTemplates);
	for (size_t i = 0; i < numTemplates; i++) {
		managementTemplateIds[i] = TemplateContainer::getId(managementTemplateNames[i]);
	}
}

