#ifndef COMPARTMENTDATA_H_
#define COMPARTMENTDATA_H_

#include <memory>
using std::auto_ptr;
#include <vector>
using std::vector;
#include <iostream>
#ifdef USE_MPI
#include <mpi.h>
#endif
#include <set>
using std::set;
#include <map>
using std::map;
#include <string>
using std::string;

#include "CompartmentStatistics.h"
#include "Polygon.h"

#include <xercesc/util/XercesVersion.hpp>

class Management;
class SimulationData;

namespace XERCES_CPP_NAMESPACE
{
	class DOMElement;
}


class CompartmentData
{
public:
	CompartmentData(SimulationData const & sim, unsigned int id);
	~CompartmentData();

#ifdef USE_MPI
	void broadcast(int const rootId, MPI_Comm const comm);
#endif
	inline CompartmentStatistics & getCompartmentStatistics() {
		return compartmentStatistics;
	}
	inline CompartmentStatistics const & getCompartmentStatistics() const {
		return compartmentStatistics;
	}

	inline Polygon * getCompartmentPolygon() const {
		return polygon.get();
	}
	inline void setCompartmentPolygon(Polygon * poly) {
		polygon.reset(poly);
	}
	inline double getNumberOfInitialTreesFactor() const {
		return numberOfInitialTreesFactor;
	}
	inline double getAgeOfInitialTreesFactor() const {
		return ageOfInitialTreesFactor;
	}
	inline vector<unsigned int> const & getManagementTemplateIds() const {
		return managementTemplateIds;
	}
	inline unsigned int getIdNumber() const {
		return idNumber;
	}

	void loadData(XERCES_CPP_NAMESPACE::DOMElement const * compartmentData);

	inline unsigned int getForestTypeId() const {
		return forestTypeId;
	}
	inline vector<string> const & getManagementTemplateNames() const {
		return managementTemplateNames;
	}
	inline string const & getForestTypeName() const {
		return forestTypeName;
	}
#ifdef USE_MPI
	inline MPI_Comm getCompartmentCommunicator() const {
		return compartmentActiveProcesses;
	}
	inline void freeCompartmentCommunicator() {
		if (compartmentActiveProcesses!=MPI_COMM_NULL) {
			MPI_Comm_free(&compartmentActiveProcesses);
		}
	}
	inline void setCompartmentCommunicator(MPI_Comm comm) {
		freeCompartmentCommunicator();
		compartmentActiveProcesses = comm;
		int size = 0;
		if (comm != MPI_COMM_NULL) {
			MPI_Comm_size(comm, &size);
		}
		shared = (size > 1);
	}
	inline bool isShared() const {
		return shared;
	}
#else
	inline bool isShared() const {
		return false;
	}
#endif
	static void loadPolygonData(XERCES_CPP_NAMESPACE::DOMElement const * root);
	static void cleanupLoadedPolygons();

	void setUpIds();

	inline unsigned int getInitialStandAge() const {
		return initialStandAge;
	}
	inline void setInitialStandAge(unsigned int age) {
		initialStandAge = age;
	}

private:
	/*
	 * Loads all the polygon data needed for the simulation
	 *
	 */
	static set<string> findUsedPolygons(XERCES_CPP_NAMESPACE::DOMElement const * root);
	static void findPolygonFileLoop(set<string> & polygonFiles, XERCES_CPP_NAMESPACE::DOMElement * element);

	/**
	 * Returns the polygon from the given file with the given polygon ID
	 * and removes it from the internal list. The caller is responsible for
	 * deleting the returned polygon.
	 */
	static Polygon* getPolygon(string filename, unsigned int polygonId);
	friend class SimulationData;

	CompartmentStatistics		compartmentStatistics;
	auto_ptr<Polygon>			polygon;
	double						numberOfInitialTreesFactor;
	double						ageOfInitialTreesFactor;
	unsigned int				idNumber;
	unsigned int				forestTypeId;
	unsigned int				initialStandAge;
	string						forestTypeName;
	vector<unsigned int>		managementTemplateIds;
	vector<string>				managementTemplateNames;

	static map<string,map<unsigned int, Polygon*> > polygonData;
#ifdef USE_MPI
	MPI_Comm compartmentActiveProcesses;
	bool shared;
#endif

};
#endif /*COMPARTMENTDATA_H_*/
