#include "CompartmentDataContainer.h"
#include "CompartmentData.h"
#include "Log.h"
#include "SimulationData.h"
#include "TemplateContainer.h"
#include "ForestTypeContainer.h"

#ifdef USE_MPI
#include "MPIUtil.h"
#endif

#include "XercesString.h"
#include "XercesUtilities.h"
#include <xercesc/util/XMLString.hpp>
#include <xercesc/dom/DOMElement.hpp>
XERCES_CPP_NAMESPACE_USE
#include <map>
using std::map;
#include <set>
using std::set;

CompartmentDataContainer::CompartmentDataContainer() {

}

CompartmentDataContainer::~CompartmentDataContainer() {
	size_t const numDatas = size();
	for (size_t index=0; index<numDatas; index++) {
		delete compartmentDatas[index];
	}
	compartmentDatas.clear();
}

#ifdef USE_MPI
void CompartmentDataContainer::broadcast(int const rootId, MPI_Comm const comm, SimulationData const & sim) {
	unsigned int numDatas = size();
	MPIUtil::broadcast(numDatas, rootId, comm);
	Log::log(1, "Number of compartment datas %d", numDatas);
	for (unsigned int index=0; index<numDatas; index++) {
		Log::log(2, "Broadcasting compartmentData index %d", index);
		if (MPIUtil::mpiId()==rootId) {
			compartmentDatas[index]->broadcast(rootId, comm);
		} else {
			CompartmentData * const compData = new CompartmentData(sim,0);
			compData->broadcast(rootId, comm);
			add(compData);
		}
		// Barrier needed here because broadcast is non-blocking for the root process.
		// That means that root broadcasts everything before the other processes
		// have received even half of the data, filling the internal buffers of MPI.
		MPI_Barrier(comm);
	}
}
#endif

CompartmentData * CompartmentDataContainer::find(unsigned int const compartmentId) const {
	size_t const numDatas = size();
	for (size_t index=0; index<numDatas; index++) {
		if (compartmentDatas[index]->getIdNumber() == compartmentId) {
			return compartmentDatas[index];
		}
	}
	return NULL;
}

void CompartmentDataContainer::loadData(DOMElement const * compartments, SimulationData const & sim) {

	unsigned int compartmentDataId = 0;
	set<unsigned int> usedIdNumbers;

	// Load polygon data once and for all
	CompartmentData::loadPolygonData(compartments);

	// eUnrecognized denotes an unrecognized tag.
	enum Tags {eUnrecognized, eCompartment};

	map<string, int> tagMap;
	tagMap["compartment"] = eCompartment;

	for (DOMElement * element = XercesUtilities::getFirstChildDOMElement(compartments); element != 0; element
			= XercesUtilities::getNextSiblingDOMElement(element)) {
		XercesString tagName = XMLString::replicate(element->getTagName());
		switch (tagMap[tagName.toString()]) {
		case (eCompartment): {
			Log::log(2, "Loading compartment parameters");
			CompartmentData * compartmentData = new CompartmentData(sim, compartmentDataId);
			compartmentData->loadData(element);
			unsigned int id = compartmentData->getIdNumber();
			if (usedIdNumbers.find(id) != usedIdNumbers.end()) {
				Log::logError(-1, "A compartment with id %u already exists!", id);
				throw "Same compartment id used multiple times";
			}
			add(compartmentData);
			compartmentDataId++;
			usedIdNumbers.insert(id);
			break;
		}
		default: // Handles eUnrecognized.
		{
			throw "Unrecognized XML-element: "+tagName.toString();
		}
		}
	}

	CompartmentData::cleanupLoadedPolygons();
}

void CompartmentDataContainer::setupIds() {
	size_t const numDatas = size();
	for (size_t index=0; index<numDatas; index++) {
		compartmentDatas[index]->setUpIds();
	}
}

