#include "SimulationData.h"
#include "InitialCompartmentGenerator.h"
#include "parameters.h"
#include "Log.h"
#include "StandardNormalRandomNumberGenerator.h"
#include "UniformRandomNumberGenerator.h"
#include "RegenerationModel.h"

#include "XercesString.h"
#include "XercesUtilities.h"
#include <xercesc/util/XMLString.hpp>
#include <xercesc/dom/DOMElement.hpp>
XERCES_CPP_NAMESPACE_USE

#include <map>
using std::map;

#include <algorithm>
using std::max;

#include <sys/stat.h>

SimulationData::SimulationData()
: growthDatas(TreeMarks::NUMBER_OF_MARKS)
, mortalityDatas(TreeMarks::NUMBER_OF_MARKS)
, reproductionDatas(TreeMarks::NUMBER_OF_MARKS)
, regenerationModelData()
, compartmentGenerator(*this)
, numberOfReplications(0)
, iterationLength(1)
, numberOfYears(0)
, saveFinalTrees(false)
, saveIterationTrees(false)
, iterationLoadBalancing(false)
, iterationLoadBalancingDetectionPct(0)
, correlationDistanceFactor(1.0)
{
}

SimulationData::~SimulationData() {
}
#ifdef USE_MPI
void SimulationData::broadcast(int const rootId, MPI_Comm const comm) {
	MPIUtil::broadcast(simulationName, rootId, comm);
	MPIUtil::broadcast(initialStandFilename, rootId, comm);
	ageClassHandler.broadcast(rootId, comm);
	sizeClassHandler.broadcast(rootId, comm);
	heightClassHandler.broadcast(rootId, comm);
	for (size_t index=0; index < TreeMarks::NUMBER_OF_MARKS; index++) {
		growthDatas[index].broadcast(rootId, comm);
		mortalityDatas[index].broadcast(rootId, comm);
		reproductionDatas[index].broadcast(rootId, comm);
		// Barrier needed here because broadcast is non-blocking for the root process.
		// That means that root broadcasts everything before the other processes
		// have received even half of the data, filling the internal buffers of MPI.
		MPI_Barrier(comm);
	}
	regenerationModelData.broadcast(rootId, comm);
	MPIUtil::broadcast(numberOfReplications, rootId, comm);
	MPIUtil::broadcast(iterationLength, rootId, comm);
	MPIUtil::broadcast(numberOfYears, rootId, comm);
	MPIUtil::broadcast(saveFinalTrees, rootId, comm);
	MPIUtil::broadcast(saveIterationTrees, rootId, comm);
	MPIUtil::broadcast(iterationLoadBalancing, rootId, comm);
	MPIUtil::broadcast(iterationLoadBalancingDetectionPct, rootId, comm);

	compartmentGenerator.broadcast(rootId, comm);

	// Special case for variables that are not actually in SimulationData...
	StandardNormalRandomNumberGenerator::broadcast(rootId, comm);
	UniformRandomNumberGenerator::broadcast(rootId, comm);
}
#endif

#ifdef INCLUDE_ENVIRONMENTAL_EFFECTS
bool SimulationData::hasReproductionEnvironmentalEffects() const {
	for (unsigned int index = 0; index < TreeMarks::NUMBER_OF_MARKS; ++index) {
		if (getReproductionData(index).hasEnvironmentalEffects())
			return true;
	}
	return false;
}

bool SimulationData::hasGrowthEnvironmentalEffects() const {
	for (unsigned int index = 0; index < TreeMarks::NUMBER_OF_MARKS; ++index) {
		if (getGrowthData(index).hasEnvironmentalEffects())
			return true;
	}
	return false;
}

bool SimulationData::hasMortalityEnvironmentalEffects() const {
	for (unsigned int index = 0; index < TreeMarks::NUMBER_OF_MARKS; ++index) {
		if (getMortalityData(index).hasEnvironmentalEffects())
			return true;
	}
	return false;
}
#endif

string SimulationData::getOutputDirectory(bool appendTrailingSlash) const {
	if (appendTrailingSlash && outputDirectory.length() > 0) {
		return outputDirectory + "/";
	} else {
		return outputDirectory;
	}
}

void SimulationData::setOutputDirectory(string const & directory) {
	outputDirectory = directory;
	createOutputDirectory();
	Log::setDirectory(outputDirectory);
}

void SimulationData::loadData(DOMElement const * simulationData) {
	// eUnrecognized denotes an unrecognized tag.
	enum Tags {eUnrecognized, eName, eRNGLimitingFactor, eRNGSeed,
		eNumberOfReplications, eIterationLength, eNumberOfYears, eSaveTreeFiles,
		eCorrelationDistanceFactor,
		eSizeClasses, eAgeClasses, eHeightClasses, eInitialCompartment, eRegenerationParameters,
		eReproductionParameters, eGrowthParameters, eMortalityParameters,
		eIterationLoadBalancing, eOutputDirectory,
		eIgnored};

	map<string, int> tagMap;
	tagMap["name"] = eName;
	tagMap["standardNormalLimitingFactor"] = eRNGLimitingFactor;
	tagMap["rngSeed"] = eRNGSeed;
	tagMap["numberOfReplications"] = eNumberOfReplications;
	tagMap["iterationLength"] = eIterationLength;
	tagMap["numberOfYears"] = eNumberOfYears;
	tagMap["correlationDistanceFactor"] = eCorrelationDistanceFactor;
	tagMap["saveTreeFiles"] = eSaveTreeFiles;
	tagMap["ageClasses"] = eAgeClasses;
	tagMap["sizeClasses"] = eSizeClasses;
	tagMap["heightClasses"] = eHeightClasses;
	tagMap["initialCompartment"] = eInitialCompartment;
	tagMap["regenerationParameters"] = eRegenerationParameters;
	tagMap["reproductionParameters"] = eReproductionParameters;
	tagMap["growthParameters"] = eGrowthParameters;
	tagMap["mortalityParameters"] = eMortalityParameters;
	tagMap["iterationLoadBalancing"] = eIterationLoadBalancing;
	tagMap["outputDirectory"] = eOutputDirectory;
	tagMap["assortmentsOfAllTrees"] = eIgnored;

	for (DOMElement * element = XercesUtilities::getFirstChildDOMElement(simulationData); element != 0; element
			= XercesUtilities::getNextSiblingDOMElement(element)) {
		XercesString tagName = XMLString::replicate(element->getTagName());
		switch (tagMap[tagName.toString()]) {
		case (eName): {
			setSimulationName(XercesUtilities::getFirstChildTextContentAsString(element));
			Log::log(2, "Name: %s", getSimulationName().c_str());
			break;
		}
		case (eRNGLimitingFactor): {
			StandardNormalRandomNumberGenerator::setLimitingFactor(XercesUtilities::getFirstChildTextContentAsDouble(element));
			Log::log(2, "Limiting factor for the standard normal random number generator: %.2f",
					StandardNormalRandomNumberGenerator::getLimitingFactor());
			break;
		}
		case (eRNGSeed): {
			double const seed = XercesUtilities::getFirstChildTextContentAsDouble(element);
			UniformRandomNumberGenerator::storeSeed(seed);
			Log::log(2, "Seed for the random number generator: %.2f",seed);
			break;
		}
		case (eNumberOfReplications): {
			setNumberOfReplications(max(1,XercesUtilities::getFirstChildTextContentAsInt(element)));
			Log::log(2, "Number of replications: %d", getNumberOfReplications());
			break;
		}
		case (eIterationLength): {
			setIterationLength(max(1,XercesUtilities::getFirstChildTextContentAsInt(element)));
			Log::log(2, "Iteration length: %d", getIterationLength());
			break;
		}
		case (eNumberOfYears): {
			setNumberOfYears(max(0,XercesUtilities::getFirstChildTextContentAsInt(element)));
			Log::log(2, "Number of years to simulate: %d", getNumberOfYears());
			break;
		}
		case (eCorrelationDistanceFactor): {
			setCorrelationDistanceFactor(max(1.0, XercesUtilities::getFirstChildTextContentAsDouble(element)));
			Log::log(2, "Local correlation distance is %f times the competition radius", getCorrelationDistanceFactor());
			break;
		}
		case (eSaveTreeFiles): {
			int saveTrees = XercesUtilities::getFirstChildTextContentAsInt(element);
			if (saveTrees>0) {
				setSaveFinalTrees(true);
				int saveIterations = XercesUtilities::getAttributeValueAsInt(element, "iterations");
				if (saveIterations > 0) {
					setSaveIterationTrees(true);
					Log::log(2, "Writing trees to files in every iteration");
				} else {
					Log::log(2, "Writing trees to files at the end of the simulation");
				}
			} else {
				Log::log(2, "Not writing any trees to files");
			}
			break;
		}
		case (eAgeClasses): {
			ageClassHandler.loadData(element);
			break;
		}
		case (eSizeClasses): {
			sizeClassHandler.loadData(element);
			break;
		}
		case (eHeightClasses): {
			heightClassHandler.loadData(element);
			break;
		}
		case (eInitialCompartment): {
			loadCompartmentGenerator(element);
			break;
		}
		case (eRegenerationParameters): {
			loadRegenerationParameters(element);
			break;
		}
		case (eReproductionParameters): {
			Log::log(2,"Loading reproduction parameters");
			loadMarks(element,REPRODUCTION);
			break;
		}
		case (eGrowthParameters): {
			Log::log(2,"Loading growth parameters");
			loadMarks(element,GROWTH);
			break;
		}
		case (eMortalityParameters): {
			Log::log(2,"Loading mortality parameters");
			loadMarks(element,MORTALITY);
			break;
		}
		case (eIterationLoadBalancing): {
			int const iterationBalancing = XercesUtilities::getFirstChildTextContentAsInt(element);
			if(iterationBalancing != 0) {
				iterationLoadBalancing = true;
				iterationLoadBalancingDetectionPct = XercesUtilities::getAttributeValueAsInt(element, "imbalancePct");
				if (iterationLoadBalancingDetectionPct > 0) {
					Log::log(2,"Iteration load balancing enabled using imbalancePct: %.0d",iterationLoadBalancingDetectionPct);
				} else {
					Log::log(2,"Iteration load balancing enabled in every iteration");
				}
			}
			break;
		}
		case (eOutputDirectory): {
			// Do nothing, handled separately
			break;
		}
		case (eIgnored): {
			break;
		}
		default: // Handles eUnrecognized.
		{
			throw "Unrecognized XML-element: "+tagName.toString();
		}
		}
	}
}
void SimulationData::loadMarks(DOMElement const * marks, DataType const type) {
	enum Tags2 {eUnrecognized, eMark};
	map<string, int> tagMap;
	tagMap["mark"] = eMark;

	for (DOMElement * element = XercesUtilities::getFirstChildDOMElement(marks); element != 0; element
			= XercesUtilities::getNextSiblingDOMElement(element)) {
		XercesString tagName = XMLString::replicate(element->getTagName());
		switch (tagMap[tagName.toString()]) {
		case (eMark): {
			string const name = XercesUtilities::getAttributeValueAsString(element, "name");
			unsigned int const markId = TreeMarks::getMarkId(name);
			if(type==REPRODUCTION) {
				reproductionDatas[markId].loadData(element);
			}
			else if(type==MORTALITY) {
				mortalityDatas[markId].loadData(element);
			}
			else if(type==GROWTH) {
				growthDatas[markId].loadData(element);
			}
			break;
		}
		default: // Handles eUnrecognized.
		{
			throw "Unrecognized XML-element: "+tagName.toString()+", expected 'mark name=\"...\"' at this level.";
		}
		}
	}
}

void SimulationData::loadCompartmentGenerator(DOMElement const * initialCompartment) {
	// eUnrecognized denotes an unrecognized tag.
	enum Tags {eUnrecognized, eGenerate, eLoad};

	map<string, int> tagMap;
	tagMap["generate"] = eGenerate;
	tagMap["load"] = eLoad;

	for (DOMElement * element = XercesUtilities::getFirstChildDOMElement(initialCompartment); element != 0; element
			= XercesUtilities::getNextSiblingDOMElement(element)) {
		XercesString tagName = XMLString::replicate(element->getTagName());
		switch (tagMap[tagName.toString()]) {
		case (eGenerate): {
			// Log::log(3, "Will generate new compartment");
			compartmentGenerator.loadData(element);
			break;
		}
		case (eLoad): {
			Log::log(3, "Will load trees from file");
			setInitialStandFilename(XercesUtilities::getFirstChildTextContentAsString(element));
			break;
		}
		default: // Handles eUnrecognized.
		{
			throw "Unrecognized XML-element: "+tagName.toString();
		}
		}
	}
}

void SimulationData::loadRegenerationParameters(DOMElement const * coefficients) {
	// eUnrecognized denotes an unrecognized tag.
	enum Tags { eUnrecognized, eVariable };

	map<string, int> tagMap;
	tagMap["variable"] = eVariable;

	for(DOMElement * element = XercesUtilities::getFirstChildDOMElement(coefficients); element != 0; element = XercesUtilities::getNextSiblingDOMElement(element)) {
		XercesString tagName = XMLString::replicate(element->getTagName());
		switch(tagMap[tagName.toString()])
		{
			case(eVariable):
			{
				string const name = XercesUtilities::getAttributeValueAsString(element,"name");
				unsigned int const variableId = RegenerationModel::getVariableId(name);
				regenerationModelData.setCoefficient(variableId, XercesUtilities::getFirstChildTextContentAsDouble(element));
				break;
			}
			default:	// Handles eUnrecognized.
			{
				throw "Unrecognized XML-element: "+tagName.toString();
			}
		}
	}
}

void SimulationData::createOutputDirectory() {
#ifdef USE_MPI
	if (MPIUtil::mpiId() != 0) {
		return;
	}
#endif
	string directory = getOutputDirectory();
	if (directory.length() == 0) // No output directory used
		return;
	struct stat st;
	if (stat(directory.c_str(), &st) != 0) {
		// The directory doesn't exist. Create it.
		if (mkdir(directory.c_str(), 0777) != 0) {
			throw "Unable to create output directory '" + directory + "'";
		}
	}
}

