#ifndef SIMULATIONDATA_H_
#define SIMULATIONDATA_H_

#include <cassert>
#include <memory>
using std::auto_ptr;
#include <string>
using std::string;
#include <cmath>
using std::ceil;

#include "TreeMarks.h"
#include "SizeClassHandler.h"
#include "AgeClassHandler.h"
#include "HeightClassHandler.h"
#include "GrowthData.h"
#include "MortalityData.h"
#include "ReproductionData.h"
#include "RegenerationModelData.h"
#include "InitialCompartmentGenerator.h"
#ifdef USE_MPI
#include "MPIUtil.h"
#endif

#include <xercesc/util/XercesVersion.hpp>

namespace XERCES_CPP_NAMESPACE
{
	class DOMElement;
}

/**
 * General parameters for a simulation.
 */
class SimulationData
{
public:
	SimulationData();
	~SimulationData();

#ifdef USE_MPI
	void broadcast(int const rootId, MPI_Comm const comm);
#endif
	inline string const & getSimulationName() const {
		return simulationName;
	}
	inline string const & getInitialStandFilename() const {
		return initialStandFilename;
	}
	inline void setSimulationName(string const & name) {
		this->simulationName = name;
	}
	inline void setInitialStandFilename(string const & name) {
		initialStandFilename = name;
	}

	string getOutputDirectory(bool appendTrailingSlash = false) const;
	void setOutputDirectory(string const & directory);

	inline unsigned int getNumberOfReplications() const {
		return numberOfReplications;
	}
	inline unsigned int getNumberOfIterations() const {
		double numIters = ceil( (double)getNumberOfYears() / (double)getIterationLength() );
		return (unsigned int)numIters;
	}
	inline unsigned int getIterationLength() const {
		return iterationLength;
	}
	inline unsigned int getNumberOfYears() const {
		return numberOfYears;
	}
	inline double getCorrelationDistanceFactor() const {
		return correlationDistanceFactor;
	}
	inline void setNumberOfReplications(unsigned int reps) {
		this->numberOfReplications = reps;
	}
	inline void setNumberOfYears(unsigned int years) {
		this->numberOfYears = years;
	}
	inline void setIterationLength(unsigned int length) {
		this->iterationLength = length;
	}
	inline void setCorrelationDistanceFactor(double factor) {
		correlationDistanceFactor = factor;
	}
	inline SizeClassHandler const & getSizeClassHandler() const {
		return sizeClassHandler;
	}
	inline AgeClassHandler const & getAgeClassHandler() const {
		return ageClassHandler;
	}
	inline HeightClassHandler const & getHeightClassHandler() const {
		return heightClassHandler;
	}
	inline bool isSaveFinalTrees() const {
		return saveFinalTrees;
	}
	inline void setSaveFinalTrees(bool s) {
		saveFinalTrees = s;
	}
	inline bool isSaveIterationTrees() const {
		return saveIterationTrees;
	}
	inline void setSaveIterationTrees(bool s) {
		saveIterationTrees = s;
	}
	inline ReproductionData const & getReproductionData(unsigned int mark) const {
		return reproductionDatas[mark];
	}
	inline GrowthData const & getGrowthData(unsigned int mark) const {
		return growthDatas[mark];
	}
	inline MortalityData const & getMortalityData(unsigned int mark) const {
		return mortalityDatas[mark];
	}
	inline RegenerationModelData const & getRegenerationModelData() const {
		return regenerationModelData;
	}
	inline InitialCompartmentGenerator const & getCompartmentGenerator() const {
		return compartmentGenerator;
	}
	inline bool hasForestGrowth() const {
		for (size_t index=0; index < TreeMarks::NUMBER_OF_MARKS; index++) {
			if(growthDatas[index].hasTreeGrowth())
				return true;
		}
		return false;
	}
	inline bool hasForestMortality() const {
		for (size_t index=0; index < TreeMarks::NUMBER_OF_MARKS; index++) {
			if (mortalityDatas[index].hasRegularTreeMortality() || mortalityDatas[index].hasIrregularTreeMortality())
				return true;
		}
		return false;
	}
	inline bool hasIterationLoadBalancing() const {
		return iterationLoadBalancing;
	}
	inline unsigned int getIterationLoadBalancingDetectionPct() const {
		return iterationLoadBalancingDetectionPct;
	}

#ifdef INCLUDE_ENVIRONMENTAL_EFFECTS
	bool hasReproductionEnvironmentalEffects() const;

	bool hasGrowthEnvironmentalEffects() const;

	bool hasMortalityEnvironmentalEffects() const;
#endif

	/**
	 * Loads XML data for the SimulationData.
	 *
	 * Precondition:	@simulationData must not be null.
	 */
	void loadData(XERCES_CPP_NAMESPACE::DOMElement const * simulationData);
private:
	enum DataType { REPRODUCTION, MORTALITY, GROWTH };

	void loadMarks(XERCES_CPP_NAMESPACE::DOMElement const * marks, DataType const type);

	void loadCompartmentGenerator(XERCES_CPP_NAMESPACE::DOMElement const * initialCompartment);

	void loadRegenerationParameters(XERCES_CPP_NAMESPACE::DOMElement const * coefficients);

	void createOutputDirectory();

	string						simulationName;
	string						initialStandFilename;
	string						outputDirectory;
	SizeClassHandler			sizeClassHandler;
	AgeClassHandler				ageClassHandler;
	HeightClassHandler			heightClassHandler;
	vector<GrowthData>			growthDatas;
	vector<MortalityData>		mortalityDatas;
	vector<ReproductionData>	reproductionDatas;
	RegenerationModelData		regenerationModelData;
	InitialCompartmentGenerator	compartmentGenerator;
	unsigned int				numberOfReplications;
	unsigned int				iterationLength;
	unsigned int				numberOfYears;
	bool						saveFinalTrees,saveIterationTrees,iterationLoadBalancing;
	unsigned int 				iterationLoadBalancingDetectionPct;
	double						correlationDistanceFactor;
};

#endif /*SIMULATIONDATA_H_*/
