#ifdef USE_MPI
#include <mpi.h>
#endif

#include <iostream>
#include <exception>
#include <string>
#include "Log.h"

using std::cerr;
using std::cout;
using std::endl;
using std::exception;
using std::string;

#include <gsl/gsl_errno.h>

#include "Simulation.h"
#ifdef WITH_TESTS
#include "TestRunner.h"
#endif

void printLogMessageCount() {
	unsigned int warningMsgs = Log::getMessageCount(Log::WARNING);
	unsigned int errorMsgs = Log::getMessageCount(Log::ERROR);
	unsigned int failureMsgs = Log::getMessageCount(Log::FAILURE);
#ifdef USE_MPI
	MPIUtil::sum(warningMsgs, MPIUtil::getActiveCommunicator());
	MPIUtil::sum(errorMsgs, MPIUtil::getActiveCommunicator());
	MPIUtil::sum(failureMsgs, MPIUtil::getActiveCommunicator());
	if (MPIUtil::mpiId() == 0) {
#endif
		if (warningMsgs > 0 || errorMsgs > 0 || failureMsgs > 0) {
			cout << "There were " << warningMsgs << " ";
			cout << (warningMsgs == 1? "warning": "warnings");
			cout << ", " << errorMsgs << " ";
			cout << (errorMsgs == 1? "error": "errors");
			cout << " and " << failureMsgs << " ";
			cout << (failureMsgs == 1? "failure": "failures");
			cout << ". ";
#ifdef USE_MPI
			cout << "Check the log files for additional information." << endl;
#else
			cout << "Check the log file for additional information." << endl;
#endif
		}
#ifdef USE_MPI
	}
#endif
}

/* Suswood main method */
int main(int argc, char* argv[]) {

#ifdef WITH_TESTS
	if (argc == 2 && strcmp(argv[1],"--test") == 0) {
		TestRunner::runTests();
		exit(0);
	}
#endif
	if (argc == 2 && strcmp(argv[1], "--version") == 0) {
		cout << "SPATE-HPC - v." << VERSION << endl;
		return 0;
	}

	if (argc != 2) {
		cerr << "Usage: spate-hpc <parameterFile>\n";
		exit(1);
	}
	
	// Turn off GSL's default error handler.
	// Return codes are checked explicitly
	gsl_set_error_handler_off();

	try {
		Simulation simulation;
		// Read parameters.
		string parameterFile(argv[1]);
		simulation.loadInputParameters(parameterFile);

		simulation.run();
	} catch (string s) {
		cerr << "Error: " << s << endl;
		cerr << "Simulation aborted" << endl;
		Log::logFailure(-1, s.c_str());
#ifdef USE_MPI
		MPI_Abort(MPI_COMM_WORLD,42);
#endif
	} catch (char const * msg) {
		cerr << "Error: " << msg << endl;
		cerr << "Simulation aborted" << endl;
		Log::logFailure(-1, msg);
#ifdef USE_MPI
		MPI_Abort(MPI_COMM_WORLD,42);
#endif
	} catch (exception const & e) {
		cerr << "Exception: " << e.what() << endl;
		cerr << "Simulation aborted" << endl;
		Log::logFailure(-1, e.what());
#ifdef USE_MPI
		MPI_Abort(MPI_COMM_WORLD,42);
#endif
	} catch (...) {
		cerr << "Unknown error.\n";
		cerr << "Simulation aborted" << endl;
		Log::logFailure(-1, "Unknown error");
#ifdef USE_MPI
		MPI_Abort(MPI_COMM_WORLD,42);
#endif
	}
	printLogMessageCount();
	Log::finalize();
#ifdef USE_MPI
	MPI_Finalize();
#endif

	return 0;
}
