#include "ClearCut.h"
#include "Compartment.h"
#include "CompartmentData.h"
#include "TreeContainer.h"
#include "Tree.h"
#include "Log.h"
#include "PoissonRandomNumberGenerator.h"
#include "classify.h"

#include "XercesString.h"
#include "XercesUtilities.h"
#include <xercesc/util/XMLString.hpp>
#include <xercesc/dom/DOMElement.hpp>
XERCES_CPP_NAMESPACE_USE

#ifdef COMPARISONSEEDING
#include "Timer.h"
#include "PointerLessThan.h"
#include "UniformRandomNumberGenerator.h"
#include "CompartmentCommunicator.h"
#endif

#include <map>
using std::map;
#include <algorithm>
using std::min;
using std::max;
#include <limits>
using std::numeric_limits;
#include <sstream>
using std::ostringstream;
#include <iomanip>
using std::fixed;

ClearCut::ClearCut()
: treesPerHectareLeft(100)
, forcedDistanceLimit(0.0)
, sparedTreesMinDiameter(numeric_limits<double>::min())
, sparedTreesMaxDiameter(numeric_limits<double>::max())
{

}

ClearCut::~ClearCut() {

}

void ClearCut::apply(Compartment * compartment, TreeContainer * treeContainer) {
	CompartmentData const * const compData = compartment->getCompartmentData();
	if(!compData->getCompartmentPolygon()->isActive()) {
		return;
	}

	unsigned int const compartmentId = compData->getIdNumber();
	Log::log(4,"Applying clear cut to compartment id %d",compartmentId);
	vector<Tree*> validTrees;
	getUnmarkedLivingTrees(compartment,validTrees);

#ifdef COMPARISONSEEDING
#ifdef USE_MPI
	vector<Tree*> foreignTrees;
	MPI_Comm comm = compData->getCompartmentCommunicator();
	Timer broadcasting;
	Log::log(5,"Transferring trees for comparisonseeding");
	CompartmentCommunicator::transferTrees(comm,validTrees,foreignTrees);
	Log::log(5,"Transferring took %.2f seconds. Received %d foreign trees.",broadcasting.getElapsed(),foreignTrees.size());
	validTrees.insert(validTrees.end(),foreignTrees.begin(),foreignTrees.end());
#endif
	Log::log(5,"Sorting valid trees");
	std::sort(validTrees.begin(),validTrees.end(),PointerLessThan<Tree>());
	Log::log(5,"Valid trees sorted");

	double const areaInHectares = compartment->getPolygon()->getFullArea() / 10000.0;
#else
	double const areaInHectares = compartment->getPolygon()->getActiveArea() / 10000.0;
#endif
	if (areaInHectares == 0.0) {
		Log::log(5, "Compartment id %d has no active area within this process. No clear cut performed.", compartmentId);
		return;
	}

	double const treesPerHectare = (double)validTrees.size() / areaInHectares;

	if (treesPerHectare > treesPerHectareLeft) {
		TreeContainer treesToKeep(false);
		// Set neighbour distance to at least the forced distance limit.
		treesToKeep.setLimitingValue(forcedDistanceLimit+0.1);
		// Adding a dummy compartment so the rectanglecontainer can work with the polygon.
		Compartment dummy(* compartment->getCompartmentData());
		treesToKeep.addCompartment(&dummy);

		double const preciseNumberOfTreesToKeep = treesPerHectareLeft * areaInHectares;
#ifdef COMPARISONSEEDING
		UniformRandomNumberGenerator::setSeed(compartmentId*41.2);
#endif
		unsigned int const numberOfTreesToKeep = PoissonRandomNumberGenerator::get(preciseNumberOfTreesToKeep);
		float minDiameter, maxDiameter;
		findMinMaxDiameters(validTrees, minDiameter, maxDiameter);

		vector<float> sizeClassBounds;
		setupSizeClassBounds(minDiameter, maxDiameter, sizeClassBounds);

		int largestClass = classify(sizeClassBounds, sparedTreesMaxDiameter);
		// Ensure largestClass doesn't fall out of range
		if (largestClass < 0)
			largestClass = 0;
		if (largestClass >= (int)sizeClassBounds.size() - 1)
			largestClass = (int)sizeClassBounds.size() - 2;

		int smallestClass = classify(sizeClassBounds, sparedTreesMinDiameter);
		// Ensure smallestClass doesn't fall out of range
		if (smallestClass < 0)
			smallestClass = 0;
		if (smallestClass >= (int)sizeClassBounds.size() - 1)
			smallestClass = (int)sizeClassBounds.size() - 2;

		// Loop accepting a larger range of trees in each iteration.
		for (int sizeClass = largestClass; sizeClass >= smallestClass; sizeClass--) {
			float const diameterLimit = max(sizeClassBounds[sizeClass], sparedTreesMinDiameter);
			findTreesToKeep(validTrees, treesToKeep, numberOfTreesToKeep, diameterLimit);
		}
		Log::log(5, "Kept %d trees (out of %d wanted)", treesToKeep.size(), numberOfTreesToKeep);
		if (treesToKeep.size() < numberOfTreesToKeep) {
			Log::log(5, "Could not achieve %d trees per hectare that stand more than %.2f meters apart.", treesPerHectareLeft, forcedDistanceLimit);
			Log::log(5, "Use a lower number of trees per hectare, a lower forced distance between trees or a wider diameter range.");
		}
		markTreesForRemoval(compartment,treesToKeep);
	} else {
		ostringstream ss;
		ss.precision(2);
		ss << "Wanted to keep " << treesPerHectareLeft << " trees per hectare although there are only ";
		ss << fixed << treesPerHectare << " trees per hectare. No clear cut performed.",
		throw ss.str();
	}
#ifdef USE_MPI
#ifdef COMPARISONSEEDING
	unsigned int const numForeignTrees = foreignTrees.size();
	for(unsigned int index=0; index<numForeignTrees; index++) {
		delete foreignTrees[index];
	}
#endif
#endif
}

void ClearCut::findTreesToKeep(vector<Tree*> & validTrees, TreeContainer & treesToKeep, unsigned int const numToKeep, float const dbhLimit) const {
	unsigned int const numValidTrees = validTrees.size();
	for (unsigned int index=0; (index < numValidTrees) && (treesToKeep.size() < numToKeep); index++) {
		Tree * const curTree = validTrees[index];
		float const diameter = curTree->getMark(TreeMarks::DIAMETER);
		// Check if tree is big enough to keep.
		if (diameter >= dbhLimit && diameter < sparedTreesMaxDiameter) {
			bool lonely = true;
			if (forcedDistanceLimit > 0.0) {
				vector<Tree*> const & neighbours = treesToKeep.getTreesCloseTo(curTree);
				unsigned int const numNeighbours = neighbours.size();
				for (unsigned int otherTreeIndex=0; otherTreeIndex < numNeighbours; otherTreeIndex++) {
					Tree * const otherTree = neighbours[otherTreeIndex];
					double const distance = curTree->distanceTo(*otherTree);
					if (distance <= forcedDistanceLimit) {
						lonely = false;
						break;
					}
				}
			}
			if (lonely) {
				// Accept tree as seeding tree
				treesToKeep.add(curTree);
				// "Remove" the tree from the valid trees vector.
				// Erase operation is time consuming and would move the elements of the vector we are looping over.
				validTrees[index] = NULL;
			}
		}
	}
	// Rebuild the validTrees vector, excluding all nullpointers set above.
	Tree * const toDelete = NULL;
	validTrees.erase(remove(validTrees.begin(),validTrees.end(),toDelete),validTrees.end());
}

void ClearCut::markTreesForRemoval(Compartment const * const compartment, TreeContainer & treesToKeep) const {
	// Mark all trees for removal except the ones we should keep
	for(Tree * curTree = compartment->getFirstTree(); curTree!=NULL; curTree = curTree->getNextTreeInCompartment()) {
		bool shouldBeRemoved = true;
		vector<Tree*> const & neighbours = treesToKeep.getTreesCloseTo(curTree);
		unsigned int const numNeighbours = neighbours.size();
		for(unsigned int otherTreeIndex=0; otherTreeIndex < numNeighbours; otherTreeIndex++) {
			Tree * const otherTree = neighbours[otherTreeIndex];
			if(curTree == otherTree) {
				shouldBeRemoved = false;
				break;
			}
		}
		if(shouldBeRemoved) {
			curTree->markForRemoval();
		}
	}
}

void ClearCut::getUnmarkedLivingTrees(Compartment const * const compartment, vector<Tree*> & unMarkedLivingTrees) const {
	for(Tree * curTree = compartment->getFirstTree(); curTree!=NULL; curTree = curTree->getNextTreeInCompartment()) {
		if(!curTree->isMarked() && !curTree->isDead()) {
			unMarkedLivingTrees.push_back(curTree);
		}
	}
}

void ClearCut::findMinMaxDiameters(vector<Tree*> const & validTrees, float & minDiameter, float & maxDiameter) const {
	unsigned int const numberOfValidTrees = validTrees.size();
	minDiameter = numeric_limits<float>::max();
	maxDiameter = numeric_limits<float>::min();
	for(unsigned int index=0; index<numberOfValidTrees; index++) {
		minDiameter = min(minDiameter,validTrees[index]->getMark(TreeMarks::DIAMETER));
		maxDiameter = max(maxDiameter,validTrees[index]->getMark(TreeMarks::DIAMETER));
	}
}

void ClearCut::loadData(DOMElement const * management) {
	// eUnrecognized denotes an unrecognized tag.
	enum Tags { eUnrecognized, eTreesLeft, eForcedDistance, eClassWidthInPercent, eSparedTreesMinDiameter, eSparedTreesMaxDiameter };

	map<string, int> tagMap;
	tagMap["treesLeftPerHectare"] = eTreesLeft;
	tagMap["forcedTreeDistance"] = eForcedDistance;
	tagMap["classWidthInPercent"] = eClassWidthInPercent;
	tagMap["sparedTreesMinDiameter"] = eSparedTreesMinDiameter;
	tagMap["sparedTreesMaxDiameter"] = eSparedTreesMaxDiameter;

	for (DOMElement * element = XercesUtilities::getFirstChildDOMElement(management); element != 0; element = XercesUtilities::getNextSiblingDOMElement(element)) {
		XercesString tagName = XMLString::replicate(element->getTagName());
		switch (tagMap[tagName.toString()])
		{
			case (eTreesLeft):
			{
				treesPerHectareLeft = max(0, XercesUtilities::getFirstChildTextContentAsInt(element));
				break;
			}
			case (eForcedDistance):
			{
				forcedDistanceLimit = max(0.0,XercesUtilities::getFirstChildTextContentAsDouble(element));
				break;
			}
			case (eClassWidthInPercent):
			{
				Log::logWarning(-1, "The parameter classWidthInPercent is obsolete and has no effect.");
				break;
			}
			case (eSparedTreesMinDiameter):
			{
				sparedTreesMinDiameter = XercesUtilities::getFirstChildTextContentAsDouble(element);
				break;
			}
			case (eSparedTreesMaxDiameter):
			{
				sparedTreesMaxDiameter = XercesUtilities::getFirstChildTextContentAsDouble(element);
				break;
			}
			default:	// Handles eUnrecognized.
			{
				loadSizeClassData(element);
			}
		}
	}
}

