#include "LinearTailWeibullFunction.h"
#include "Log.h"

#include <cassert>
#include <cmath>
#include <sstream>
#include <ios>
#include <iomanip>
using namespace std;

LinearTailWeibullFunction::LinearTailWeibullFunction()
: x0(numeric_limits<double>::infinity())
, gradient(0.0)
, yIntercept(0.0) {
}

LinearTailWeibullFunction::LinearTailWeibullFunction(
		double scale, double shape, double origin)
: WeibullFunction(scale, shape, origin)
, x0(numeric_limits<double>::infinity())
, gradient(0.0)
, yIntercept(0.0) {
}

double LinearTailWeibullFunction::value(double x) const {
	if (x < x0) {
		return WeibullFunction::value(x);
	} else {
		return max((x - x0) * gradient + yIntercept, 0.0);
	}
}

double LinearTailWeibullFunction::derivative(double x) const {
	if (x < x0) {
		return WeibullFunction::derivative(x);
	} else if (x >= upperBound()) {
		return 0.0;
	} else {
		return gradient;
	}
}

double LinearTailWeibullFunction::upperBound(double tolerance) const {
	if (gradient == 0.0 || yIntercept == 0.0) {
		return WeibullFunction::upperBound(tolerance);
	} else {
		return x0 - yIntercept / gradient;
	}
}

void LinearTailWeibullFunction::shift(double offset) {
	WeibullFunction::shift(offset);
	x0 += offset;
}

void LinearTailWeibullFunction::linearizeTail(double classWidth, double ratio) {
	if (ratio == numeric_limits<double>::infinity()) {
		x0 = numeric_limits<double>::infinity();
		gradient = 0.0;
		yIntercept = 0.0;
	}
	double const tolerance = 0.0001;
	double const maxX = maximum();
	double x1 = (floor(maxX / classWidth) + 0.5) * classWidth;
	double x2 = x1 + classWidth;
	double slope1 = WeibullFunction::derivative(x1);
	double slope2 = WeibullFunction::derivative(x2);
	for (int i = 0; i < 1000; i++) {
		if (slope1 < 0.0 && slope2 < 0.0 && slope1 / slope2 > ratio) {
			x0 = x1;
			gradient = slope1;
			yIntercept = WeibullFunction::value(x1);
			return;
		}
		if (WeibullFunction::value(x2) <= tolerance) {
			x0 = numeric_limits<double>::infinity();
			gradient = 0.0;
			yIntercept = 0.0;
			return;
		}
		x1 = x2;
		x2 = x2 + classWidth;
		slope1 = slope2;
		slope2 = WeibullFunction::derivative(x2);
	}
	x0 = numeric_limits<double>::infinity();
	gradient = 0.0;
	yIntercept = 0.0;
	Log::logWarning(-1, "Could not find a linear tail to Weibull function. \
Weibull parameters: scale = %.2f, shape = %.2f, origin = %.2f. \
Linear tail ratio = %.2f. \
Last values tested: x1 = %.2f, x2 = %.2f with slope ratio = %.2f/%.2f = %.2f",
			_scale, _shape, _origin, ratio,
			x1, x2, slope1, slope2, slope1 / slope2);
}

string LinearTailWeibullFunction::toString() const {
	ostringstream ostr;
	ostr << WeibullFunction::toString();
	if (x0 == numeric_limits<double>::infinity()) {
		ostr << ", no linear tail";
	} else {
		ostr.precision(2);
		ostr.setf(ios::fixed, ios::floatfield);
		ostr << ", linear tail y = (x " << showpos << -x0 << ") * ";
		ostr.precision(6);
		ostr << gradient << " ";
		ostr.precision(4);
		ostr << showpos << yIntercept;
	}
	return ostr.str();
}

