#include "LowThinning.h"

#include "Compartment.h"
#include "CompartmentData.h"
#include "TreeContainer.h"
#include "Tree.h"
#include "RandomNumber.h"
#include "Compartment.h"
#include "Log.h"
#include "WeibullFunction.h"
#include "classify.h"
#ifdef COMPARISONSEEDING
#include "UniformRandomNumberGenerator.h"
#endif

#ifdef CALCULATION_INSPECTION
#include "CalculationInspector.h"
#endif

#include "XercesString.h"
#include "XercesUtilities.h"
#include <xercesc/util/XMLString.hpp>
#include <xercesc/dom/DOMElement.hpp>
XERCES_CPP_NAMESPACE_USE

#include <map>
using std::map;
#include <algorithm>
using std::max;
#include <cmath>
using std::min;
using std::max;

LowThinning::LowThinning() : WeibullThinning(SHIFT_RIGHT) { 
}

LowThinning::~LowThinning() {

}

void LowThinning::initIterativeThinning(Compartment * compartment, TreeContainer * treeContainer) {
	CompartmentData const * const compData = compartment->getCompartmentData();
	Log::log(4, "Applying low thinning to compartment id %d", compData->getIdNumber());
	sizeClassBounds.clear();
	setupSizeClassBounds(0.0, maxDiameter, sizeClassBounds);
}

void LowThinning::applyIteration(
		Compartment * compartment, TreeContainer * treeContainer,
		WeibullFunction const & weibull1, WeibullFunction const & weibull2,
		double const intersectionPoint) {
#ifdef CALCULATION_INSPECTION
	CalculationInspector& inspector = CalculationInspector::Instance();
#endif
	// Find the largest class to include in the thinning
	int largestClass = classify(sizeClassBounds, intersectionPoint);
	// Ensure largestClass is within bounds
	largestClass = min(max(largestClass, 0), (int)(sizeClassBounds.size() - 2));

	// Apply thinning rules to all thinning classes, beginning from the largest class.
	for (int i = 0; i <= largestClass; ++i) {
#ifdef CALCULATION_INSPECTION
		if (inspector.shouldInspectWeibullIteration())
			inspector.print("  Thinning class %.0f - %.0f:", sizeClassBounds[i], sizeClassBounds[i+1]);
#endif
		vector<Tree*> classTrees; // Trees in the current class (this process only)
		unsigned int numberOfTreesInClass; // Total number of trees in the class (all processes)
		classifyTrees(compartment, sizeClassBounds, i, classTrees, numberOfTreesInClass);
		double const sizeClassCentre = (sizeClassBounds[i] + sizeClassBounds[i+1]) * 0.5;
		double const width = sizeClassBounds[i+1] - sizeClassBounds[i];
		// Calculate the value of the second Weibull curve at the centre of the class.
		// Scale it so its height corresponds to that of the simulated data.
		double const scaledWeibull2 = weibull2.value(sizeClassCentre) * (double)compartment->getTotalSize() * width;
		double const thinningRate = numberOfTreesInClass > 0? 1.0 - scaledWeibull2 / (double)numberOfTreesInClass: 0.0;
#ifdef CALCULATION_INSPECTION
		if (inspector.shouldInspectWeibullIteration())
			inspector.print("    Number of trees in the class = %d. Scaled Weibull2 = %.1f. Thinning rate = %.2f",
				numberOfTreesInClass, scaledWeibull2, thinningRate);
#endif
		if (thinningRate > 0.0) {
			applyLowThinning(compartment, classTrees, thinningRate);
		} else {
#ifdef CALCULATION_INSPECTION
			if (inspector.shouldInspectWeibullIteration())
				inspector.print("    No thinning applied.");
#endif
		}
	}
}

void LowThinning::applyLowThinning(Compartment * compartment,
		vector<Tree*> const & classTrees, double const thinningRate) {
#ifdef CALCULATION_INSPECTION
	CalculationInspector& inspector = CalculationInspector::Instance();
#endif
	unsigned int numberOfTrees = classTrees.size();
	unsigned int numberOfTreesCut = 0;
	// Loop over all trees in the class
	for (unsigned int curTreeIndex = 0; curTreeIndex < numberOfTrees; ++curTreeIndex) {
		Tree * const curTree = classTrees[curTreeIndex];
		if (!curTree->isDead() && !curTree->isMarked()) {
#ifdef COMPARISONSEEDING
			UniformRandomNumberGenerator::setSeeds(curTree,7);
#endif
			double const u = RandomNumber::get();
			if (thinningRate >= u) {
				curTree->markForRemoval();
				numberOfTreesCut++;
			}
		}
	}
#ifdef CALCULATION_INSPECTION
	if (inspector.shouldInspectWeibullIteration())
		inspector.print("    Trees cut: %u out of %u", numberOfTreesCut, numberOfTrees);
#endif
}

void LowThinning::loadData(DOMElement const * management) {
	// eUnrecognized denotes an unrecognized tag.
	enum Tags { eUnrecognized };

	map<string, int> tagMap;

	for (DOMElement * element = XercesUtilities::getFirstChildDOMElement(management); element != 0; element
			= XercesUtilities::getNextSiblingDOMElement(element)) {
		XercesString tagName = XMLString::replicate(element->getTagName());
		switch (tagMap[tagName.toString()]) {
			default: // Handles eUnrecognized.
			{
				loadWeibullData(element);
			}
		}
	}
}

