#include "Management.h"
#include "Compartment.h"
#include "Tree.h"
#include "SpeciesContainer.h"

#include <limits>
using std::numeric_limits;
#include <algorithm>
using std::min;
using std::max;
#include <string>
using std::string;

#include "XercesString.h"
#include "XercesUtilities.h"
#include <xercesc/util/XMLString.hpp>
#include <xercesc/dom/DOMElement.hpp>
XERCES_CPP_NAMESPACE_USE

#include <map>
using std::map;

Management::Management()
: secondaryModelValue(0.0)
, removalPerHectare(0.0)
{
}

Management::~Management()
{
}

void Management::findMinMaxDiameters(Compartment const * const compartment, double & minDiameter, double & maxDiameter) {
	minDiameter = numeric_limits<double>::max();
	maxDiameter = numeric_limits<double>::min();
	for (Tree const * curTree = compartment->getFirstTree(); curTree!=NULL; curTree=curTree->getNextTreeInCompartment()) {
		if (!curTree->isMarked() && !curTree->isDead()) {
			double const diameter = curTree->getMark(TreeMarks::DIAMETER);
			minDiameter = min(minDiameter, diameter);
			maxDiameter = max(maxDiameter, diameter);
		}
	}
#ifdef USE_MPI
	CompartmentData const * const compData = compartment->getCompartmentData();
	assert(compartment->getPolygon()->isActive());
	MPIUtil::min(minDiameter, compData->getCompartmentCommunicator());
	MPIUtil::max(maxDiameter, compData->getCompartmentCommunicator());
#endif
}

void Management::loadCommonData(DOMElement const * common) {
	// eUnrecognized denotes an unrecognized tag.
	enum Tags { eUnrecognized, eSecondaryModelValue, eRemovalPerHectare, eSpeciesToCut };
	
	map<string, int> tagMap;
	tagMap["secondaryModelValue"] = eSecondaryModelValue;
	tagMap["removalPerHectare"] = eRemovalPerHectare;
	tagMap["speciesToCut"] = eSpeciesToCut;

	XercesString tagName = XMLString::replicate(common->getTagName());
	switch(tagMap[tagName.toString()])
	{
		case (eRemovalPerHectare): {
			setRemovalPerHectare(max(0.0, XercesUtilities::getFirstChildTextContentAsDouble(common)));
			break;
		}
		case (eSecondaryModelValue):
		{
			secondaryModelValue = XercesUtilities::getFirstChildTextContentAsDouble(common);
			break;
		}
		case (eSpeciesToCut):
		{
			string name = XercesUtilities::getFirstChildTextContentAsString(common);
			unsigned int id = SpeciesContainer::getId(name);
			speciesToCut.insert(id);
			break;
		}
		default:	// Handles eUnrecognized.
		{
			throw "Unrecognized XML-element: "+tagName.toString();
		}
	}
}

bool Management::shouldCutSpecies(unsigned int speciesId) const {
	if (speciesToCut.empty()) { // If no species are given then all species are cut
		return true;
	}
	return speciesToCut.find(speciesId) != speciesToCut.end();
}

