#include "ManagementTemplate.h"

#include "Management.h"
#include "SelectiveThinning.h"
#include "DimensionCutting.h"
#include "EnergywoodCutting.h"
#include "LowThinning.h"
#include "ClearCut.h"
#include "NullManagement.h"
#include "Compartment.h"
#include "Tree.h"
#include "EvaluatorFunctions.h"
#include "VolumeModel.h"
#include "Log.h"
#include "TemplateContainer.h"
#include "BasalAreaModel.h"

#ifdef USE_MPI
#include "MPIUtil.h"
#endif

#include "XercesString.h"
#include "XercesUtilities.h"
#include <xercesc/util/XMLString.hpp>
#include <xercesc/dom/DOMElement.hpp>
XERCES_CPP_NAMESPACE_USE
#include <map>
using std::map;
#include <cmath>
using std::fmod;
#include <limits>

ManagementTemplate::ManagementTemplate(string const & name)
: method(auto_ptr<Management>(new NullManagement()))
, clearcut(auto_ptr<Management>(new NullManagement()))
, lowerBasalArea(std::numeric_limits<double>::max())
, lowerCompartmentVolume(std::numeric_limits<double>::max())
, thinningIntervalInYears(0)
, yearsBetweenClearCuts(0)
, templateName(name)
{

}

ManagementTemplate::~ManagementTemplate() {

}

void ManagementTemplate::apply(unsigned int iteration, unsigned int iterationLength, Compartment * compartment, TreeContainer * treeContainer) const {
	unsigned int const templateId = TemplateContainer::getId(templateName);

	if (!compartment->getPolygon()->isActive()) {
		return;
	}

	double const fullAreaInHectares = compartment->getPolygon()->getFullArea() / 10000.0;
	if (getMethod()->getObjectType() != eNULLM) {
		// We cannot use the values for basal area and volume stored in the
		// compartment object, since these might include basal area and
		// volume of trees that are dead or removed.
		double totalBasalArea = 0.0;
		double totalVolume = 0.0;
		calculateCharacteristicsOfLivingTrees(compartment, totalBasalArea, totalVolume);

		double const basalAreaPerHectare = totalBasalArea / fullAreaInHectares;
		double const compVolumePerHectare = totalVolume / fullAreaInHectares;

		bool const noThinningYetPerformed = compartment->getManagementIteration(templateId) == 0;
		unsigned int const yearsSinceLastThinning = (iteration - compartment->getManagementIteration(templateId)) * iterationLength;
		if (noThinningYetPerformed || yearsSinceLastThinning >= getThinningIntervalInYears()) {
			if ((basalAreaPerHectare > getLowerBasalArea()) || (compVolumePerHectare > getLowerCompartmentVolume())) {
				Log::log(3, "Compartment %d, template '%s': Basal area is %.2f m^2/ha and volume is %.2f m^3/ha. Applying management.",
						compartment->getCompartmentData()->getIdNumber(),
						templateName.c_str(),
						totalBasalArea / fullAreaInHectares,
						totalVolume / fullAreaInHectares);
				try {
					getMethod()->apply(compartment, treeContainer);
					compartment->setManagementIteration(templateId, iteration);
				} catch (const char * msg) {
					Log::logWarning(-1, msg);
					Log::logWarning(3, "Management in compartment %d failed.",
							compartment->getCompartmentData()->getIdNumber());
				}
			} else {
				Log::log(3, "Compartment %d, template '%s': Basal area is %.2f m^2/ha and volume is %.2f m^3/ha. Postponing management.",
						compartment->getCompartmentData()->getIdNumber(),
						templateName.c_str(),
						totalBasalArea / fullAreaInHectares,
						totalVolume / fullAreaInHectares);
			}
		}
	}

	if (getClearCut()->getObjectType() != eNULLM) {
		double totalBasalArea = 0.0;
		double totalVolume = 0.0;
		calculateCharacteristicsOfLivingTrees(compartment, totalBasalArea, totalVolume);

		double const basalAreaPerHectare = totalBasalArea / fullAreaInHectares;
		double const compVolumePerHectare = totalVolume / fullAreaInHectares;

		unsigned int const initialStandAge = compartment->getCompartmentData()->getInitialStandAge();

		unsigned int yearsSinceLastClearCut = 0;
		if (compartment->getClearCutIteration() == 0) {
			// Clear cut has not previously been carried out.
			yearsSinceLastClearCut = initialStandAge + iteration * iterationLength;
		} else {
			yearsSinceLastClearCut = (iteration - compartment->getClearCutIteration()) * iterationLength;
		}

		if (getYearsBetweenClearCuts() == 0) {
			Log::logWarning(3, "Clear cut specified in management template but the number of years between clear cuts is not specified.");
			Log::logWarning(3, "Not applying clear cut.");
			return;
		}

		if (yearsSinceLastClearCut >= getYearsBetweenClearCuts()) {
			if ((basalAreaPerHectare > getLowerBasalArea()) || (compVolumePerHectare > getLowerCompartmentVolume())) {
				Log::log(3, "Compartment %d, template '%s': Basal area is %.2f m^2/ha and volume is %.2f m^3/ha. Applying clear cut.",
						compartment->getCompartmentData()->getIdNumber(),
						templateName.c_str(),
						totalBasalArea / fullAreaInHectares,
						totalVolume / fullAreaInHectares);
				try {
					getClearCut()->apply(compartment, treeContainer);
					compartment->setClearCutIteration(iteration);
				} catch (const string & msg) {
					Log::logWarning(-1, msg.c_str());
					Log::logWarning(3, "Clear cut in compartment %d failed.",
							compartment->getCompartmentData()->getIdNumber());
				}
			} else {
				Log::log(3, "Compartment %d, template '%s': Basal area is %.2f m^2/ha and volume is %.2f m^3/ha. Postponing clear cut.",
						compartment->getCompartmentData()->getIdNumber(),
						templateName.c_str(),
						totalBasalArea / fullAreaInHectares,
						totalVolume / fullAreaInHectares);
			}
		}
	}
}

double ManagementTemplate::getMethodSecondaryModelValue() const {
	return getMethod()->getSecondaryModelValue();
}

#ifdef USE_MPI
void ManagementTemplate::broadcast(int const rootId, MPI_Comm const comm) {

	MPIUtil::broadcast(lowerBasalArea, rootId, comm);
	MPIUtil::broadcast(lowerCompartmentVolume, rootId, comm);
	MPIUtil::broadcast(templateName, rootId, comm);
	MPIUtil::broadcast(thinningIntervalInYears, rootId, comm);
	MPIUtil::broadcast(yearsBetweenClearCuts, rootId, comm);

	MObjectType mType = method->getObjectType();
	int mTypeAsInt = (int)mType;
	MPIUtil::broadcast(mTypeAsInt, rootId, comm);
	if (MPIUtil::mpiId() != rootId) {
		MObjectType receiveMType = (MObjectType)mTypeAsInt;
		switch (receiveMType) {
		case eSELECTIVEM: {
			method.reset(new SelectiveThinning());
			break;
		}
		case eDIMENSIONM: {
			method.reset(new DimensionCutting());
			break;
		}
		case eENERGYWOODM: {
			method.reset(new EnergywoodCutting());
			break;
		}
		case eLOWTHINM: {
			method.reset(new LowThinning());
			break;
		}
		default: {
			method.reset(new NullManagement());
			break;
		}
		}
	}
	method->broadcast(rootId, comm);

	MObjectType cType = clearcut->getObjectType();
	int cTypeAsInt = (int)cType;
	MPIUtil::broadcast(cTypeAsInt, rootId, comm);
	if (MPIUtil::mpiId() != rootId) {
		MObjectType receiveCType = (MObjectType)cTypeAsInt;
		switch (receiveCType) {
		case eCLEARCUTM: {
			clearcut.reset(new ClearCut());
			break;
		}
		default: {
			clearcut.reset(new NullManagement());
			break;
		}
		}
	}
	clearcut->broadcast(rootId, comm);
}
#endif

void ManagementTemplate::loadData(DOMElement const * managementTemplate) {

	// eUnrecognized denotes an unrecognized tag.
	enum Tags {eUnrecognized, eMethod, eClearCut, eYearsBetweenClearCuts, eLowerBasalArea, eLowerCompartmentVolume, eMinimumVolume, eThinningInterval};

	map<string, int> tagMap;
	tagMap["managementMethod"] = eMethod;
	tagMap["clearCut"] = eClearCut;
	tagMap["yearsBetweenClearCuts"] = eYearsBetweenClearCuts;
	tagMap["lowerBasalArea"] = eLowerBasalArea;
	tagMap["lowerCompartmentVolume"] = eLowerCompartmentVolume;
	tagMap["minimumVolumePerHectare"] = eMinimumVolume;
	tagMap["thinningIntervalInYears"] = eThinningInterval;

	double minimumRemainingVolume = 0.0;
	for (DOMElement * element = XercesUtilities::getFirstChildDOMElement(managementTemplate); element != 0; element
			= XercesUtilities::getNextSiblingDOMElement(element)) {
		XercesString tagName = XMLString::replicate(element->getTagName());
		switch (tagMap[tagName.toString()]) {
		case (eMethod): {
			loadManagementMethod(element);
			break;
		}
		case (eClearCut): {
			Log::log(3, "Includes clear cut");
			clearcut.reset(new ClearCut());
			clearcut->loadData(element);
			break;
		}
		case (eYearsBetweenClearCuts): {
			yearsBetweenClearCuts = XercesUtilities::getFirstChildTextContentAsInt(element);
			break;
		}
		case (eLowerBasalArea): {
			lowerBasalArea = XercesUtilities::getFirstChildTextContentAsDouble(element);
			break;
		}
		case (eLowerCompartmentVolume): {
			lowerCompartmentVolume = XercesUtilities::getFirstChildTextContentAsDouble(element);
			break;
		}
		case (eMinimumVolume): {
			minimumRemainingVolume = XercesUtilities::getFirstChildTextContentAsDouble(element);
			break;
		}
		case (eThinningInterval): {
			thinningIntervalInYears = XercesUtilities::getFirstChildTextContentAsInt(element);
			break;
		}
		default: // Handles eUnrecognized.
		{
			throw "Unrecognized XML-element: "+tagName.toString();
		}
		}
	}

	if (minimumRemainingVolume > 0.0 && lowerCompartmentVolume == std::numeric_limits<double>::max()) {
		lowerCompartmentVolume = minimumRemainingVolume + getMethod()->getRemovalPerHectare();
	}
}

void ManagementTemplate::calculateCharacteristicsOfLivingTrees(Compartment const * compartment,
		double & basalArea, double & volume) const {
	basalArea = 0.0;
	volume = 0.0; 
	for (Tree const * tree = compartment->getFirstTree(); tree != NULL; tree = tree->getNextTreeInCompartment()) {
		if (!tree->isMarked() && !tree->isDead()) {
			basalArea += BasalAreaModel::calculate(*tree);
			volume += VolumeModel::calculateStemVolumeOf(tree);
		}
	}
#ifdef USE_MPI
	CompartmentData const * const compData = compartment->getCompartmentData();
	MPIUtil::sum(basalArea, compData->getCompartmentCommunicator());
	MPIUtil::sum(volume, compData->getCompartmentCommunicator());
#endif
}

void ManagementTemplate::loadManagementMethod(DOMElement const * managementMethod) {
	// eUnrecognized denotes an unrecognized tag.
	enum Tags {eUnrecognized, eSelectiveThinning, eDimensionCutting, eEnergywoodCutting, eLowThinning};

	map<string, int> tagMap;
	tagMap["selectiveThinning"] = eSelectiveThinning;
	tagMap["dimensionCutting"] = eDimensionCutting;
	tagMap["energywoodCutting"] = eEnergywoodCutting;
	tagMap["lowThinning"] = eLowThinning;

	for (DOMElement * element = XercesUtilities::getFirstChildDOMElement(managementMethod); element != 0; element
			= XercesUtilities::getNextSiblingDOMElement(element)) {
		XercesString tagName = XMLString::replicate(element->getTagName());
		switch (tagMap[tagName.toString()]) {
		case eSelectiveThinning: {
			Log::log(3, "Selective thinning");
			method.reset(new SelectiveThinning());
			break;
		}
		case eDimensionCutting: {
			Log::log(3, "Dimension cutting");
			method.reset(new DimensionCutting());
			break;
		}
		case eEnergywoodCutting: {
			Log::log(3, "Energy wood cutting");
			method.reset(new EnergywoodCutting());
			break;
		}
		case eLowThinning: {
			Log::log(3, "Low thinning");
			method.reset(new LowThinning());
			break;
		}
		default: // Handles eUnrecognized.
		{
			throw "Unrecognized XML-element: "+tagName.toString();
		}
		}
		method->loadData(element);
	}
}
