#include "SelectiveThinning.h"
#include "TreeContainer.h"
#include "RandomNumber.h"
#include "Polygon.h"
#include "Compartment.h"
#include "classify.h"
#include "CompartmentData.h"
#include "Tree.h"
#include "Log.h"
#include "LimitFinder.h"
#include "WeibullFunction.h"
#include "PoissonRandomNumberGenerator.h"
#ifdef COMPARISONSEEDING
#include "UniformRandomNumberGenerator.h"
#endif

#ifdef CALCULATION_INSPECTION
#include "CalculationInspector.h"
#endif

#include "XercesString.h"
#include "XercesUtilities.h"
#include <xercesc/util/XMLString.hpp>
#include <xercesc/dom/DOMElement.hpp>
XERCES_CPP_NAMESPACE_USE

#include <cmath>
#include <map>
#include <utility>
#include <iterator>
#include <algorithm>
#include <functional>
using namespace std;

SelectiveThinning::SelectiveThinning()
: WeibullThinning(SHIFT_LEFT)
, thinningLevel(0.5), densityCoefficient(100.0)
{
}

SelectiveThinning::~SelectiveThinning()
{
}

void SelectiveThinning::initIterativeThinning(Compartment * compartment, TreeContainer * treeContainer) {
	CompartmentData const * const compData = compartment->getCompartmentData();
	Log::log(4,"Applying selective thinning (Stochastic thinning with fixed probabilities) to compartment id %d",compData->getIdNumber());
	sizeClassBounds.clear();
	setupSizeClassBounds(0.0, maxDiameter, sizeClassBounds);
}

void SelectiveThinning::applyIteration(
		Compartment * compartment, TreeContainer * treeContainer,
		WeibullFunction const & weibull1, WeibullFunction const & weibull2,
		double const intersectionPoint) {
#ifdef CALCULATION_INSPECTION
	CalculationInspector& inspector = CalculationInspector::Instance();
#endif

	// Find the smallest class to include in the thinning
	int smallestClass = classify(sizeClassBounds, intersectionPoint);
	// Ensure smallestClass is within bounds
	smallestClass = min(max(smallestClass, 0), (int)(sizeClassBounds.size() - 2));

	// Stochastic thinning with fixed probabilities.
	// Apply thinning rules to all thinning classes, beginning from the largest class.
	for (int i = sizeClassBounds.size() - 2; i >= smallestClass; --i) {
#ifdef CALCULATION_INSPECTION
		if (inspector.shouldInspectWeibullIteration())
			inspector.print("  Thinning class %.0f - %.0f:", sizeClassBounds[i], sizeClassBounds[i+1]);
#endif
		vector<Tree*> classTrees; // Trees in the current class (this process only)
		unsigned int numberOfTreesInClass; // Total number of trees in the class (all processes)
		classifyTrees(compartment, sizeClassBounds, i, classTrees, numberOfTreesInClass);
		double const sizeClassCentre = (sizeClassBounds[i] + sizeClassBounds[i+1]) * 0.5;
		double const width = sizeClassBounds[i+1] - sizeClassBounds[i];
		// Calculate the value of the second Weibull curve at the centre of the class.
		// Scale it so its height corresponds to that of the simulated data.
		double const scaledWeibull2 = weibull2.value(sizeClassCentre) * (double)compartment->getTotalSize() * width;
		double const thinningRate = numberOfTreesInClass > 0? 1.0 - scaledWeibull2 / (double)numberOfTreesInClass: 0.0;
#ifdef CALCULATION_INSPECTION
		if (inspector.shouldInspectWeibullIteration())
			inspector.print("    Number of trees in the class = %d. Scaled Weibull2 = %.1f. Thinning rate = %.2f",
				numberOfTreesInClass, scaledWeibull2, thinningRate);
#endif
		if (thinningRate > 0.0) {
			applyStochasticThinningWithFixedProbabilities(compartment, treeContainer, classTrees, thinningRate);
		} else {
#ifdef CALCULATION_INSPECTION
			if (inspector.shouldInspectWeibullIteration())
				inspector.print("    No thinning applied.");
#endif
		}
	}
}

void SelectiveThinning::applyStochasticThinningWithFixedProbabilities(
		Compartment * compartment, TreeContainer * trees,
		vector<Tree*> const & classTrees, double const thinningRate)
{
#ifdef CALCULATION_INSPECTION
	CalculationInspector& inspector = CalculationInspector::Instance();
#endif

	LimitFinder & limitFinder = LimitFinder::Instance();
	limitFinder.setLimitingMode(limitedByCompRad);
	double const competitionRadius = limitFinder.getLargestLimitingValue();
	trees->setLimitingValue(competitionRadius);
	double const squaredCompetitionRadius = competitionRadius*competitionRadius;

	double const standDensity = (double)compartment->getTotalSize() / compartment->getPolygon()->getFullArea();

	unsigned int numberOfTrees = classTrees.size();
	vector< pair<double, Tree*> > orderedTrees(numberOfTrees);
	// Loop over all trees in the class and associate each one with a thinning probability
	for (unsigned int curTreeIndex = 0; curTreeIndex < numberOfTrees; ++curTreeIndex) {
		Tree * const curTree = classTrees[curTreeIndex];
		unsigned int numberOfNeighbours = 0;
		vector<Tree*> neighbours = trees->getTreesCloseTo(curTree);
		size_t const numNeighbours = neighbours.size();
		for (size_t index=0; index < numNeighbours; index++) {
			Tree const * const otherTree = neighbours[index];
			if (!otherTree->isMarked() && !otherTree->isDead()) {
				double const squaredTreeDistance = curTree->squaredDistanceTo(*otherTree);
				if (0.0 < squaredTreeDistance && squaredTreeDistance <= squaredCompetitionRadius) {
					numberOfNeighbours++;
				}
			}
		}
		double const localDensity = (double)(numberOfNeighbours + 1) / (M_PI * squaredCompetitionRadius);

		double const exponent = thinningLevel + densityCoefficient * log(localDensity/standDensity);
		double const thinningProbability = 1.0 / (1.0 + exp(-exponent));
		orderedTrees[curTreeIndex] = pair<double, Tree*>(thinningProbability, curTree);
#ifdef CALCULATION_INSPECTION
		if (inspector.shouldInspectWeibullProbability()) {
			inspector.print("      Tree at (%.2f, %.2f) has thinning probability logit(%f) = %.2f + %.2f * log(%.2f/%.2f)",
					curTree->getX(), curTree->getY(), thinningProbability,
					thinningLevel, densityCoefficient, localDensity, standDensity);
		}
#endif
	}
	// Sort the trees in descending order according to thinning probability.
	sort(orderedTrees.begin(), orderedTrees.end(), std::greater< pair<double, Tree*> >());

	double const numberOfTreesToCutMean = thinningRate * (double)numberOfTrees;
	unsigned int numberOfTreesToCut = PoissonRandomNumberGenerator::get(numberOfTreesToCutMean);
	unsigned int numberOfTreesCut = 0;
	// Loop numberOfTreesToCut times. At each iteration, take the tree with
	// the highest thinning probability and compare it to a uniform random
	// number. If the tree is cut, move on to the next iteration.
	// Otherwise test the tree with the second highest thinning probability
	// etc.
	for (unsigned int i = 0; i < numberOfTreesToCut; ++i) {
		for (unsigned int curTreeIndex = 0; curTreeIndex < numberOfTrees; ++curTreeIndex) {
			double const thinningProbability = orderedTrees[curTreeIndex].first;
			Tree * const curTree = orderedTrees[curTreeIndex].second;
			if (curTree->isDead() || curTree->isMarked()) // This tree is dead or already marked for removal
				continue;
#ifdef COMPARISONSEEDING
			UniformRandomNumberGenerator::setSeeds(curTree,13.37);
#endif
			double const u = RandomNumber::get();
			if (thinningProbability >= u) {
				curTree->markForRemoval();
				numberOfTreesCut++;
				break;
			}
		}

	}
#ifdef CALCULATION_INSPECTION
	if (inspector.shouldInspectWeibullIteration())
		inspector.print("    Trees to cut (mean %.1f): %u. Trees actually cut: %u out of %u",
				numberOfTreesToCutMean, numberOfTreesToCut, numberOfTreesCut, numberOfTrees);
#endif
}

void SelectiveThinning::loadData(DOMElement const * management)
{
	// eUnrecognized denotes an unrecognized tag.
	enum Tags { eUnrecognized, eThinningLevel, eDensityCoefficient };

	map<string, int> tagMap;
	tagMap["thinningLevel"] = eThinningLevel;
	tagMap["densityCoefficient"] = eDensityCoefficient;

	for(DOMElement * element = XercesUtilities::getFirstChildDOMElement(management); element != 0; element = XercesUtilities::getNextSiblingDOMElement(element)) {
		XercesString tagName = XMLString::replicate(element->getTagName());
		switch(tagMap[tagName.toString()])
		{
			case(eThinningLevel):
			{
				thinningLevel = XercesUtilities::getFirstChildTextContentAsDouble(element);
				break;
			}
			case(eDensityCoefficient): {
				densityCoefficient = XercesUtilities::getFirstChildTextContentAsDouble(element);
				break;
			}
			default:	// Handles eUnrecognized.
			{
				loadWeibullData(element);
			}
		}
	}
}

