#include "SizeClassThinning.h"
#include "Compartment.h"
#include "Tree.h"
#include "classify.h"
#include "Management.h"
#include "Log.h"
#include <cassert>
#include <cmath>

#include "XercesString.h"
#include "XercesUtilities.h"
#include <xercesc/util/XMLString.hpp>
#include <xercesc/dom/DOMElement.hpp>
XERCES_CPP_NAMESPACE_USE

SizeClassThinning::SizeClassThinning() :
	classWidthInCm(4.0) {

}

SizeClassThinning::~SizeClassThinning() {
}

void SizeClassThinning::loadSizeClassData(DOMElement const * element) {
	// eUnrecognized denotes an unrecognized tag.
	enum Tags { eUnrecognized, eClassWidthInCm };
	
	map<string, int> tagMap;
	tagMap["classWidthInCm"] = eClassWidthInCm;

	XercesString tagName = XMLString::replicate(element->getTagName());
	switch(tagMap[tagName.toString()])
	{
		case (eClassWidthInCm): {
			double classWidth = XercesUtilities::getFirstChildTextContentAsDouble(element);
			if (classWidth <= 0.0) {
				Log::logError(-1, "Class width must be a positive number, %.2f given", classWidth);
			}
			classWidthInCm = classWidth;
			break;
		}
		default:	// Handles eUnrecognized.
		{
			loadCommonData(element);
		}
	}

}

void SizeClassThinning::setupSizeClassBounds(double minDiameter, double maxDiameter, vector<float> & sizeClassBounds) const {
	assert(classWidthInCm > 0.0);
	// Offset the maximum diameter so we get an interval evenly divisible
	// by the class width
	maxDiameter = std::ceil((maxDiameter - minDiameter) / classWidthInCm) * classWidthInCm;
	sizeClassBounds.push_back(minDiameter);
	for (double diameter = minDiameter + classWidthInCm; diameter <= maxDiameter; diameter += classWidthInCm) {
		sizeClassBounds.push_back(diameter);
	}
	if (sizeClassBounds.size() == 1) {
		// There must be at least two size class bounds
		sizeClassBounds.push_back(maxDiameter);
	}
}

void SizeClassThinning::setupSizeClassBounds(Compartment const * const compartment, vector<float> & sizeClassBounds) const {
	double minDiameter, maxDiameter;
	Management::findMinMaxDiameters(compartment, minDiameter, maxDiameter);
	setupSizeClassBounds(minDiameter, maxDiameter, sizeClassBounds);
}

void SizeClassThinning::classifyTrees(
		Compartment const * const compartment,
		vector<float> const & sizeClassBounds,
		int sizeClass,
		vector<Tree*> & treesInClass,
		unsigned int & totalNumberOfTreesInClass) const {
	totalNumberOfTreesInClass = 0;
	treesInClass.clear();
	for (Tree * curTree = compartment->getFirstTree(); curTree != NULL; curTree = curTree->getNextTreeInCompartment()) {
		if (!curTree->isDead() && !curTree->isMarked()) {
			if (classify(sizeClassBounds, curTree->getMark(TreeMarks::DIAMETER)) == sizeClass) {
				totalNumberOfTreesInClass++;
				treesInClass.push_back(curTree);
			}
		}
	}

#ifdef USE_MPI
	CompartmentData const * const compData = compartment->getCompartmentData();
	MPIUtil::sum(totalNumberOfTreesInClass, compData->getCompartmentCommunicator());
#endif
}

