#include "TemplateContainer.h"
#include "ManagementTemplate.h"
#include "Log.h"

#ifdef USE_MPI
#include "MPIUtil.h"
#endif

#include "XercesString.h"
#include "XercesUtilities.h"
#include <xercesc/util/XMLString.hpp>
#include <xercesc/dom/DOMElement.hpp>
XERCES_CPP_NAMESPACE_USE
#include <map>
using std::map;

vector<ManagementTemplate*> TemplateContainer::managementTemplates;

void TemplateContainer::Destroy() {
	unsigned int const numTemplates = managementTemplates.size();
	for(unsigned int index=0; index<numTemplates; index++) {
		delete managementTemplates[index];
	}
	managementTemplates.clear();
}

#ifdef USE_MPI
void TemplateContainer::broadcast(int const rootId, MPI_Comm const comm) {
	unsigned int numTemplates = managementTemplates.size();
	MPIUtil::broadcast(numTemplates,rootId,comm);
	Log::log(1,"Number of management templates %d",numTemplates);
	for(unsigned int index=0; index<numTemplates; index++) {
		ManagementTemplate * managementTemplate;
		Log::log(2,"Broadcasting management template index %d",index);
		if(MPIUtil::mpiId()==rootId) {
			managementTemplate = managementTemplates[index];
			string name = managementTemplate->getName();
			MPIUtil::broadcast(name, rootId, comm);
			managementTemplate->broadcast(rootId,comm);
		}
		else {
			string name;
			MPIUtil::broadcast(name, rootId, comm);
			managementTemplate = new ManagementTemplate(name);
			managementTemplate->broadcast(rootId,comm);
			managementTemplates.push_back(managementTemplate);
		}
	}
}
#endif

void TemplateContainer::add(ManagementTemplate * managementTemplate) {
	if(managementTemplate == NULL) {
		Log::logError(-1,"Trying to add a NULL managementTemplate to the TemplateContainer!");
	}
	else {
		string const & name = managementTemplate->getName();
		if(silentFind(name) == NULL) {
			managementTemplates.push_back(managementTemplate);
		}
		else {
			Log::logWarning(-1,"Management template with name %s already loaded from input file! The later one is not added.",name.c_str());
		}
	}
}

ManagementTemplate * TemplateContainer::get(string const & name) {
	ManagementTemplate * managementTemplate = silentFind(name);
	if(managementTemplate==NULL) {
		Log::logWarning(-1,"No management template named %s loaded from input file, setting up a NullManagement",name.c_str());
		managementTemplate = new ManagementTemplate(name);
		add(managementTemplate);
	}
	return managementTemplate;
}
ManagementTemplate * TemplateContainer::silentFind(string const & name) {
	unsigned int const numTemplates = managementTemplates.size();
	for(unsigned int index=0; index<numTemplates; index++) {
		ManagementTemplate * const curTemplate = managementTemplates[index];
		if(curTemplate->getName() == name)
			return curTemplate;
	}
	return NULL;
}
unsigned int TemplateContainer::getId(string const & name) {
	// This function makes sure a management template with this name exists.
	ManagementTemplate * managementTemplate = get(name);

	// Find the id of the management template
	unsigned int const numTemplates = size();
	for(unsigned int index=0; index<numTemplates; index++) {
		if(managementTemplates[index]->getName() == managementTemplate->getName())
			return index;
	}
	throw "management template "+name+" not found!";
}

void TemplateContainer::loadData(DOMElement const * templates) {

	// eUnrecognized denotes an unrecognized tag.
	enum Tags {eUnrecognized, eManagementTemplate};

	map<string, int> tagMap;
	tagMap["managementTemplate"] = eManagementTemplate;

	for (DOMElement * element = XercesUtilities::getFirstChildDOMElement(templates); element != 0; element
			= XercesUtilities::getNextSiblingDOMElement(element)) {
		XercesString tagName = XMLString::replicate(element->getTagName());
		switch (tagMap[tagName.toString()]) {
		case (eManagementTemplate): {
			string const name = XercesUtilities::getAttributeValueAsString(element,"name");
			Log::log(2, "Loading a management template with name %s",name.c_str());
			ManagementTemplate * managementTemplate = new ManagementTemplate(name);
			managementTemplate->loadData(element);
			TemplateContainer::add(managementTemplate);
			break;
		}
		default: // Handles eUnrecognized.
		{
			throw "Unrecognized XML-element: "+tagName.toString();
		}
		}
	}

}
