#include "WeibullFunction.h"
#include "Log.h"

#include <cassert>
#include <cmath>
#include <sstream>
#include <ios>
#include <iomanip>
using namespace std;

WeibullFunction::WeibullFunction()
: _scale(0.0)
, _shape(0.0)
, _origin(0.0) {
}

WeibullFunction::WeibullFunction(double scale, double shape, double origin)
: _scale(scale)
, _shape(shape)
, _origin(origin) {
}

WeibullFunction::~WeibullFunction() {
}

double WeibullFunction::value(double x) const {
	double const k = _shape;
	double const l = _scale;
	double const xs = x - _origin;
	assert(k > 0.0);
	assert(l > 0.0);
	if (xs < 0.0 || (xs == 0.0 && k < 1.0)) {
		return 0.0;
	}
	double const tmp = pow(xs / l, k);
	double const value = (k/l) * tmp * l/xs * exp(-tmp);
	assert(value >= 0.0);
	return value;
}

double WeibullFunction::maximum() const {
	if (_shape > 1.0) {
		// Root of derivative
		return _scale * pow((_shape - 1.0) / _shape, 1.0 / _shape) + _origin;
	} else if (_shape == 1.0) {
		return _origin;
	} else {
		// Infinite at x = origin
		double const tolerance = 0.0001;
		return _origin + tolerance;
	}
}

double WeibullFunction::derivative(double x) const {
	double const k = _shape;
	double const l = _scale;
	double const xs = x - _origin;
	assert(k > 0.0);
	assert(l > 0.0);
	if (xs <= 0.0) {
		return 0.0;
	}
	double const tmp = pow(xs / l, k);
	return ( (k*k-k)/(xs*xs) * tmp - k*k/(xs*xs) * tmp*tmp ) * exp(-tmp);
}

double WeibullFunction::upperBound(double tolerance) const {
	return _scale * pow(-log(tolerance * 0.1), 1.0 / _shape) + _origin;
}

void WeibullFunction::shift(double offset) {
	_origin += offset;
}

string WeibullFunction::toString() const {
	ostringstream ostr;
	ostr.precision(2);
	ostr.setf(ios::fixed, ios::floatfield);
	ostr << "scale = " << _scale << ", ";
	ostr << "shape = " << _shape << ", ";
	ostr << "origin = " << _origin;
	return ostr.str();
}

