#include "WeibullThinning.h"
#include "Compartment.h"
#include "TreeContainer.h"
#include "Tree.h"
#include "Log.h"
#include "Management.h"
#include "VolumeModel.h"
#include "RandomNumber.h"
#include "LinearTailWeibullFunction.h"

#ifdef CALCULATION_INSPECTION
#include "CalculationInspector.h"
#endif

#include "XercesString.h"
#include "XercesUtilities.h"
#include <xercesc/util/XMLString.hpp>
#include <xercesc/dom/DOMElement.hpp>
XERCES_CPP_NAMESPACE_USE

#include <stdint.h>
#include <algorithm>
#include <cassert>
#include <cmath>
#include <limits>
using namespace std;

WeibullThinning::WeibullThinning(Direction direction)
: direction(direction)
, removalTolerance(0.05)
, linearTailRatio(numeric_limits<double>::infinity()) {
}

void WeibullThinning::apply(Compartment * compartment, TreeContainer * treeContainer) {
	CompartmentData const * const compData = compartment->getCompartmentData();
	if (!compData->getCompartmentPolygon()->isActive()) {
		return;
	}

	findMinMaxDiameters(compartment, minDiameter, maxDiameter);
	initIterativeThinning(compartment, treeContainer);
	LinearTailWeibullFunction weibull1;
	try {
		WeibullThinning::estimateWeibullFromSimulated(compartment, weibull1);
	} catch (char const * msg) {
		Log::logWarning(4, "Failed to estimate a Weibull function from simulated data. Thinning not applied!");
		return;
	}
	applyThinningIteratively(compartment, treeContainer, weibull1);
}

void WeibullThinning::loadWeibullData(XERCES_CPP_NAMESPACE::DOMElement const * element) {
	// eUnrecognized denotes an unrecognized tag.
	enum Tags { eUnrecognized, eThinningMargin, eSlopeRatioLimit };

	map<string, int> tagMap;
	tagMap["removalTolerance"] = eThinningMargin;
	tagMap["linearTailRatio"] = eSlopeRatioLimit;

	XercesString tagName = XMLString::replicate(element->getTagName());
	switch (tagMap[tagName.toString()]) {
		case (eThinningMargin):
		{
			removalTolerance = XercesUtilities::getFirstChildTextContentAsDouble(element);
			if (removalTolerance <= 0.0 || removalTolerance >= 1.0) {
				Log::logWarning(-1, "Removal tolerance out of range: %.2f. Using default (0.05).", removalTolerance);
				removalTolerance = 0.05;
			}
			break;
		}
		case (eSlopeRatioLimit):
		{
			linearTailRatio = XercesUtilities::getFirstChildTextContentAsDouble(element);
			break;
		}
		default: // Handles eUnrecognized.
		{
			loadSizeClassData(element);
		}
	}
}

void WeibullThinning::applyThinningIteratively(Compartment * compartment, TreeContainer * treeContainer,
		LinearTailWeibullFunction const & weibull1) {
	LinearTailWeibullFunction weibull2 = weibull1;

#ifdef CALCULATION_INSPECTION
	CalculationInspector& inspector = CalculationInspector::Instance();
#endif

	double const totalArea = compartment->getPolygon()->getFullArea() / 10000.0;
	double wantedRemoval = removalPerHectare * totalArea;
	if (wantedRemoval == 0.0) {
		Log::log(5, "No thinning applied.");
		return;
	}

	double const totalVolume = compartment->getTotalVolume();
	if (totalVolume == 0.0) {
		Log::log(5, "Total compartment volume is 0.0 m^3. Cannot perform thinning.");
		return;
	}
	if (wantedRemoval > totalVolume) wantedRemoval = totalVolume;
	double const absoluteMargin = wantedRemoval * removalTolerance;
	Log::log(5, "Want to cut %.2f +/- %.2f m^3 (%.2f +/- %.2f m^3/ha).",
			wantedRemoval, absoluteMargin,
			wantedRemoval / totalArea, absoluteMargin / totalArea);

	double intersectionPoint;
	unsigned int interSectionPointTests = 0;
	unsigned int thinningIterations = 0;

	// Use the bisection method to find a new origin value which removes the
	// target volume.
	// Start with the interval given by 'left'--'right', where 'left' removes
	// smallest volume and 'right' removes the largest.
	double left = 0.0, right = 0.0, midpoint = 0.0;
	double volume = 0.0;
	switch (direction) {
	case SHIFT_LEFT:
		left = weibull1.origin();
		right = weibull1.origin() - (maxDiameter - weibull1.origin());
		break;
	case SHIFT_RIGHT:
		left = weibull1.origin();
		right = maxDiameter;
		break;
	}
	// The initial midpoint is the shifted location which gives an overlapping area
	// corresponding to the percentage we want to remove.
	// This might or might not be a good starting point.
	Log::log(5, "Initial guess:");
	guessInitialSecondWeibull(compartment, wantedRemoval, weibull1, weibull2);
	midpoint = weibull2.origin();
	for (thinningIterations = 1; thinningIterations < 21; thinningIterations++) {
#ifdef CALCULATION_INSPECTION
		if (inspector.shouldInspectWeibullIteration())
			inspector.print("Thinning iteration %d:", thinningIterations);
#endif
		if (thinningIterations > 1)
			discardThinningResult(compartment);
		weibull2.shift(midpoint - weibull2.origin());
		try {
			intersectionPoint = findIntersectionPoint(weibull1, weibull2, interSectionPointTests);
		} catch (char const * msg) {
			intersectionPoint = weibull1.origin();
		}
#ifdef CALCULATION_INSPECTION
		if (inspector.shouldInspectWeibullIteration()) {
			inspector.print("  Shifted Weibull curve: %s", weibull2.toString().c_str());
			inspector.print("  Intersection point = %.2f", intersectionPoint);
			inspector.print("  Bisection method: left = %.2f, right = %.2f, midpoint = %.2f", left, right, midpoint);
		}
#endif
		applyIteration(compartment, treeContainer, weibull1, weibull2, intersectionPoint);
		volume = calculateThinningVolume(compartment);
#ifdef CALCULATION_INSPECTION
		if (inspector.shouldInspectWeibullIteration())
			inspector.print("  Resulting thinning volume = %.2f m^3", volume);
#endif
		if (abs(volume - wantedRemoval) <= absoluteMargin) {
			// Resulting volume is close enough to target
			break;
		} else if (abs(left - right) < 0.05) {
			// The interval is very narrow. It is unlikely we'll ever reach the target volume.
			if (volume < wantedRemoval) {
				Log::logWarning(5, "Could not reach the target volume. Accepting a smaller one.");
				break;
			} else {
				discardThinningResult(compartment);
				throw "Could not reach the target volume. Giving up.";
			}
		} else if (volume < wantedRemoval) {
			left = midpoint;
		} else {
			right = midpoint;
		}
		midpoint = (left + right) * 0.5;
	}
	if (thinningIterations == 21) {
		Log::logWarning(5, "Could not reach the target volume after 20 iterations. Giving up.");
		discardThinningResult(compartment);
	} else {
		Log::log(5, "Accepting the removal volume %.2f m^3 (%.2f m^3/ha).", volume, volume / totalArea);
		Log::log(5, "Estimated a new weibull function with %s", weibull2.toString().c_str());
		Log::log(5, "Intersection point = %.2f.", intersectionPoint);
		Log::log(5, "Performed %u thinning iterations and tested %u intersection points.", thinningIterations, interSectionPointTests);
	}
}

void WeibullThinning::initIterativeThinning(Compartment * compartment, TreeContainer * treeContainer) {
}

void WeibullThinning::calculateTreeVolumes(Compartment const * compartment, vector<double> & volumes) const {
	size_t const numberOfClasses = 100;
	volumes.resize(numberOfClasses, 0.0f);
	double const classWidth = (maxDiameter - minDiameter) / (double)numberOfClasses;
	for (Tree const * tree = compartment->getFirstTree(); tree != NULL; tree = tree->getNextTreeInCompartment()) {
		if (!tree->isMarked() && !tree->isDead()) {
			double const t = floor((tree->getMark(TreeMarks::DIAMETER) - minDiameter) / classWidth);
			size_t const c = min((size_t)max(t, 0.0), numberOfClasses - 1);
			volumes[c] += VolumeModel::calculateStemVolumeOf(tree);
		}
	}
#ifdef USE_MPI
	CompartmentData const * const compData = compartment->getCompartmentData();
	MPIUtil::sum(volumes, compData->getCompartmentCommunicator());
#endif
}

void WeibullThinning::estimateWeibullFromSimulated(
		Compartment const * const compartment, LinearTailWeibullFunction & weibull) const {
	uint64_t const totalTrees = compartment->getTotalSize();
	// calculate mean and min diameter
	double meanDiam = 0.0;
	double squaredDiamSum = 0.0;
	for (Tree const * curTree = compartment->getFirstTree(); curTree!=NULL; curTree=curTree->getNextTreeInCompartment()) {
		double const diameter = curTree->getMark(TreeMarks::DIAMETER);
		meanDiam += diameter;
		squaredDiamSum += diameter*diameter;
	}
#ifdef USE_MPI
	assert(compartment->getPolygon()->isActive());
	CompartmentData const * const compData = compartment->getCompartmentData();
	MPIUtil::sum(meanDiam, compData->getCompartmentCommunicator());
	MPIUtil::sum(squaredDiamSum, compData->getCompartmentCommunicator());
#endif
	meanDiam /= (double)totalTrees;

	// calculate stdev
	double const stdev = sqrt( (squaredDiamSum / ((double)totalTrees - 1.0)) - (double)totalTrees / ((double)totalTrees - 1.0) * meanDiam*meanDiam );

	Log::log(5, "Estimating weibull function with input parameters: mean = %.3f, stdev = %.3f, min = %.3f", meanDiam,
			stdev, minDiameter);
	double scale = 0.0;
	double shape = 0.0;
	double origin = 0.0;
	bool const success = weibullEstimation(meanDiam, stdev, minDiameter, scale, shape, origin);

	if (!success) {
		Log::logWarning(5, "Estimation of weibull function failed! Input parameters out of range. mean = %.3f, stddev = %.3f, min = %.3f", meanDiam, stdev, minDiameter);
		throw "Weibull estimation failed. Input parameters out of range.";
	}
	weibull = LinearTailWeibullFunction(scale, shape, origin);
	weibull.linearizeTail(classWidthInCm, linearTailRatio);
	Log::log(5, "Estimated a weibull function with %s", weibull.toString().c_str());
}

void WeibullThinning::guessInitialSecondWeibull(
		Compartment const * const compartment,
		double const wantedRemoval,
		WeibullFunction const & weibull1,
		WeibullFunction & weibull2) const {
	weibull2 = weibull1;
	vector<double> volumes;
	calculateTreeVolumes(compartment, volumes);
	int const numberOfClasses = volumes.size();

	double const absoluteMargin = wantedRemoval * removalTolerance;
	double const classWidth = (maxDiameter - minDiameter) / (double)numberOfClasses;
	double volumeSum = 0.0;
	double distance = 0.0;
	switch (direction) {
	case SHIFT_LEFT: {
		for (int i = numberOfClasses - 1; i >= 0; i--) {
			volumeSum += volumes[i];
			if (volumeSum >= wantedRemoval - absoluteMargin) {
				distance = (double)(numberOfClasses - i) * classWidth;
				break;
			}
		}
		weibull2.shift(-distance);
		break;
	}
	case SHIFT_RIGHT: {
		for (int i = 0; i < numberOfClasses; i++) {
			volumeSum += volumes[i];
			if (volumeSum >= wantedRemoval - absoluteMargin) {
				distance = (double)(i + 1) * classWidth;
				break;
			}
		}
		weibull2.shift(distance);
		break;
	}
	}
	Log::log(6, "Estimated a new weibull function with %s", weibull2.toString().c_str());
}

double WeibullThinning::findIntersectionPoint(
		WeibullFunction const & weibull1, WeibullFunction const & weibull2,
		unsigned int & interSectionPointTests) {
	double const tolerance = 0.0001;

	if (weibull1.shape() <= 1.0 && weibull2.shape() <= 1.0) {
		// Negative exponential curves.
		return max(weibull1.origin(), weibull2.origin());
	}
	
	// Use the bisection method to find the intersection point
	// The intersection point should be bounded by the maxima of the two curves.
	double left = weibull1.maximum();
	double right = weibull2.maximum();
	if (left == right) { // Curves are identical
		return left;
	}
	double midpoint;
	double f_left = weibull1.value(left) - weibull2.value(left);
	double f_midpoint;
	for (int i = 0; i < 200; i++) {
		midpoint = (left + right) * 0.5; 
		if (abs(midpoint - left) <= tolerance) {
			return midpoint;
		}
		interSectionPointTests++;
		f_midpoint = weibull1.value(midpoint) - weibull2.value(midpoint);
		if (f_left * f_midpoint > 0.0) {
			left = midpoint;
			f_left = f_midpoint;
		} else {
			right = midpoint;
		}
	}
	Log::logWarning(6, "Failed to find an intersection point between Weibull curves (scale = %.2f, shape = %.2f, origin = %.2f) and (scale = %.2f, shape = %.2f, origin = %.2f).",
		weibull1.scale(), weibull1.shape(), weibull1.origin(),
		weibull2.scale(), weibull2.shape(), weibull2.origin());
	throw "Failed to find an intersection point";
}

bool WeibullThinning::weibullEstimation(double const mean, double const stdev, double const orign, double & outScale,
		double & outShape, double & outOrign) {
	double const clw = 2.0;
	double a = orign;
	if (a > clw) a = a - clw;
	
	double temp = stdev / (mean - a);
	if (temp > 1.2 || temp <= 0.0) {
		return false;
	}

	outOrign = orign;

	temp = temp*(((((0.007454537*temp*temp-0.08354348)*temp+0.153109251)*temp-0.001946641)*temp-0.22000991)*(1.0-temp)
			*(1.0-temp)+1.0);
	outShape = 1.0 / temp;
	if (temp <= 1.0) {
		outScale = 1.0;
	} else {
		outScale = temp;
		temp -= 1.0;
	}
	outScale = ((((((((0.035868343*temp-0.193527818)*temp+0.482199394)*temp-0.756704078)*temp+0.918206857)*temp
			-0.897056937)*temp+0.988205891)*temp-0.577191652)*temp+1.0)*outScale;
	outScale = (mean - a) / outScale;
	return true;
}

double WeibullThinning::calculateThinningVolume(Compartment const * compartment) {
	double volume = 0.0;
	for (Tree const * tree = compartment->getFirstTree(); tree != NULL; tree = tree->getNextTreeInCompartment()) {
		if (tree->isMarked() && !tree->isDead()) {
			volume += VolumeModel::calculateStemVolumeOf(tree);
		}
	}
#ifdef USE_MPI
	CompartmentData const * const compData = compartment->getCompartmentData();
	MPIUtil::sum(volume, compData->getCompartmentCommunicator());
#endif
	return volume;
}

void WeibullThinning::discardThinningResult(Compartment * compartment) {
	for (Tree * tree = compartment->getFirstTree(); tree != NULL; tree = tree->getNextTreeInCompartment()) {
		if (tree->isMarked() && !tree->isDead())
			tree->unmarkRemoval();
	}
}

