#ifndef WEIBULLTHINNING_H_
#define WEIBULLTHINNING_H_

#include "Management.h"
#include "SizeClassThinning.h"

#include <vector>
using std::vector;
#include <cassert>

class WeibullFunction;
class LinearTailWeibullFunction;

class WeibullThinning: public SizeClassThinning {
public:
	enum Direction { SHIFT_LEFT, SHIFT_RIGHT };

	explicit WeibullThinning(Direction direction);

#ifdef USE_MPI
	inline virtual void broadcast(int const rootId, MPI_Comm const comm) {
		SizeClassThinning::broadcast(rootId, comm);
		int dir = direction;
		MPIUtil::broadcast(dir, rootId, comm);
		direction = (Direction)dir;
		MPIUtil::broadcast(removalTolerance, rootId, comm);
		MPIUtil::broadcast(linearTailRatio, rootId, comm);
	}
#endif
	
	/**
	 * Applies Weibull thinning iteratively until the desired volume within the
	 * given error margin has been reached. Subclasses must implement the
	 * applyThinningIteratively() method.
	 */
	virtual void apply(Compartment * compartment, TreeContainer * treeContainer);

	/**
	 * Iterative thinning is applied until the target volume within a certain
	 * error margin has been reached. The relative margin is provided with
	 * this method.
	 */
	inline void setRemovalTolerance(double margin) {
		assert(margin > 0.0);
		removalTolerance = margin;
	}

protected:
	virtual void initIterativeThinning(Compartment * compartment, TreeContainer * treeContainer);

	virtual void applyIteration(
			Compartment * compartment, TreeContainer * treeContainer,
			WeibullFunction const & weibull1, WeibullFunction const & weibull2,
			double const intersectionPoint) = 0;
	
	void loadWeibullData(XERCES_CPP_NAMESPACE::DOMElement const * element);
	
	Direction direction;
	double minDiameter, maxDiameter;
private:
	/**
	 * Estimate a Weibull distribution function from simulated data.
	 */
	void estimateWeibullFromSimulated(
			Compartment const * const compartment,
			LinearTailWeibullFunction & weibull) const;

	void guessInitialSecondWeibull(
			Compartment const * const compartment,
			double const wantedRemoval,
			WeibullFunction const & weibull1,
			WeibullFunction & weibull2) const;
	
	void applyThinningIteratively(
			Compartment * compartment, TreeContainer * treeContainer,
			LinearTailWeibullFunction const & weibull1);

	void calculateTreeVolumes(Compartment const * compartment, vector<double> & volumes) const;

	static double findIntersectionPoint(
			WeibullFunction const & weibull1, WeibullFunction const & weibull2,
			unsigned int & interSectionPointTests);

	/**
	 * 3 parameter Weibull distribution function estimation.
	 * (Moment estimates for Weibull parameters. Garcia, O., N.Z.J.For.Sci. 11,304-306,1981.)
	 * stdev: square root of variance computed from data
	 * mean: mean computed from data
	 * orign and outOrign: origin (0.0) or the smallest diameter in the data (if orign==0.0, this will be a 2 parameter Weibull)
	 * outScale: scale parameter
	 * outShape: shape parameter
	 * 
	 * returns false if arguments are out of range, true if OK
	 */
	static bool weibullEstimation(double const mean, double const stdev, double const orign, double & outScale, double & outShape, double & outOrign);

	static double calculateThinningVolume(Compartment const * compartment);

	static void discardThinningResult(Compartment * compartment);

	double removalTolerance;
	double linearTailRatio;
};

#endif /*WEIBULLTHINNING_H_*/

