#ifdef USE_MPI
#include "CompartmentCommunicator.h"
#include "ForestSplitter.h"
#include "Log.h"
#include "CompartmentDataContainer.h"
#include "Compartment.h"
#include "TreeSender.h"
#include "Tree.h"

#include <algorithm>
using std::find;
#include <cassert>

#ifdef COMPARISONSEEDING
#include "TreeContainer.h"
#endif

void CompartmentCommunicator::combineAverage(CompartmentData const * const compartmentData,
		unsigned int & contributions, Vector & averages) {
	if (compartmentData!=NULL) {
		const MPI_Comm compartmentCommunicator = compartmentData->getCompartmentCommunicator();
		MPIUtil::sum(averages, compartmentCommunicator);
		MPIUtil::sum(contributions, compartmentCommunicator);
	} else {
		Log::logError(-1, "CompartmentCommunicator::combineAverage given NULL compartmentData");
		throw "CompartmentCommunicator::combineAverage given NULL compartmentData";
	}
}

void CompartmentCommunicator::max(unsigned int compartmentId, unsigned int & value) {
	CompartmentData const * const thisCompartmentData = findCompartmentData(compartmentId);
	if (thisCompartmentData!=NULL) {
		const MPI_Comm compartmentCommunicator = thisCompartmentData->getCompartmentCommunicator();
		MPIUtil::max(value, compartmentCommunicator);
	} else {
		Log::logError(-1,"CompartmentCommunicator::max(compartmentId, unsigned int) could not resolve compartmentData for compartmentId %u",
				compartmentId);
		throw "CompartmentCommunicator::max(compartmentId, unsigned int) could not resolve compartmentData";
	}
}

CompartmentData const * CompartmentCommunicator::findCompartmentData(unsigned int compartmentId) {
	ForestSplitter const & forestSplitter = ForestSplitter::Instance();
	CompartmentDataContainer const & dataContainer = forestSplitter.getDataContainer();
	return dataContainer.find(compartmentId);
}

#ifdef COMPARISONSEEDING
void CompartmentCommunicator::transferCompartmentTrees(TreeContainer & allTrees, Compartment * c,
		vector<Tree*> & acceptedReceivedTrees) {

	vector<Tree*> treesToSendVector;

	for (Tree * t = c->getFirstTree(); t != NULL; t = t->getNextTreeInCompartment()) {
		treesToSendVector.push_back(t);
	}

	// We must transfer the neighbours to the trees on the border for correct results
	vector<vector<Tree*> const *> rect;
	for (unsigned int i=0; i < treesToSendVector.size(); i++) {
		Tree const * const t = treesToSendVector[i];

		vector<Tree*> const & neighbours = allTrees.getTreesCloseTo(t);
		bool found=false;
		for (unsigned int j=0; j < rect.size(); j++) {
			if (rect[j] == &neighbours) {
				found = true;
				break;
			}
		}
		if (!found) {
			rect.push_back(&neighbours);
		}
	}
	vector<Tree*> treesOutsideCompartment;
	for (unsigned int j=0; j < rect.size(); j++) {
		// Log::log(-1, "Adding trees from rectangle %d/%d", j+1, rect.size());
		vector<Tree*> const * v = rect[j];
		for (unsigned int k=0; k < v->size(); k++) {
			Tree * const curTree = (*v)[k];

			// Add only the trees that aren't part of the compartment because the compartment is already added.
			if (c != curTree->getParentCompartment()) {
				bool added = false;
				for (unsigned int l=0; l<treesOutsideCompartment.size(); l++) {
					Tree const * const addedTree = treesOutsideCompartment[l];
					if (addedTree == curTree) {
						added = true;
						break;
					}
				}
				if (!added) {
					treesOutsideCompartment.push_back(curTree);
				}
			}
		}
	}
	treesToSendVector.insert(treesToSendVector.end(), treesOutsideCompartment.begin(), treesOutsideCompartment.end());
	treesOutsideCompartment.clear();

	MPI_Comm const comm = c->getCompartmentData()->getCompartmentCommunicator();
	const unsigned int compartmentId = c->getCompartmentData()->getIdNumber();

	Log::log(3, "Transferring compartment %d trees", compartmentId);

	vector<Tree*> receivedTrees;
	transferTrees(comm, treesToSendVector, receivedTrees);

	size_t const numReceived = receivedTrees.size();
	Log::log(3, "Transfer complete, %d trees received", numReceived);
	for (size_t index=0; index<numReceived; index++) {
		Tree * receivedTree = receivedTrees[index];
		// Don't accept the same tree more than once!
		// The same tree can come from more than one process.
		size_t const numberOfAcceptedTrees = acceptedReceivedTrees.size();
		for (size_t acceptedTreeIndex=0; acceptedTreeIndex<numberOfAcceptedTrees; acceptedTreeIndex++) {
			Tree const * const acceptedTree = acceptedReceivedTrees[acceptedTreeIndex];
			if (acceptedTree->getX() == receivedTree->getX() && acceptedTree->getY() == receivedTree->getY()
					&& acceptedTree->getMark(TreeMarks::DIAMETER) == receivedTree->getMark(TreeMarks::DIAMETER)) {
				// we already have this tree.
				delete receivedTree;
				receivedTree = NULL;
				receivedTrees[index] = NULL;
				break;
			}
		}
		if (receivedTree != NULL) {
			acceptedReceivedTrees.push_back(receivedTree);
		}
	}
	vector<Tree*>().swap(receivedTrees);
	Log::log(3, "%d unique trees accepted.", acceptedReceivedTrees.size());
}
void CompartmentCommunicator::transferTrees(MPI_Comm comm, vector<Tree*> const & sendTrees,
		vector<Tree*> & receivedTrees) {
	assert(comm != MPI_COMM_NULL);

	Treestruct *sendTreeBuffer = new Treestruct[TreeSender::BUFFSIZE];
	Treestruct *receiveTreeBuffer = new Treestruct[TreeSender::BUFFSIZE];

	size_t const numTreesToSend = sendTrees.size();

	int myIndex, numProcesses;
	MPI_Comm_rank(comm, &myIndex);
	MPI_Comm_size(comm, &numProcesses);
	Log::log(5, "Processes in communicator: %d.", numProcesses);
	for (int processIndex=0; processIndex<numProcesses; processIndex++) {
		int numToGetFromSender = numTreesToSend;
		MPI_Bcast(&numToGetFromSender, 1, MPI_UNSIGNED, processIndex, comm);

		if (processIndex == myIndex) {
			Log::log(6, "Process index %d sending %d trees.", myIndex, numTreesToSend);
			// Sender
			size_t indexOfTreeToSend = 0;
			while (indexOfTreeToSend < numTreesToSend) {
				size_t numTreesInThisSend = 0;
				// Fill the buffer.
				for (size_t index=0; index < std::min(TreeSender::BUFFSIZE, numTreesToSend) && indexOfTreeToSend
						< numTreesToSend; index++) {
					sendTreeBuffer[index] = TreeSender::createBufferFromTree(sendTrees[indexOfTreeToSend]);
					indexOfTreeToSend++;
					numTreesInThisSend++;
				}
				Log::log(7, "Sending %d trees", numTreesInThisSend);
				MPI_Bcast(sendTreeBuffer, numTreesInThisSend, MPIUtil::getMPITreeStruct(), processIndex, comm);
				MPI_Barrier(comm);
			}
		} else {
			// Receiver
			size_t const numTreesToReceive = numToGetFromSender;
			receivedTrees.reserve(receivedTrees.size() + numTreesToReceive);
			Log::log(6, "Process index %d receiving %d trees from process index %d.", myIndex, numTreesToReceive,
					processIndex);

			size_t indexOfTreeToReceive = 0;
			while (indexOfTreeToReceive < numTreesToReceive) {
				size_t numTreesInThisReceive = std::min(TreeSender::BUFFSIZE, numTreesToReceive - indexOfTreeToReceive);
				Log::log(7, "Receiving %d trees", numTreesInThisReceive);
				MPI_Bcast(receiveTreeBuffer, numTreesInThisReceive, MPIUtil::getMPITreeStruct(), processIndex, comm);
				MPI_Barrier(comm);
				// Read the buffer
				for (size_t index=0; index<numTreesInThisReceive; index++) {
					receivedTrees.push_back(TreeSender::createTreeFromBuffer(receiveTreeBuffer[index]));
					indexOfTreeToReceive++;
				}
			}
		}
	}

	delete[] sendTreeBuffer;
	delete[] receiveTreeBuffer;
}

#endif // COMPARISONSEEDING
#endif // USE_MPI
