#ifdef USE_MPI
#include "ForestSplitter.h"
#include "CompartmentDataContainer.h"
#include "CompartmentData.h"
#include "Polygon.h"
#include "Log.h"
#include "LimitFinder.h"
#include "roundToInt.h"
#include "MPIUtil.h"
#include "TreeContainer.h"
#include "Compartment.h"

#include <cassert>
#include <iostream>
using std::ios_base;
using std::cout;
#include <sstream>
using std::stringstream;
#include <algorithm>
using std::max;
using std::min;
#include <limits>
using std::numeric_limits;
#include <cmath>
using std::abs;
#include <stdint.h>

ForestSplitter * ForestSplitter::singleton = NULL;

ForestSplitter & ForestSplitter::Instance(CompartmentDataContainer const & dataCont) {
	if (singleton == NULL)
		singleton = new ForestSplitter(dataCont);
	return *singleton;
}
ForestSplitter & ForestSplitter::Instance() {
	// unsafe method, may return reference to null.
	assert(singleton!=NULL);
	return *singleton;
}

void ForestSplitter::Destroy() {
	delete singleton;
	singleton = NULL;
}

int ForestSplitter::getDestination(Point const & pos, unsigned int const compartmentId) {
	vector<int> candidates = getCompartmentProcesses(compartmentId);

	for (vector<int>::const_iterator idit = candidates.begin(); idit != candidates.end(); idit++) {
		ProcessCoordinates const & pc = processCoordinates[*idit];
		if (pc.contains(pos))
			return *idit;
	}
	return -1;
}

ForestSplitter::ForestSplitter(CompartmentDataContainer const & dataCont) :
	dataContainer(dataCont) {
	// Find out bounding box for area
	forestMinCoord = Point(numeric_limits<double>::max(), numeric_limits<double>::max());
	forestMaxCoord = Point(numeric_limits<double>::min(), numeric_limits<double>::min());

	for(CompartmentDataContainer::const_iterator compDataIter = dataContainer.begin(); compDataIter!=dataContainer.end(); ++compDataIter) {
		CompartmentData const * const compartmentData = *compDataIter;
		Polygon const * const p = compartmentData->getCompartmentPolygon();
		Point const pMinCoord = p->getMinCoordinate();
		Point const pMaxCoord = p->getMaxCoordinate();

		forestMinCoord.x = min(forestMinCoord.x, pMinCoord.x);
		forestMinCoord.y = min(forestMinCoord.y, pMinCoord.y);
		forestMaxCoord.x = max(forestMaxCoord.x, pMaxCoord.x);
		forestMaxCoord.y = max(forestMaxCoord.y, pMaxCoord.y);
	}
}

ForestSplitter::~ForestSplitter() {
}

void ForestSplitter::split(TreeContainer * const trees) {
	if (trees != NULL) {
		storeCompartmentOldProcesses();
		resetForestAndNeighbours();
	}
	splitForest(trees);
	findNeighbours();
	createCompartmentsActiveCommunicators();
}

void ForestSplitter::resetForestAndNeighbours() {
	for(CompartmentDataContainer::const_iterator compDataIter = dataContainer.begin(); compDataIter!=dataContainer.end(); ++compDataIter) {
		CompartmentData const * const compartmentData = *compDataIter;
		Polygon * const p = compartmentData->getCompartmentPolygon();
		p->reset();
	}
	neighbourProcesses.clear();
	compartmentProcesses.clear();
}

bool ForestSplitter::isFirstProcessInCompartment(unsigned int compartmentId) {
	return getCompartmentProcesses(compartmentId)[0] == MPIUtil::mpiId();
}

void ForestSplitter::setupProcessCoordinates() {
	int const myNumber = MPIUtil::mpiId();
	int const totalNumber = MPIUtil::mpiSize();

	/* TODO: maybe improve the way number of rows and number of columns are set up.
	 */
	const double aspectRatio = (forestMaxCoord.x - forestMinCoord.x) / (forestMaxCoord.y - forestMinCoord.y);
	Log::log(1, "Forest aspect ratio: %.2f", aspectRatio);

	// Setup the number of rows and columns the processes are divided into.
	// Starting with at least a 1*1 grid.
	const double rootTotalNumber = sqrt((double) totalNumber);
	const double rootAspectRatio = sqrt(aspectRatio);
	numberOfForestRows = max(1, roundToInt((1.0 / rootAspectRatio) * rootTotalNumber));
	numberOfForestColumns = max(1, roundToInt(rootAspectRatio * rootTotalNumber));

	// Try to include more processes in the grid by adding columns or rows
	while (numberOfForestRows * numberOfForestColumns < totalNumber) {
		if (numberOfForestRows > numberOfForestColumns)
			numberOfForestRows++;
		else
			numberOfForestColumns++;
	}
	// Remove columns or row from the grid if we don't have enough processes
	while (numberOfForestRows * numberOfForestColumns > totalNumber) {
		if (numberOfForestRows > numberOfForestColumns)
			numberOfForestRows--;
		else
			numberOfForestColumns--;
	}

	// Find out this process position in the grid
	myColumn = myNumber / numberOfForestRows;
	myRow = myNumber % numberOfForestRows;

	Log::log(1, "My column: %d/%d My row: %d/%d", myColumn + 1, numberOfForestColumns, myRow + 1, numberOfForestRows);
}

void ForestSplitter::splitForest(TreeContainer const * const trees) {
	int const totalNumber = MPIUtil::mpiSize();

	// Find out the x-boundaries of this process active area
	double xFractionToLeft = (double) myColumn * (double) numberOfForestRows / (double) totalNumber;
	double xFractionToLeftIncludingOwn = (double) (myColumn + 1) * (double) numberOfForestRows / (double) totalNumber;

	bool const inExtraColumn = (myColumn >= numberOfForestColumns);
	if (inExtraColumn) {
		xFractionToLeftIncludingOwn = 1.0;
	}

	activeMinCoord.x = findSplitCoordinate(KEEP_LEFT, xFractionToLeft, trees);
	activeMaxCoord.x = findSplitCoordinate(KEEP_LEFT, xFractionToLeftIncludingOwn, trees);

#ifdef SPLIT_USING_BOUNDING_BOX
	activeMinCoord.x = (forestMaxCoord.x-forestMinCoord.x)*xFractionToLeft;
	activeMaxCoord.x = (forestMaxCoord.x-forestMinCoord.x)*xFractionToLeftIncludingOwn;
#endif

	Log::log(-1, "Boundaries in x-direction: %.2f to %.2f", activeMinCoord.x, activeMaxCoord.x);

	LimitFinder & limitFinder = LimitFinder::Instance();
	limitFinder.reset();
	if (trees != NULL) {
		// Log::log(-1, "We have %d trees", trees->size());
		limitFinder.update(*trees);
	}
	limitFinder.setLimitingMode(limitedByAll);
	double largestLimitingValue = limitFinder.getLargestLimitingValue();
	// Log::log(-1, "Limiting value returned from LimitFinder: %.2f", largestLimitingValue);

	// Setting a reasonable minimum for the size of the splitted area.
	largestLimitingValue = max(largestLimitingValue * 2.0, 100.0);

	// Check if the active area is still sane
	if (activeMaxCoord.x-activeMinCoord.x < largestLimitingValue && (numberOfForestRows != 1 || numberOfForestColumns != 1)) {
		Log::logError(-1, "Active area too small");
		Log::logError(-1, "Total area simulated is too small for this number of processes");
		Log::logError(-1, "Width of the forest: %.2f", forestMaxCoord.x - forestMinCoord.x);
		Log::logError(-1, "Width of this area: %.2f", activeMaxCoord.x - activeMinCoord.x);
		Log::logError(-1, "Minimum width accepted: %.2f", largestLimitingValue);
		throw "Active area too small for this number of processes\n";
	}

	// Divide the polygons of all compartments according to the x-boundaries
	for(CompartmentDataContainer::const_iterator compDataIter = dataContainer.begin(); compDataIter!=dataContainer.end(); ++compDataIter) {
		CompartmentData const * const compartmentData = *compDataIter;
		Polygon * const p = compartmentData->getCompartmentPolygon();
		p->dividePolygon(KEEP_RIGHT, activeMinCoord.x);
		p->dividePolygon(KEEP_LEFT, activeMaxCoord.x);
	}

	int rowsInThisColumn = numberOfForestRows;
	if (inExtraColumn) {
		rowsInThisColumn = totalNumber - numberOfForestRows * numberOfForestColumns;
	}
	// Find out the y-boundaries of this process active area
	double yFractionAbove = (double) myRow / (double) rowsInThisColumn;
	double yFractionAboveIncludingOwn = (double) (myRow + 1) / (double) rowsInThisColumn;

	activeMinCoord.y = findSplitCoordinate(KEEP_ABOVE, yFractionAbove, trees);
	activeMaxCoord.y = findSplitCoordinate(KEEP_ABOVE, yFractionAboveIncludingOwn, trees);

#ifdef SPLIT_USING_BOUNDING_BOX
	activeMinCoord.y = (forestMaxCoord.y-forestMinCoord.y)*yFractionAbove;
	activeMaxCoord.y = (forestMaxCoord.y-forestMinCoord.y)*yFractionAboveIncludingOwn;
#endif

	Log::log(-1, "Boundaries in y-direction: %.2f to %.2f", activeMinCoord.y, activeMaxCoord.y);

	// Check if the active area is still sane
	if (activeMaxCoord.y-activeMinCoord.y < largestLimitingValue && (numberOfForestRows != 1 || numberOfForestColumns != 1)) {
		Log::logError(-1, "Active area too small");
		Log::logError(-1, "Total area simulated is too small for this number of processes");
		Log::logError(-1, "Height of the forest: %.2f", forestMaxCoord.y - forestMinCoord.y);
		Log::logError(-1, "Height of this area: %.2f", activeMaxCoord.y - activeMinCoord.y);
		Log::logError(-1, "Minimum height accepted: %.2f", largestLimitingValue);
		throw "Active area too small for this number of processes\n";
	}

	double area = 0.0;
	// Divide the polygons of all compartments according to the y-boundaries
	for(CompartmentDataContainer::const_iterator compDataIter = dataContainer.begin(); compDataIter!=dataContainer.end(); ++compDataIter) {
		CompartmentData const * const compartmentData = *compDataIter;
		Polygon * const p = compartmentData->getCompartmentPolygon();
		p->dividePolygon(KEEP_BELOW, activeMinCoord.y);
		p->dividePolygon(KEEP_ABOVE, activeMaxCoord.y);
		area += p->getActiveArea();
	}
	Log::log(-1, "Active area for this process: %.2f", area);
}

void ForestSplitter::findNeighbours() {
	// Create a vector of process-coordinate objects.
	Log::log(-1, "Broadcasting split coordinates of all processes");
	processCoordinates = broadcastProcessCoordinates();

	// Find and store the neighbour processes left, right, above and below of this process.
	Log::log(-1, "Finding neighbours left, right, above and below");
	findDirectionNeighbours(processCoordinates);

	// Find and store all processes working on each compartment.
	Log::log(-1, "Finding out for each compartment which processes are involved");
	findCompartmentNeighbours();
}

void ForestSplitter::createCompartmentsActiveCommunicators() {
	// Go through each CompartmentData
	const MPI_Comm activeProcesses = MPIUtil::getActiveCommunicator();
	const int mpiId = MPIUtil::mpiId();
	Log::log(-1, "Setting up communicators for all compartments");
	for(CompartmentDataContainer::const_iterator compDataIter = dataContainer.begin(); compDataIter!=dataContainer.end(); ++compDataIter) {
		CompartmentData * const compartmentData = *compDataIter;
		Polygon const * const polygon = compartmentData->getCompartmentPolygon();
		// Determine if process should be in this communicator.
		int const color = (polygon->isActive()) ? (int) compartmentData->getIdNumber() : MPI_UNDEFINED;
		// Create new communicator
		MPI_Comm curCompartmentCommunicator;
		MPI_Comm_split(activeProcesses, color, mpiId, &curCompartmentCommunicator);
		// Store communicator under CompartmentData
		compartmentData->setCompartmentCommunicator(curCompartmentCommunicator);
	}
}

double ForestSplitter::findSplitCoordinate(DivideType const divideType, double const pctRemaining,
		TreeContainer const * const trees) {

	// Special cases 0% or 100% means no split needed, return minCoord or maxCoord.
	if (pctRemaining <= 0.0) {
		if (divideType == KEEP_LEFT || divideType == KEEP_RIGHT) {
			return forestMinCoord.x;
		} else {
			return forestMinCoord.y;
		}
	} else if (pctRemaining >= 1.0) {
		if (divideType == KEEP_LEFT || divideType == KEEP_RIGHT) {
			return forestMaxCoord.x;
		} else {
			return forestMaxCoord.y;
		}
	}
	// The early returns above mean that no communication can take place after this point
	// or some processes will skip it.

	// variables needed
	double d;
	double splitCoord;

	// "Binary search" to find the correct split coordinate
	// Start by chosing the split coordinate in the middle of the boundaries.
	if (divideType == KEEP_LEFT || divideType == KEEP_RIGHT) {
		d = (forestMaxCoord.x - forestMinCoord.x) * 0.5;
		splitCoord = forestMinCoord.x + d;
	} else {
		d = (forestMaxCoord.y - forestMinCoord.y) * 0.5;
		splitCoord = forestMinCoord.y + d;
	}
	// Log::log( -1, "Splitting compartments (%d) using starting d=%.2f. Wanted pct: %.2f", divideType, d, pctRemaining);

	if (trees == NULL) {
		// Splitting with equal area as condition
		Log::log(-1, "Splitting based on area");
		// calculate the current active area of this process and the wanted area.
		double const wantedArea = calculateCurrentActiveProcessArea() * pctRemaining;
		double const tolerance = 0.1; // tolerance unit is square meters
		while (true) {
			// test split all compartments at this coordinate to see what area we would get with this split
			double const newArea = calculateDividedProcessArea(divideType, splitCoord);
			//		Log::log(-1,"Wanted area: %.2f, newArea: %.2f, split at %.2f, d: %.2f",	wantedArea, newArea, splitCoord, d);

			// Check if a split at the split coordinate resulted in the wanted area
			if (std::abs(newArea - wantedArea) < tolerance) {
				return splitCoord;
			}

			// if area wasn't correct, try with another split coordinate
			double const d2 = d * 0.5;
			if (d2 == d) {
				// Accept area if d cannot be halved anymore
				Log::log(-1, "Forestsplitter ran out of precision, accepting area %.2f", newArea);
				return splitCoord;
			}

			d = d2;
			if (newArea > wantedArea) {
				splitCoord -= d;
			} else {
				splitCoord += d;
			}
		}
	} else {
#ifndef NUMBER_OF_TREES_AS_WEIGHT
		Log::log(-1, "Splitting based on weight (trees*density)");
#else
		// Splitting with similar number of trees as condition
		Log::log(-1, "Splitting based on weight (trees)");
#endif

		// calculate the current number of trees on the active area and the wanted number of trees in this process.
		double const requestedWeight = calculateWeightForActiveArea(trees) * pctRemaining;
		const double tolerance = 1.0; // tolerance unit depends on used weighing method
		while (true) {
			// test split all compartments at this coordinate to see what number of trees we would get with this split
			double const currentWeight = calculateWeightForDividedArea(divideType, splitCoord, trees);

//			Log::log(-1,"Requested weight: %.2f, weight at %.1f: %.2f",requestedWeight,splitCoord,currentWeight);

			// Check if a split at the split coordinate resulted in the wanted number of trees
			if (std::abs(currentWeight - requestedWeight) < tolerance) {
//				Log::log(-1, "Using weight: %.2f. Splitting at %.1f", currentWeight, splitCoord);
				return splitCoord;
			}

			// if newNumTrees wasn't correct, try with another split coordinate
			double const d2 = d * 0.5;
			if (d2 == d) {
				// Accept newNumTrees if d cannot be halved anymore
				Log::log(-1, "Forestsplitter ran out of precision, accepting weight: %.2f", currentWeight);
				return splitCoord;
			}

			d = d2;
			if (currentWeight > requestedWeight) {
				splitCoord -= d;
			} else {
				splitCoord += d;
			}
		}
	}
}

double ForestSplitter::calculateCurrentActiveProcessArea() const {
	double area = 0.0;
	for(CompartmentDataContainer::const_iterator compDataIter = dataContainer.begin(); compDataIter!=dataContainer.end(); ++compDataIter) {
		CompartmentData const * const compartmentData = *compDataIter;
		Polygon const * const p = compartmentData->getCompartmentPolygon();
		area += p->getActiveArea();
	}
	return area;
}

double ForestSplitter::calculateDividedProcessArea(DivideType const divideType, double const splitCoord) const {
	double newArea = 0.0;
	// test split all compartments at this coordinate to see what area we would get with this split
	for(CompartmentDataContainer::const_iterator compDataIter = dataContainer.begin(); compDataIter!=dataContainer.end(); ++compDataIter) {
		CompartmentData const * const compartmentData = *compDataIter;
		Polygon const * const p = compartmentData->getCompartmentPolygon();
		newArea += p->getDividedArea(divideType, splitCoord);
	}
	return newArea;
}

double ForestSplitter::calculateWeightForActiveArea(TreeContainer const * const trees) {
	double weightForActive = 0.0;

	for (TreeContainer::const_comp_iterator compIter = trees->compBegin(); compIter != trees->compEnd(); ++compIter) {
		Compartment const * const curCompartment = *compIter;
		Polygon const * const p = curCompartment->getPolygon();
		double areaWeight = calculateCompartmentWeight(curCompartment);

		double weight = areaWeight * (p->getActiveArea() / p->getFullArea());

		weightForActive += weight;
	}

return weightForActive;
}

double ForestSplitter::calculateCompartmentWeight(Compartment const * const compartment) {
	double treesInCompartment = compartment->getTotalSize();
	double compartmentWeight;

#ifndef NUMBER_OF_TREES_AS_WEIGHT
	Polygon const * const p = compartment->getPolygon();
	double compartmentDensity = treesInCompartment / p->getFullArea();

	// Weight = T*D*(active area in pct)
	compartmentWeight = treesInCompartment * compartmentDensity;
//	Log::log(-1,"Compartment %d density: %.1f, trees: %.1f, weight: %.1f",compartment->getCompartmentData()->getIdNumber(),compartmentDensity, treesInCompartment, compartmentWeight);
#else
	compartmentWeight = treesInCompartment;
#endif

	return compartmentWeight;
}

double ForestSplitter::calculateWeightForDividedArea(DivideType const divideType, double const splitCoord,
		TreeContainer const * const trees) {
	// test split all compartments at this coordinate to see what number of trees we would get with this split
	double weight = 0.0;
	for (TreeContainer::const_comp_iterator compIter = trees->compBegin(); compIter != trees->compEnd(); ++compIter) {
		Compartment const * const curCompartment = *compIter;
		Polygon const * const p = curCompartment->getPolygon();

		double compartmentWeight = calculateCompartmentWeight(curCompartment);

		weight += compartmentWeight * (p->getDividedArea(divideType, splitCoord)
				/ p->getFullArea());
	}
	return weight;
}

vector<ProcessCoordinates> ForestSplitter::broadcastProcessCoordinates() {
	int const totalNumber = MPIUtil::mpiSize();

	// Buffers for the split coords of all processes.
	double * receivebuffer = new double[4 * totalNumber];

	// Setting the values this process will broadcast.
	double * sendbuffer = new double[4];
	sendbuffer[0] = activeMinCoord.x;
	sendbuffer[1] = activeMinCoord.y;
	sendbuffer[2] = activeMaxCoord.x;
	sendbuffer[3] = activeMaxCoord.y;

	MPI_Comm activeCommunicator = MPIUtil::getActiveCommunicator();

	MPI_Allgather(sendbuffer, 4, MPI_DOUBLE, receivebuffer, 4, MPI_DOUBLE, activeCommunicator);

	vector<ProcessCoordinates> coordinates;
	coordinates.reserve(totalNumber);
	for (int i = 0; i < totalNumber; i++) {
		int const offset = i * 4;
		Point minCoord(receivebuffer[offset], receivebuffer[offset + 1]);
		Point maxCoord(receivebuffer[offset + 2], receivebuffer[offset + 3]);
		coordinates.push_back(ProcessCoordinates(minCoord, maxCoord, i));
	}

	delete[] receivebuffer;
	delete[] sendbuffer;

	return coordinates;
}

void ForestSplitter::findDirectionNeighbours(vector<ProcessCoordinates> const & processCoords) {
	// int const myNumber = MPIUtil::mpiId();
	int const totalNumber = MPIUtil::mpiSize();

	LimitFinder & limitFinder = LimitFinder::Instance();

	// The value here should be large enough for the simulation and the statistics,
	// and also for the tree diameters we don't yet know for sure.
	double const largestNeeded = max(limitFinder.getLargestLimitingValue(), 50.0);

	for (int i = 0; i < totalNumber; i++) {
		// Finding all surrounding processes.
		ProcessCoordinates const & otherProcess = processCoords[i];
		if ( (otherProcess.start.y <= activeMaxCoord.y+largestNeeded) && (otherProcess.end.y >= activeMinCoord.y-largestNeeded)) {
			// Finding LEFT
			if (otherProcess.end.x == activeMinCoord.x) {
				// Log::log(-1, "Process %d is a LEFT neighbour to process %d", otherProcess.processId, myNumber);
				neighbourProcesses[ProcessDirection::left].push_back(otherProcess.processId);
			}
			// Finding RIGHT
			if (otherProcess.start.x == activeMaxCoord.x) {
				// Log::log(-1, "Process %d is a RIGHT neighbour to process %d", otherProcess.processId, myNumber);
				neighbourProcesses[ProcessDirection::right].push_back(otherProcess.processId);
			}
			// Finding the ones ABOVE and BELOW
			if (otherProcess.start.x == activeMinCoord.x && otherProcess.end.x == activeMaxCoord.x) {
				// Finding the one ABOVE
				if (otherProcess.end.y == activeMinCoord.y) {
					// Log::log(-1, "Process %d is ABOVE process %d", otherProcess.processId, myNumber);
					neighbourProcesses[above].push_back(otherProcess.processId);
				}
				// Finding the one BELOW
				if (otherProcess.start.y == activeMaxCoord.y) {
					// Log::log(-1, "Process %d is BELOW process %d", otherProcess.processId, myNumber);
					neighbourProcesses[below].push_back(otherProcess.processId);
				}
			}
		}
	}
}

void ForestSplitter::findCompartmentNeighbours() {
	int const myNumber = MPIUtil::mpiId();
	int const totalNumber = MPIUtil::mpiSize();

	MPI_Comm activeCommunicator = MPIUtil::getActiveCommunicator();

	size_t const numberOfCompartments = dataContainer.size();

	char * activeCompartmentsBuffer = new char[numberOfCompartments];

	for (int processId = 0; processId < totalNumber; processId++) {
		if (processId == myNumber) {
			// Sending
			for (CompartmentDataContainer::const_iterator compDataIter = dataContainer.begin(); compDataIter!=dataContainer.end(); ++compDataIter) {
				Polygon const * const polygon = (*compDataIter)->getCompartmentPolygon();
				size_t const index = std::distance(dataContainer.begin(), compDataIter);
				// typecast bool to char
				activeCompartmentsBuffer[index] = (char) polygon->isActive();
			}
		}

		MPI_Bcast(activeCompartmentsBuffer, numberOfCompartments, MPI_CHAR, processId, activeCommunicator);

		for (CompartmentDataContainer::const_iterator compDataIter = dataContainer.begin(); compDataIter!=dataContainer.end(); ++compDataIter) {
			size_t const index = std::distance(dataContainer.begin(), compDataIter);
			size_t const compartmentId = (*compDataIter)->getIdNumber();
			// typecast char back to bool
			if ((bool) activeCompartmentsBuffer[index]) {
				compartmentProcesses[compartmentId].push_back(processId);
			}
		}
	}

	delete[] activeCompartmentsBuffer;
}

bool ForestSplitter::needRebalancing(TreeContainer const & ownTrees, unsigned int balancePct) {
	if (balancePct > 0) {
		unsigned int needBalancing = 0;
		double ownWeight = 0.0;
		double minWeight, maxWeight;

		ownWeight = calculateWeightForActiveArea(&ownTrees);

		// Check if we need to rebalance
		MPI_Reduce(&ownWeight, &minWeight, 1, MPI_DOUBLE, MPI_MIN, 0, MPI_COMM_WORLD);
		MPI_Reduce(&ownWeight, &maxWeight, 1, MPI_DOUBLE, MPI_MAX, 0, MPI_COMM_WORLD);

		if (MPIUtil::mpiId() == 0) {
			double maxToMinPct = (maxWeight - minWeight) / minWeight*100.0;
//			Log::log(-1,"Max weight: %d, Min weight: %d, Difference pct: %.2f", maxWeight, minWeight, maxToMinPct);

			if (maxToMinPct> balancePct) {
				// Need to do rebalancing
				needBalancing = 1;
			}

		}

		MPI_Bcast(&needBalancing, 1, MPI_UNSIGNED, 0, MPI_COMM_WORLD);

		return (needBalancing> 0);
	} else {
		return true;
	}
}

bool ForestSplitter::isActiveAreaTooSmall(TreeContainer const & ownTrees) {
	unsigned int needBalancing = 0;
	if (numberOfForestRows != 1 || numberOfForestColumns != 1) {
		// Check that the side length of the active process area is still at least
		// twice the competition radius times the local correlation factor
		LimitFinder & limitFinder = LimitFinder::Instance();
		limitFinder.reset();
		limitFinder.update(ownTrees);
		limitFinder.setLimitingMode(limitedByAll);
		double const largestLimitingValue = limitFinder.getLargestLimitingValue();
		if (activeMaxCoord.x - activeMinCoord.x < largestLimitingValue) {
			needBalancing = 1;
		}
		MPIUtil::sum(needBalancing);
	}
	return needBalancing > 0;
}

ProcessCoordinates::ProcessCoordinates(Point const & minCoord, Point const & maxCoord, int id) :
	start(minCoord), end(maxCoord), processId(id) {
}

ProcessCoordinates::~ProcessCoordinates() {

}

#endif /* USE_MPI */
