#ifdef USE_MPI
#ifndef FORESTSPLITTER_H_
#define FORESTSPLITTER_H_

#include <vector>
using std::vector;
#include <map>
using std::map;
#include <mpi.h>

#include "Polygon.h"
#include "Compartment.h"

class CompartmentDataContainer;
class ProcessCoordinates;
class TreeContainer;

namespace ProcessDirection {
	enum Direction { left, right, above, below };
}
using namespace ProcessDirection;

class ForestSplitter {
public:
	static ForestSplitter & Instance(CompartmentDataContainer const & dataCont);
	static ForestSplitter & Instance();
	static void Destroy();

	void setupProcessCoordinates();
	void split(TreeContainer * const trees = NULL);
	bool isFirstProcessInCompartment(unsigned int compartmentId);
	/* getNeighbourProcesses(direction) returns a vector of the process ids in that direction
	 * never returns null.
	 */
	inline vector<int> const getNeighbourProcesses(Direction dir) const {
		if(neighbourProcesses.find(dir)!=neighbourProcesses.end())
			return neighbourProcesses.find(dir)->second;
		return vector<int>();
	}
	/* getCompartmentProcesses(compartmentId) returns a vector of the processes having a part of that compartment,
	 * including the process itself if that is the case.
	 * never returns null.
	 */
	inline vector<int> const getCompartmentProcesses(size_t compartmentId) const {
		if(compartmentProcesses.find(compartmentId)!=compartmentProcesses.end())
			return compartmentProcesses.find(compartmentId)->second;
		return vector<int>();
	}
	inline vector<int> const getCompartmentOldProcesses(size_t compartmentId) const {
		if(compartmentOldProcesses.find(compartmentId)!=compartmentOldProcesses.end())
			return compartmentOldProcesses.find(compartmentId)->second;
		return vector<int>();
	}
	inline Point const & getMinCoordinate() const { return activeMinCoord; }
	inline Point const & getMaxCoordinate() const { return activeMaxCoord; }
	inline CompartmentDataContainer const & getDataContainer() const { return dataContainer; }
	int getDestination(Point const & pos, unsigned int const compartmentId);
	bool needRebalancing(TreeContainer const & ownTrees, unsigned int balancePct);
	bool isActiveAreaTooSmall(TreeContainer const & ownTrees);
private:
	static ForestSplitter * singleton;

	CompartmentDataContainer const & dataContainer;
	map<Direction, vector<int> > neighbourProcesses; // map direction to a list of process ids.
	map<size_t, vector<int> > compartmentProcesses; // map compartment id to a list of process ids.
	map<size_t, vector<int> > compartmentOldProcesses; // map compartment id to a list of process ids.
	Point forestMinCoord, forestMaxCoord; // Min and max coordinate of the whole simulation area "bounding box"
	Point activeMinCoord, activeMaxCoord; // Min and max coordinates for the own part of the area
	int myColumn, myRow, numberOfForestRows, numberOfForestColumns;

	vector<ProcessCoordinates> processCoordinates;

	ForestSplitter(CompartmentDataContainer const & dataCont);
	~ForestSplitter();

	inline void storeCompartmentOldProcesses() {
		compartmentOldProcesses = compartmentProcesses;
	}
	void resetForestAndNeighbours();
	void splitForest(TreeContainer const * const trees);
	void findNeighbours();
	void createCompartmentsActiveCommunicators();
	double findSplitCoordinate(DivideType const divideType, double const pctAboveLeft, TreeContainer const * const trees);
	double calculateCurrentActiveProcessArea() const;
	double calculateDividedProcessArea(DivideType const divideType, double const splitCoord) const;

	inline double calculateCompartmentWeight(Compartment const * const compartment);
	inline double calculateWeightForActiveArea(TreeContainer const * const trees);
	inline double calculateWeightForDividedArea(DivideType const divideType, double const splitCoord, TreeContainer const * const trees);

	vector<ProcessCoordinates> broadcastProcessCoordinates();
	void findDirectionNeighbours(vector<ProcessCoordinates> const & processCoords);
	void findCompartmentNeighbours();
};

class ProcessCoordinates {
public:

	ProcessCoordinates(Point const & minCoord, Point const & maxCoord, int id);
	~ProcessCoordinates();
	inline bool contains(Point const & pos) const {
		return (start <= pos && pos < end);
	}
	Point start, end;
	int processId;
};

#endif /*FORESTSPLITTER_H_*/
#endif /*USE_MPI */
