#ifdef USE_MPI

#include "MPIUtil.h"
#include "Log.h"
#include "Vector.h"
#include "TreeMarks.h"
#include "Tree.h"

int MPIUtil::staticMpiId = -1;
int MPIUtil::staticMpiSize = -1;
MPI_Datatype MPIUtil::MPI_Treestruct(MPI_DATATYPE_NULL);
size_t MPIUtil::numberOfBroadcastsDone = 0;
size_t MPIUtil::numberOfSumsDone = 0;

void MPIUtil::sum(Vector & v, MPI_Comm const communicator) {
	const unsigned int size = v.size();
	double *sendBuffer = v.getDataPointer();
	double *receiveBuffer = new double[size];

	int count = size;
	MPI_Allreduce(sendBuffer, receiveBuffer, count, MPI_DOUBLE, MPI_SUM, communicator);

	for (unsigned int i = 0; i < size; i++) {
		v[i] = receiveBuffer[i];
	}
	delete[] receiveBuffer;
	numberOfSumsDone++;
}
void MPIUtil::sum(vector<double> & v, MPI_Comm const communicator) {
	const unsigned int size = v.size();
	double *sendBuffer = new double[size];
	double *receiveBuffer = new double[size];

	for (unsigned int i = 0; i < size; i++) {
		sendBuffer[i] = v[i];
	}
	int count = size;
	MPI_Allreduce(sendBuffer, receiveBuffer, count, MPI_DOUBLE, MPI_SUM, communicator);

	for (unsigned int i = 0; i < size; i++) {
		v[i] = receiveBuffer[i];
	}
	delete[] sendBuffer;
	delete[] receiveBuffer;
	numberOfSumsDone++;
}
void MPIUtil::sum(vector<int> & v, MPI_Comm const communicator) {
	const unsigned int size = v.size();
	int *sendBuffer = new int[size];
	int *receiveBuffer = new int[size];

	for (unsigned int i = 0; i < size; i++) {
		sendBuffer[i] = v[i];
	}
	int count = size;
	MPI_Allreduce(sendBuffer, receiveBuffer, count, MPI_INT, MPI_SUM, communicator);

	for (unsigned int i = 0; i < size; i++) {
		v[i] = receiveBuffer[i];
	}
	delete[] sendBuffer;
	delete[] receiveBuffer;
	numberOfSumsDone++;
}
void MPIUtil::sum(vector<uint64_t> & v, MPI_Comm const communicator) {
	const unsigned int size = v.size();

	// The buffers have to be of type double here as some MPI-implementations
	// apparently don't support operator MPI_SUM on MPI_LONG_LONG...
	// This might lose precision on very large numbers.

	double *sendBuffer = new double[size];
	double *receiveBuffer = new double[size];

	for (unsigned int i = 0; i < size; i++) {
		sendBuffer[i] = (double)v[i];
	}
	int count = size;
	MPI_Allreduce(sendBuffer, receiveBuffer, count, MPI_DOUBLE, MPI_SUM, communicator);

	for (unsigned int i = 0; i < size; i++) {
		v[i] = (uint64_t)receiveBuffer[i];
	}
	delete[] sendBuffer;
	delete[] receiveBuffer;
	numberOfSumsDone++;
}
void MPIUtil::sum(vector<unsigned int> & v, MPI_Comm const communicator) {
	const unsigned int size = v.size();
	unsigned int *sendBuffer = new unsigned int[size];
	unsigned int *receiveBuffer = new unsigned int[size];

	for (unsigned int i = 0; i < size; i++) {
		sendBuffer[i] = v[i];
	}
	int count = size;
	MPI_Allreduce(sendBuffer, receiveBuffer, count, MPI_UNSIGNED, MPI_SUM, communicator);

	for (unsigned int i = 0; i < size; i++) {
		v[i] = receiveBuffer[i];
	}
	delete[] sendBuffer;
	delete[] receiveBuffer;
	numberOfSumsDone++;
}
void MPIUtil::sum(int & integer, MPI_Comm const communicator) {
	// Some MPI implementations demand separate variables for send and receive.
	int receive = 0;
	MPI_Allreduce(&integer, &receive, 1, MPI_INT, MPI_SUM, communicator);
	integer = receive;
	numberOfSumsDone++;
}
void MPIUtil::sum(uint64_t & integer, MPI_Comm const communicator) {
	// Some MPI implementations demand separate variables for send and receive.
	// The buffers have to be of type double here as some MPI-implementations
	// apparently don't support operator MPI_SUM on MPI_LONG_LONG...
	// This might lose precision on very large numbers.
	double send = (double)integer;
	double receive = 0.0;
	MPI_Allreduce(&send, &receive, 1, MPI_DOUBLE, MPI_SUM, communicator);
	integer = (uint64_t)receive;
	numberOfSumsDone++;
}
void MPIUtil::sum(unsigned int & integer, MPI_Comm const communicator) {
	// Some MPI implementations demand separate variables for send and receive.
	unsigned int receive = 0;
	MPI_Allreduce(&integer, &receive, 1, MPI_UNSIGNED, MPI_SUM, communicator);
	integer = receive;
	numberOfSumsDone++;
}
void MPIUtil::sum(double & value, MPI_Comm const communicator) {
	// Some MPI implementations demand separate variables for send and receive.
	double receive = 0.0;
	MPI_Allreduce(&value, &receive, 1, MPI_DOUBLE, MPI_SUM, communicator);
	value = receive;
	numberOfSumsDone++;
}

void MPIUtil::min(double & lowerBound, MPI_Comm const communicator) {
	double receive = 0.0;
	MPI_Allreduce(&lowerBound, &receive, 1, MPI_DOUBLE, MPI_MIN, communicator);
	lowerBound = receive;
}
void MPIUtil::max(double & upperBound, MPI_Comm const communicator) {
	double receive = 0.0;
	MPI_Allreduce(&upperBound, &receive, 1, MPI_DOUBLE, MPI_MAX, communicator);
	upperBound = receive;
}
void MPIUtil::max(unsigned int & upperBound, MPI_Comm const communicator) {
	unsigned int receive = 0;
	MPI_Allreduce(&upperBound, &receive, 1, MPI_UNSIGNED, MPI_MAX, communicator);
	upperBound = receive;
}
void MPIUtil::allgather(uint64_t const & integer, vector<uint64_t> & integers, MPI_Comm const communicator) {
	int numProcs;
	MPI_Comm_size(communicator,&numProcs);
	uint64_t sendbuffer = integer;
	uint64_t * receivebuffer = new uint64_t[numProcs];
	MPI_Allgather(&sendbuffer, 1, MPI_UNSIGNED_LONG_LONG, receivebuffer, 1, MPI_UNSIGNED_LONG_LONG, communicator);
	integers.resize(numProcs,0);
	for(int i=0; i<numProcs; i++) {
		integers[i] = receivebuffer[i];
	}
	delete[] receivebuffer;
}
void MPIUtil::broadcast(unsigned char & value, int rootId, MPI_Comm const communicator) {
	MPI_Bcast(&value, 1, MPI_UNSIGNED_CHAR, rootId, communicator);
	numberOfBroadcastsDone++;
}
void MPIUtil::broadcast(unsigned int & value, int rootId, MPI_Comm const communicator) {
	MPI_Bcast(&value, 1, MPI_UNSIGNED, rootId, communicator);
	numberOfBroadcastsDone++;
}
void MPIUtil::broadcast(int & value, int rootId, MPI_Comm const communicator) {
	MPI_Bcast(&value, 1, MPI_INT, rootId, communicator);
	numberOfBroadcastsDone++;
}
void MPIUtil::broadcast(float & value, int rootId, MPI_Comm const communicator) {
	MPI_Bcast(&value, 1, MPI_FLOAT, rootId, communicator);
	numberOfBroadcastsDone++;
}
void MPIUtil::broadcast(double & value, int rootId, MPI_Comm const communicator) {
	MPI_Bcast(&value, 1, MPI_DOUBLE, rootId, communicator);
	numberOfBroadcastsDone++;
}
void MPIUtil::broadcast(bool & value, int rootId, MPI_Comm const communicator) {
	char caststuff = (char)value;
	MPI_Bcast(&caststuff, 1, MPI_CHAR, rootId, communicator);
	value = (bool)caststuff;
	numberOfBroadcastsDone++;
}
void MPIUtil::broadcast(vector<string> & strings, int rootId, MPI_Comm const communicator) {
	unsigned int numStrings = strings.size();
	broadcast(numStrings, rootId, communicator);

	if (mpiId() != rootId)
		strings.resize(numStrings);
	for (unsigned int index=0; index<numStrings; index++) {
		broadcast(strings[index], rootId, communicator);
	}
}
void MPIUtil::broadcast(string & str, int rootId, MPI_Comm const communicator) {
	unsigned int stringSize = str.size();
	broadcast(stringSize, rootId, communicator);

	int count = stringSize+1;
	if (mpiId() == rootId) {
		char * const cstr = (char*)str.c_str();
		MPI_Bcast(cstr, count, MPI_CHAR, rootId, communicator);
	} else {
		char * const buffer = new char[count];
		MPI_Bcast(buffer, count, MPI_CHAR, rootId, communicator);
		str = buffer;
		delete[] buffer;
	}
	numberOfBroadcastsDone++;
}
void MPIUtil::broadcast(Vector & values, int const rootId, MPI_Comm const communicator) {
	unsigned int vecSize = values.size();
	broadcast(vecSize, rootId, communicator);

	if (mpiId() != rootId) {
		values = Vector(vecSize);
	}
	double * ptr = values.getDataPointer();
	int count = vecSize;
	MPI_Bcast(ptr, count, MPI_DOUBLE, rootId, communicator);
	numberOfBroadcastsDone++;
}
void MPIUtil::broadcast(vector<double> & values, int const rootId, MPI_Comm const communicator) {
	unsigned int vecSize = values.size();
	broadcast(vecSize, rootId, communicator);

	if (mpiId() != rootId) {
		values.resize(vecSize);
	}
	double * ptr = &values[0];
	int count = vecSize;
	MPI_Bcast(ptr, count, MPI_DOUBLE, rootId, communicator);
	numberOfBroadcastsDone++;
}
void MPIUtil::broadcast(vector<float> & values, int const rootId, MPI_Comm const communicator) {
	unsigned int vecSize = values.size();
	broadcast(vecSize, rootId, communicator);

	if (mpiId() != rootId) {
		values.resize(vecSize);
	}
	float * ptr = &values[0];
	int count = vecSize;
	MPI_Bcast(ptr, count, MPI_FLOAT, rootId, communicator);
	numberOfBroadcastsDone++;
}
void MPIUtil::broadcast(vector<int> & values, int const rootId, MPI_Comm const communicator) {
	unsigned int vecSize = values.size();
	broadcast(vecSize, rootId, communicator);

	if (mpiId() != rootId) {
		values.resize(vecSize);
	}
	int * ptr = &values[0];
	int count = vecSize;
	MPI_Bcast(ptr, count, MPI_INT, rootId, communicator);
	numberOfBroadcastsDone++;
}
void MPIUtil::broadcast(vector<unsigned int> & values, int const rootId, MPI_Comm const communicator) {
	unsigned int vecSize = values.size();
	broadcast(vecSize, rootId, communicator);

	if (mpiId() != rootId) {
		values.resize(vecSize);
	}
	unsigned int * ptr = &values[0];
	int count = vecSize;
	MPI_Bcast(ptr, count, MPI_UNSIGNED, rootId, communicator);
	numberOfBroadcastsDone++;
}

void MPIUtil::registerMPIType() {
	// A tree is made up of 2 doubles (the coordinate) some floats (the marks mostly) and 3 unsigned int (the speciesId, age and compartmentId).
	// This must be consistent with the Treestruct in TreeSender.h

	int blocklens[3]; // array for the number of each datatype
	MPI_Aint offset[3]; // array for the offset of each block of datatype
	MPI_Datatype type[3]; // array for the mpi-type of each datatype
	MPI_Aint extent; // temp variable for extent

	blocklens[0] = 2; // Number of doubles
	offset[0] = 0;
	type[0] = MPI_DOUBLE;

	MPI_Type_extent(MPI_DOUBLE, &extent); // get the extent of an mpi-double
#ifdef INCLUDE_ENVIRONMENTAL_EFFECTS
	blocklens[1] = TreeMarks::NUMBER_OF_MARKS * (1 + Tree::NUMBER_OF_RND_NUM_SETS) + 2 + 2 + 1; // Number of floats
#else
	blocklens[1] = TreeMarks::NUMBER_OF_MARKS + 2 + 1; // Number of floats
#endif
	offset[1] = blocklens[0]*extent; // offset of floats due to the doubles
	type[1] = MPI_FLOAT;

	MPI_Type_extent(MPI_FLOAT, &extent); // get the extent of an mpi-float
	blocklens[2] = 3; // Number of unsigned ints
	offset[2] = offset[1] + blocklens[1]*extent;
	type[2] = MPI_UNSIGNED;

	MPI_Type_struct(3, blocklens, offset, type, &MPI_Treestruct);
	MPI_Type_commit(&MPI_Treestruct);
}

#endif
