#ifdef USE_MPI

#include "TreeSender.h"
#include "TreeContainer.h"
#include "Compartment.h"
#include "LimitFinder.h"
#include "ForestSplitter.h"
#include "Log.h"
#include "MPIUtil.h"
#include "CompartmentCommunicator.h"

#include <algorithm>
using std::min;

size_t TreeSender::BUFFSIZE = 100000;

vector<Tree*> TreeSender::transferForeignTrees(TreeContainer & allTrees) {
	LimitFinder & limitFinder = LimitFinder::Instance();
	limitFinder.reset();
	limitFinder.update(allTrees);
	limitFinder.setLimitingMode(limitedByAll);
	double const limitingValue = limitFinder.getLargestLimitingValue();
	allTrees.setLimitingValue(limitingValue);

	map<Direction, vector<Tree const *> > treesToSend;
	setupTreesToSend(allTrees, treesToSend);

	vector<Tree*> acceptedReceivedTrees;
	transferTreesInDirection(treesToSend, acceptedReceivedTrees, ProcessDirection::left);
	transferTreesInDirection(treesToSend, acceptedReceivedTrees, ProcessDirection::right);
	transferTreesInDirection(treesToSend, acceptedReceivedTrees, above);
	transferTreesInDirection(treesToSend, acceptedReceivedTrees, below);

	return acceptedReceivedTrees;
}

Treestruct TreeSender::createBufferFromTree(Tree const * tree) {
	Treestruct out;
	out.x = tree->getX();
	out.y = tree->getY();
	out.speciesId = tree->getSpeciesId();
	out.age = tree->getAge();
	for (unsigned int markIndex = 0; markIndex < TreeMarks::NUMBER_OF_MARKS; markIndex++) {
		out.marks[markIndex] = tree->getMark(markIndex);
#ifdef INCLUDE_ENVIRONMENTAL_EFFECTS
		for (unsigned int setIndex = 0; setIndex < Tree::NUMBER_OF_RND_NUM_SETS; setIndex++) {
			out.randomNumbers[setIndex * TreeMarks::NUMBER_OF_MARKS + markIndex] = tree->getRandomNumber(setIndex, markIndex);
		}
#endif
	}
#ifdef INCLUDE_ENVIRONMENTAL_EFFECTS
	out.processX = tree->getProcessXPosition();
	out.processY = tree->getProcessYPosition();
#endif
	out.previousDbhGrowth[0] = tree->getPreviousDbhGrowth(0);
	out.previousDbhGrowth[1] = tree->getPreviousDbhGrowth(1);
	out.initialHeterogeneity = tree->getInitialHeterogeneity();

	Compartment * compartment = tree->getParentCompartment();
	if (compartment == NULL)
		out.compartmentId = std::numeric_limits<unsigned int>::max();
	else
		out.compartmentId = compartment->getCompartmentData()->getIdNumber();

	return out;
}

Tree * TreeSender::createTreeFromBuffer(Treestruct const & in) {
	Tree * newTree = new Tree(in.speciesId, in.age, Point(in.x, in.y));
	for (unsigned int markIndex = 0; markIndex < TreeMarks::NUMBER_OF_MARKS; markIndex++) {
		newTree->setMark(markIndex, in.marks[markIndex]);
#ifdef INCLUDE_ENVIRONMENTAL_EFFECTS
		for (unsigned int setIndex = 0; setIndex < Tree::NUMBER_OF_RND_NUM_SETS; setIndex++) {
			newTree->setRandomNumber(setIndex, markIndex, in.randomNumbers[setIndex * TreeMarks::NUMBER_OF_MARKS + markIndex]);
		}
#endif
	}
#ifdef INCLUDE_ENVIRONMENTAL_EFFECTS
	newTree->setProcessXPosition(in.processX);
	newTree->setProcessYPosition(in.processY);
#endif
	newTree->setPreviousDbhGrowth(0, in.previousDbhGrowth[0]);
	newTree->setPreviousDbhGrowth(1, in.previousDbhGrowth[1]);
	newTree->setInitialHeterogeneity(in.initialHeterogeneity);

	return newTree;
}

void TreeSender::rebalanceTreesFast(TreeContainer & allTrees) {
	ForestSplitter & splitter = ForestSplitter::Instance();
	Point myMax = splitter.getMaxCoordinate();
	Point myMin = splitter.getMinCoordinate();

	Log::log(-1, "Rebalance trees started");
	Log::log(-1, "Trees in container: %d", allTrees.size());

	// Store a separate vector of treestructs to send to each process id.
	map<int, vector<Treestruct> > treesToSend;

	// Go through all trees
	for (TreeContainer::iterator treeIter = allTrees.begin(); treeIter != allTrees.end(); ++treeIter) {
		Tree const * const tree = *treeIter;
		Point const & treePos = tree->getPosition();
		if (!(myMin <= treePos && treePos < myMax)) {
			// Tree is outside the new active area and needs to be transferred
			int const destination = splitter.getDestination(treePos,
					tree->getParentCompartment()->getCompartmentData()->getIdNumber());
			if (destination < 0 || destination > (int)(MPIUtil::mpiSize())) {
				Log::logError(-1, "Tree at (%.2f, %.2f) in compartment %u has destination process %d",
						tree->getX(), tree->getY(),
						tree->getParentCompartment()->getCompartmentData()->getIdNumber(),
						destination);
				throw "Destination process id out of bounds";
			}
			treesToSend[destination].push_back(TreeSender::createBufferFromTree(tree));
			allTrees.prepareForErase(treeIter);
		}
	}
	// Rebuild the vector
	allTrees.massErase();

	Log::log(-1, "Processed all trees");
	int const numProcesses = MPIUtil::mpiSize();
	int * sendInformation = new int[numProcesses];
	int * recvCounts = new int[numProcesses];

	for (int i = 0; i < numProcesses; i++) {
		sendInformation[i] = 0;
		recvCounts[i] = 1;
	}

	for (map<int, vector<Treestruct> >::const_iterator iter = treesToSend.begin(); iter != treesToSend.end(); iter++) {
		int const destination = iter->first;
		vector<Treestruct> const & sendVector = iter->second;
		//  Log::log(-1, "Sending trees to proc %d: %d", destination, sendVector.size());
		sendInformation[destination] = sendVector.size();
	}

	Log::log(-1, "About to send tree number information");

	int toReceive;
	MPI_Reduce_scatter(sendInformation, &toReceive, recvCounts, MPI_INTEGER, MPI_SUM, MPI_COMM_WORLD);
	delete[] sendInformation;
	delete[] recvCounts;

	Log::log(-1, "Will receive: %d trees", toReceive);

	Treestruct * receiveBuffer = new Treestruct[toReceive];
	Treestruct * receivePointer = receiveBuffer;
	unsigned int received = 0;

	MPI_Request receiveRequest;

	// Start receiving data if anybody wants to send
	if (toReceive> 0) {
		MPI_Irecv(receivePointer, toReceive, MPIUtil::getMPITreeStruct(), MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD,
				&receiveRequest);
	}
	//	Log::log(-1, "Preparing to send");
	// Send all data we should send
	vector<MPI_Request> sendRequests;
	for (map<int, vector<Treestruct> >::iterator iter = treesToSend.begin(); iter != treesToSend.end(); iter++) {
		int const destination = iter->first;
		vector<Treestruct> & sendVector = iter->second;
		MPI_Request sendRequest;
		Log::log(-1, "Sending %d trees to %d", sendVector.size(), destination);
		MPI_Isend(&sendVector[0], sendVector.size(), MPIUtil::getMPITreeStruct(), destination, 1, MPI_COMM_WORLD,
				&sendRequest);
		sendRequests.push_back(sendRequest);
	}

	//	Log::log(-1, "Sends posted");
	while (toReceive> 0) {
		MPI_Status status;
		MPI_Wait(&receiveRequest, &status);
		int count;
		MPI_Get_count(&status, MPIUtil::getMPITreeStruct(), &count);
		receivePointer += count;
		received += count;
		toReceive -= count;
		// Log::log(-1, "Received %d trees. Still %d to go\n", count, toReceive);
		if (toReceive> 0) {
			MPI_Irecv(receivePointer, toReceive, MPIUtil::getMPITreeStruct(), MPI_ANY_SOURCE, MPI_ANY_TAG,
					MPI_COMM_WORLD, &receiveRequest);
		}
	}

	TreeContainer::CompartmentMap compartmentMap = allTrees.getCompartmentMap();

	// Create Tree objects from received treestructs.
	vector<Tree*> receivedTrees;
	receivedTrees.reserve(received);
	for (unsigned int i = 0; i < received; i++) {
		Treestruct const & ts = receiveBuffer[i];
		Tree * const tree = TreeSender::createTreeFromBuffer(ts);
		receivedTrees.push_back(tree);
		// Add tree to correct compartment
		Compartment * const comp = compartmentMap[ts.compartmentId];
		comp->addTree(tree);
	}
	delete[] receiveBuffer;
	// Add all received trees to tree container at once.
	// smarter allocation and doesn't do unnecessary work in rectanglecontainer
	allTrees.addAll(receivedTrees);

	Log::log(-1, "Waiting for all %d sends to complete", sendRequests.size());
	MPI_Waitall(sendRequests.size(), &sendRequests[0], MPI_STATUSES_IGNORE);
	//	Log::log(-1, "Rebalance done");
}

void TreeSender::transferTreesInDirection(map<Direction, vector<Tree const *> > const & treesToSend,
		vector<Tree*> & acceptedReceivedTrees, Direction sendDirection) {
	ForestSplitter & forestSplitter = ForestSplitter::Instance();
	LimitFinder & limitFinder = LimitFinder::Instance();
	Point const minCoordinate = forestSplitter.getMinCoordinate();
	Point const maxCoordinate = forestSplitter.getMaxCoordinate();
	const float limitingValue = limitFinder.getLargestLimitingValue();
	const MPI_Comm activeCommunicator = MPIUtil::getActiveCommunicator();

	// Send-related variables
	bool sendsToDo = true;
	vector<int> const sendProcessIds = forestSplitter.getNeighbourProcesses(sendDirection);
	const size_t numberOfNeighboursToSendTo = sendProcessIds.size();
	// Log::log(3, "Number of neighbours to send to: %u", numberOfNeighboursToSendTo);

	MPI_Request * requests = new MPI_Request[numberOfNeighboursToSendTo];

	vector<Tree const *> treesToSendVector = treesToSend.find(sendDirection)->second;
	const size_t numberOfTreesToSend = treesToSendVector.size();
	// Log::log(3, "Number of trees to send: %u", numberOfTreesToSend);

	Treestruct *sendTreeBuffer = new Treestruct[BUFFSIZE];
	size_t indexOfTreeToSend = 0;

	//Receive-related variables
	Direction receiveDirection = ProcessDirection::right;
	switch (sendDirection) {
	case (ProcessDirection::left): {
		receiveDirection = ProcessDirection::right;
		break;
	}
	case (ProcessDirection::right): {
		receiveDirection = ProcessDirection::left;
		break;
	}
	case (above): {
		receiveDirection = below;
		break;
	}
	case (below): {
		receiveDirection = above;
		break;
	}
	}
	vector<int> receiveProcessIds = forestSplitter.getNeighbourProcesses(receiveDirection);
	const size_t numberOfNeighboursToReceiveFrom = receiveProcessIds.size();

	size_t numberOfProcessesSending = numberOfNeighboursToReceiveFrom;

	Treestruct *receiveTreeBuffer = new Treestruct[BUFFSIZE];

	while (indexOfTreeToSend < numberOfTreesToSend || numberOfProcessesSending > 0 || sendsToDo) {
		int sendTag = MPIUtil::lastTransfer;
		size_t numberOfTreesInThisSend = 0;
		// Setup sends if there are any.
		// Log::log(4,"Setup sends if there are any");
		if (indexOfTreeToSend < numberOfTreesToSend) {
			sendsToDo = true;
			// Fill the buffer.
			for (size_t index = 0; index < min(BUFFSIZE, numberOfTreesToSend) && indexOfTreeToSend
					< numberOfTreesToSend; index++) {
				sendTreeBuffer[index] = createBufferFromTree(treesToSendVector[indexOfTreeToSend]);
				indexOfTreeToSend++;
				numberOfTreesInThisSend++;
			}
			// Log::log(4,"Number of trees in this send: %u",numberOfTreesInThisSend);
		}
		if (sendsToDo) {
			if (indexOfTreeToSend == numberOfTreesToSend) {
				sendTag = MPIUtil::lastTransfer;
			} else {
				sendTag = MPIUtil::moreToTransfer;
				sendsToDo = true;
			}
			// send to each neighbour.
			for (size_t processIndex = 0; processIndex < numberOfNeighboursToSendTo; processIndex++) {
				// Log::log(5, "Sending %u trees to process: %d", numberOfTreesInThisSend, sendProcessIds[processIndex]);
				MPI_Isend(sendTreeBuffer, numberOfTreesInThisSend, MPIUtil::getMPITreeStruct(),
						sendProcessIds[processIndex], sendTag, activeCommunicator, &requests[processIndex]);
			}
		}
		// Do recieves if there are any
		if (numberOfProcessesSending > 0) {
			const size_t numberOfReceivesToDoThisLoop = numberOfProcessesSending;
			// Log::log(4, "Number of processes to receive from in this loop: %u", numberOfReceivesToDoThisLoop);
			// Do all receives needed this iteration.
			for (size_t receiveNumber = 0; receiveNumber < numberOfReceivesToDoThisLoop; receiveNumber++) {
				MPI_Status status;
				// Log::log(5, "Receiving...");
				MPI_Recv(receiveTreeBuffer, BUFFSIZE, MPIUtil::getMPITreeStruct(), MPI_ANY_SOURCE, MPI_ANY_TAG,
				activeCommunicator, &status);
				if (status.MPI_TAG == MPIUtil::lastTransfer) {
					numberOfProcessesSending--;
				}
				int numberOfTreesReceived = 0;
				// int numberOfTreesAccepted = 0;
				MPI_Get_count(&status, MPIUtil::getMPITreeStruct(), &numberOfTreesReceived);
				// Log::log(5, "Received %d trees from process %d", numberOfTreesReceived, status.MPI_SOURCE);
				for (int treestructIndex = 0; treestructIndex < numberOfTreesReceived; treestructIndex++) {
					Treestruct const & curTreestruct = receiveTreeBuffer[treestructIndex];
					if (treeIsRelevant(curTreestruct, minCoordinate, maxCoordinate, limitingValue)) {
						// numberOfTreesAccepted++;
						Tree * newTree = createTreeFromBuffer(curTreestruct);
						acceptedReceivedTrees.push_back(newTree);
					}
				}
				// Log::log(5, "Accepted %d trees from process %d", numberOfTreesAccepted, status.MPI_SOURCE);
			}
		}
		// Log::log(4, "Waiting for the sends to complete before going to next iteration of the loop...");
		// Wait for the sends to complete before doing more.
		MPI_Waitall(numberOfNeighboursToSendTo, requests, MPI_STATUSES_IGNORE);
		if (sendTag == MPIUtil::lastTransfer)
			sendsToDo = false;
	}
	// Log::log(3, "Barrier for all processes after all sends and receives in this direction...");
	MPI_Barrier(activeCommunicator);
	delete[] receiveTreeBuffer;
	delete[] sendTreeBuffer;
	delete[] requests;
}

void TreeSender::setupTreesToSend(TreeContainer const & allTrees, map<Direction, vector<Tree const *> > & treesToSend) {
	ForestSplitter & forestSplitter = ForestSplitter::Instance();

	LimitFinder & limitFinder = LimitFinder::Instance();
	const float limitingValue = limitFinder.getLargestLimitingValue();
	Log::log(-1, "Limiting value for trees during transfer: %f", limitingValue);

	treesToSend[ProcessDirection::left] = vector<Tree const *> ();
	treesToSend[ProcessDirection::right] = vector<Tree const *> ();
	treesToSend[above] = vector<Tree const *> ();
	treesToSend[below] = vector<Tree const *> ();

	Point const minCoordinate = forestSplitter.getMinCoordinate();
	Point const maxCoordinate = forestSplitter.getMaxCoordinate();

	const size_t numberOfAllTrees = allTrees.size();
	for (size_t treeIndex = 0; treeIndex < numberOfAllTrees; treeIndex++) {
		Tree const * const curTree = allTrees[treeIndex];
		// Check if the tree is near the borders of this process' area
		// if it is, add the tree to the trees being sent in that direction.
		if (curTree->getX() <= (minCoordinate.x + limitingValue)) {
			treesToSend[ProcessDirection::left].push_back(curTree);
		}
		if (curTree->getX() >= (maxCoordinate.x - limitingValue)) {
			treesToSend[ProcessDirection::right].push_back(curTree);
		}
		if (curTree->getY() <= (minCoordinate.y + limitingValue)) {
			treesToSend[above].push_back(curTree);
		}
		if (curTree->getY() >= (maxCoordinate.y - limitingValue)) {
			treesToSend[below].push_back(curTree);
		}
	}
}

#endif // USE_MPI
