#ifdef USE_MPI
#ifndef TREESORTER_H_
#define TREESORTER_H_

#include "Tree.h"
#include "ForestSplitter.h"
#include <algorithm>

using std::sort;

/**
 * Class that contains static methods for sorting trees in the appropriate
 * order expected by certain parallel algorithms.
 */
class TreeSorter {
public:
	TreeSorter();

	/**
	 * Sort trees to occur in the correct order expected by the local
	 * Cholesky decomposition. I.e. trees appear according to rising
	 * y-coordinate, and the trees in the left strip come before those
	 * in the right strip.
	 *
	 * This method should be called before trees are transferred between
	 * processes. When called, trees are assigned process x and y
	 * coordinates accordingly.
	 */
	template <class TreeIterator>
	static inline void sortForLocalCholesky(TreeIterator first, TreeIterator last) {
		LocalCorrelationFunctor functor;
		assignProcessPositions(first, last);
		sort(first, last, functor);
	}

	template <class Container>
	static inline void sortForLocalCholesky(Container & container) {
		sortForLocalCholesky(container.begin(), container.end());
	}

	/**
	 * Sort neighbouring trees to appear in the order they have been
	 * processed. The trees' process x and y coordinates are used to
	 * accomplish this.
	 */
	template <class TreeIterator>
	static inline void sortProcessedNeighbours(TreeIterator first, TreeIterator last) {
		LocalCorrelationFunctor functor;
		sort(first, last, functor);
	}

	template <class Container>
	static inline void sortProcessedNeighbours(Container & container) {
		sortProcessedNeighbours(container.begin(), container.end());
	}

	class LocalCorrelationFunctor {
	public:
		bool operator()(Tree const * const tree1, Tree const * const tree2);
	};
private:
	template <class TreeIterator>
	static inline void assignProcessPositions(TreeIterator first, TreeIterator last) {
		ForestSplitter & forestSplitter = ForestSplitter::Instance();
		Point activeMinCoord = forestSplitter.getMinCoordinate();
		Point activeMaxCoord = forestSplitter.getMaxCoordinate();
		for (; first != last; ++first) {
			Tree * const tree = *first;
			float const x = (tree->getX() - activeMinCoord.x) / (activeMaxCoord.x - activeMinCoord.x);
			tree->setProcessXPosition(x);
			float const y = (tree->getY() - activeMinCoord.y) / (activeMaxCoord.y - activeMinCoord.y);
			tree->setProcessYPosition(y);
		}
	}
};

#endif /* TREESORTER_H_ */
#endif
