#include "ClusteredPointGenerator.h"
#include "PoissonRandomNumberGenerator.h"
#include "StandardNormalRandomNumberGenerator.h"
#include "Point.h"
#include "Polygon.h"
#include "roundToInt.h"
#include "Log.h"

#include "XercesString.h"
#include "XercesUtilities.h"
#include <xercesc/util/XMLString.hpp>
#include <xercesc/dom/DOMElement.hpp>
XERCES_CPP_NAMESPACE_USE

#include <map>
using std::map;

ClusteredPointGenerator::ClusteredPointGenerator()
: expectedNumberOfOffspringPoints(0)
, offspringDispersionRadius(0.0)
{
}

ClusteredPointGenerator::~ClusteredPointGenerator()
{
}

void ClusteredPointGenerator::generatePointsInside(Polygon const & shape, double const percent, vector<Point> & offspringPoints) const
{	
	vector<Point> motherPoints;
	getPointGenerator().generatePointsInside(shape, percent, motherPoints);
	
	generateOffspringPoints(shape, motherPoints, offspringPoints);
}

void ClusteredPointGenerator::generateOffspringPoints(Polygon const & polygon, vector<Point> & motherPoints, vector<Point> & offspringPoints) const
{
	for(unsigned int mother=0; mother < motherPoints.size(); ++mother) {
		// Note: Generating the poisson random numbers inside this loop causes different random number order than in Spate
		int numberOfChildren=PoissonRandomNumberGenerator::get(getExpectedNumberOfOffspringPoints());
		for(int child=0; child < numberOfChildren; ++child) {
			
			double const x = StandardNormalRandomNumberGenerator::get()*getOffspringDispersionRadius() + motherPoints[mother].x;
			double const y = StandardNormalRandomNumberGenerator::get()*getOffspringDispersionRadius() + motherPoints[mother].y;
			Point offspringPoint(x, y);
			// TODO: Child points which do not go inside this process' active polygon should possibly
			// be sent to another process. It may also be that it should be placed in an entirely
			// different neighbouring polygon. So this is very incorrect now...
			if(polygon.activeContains(offspringPoint)) {
				offspringPoints.push_back(offspringPoint);
			}
		}
	}
}

void ClusteredPointGenerator::loadData(DOMElement const * pointGenerator)
{
	// eUnrecognized denotes an unrecognized tag.
	enum Tags { eUnrecognized, eExpectedNumberOfOffspringPoints, eOffspringDispersionRadius };
	
	map<string, int> tagMap;
	tagMap["expectedNumberOfOffspringPoints"] = eExpectedNumberOfOffspringPoints;
	tagMap["offspringDispersionRadius"] = eOffspringDispersionRadius;
	
	for(DOMElement * element = XercesUtilities::getFirstChildDOMElement(pointGenerator); element != 0; element = XercesUtilities::getNextSiblingDOMElement(element)) {
		XercesString tagName = XMLString::replicate(element->getTagName());
		switch(tagMap[tagName.toString()])
		{
			case(eExpectedNumberOfOffspringPoints):
			{
				setExpectedNumberOfOffspringPoints(XercesUtilities::getFirstChildTextContentAsInt(element));
//				Log::log(5,"Expected number of offspring points: %d", getExpectedNumberOfOffspringPoints());
				break;
			}
			case(eOffspringDispersionRadius):
			{
				setOffspringDispersionRadius(XercesUtilities::getFirstChildTextContentAsDouble(element));
//				Log::log(5,"Offspring dispersion radius: %f", getOffspringDispersionRadius());
				break;
			}
			default:	// Handles eUnrecognized.
			{
				loadDatum(element);
				break;
			}
		}
	}
}
