#include "HomogenousPoissonPointGenerator.h"
#include "RandomNumber.h"
#include "PoissonRandomNumberGenerator.h"
#include "Point.h"
#include "Polygon.h"
#include "roundToInt.h"
#include "Log.h"

#include "XercesString.h"
#include "XercesUtilities.h"
#include <xercesc/util/XMLString.hpp>
#include <xercesc/dom/DOMElement.hpp>
XERCES_CPP_NAMESPACE_USE

#include <map>
using std::map;

HomogenousPoissonPointGenerator::HomogenousPoissonPointGenerator() :
	expectedNumberOfPointsPerHectare(0) {
}

HomogenousPoissonPointGenerator::~HomogenousPoissonPointGenerator() {
}

Point HomogenousPoissonPointGenerator::generatePointInside(Polygon const & polygon) const {
	return polygon.getActiveRandomPoint();
}

void HomogenousPoissonPointGenerator::generatePointsInside(Polygon const & shape, double const percent, vector<Point> & points) const {
	double const expectedPerHectareOfThisSpecies = (double)getExpectedNumberOfPointsPerHectare() * percent;
	Log::log(-1, "Expecting %.2f points per hectare", expectedPerHectareOfThisSpecies);
#ifndef COMPARISONSEEDING
	double const activeArea = shape.getActiveArea();
	double const exactAmount = expectedPerHectareOfThisSpecies * activeArea / 10000.0;
	Log::log(-1, "Expecting %.2f points on active area %.2f sqm", exactAmount, activeArea);
#else
	double const fullArea = shape.getFullArea();
	double const exactAmount = expectedPerHectareOfThisSpecies * fullArea / 10000.0;
	Log::log(-1, "Expecting %.2f points on full area %.2f sqm", exactAmount, fullArea);
#endif
	
	unsigned int const realAmount = PoissonRandomNumberGenerator::get(exactAmount);
 	
	points.reserve(realAmount); // Reserve memory for all points.

	for (unsigned int i=0; i < realAmount; ++i) {
		points.push_back(generatePointInside(shape));
	}
}

void HomogenousPoissonPointGenerator::loadData(DOMElement const * pointGenerator) {
	// eUnrecognized denotes an unrecognized tag.
	enum Tags {eUnrecognized, eExpectedNumberOfPointsPerHectare};

	map<string, int> tagMap;
	tagMap["expectedNumberOfPointsPerHectare"] = eExpectedNumberOfPointsPerHectare;

	for (DOMElement * element = XercesUtilities::getFirstChildDOMElement(pointGenerator); element != 0; element
			= XercesUtilities::getNextSiblingDOMElement(element)) {
		XercesString tagName = XMLString::replicate(element->getTagName());
		switch (tagMap[tagName.toString()]) {
		case (eExpectedNumberOfPointsPerHectare): {
			setExpectedNumberOfPointsPerHectare(XercesUtilities::getFirstChildTextContentAsInt(element));
			// Log::log(5,"Expected number of points per hectare: %d", getExpectedNumberOfPointsPerHectare());
			break;
		}
		default: // Handles eUnrecognized.
		{
			throw "Unrecognized XML-element: "+tagName.toString();
		}
		}
	}
}
