#include "InhomogenousPoissonPointGenerator.h"
#include "HomogenousPoissonPointGenerator.h"
#include "PoissonRandomNumberGenerator.h"
#include "RandomNumber.h"
#include "Point.h"
#include "Polygon.h"
#include "roundToInt.h"
#include "Log.h"

#include "loadCoefficients.h"
#include "XercesString.h"
#include "XercesUtilities.h"
#include <xercesc/util/XMLString.hpp>
#include <xercesc/dom/DOMElement.hpp>
XERCES_CPP_NAMESPACE_USE

#include <map>
using std::map;

InhomogenousPoissonPointGenerator::InhomogenousPoissonPointGenerator()
: expectedNumberOfPointsPerHectare(0)
, poissonRegressionCoefficients(POISSON_REGRESSION)
{
}

InhomogenousPoissonPointGenerator::~InhomogenousPoissonPointGenerator()
{
}

void InhomogenousPoissonPointGenerator::generatePointsInside(Polygon const & shape, double const percent, vector<Point> & points) const
{	
	double const expectedPerHectareOfThisSpecies = (double)getExpectedNumberOfPointsPerHectare() * percent;
	Log::log(-1, "Expecting %.2f points per hectare", expectedPerHectareOfThisSpecies);
#ifndef COMPARISONSEEDING
	double const activeArea = shape.getActiveArea();
	double const exactAmount = expectedPerHectareOfThisSpecies * activeArea / 10000.0;
	Log::log(-1, "Expecting %.2f points on active area %.2f sqm", exactAmount, activeArea);
#else
	double const fullArea = shape.getFullArea();
	double const exactAmount = expectedPerHectareOfThisSpecies * fullArea / 10000.0;
	Log::log(-1, "Expecting %.2f points on full area %.2f sqm", exactAmount, fullArea);
#endif
	
	unsigned int const realAmount = PoissonRandomNumberGenerator::get(exactAmount);

	Point const maxCoordinate = shape.getActiveMaxCoordinate();
	Point const minCoordinate = shape.getActiveMinCoordinate();
	double const maxLambda=supLambda(maxCoordinate.x-minCoordinate.x,maxCoordinate.y-minCoordinate.y);
	
	Vector design(6, 0.0);
	HomogenousPoissonPointGenerator	homogenousGenerator;
	
	double u=0.0, lambdaX=0.0;
	Point p(0.0, 0.0);
	
	for(unsigned int i=0; i<realAmount; ++i) {
		do {
			p=homogenousGenerator.generatePointInside(shape);
			design[0]=1.0;
			design[1]=p.x;
			design[2]=p.y;
			design[3]=p.x*p.x;
			design[4]=p.x*p.y;
			design[5]=p.y*p.y;
			u=RandomNumber::get();
			lambdaX=exp(Vector::dotProduct(design, getEstimatedPoissonRegressionCoefficients()));
		} while(u > lambdaX/maxLambda);
		points.push_back(p);
	}
}

// Identify order of a trend surface model and compute sup lambda.
//FIXME Is xSize / ySize the correct way to calculate sup ?
double InhomogenousPoissonPointGenerator::supLambda(double xSize, double ySize) const
{
	double sup=poissonRegressionCoefficients[0];
	
	if(poissonRegressionCoefficients[3] == 0.0 && poissonRegressionCoefficients[4] == 0.0 && poissonRegressionCoefficients[5] == 0.0) {
		if(poissonRegressionCoefficients[1] == 0.0 && poissonRegressionCoefficients[2] == 0.0) {
			return exp(sup);
		} else {
			if(poissonRegressionCoefficients[1] > 0.0) {
				sup += poissonRegressionCoefficients[1]*xSize;
			}
			if(poissonRegressionCoefficients[2] > 0.0) {
				sup += poissonRegressionCoefficients[2]*ySize;
			}
			return exp(sup);
		}
	} else {
		double y=(2.0*poissonRegressionCoefficients[2]*poissonRegressionCoefficients[3] - poissonRegressionCoefficients[1]*poissonRegressionCoefficients[4]) / (poissonRegressionCoefficients[4]*poissonRegressionCoefficients[4] - 4.0*poissonRegressionCoefficients[3]*poissonRegressionCoefficients[5]);
		double x=(-poissonRegressionCoefficients[1] - poissonRegressionCoefficients[4]*y) / (2.0*poissonRegressionCoefficients[3]);
		sup += poissonRegressionCoefficients[1]*x + poissonRegressionCoefficients[2]*y + poissonRegressionCoefficients[3]*x*x + poissonRegressionCoefficients[4]*x*y + poissonRegressionCoefficients[5]*y*y;
		return exp(sup);
	}
}

void InhomogenousPoissonPointGenerator::loadData(DOMElement const * pointGenerator)
{
	// eUnrecognized denotes an unrecognized tag.
	enum Tags { eUnrecognized, eExpectedNumberOfPointsPerHectare, eEstimatedPoissonRegressionCoefficients };
	
	map<string, int> tagMap;
	tagMap["expectedNumberOfPointsPerHectare"] = eExpectedNumberOfPointsPerHectare;
	tagMap["estimatedPoissonRegressionCoefficients"] = eEstimatedPoissonRegressionCoefficients;
	
	for(DOMElement * element = XercesUtilities::getFirstChildDOMElement(pointGenerator); element != 0; element = XercesUtilities::getNextSiblingDOMElement(element)) {
		XercesString tagName = XMLString::replicate(element->getTagName());
		switch(tagMap[tagName.toString()])
		{
			case(eExpectedNumberOfPointsPerHectare):
			{
				setExpectedNumberOfPointsPerHectare(XercesUtilities::getFirstChildTextContentAsInt(element));
//				Log::log(5,"Expected number of points per hectare: %d", getExpectedNumberOfPointsPerHectare());
				break;
			}
			case(eEstimatedPoissonRegressionCoefficients):
			{
//				Log::log(5,"Estimated Poisson regression coefficients loaded");
				setEstimatedPoissonRegressionCoefficients(loadCoefficients(element));
//				cout << getEstimatedPoissonRegressionCoefficients() << "\n";
				break;
			}
			default:	// Handles eUnrecognized.
			{
				throw "Unrecognized XML-element: "+tagName.toString();
			}
		}
	}
}
