#include "RegularPointGenerator.h"
#include "RandomNumber.h"
#include "Point.h"
#include "roundToInt.h"
#include "Log.h"

#include "XercesString.h"
#include "XercesUtilities.h"
#include <xercesc/util/XMLString.hpp>
#include <xercesc/dom/DOMElement.hpp>
XERCES_CPP_NAMESPACE_USE

#include <map>
using std::map;

#include <cmath>

RegularPointGenerator::RegularPointGenerator()
: density(0.0)
, interactionRadius(0.0)
{
}

RegularPointGenerator::~RegularPointGenerator()
{
}

void RegularPointGenerator::generatePointsInside(Polygon const & shape, double const percent, vector<Point> & points) const
{
	homogenousPoissonPointGenerator.generatePointsInside(shape, percent, points);
	
	for(int path=0; path < 4; ++path) {	// Path processing.
		for(unsigned int j=0; j < points.size(); ++j) {
			
			// Choose a point to discard.
			int id=roundToInt(RandomNumber::get()*(points.size() - 1));
						
			points[id]=points[0];
			
			// Generate a new point.
			double potential;	// Initialized inside the loop.
			do {
				points[0]=homogenousPoissonPointGenerator.generatePointInside(shape);
				potential=calculatePointPotential(points);
				if(density > 1.0) {
					potential /= pow(density, (int)points.size()-1);
				}
			} while(RandomNumber::get() > potential);
		}
	}
}

void RegularPointGenerator::loadData(DOMElement const * pointGenerator)
{
	// eUnrecognized denotes an unrecognized tag.
	enum Tags { eUnrecognized, eExpectedNumberOfInitialPointsPerHectare, eInteractionPotentialParameterForStrussDensity, eInteractionRadius };
	
	map<string, int> tagMap;
	tagMap["expectedNumberOfInitialPointsPerHectare"] = eExpectedNumberOfInitialPointsPerHectare;
	tagMap["interactionPotentialParameterForStrussDensity"] = eInteractionPotentialParameterForStrussDensity;
	tagMap["interactionRadius"] = eInteractionRadius;
	
	for(DOMElement * element = XercesUtilities::getFirstChildDOMElement(pointGenerator); element != 0; element = XercesUtilities::getNextSiblingDOMElement(element)) {
		XercesString tagName = XMLString::replicate(element->getTagName());
		switch(tagMap[tagName.toString()])
		{
			case(eExpectedNumberOfInitialPointsPerHectare):
			{
				homogenousPoissonPointGenerator.setExpectedNumberOfPointsPerHectare(XercesUtilities::getFirstChildTextContentAsInt(element));
//				Log::log(5,"Expected number of points per hectare for an initial point pattern: %d", homogenousPoissonPointGenerator.getExpectedNumberOfPointsPerHectare()); 
				break;
			}
			case(eInteractionPotentialParameterForStrussDensity):
			{
				setDensity(XercesUtilities::getFirstChildTextContentAsDouble(element));
//				Log::log(5,"Interaction potential parameter for the Struss density: %f", getDensity());
				break;
			}
			case(eInteractionRadius):
			{
				setInteractionRadius(XercesUtilities::getFirstChildTextContentAsDouble(element));
//				Log::log(5,"Interaction radius for neighboring points: %f", getInteractionRadius());
				break;
			}
			default:	// Handles eUnrecognized.
			{
				loadDatum(element);
				break;
			}
		}
	}
}
