#include "PoissonRandomNumberGenerator.h"
#include "RandomNumber.h"
#include "Log.h"

#include <cmath>
using std::exp;
using std::floor;
using std::sqrt;
using std::log;
using std::tan;

// Variables for Poisson random number generator with tables
double PoissonRandomNumberGenerator::plambda = -1.0; //Previous mean.
double PoissonRandomNumberGenerator::epsilon = 0.000000001; //Close enough to zero for floating point.
Cumul * PoissonRandomNumberGenerator::cumul = 0; //Cumulative probability table.
Cumul * PoissonRandomNumberGenerator::ctab[CTAB]; //Pointers to cumulative probabilities.
bool PoissonRandomNumberGenerator::init = false;

// Variables for Poisson random number generator with rejection formula
double PoissonRandomNumberGenerator::oldMean = -1.0;
double PoissonRandomNumberGenerator::almean = 0.0;
double PoissonRandomNumberGenerator::g = 0.0;
double PoissonRandomNumberGenerator::sq = 0.0;

PoissonRandomNumberGenerator::PoissonRandomNumberGenerator() {

}

PoissonRandomNumberGenerator::~PoissonRandomNumberGenerator() {

}

unsigned int PoissonRandomNumberGenerator::get(double const mean) {
	if (mean == 0.0)
		return 0;
#ifdef TABLE_BASED_POISSON
	return tableBased(mean);
#else
	return rejectionBased(mean);
#endif
}

unsigned int PoissonRandomNumberGenerator::tableBased(double const lambda) {
	// If a new mean is requested, prepare a new cumulative table.
	if (init == false || plambda != lambda) {
		init = true;
		plambda = lambda;
		initial(lambda);
	}
	// Generate the uniform number.
	double const u = RandomNumber::get();
	// Locate the corresponding cumulative probability.
	unsigned int n = (unsigned int) u * CTAB;
	for (Cumul * p = ctab[n]; p; p = p->p0) {
		n = p->n;
		if (u < p->c)
			break;
	}
	// Return the associated number.
	return n;
}

void PoissonRandomNumberGenerator::initial(double const lambda) {

	// Clear old table
	Destroy();

	// Compute the first term.
	double t = exp(-lambda);
	if (t == 0.0) {
		Log::logError(-1, "%.2f is too large as mean for poisson random number generator.", lambda);
		throw "Too large mean value for poisson random number generator.";
	}
	// Develop a table that relates the uniform probabilities to the
	// Poisson probabilities, as a linked list.
	double c = 0;
	Cumul * pw = 0;
	for (unsigned int n = 0; c < (1.0 - epsilon); n++) {
		Cumul * p;
		try {
			p = new Cumul;
		} catch (...) {
			Log::logFailure(-1, "Memory allocation failed in poisson random number generator.");
			throw "Memory allocation failed in poisson random number generator.";
		}
		if (n == 0) {
			cumul = p;
		} else {
			pw->p0 = p;
		}
		pw = p;
		c += t;
		p->c = c;
		p->n = n;
		t = t * lambda / (n + 1);
	}
	// Make sure a table was generated (verifying that lambda > 0).
	if (cumul == 0) {
		Log::logError(-1, "Internal function in poisson random number generator called with mean 0, which is invalid.");
		throw "Internal function in poisson random number generator called with mean 0, which is invalid.";
	}
	// Clear the direct access table.
	for (unsigned int n = 0; n < CTAB; n++) {
		ctab[n] = 0;
	}
	// For each entry in the list of cumulative probabilities, make
	// an entry in the direct access table.
	for (Cumul * p = cumul; p; p = p->p0) {
		unsigned int n = (unsigned int) p->c * CTAB;
		ctab[n] = p;
	}

	// Fill out all other entries of the direct access table.
	Cumul * p = cumul;
	for (unsigned int n = CTAB - 1; n >= 0; n--) {
		if (ctab[n])
			p = ctab[n];
		else
			ctab[n] = p;
	}
}
void PoissonRandomNumberGenerator::Destroy() {
	// Release any storage allocated to the previous table.
	for (Cumul * p = cumul; p;) {
		Cumul * pw = p->p0;
		delete p;
		p = pw;
	}
	cumul = 0;
}

unsigned int PoissonRandomNumberGenerator::rejectionBased(double const mean) {

	double em, t, y;

	if (mean < 12.0) {
		if (mean != oldMean) {
			oldMean = mean;
			g = exp(-mean);
		}
		em = -1.0;
		t = 1.0;
		do {
			em = em + 1.0;
			t = t * RandomNumber::get();
		} while (t > g);
	} else {
		if (mean != oldMean) {
			oldMean = mean;
			sq = sqrt(2.0 * mean);
			almean = log(mean);
			g = mean * almean - gamma(mean + 1.0);
		}
		do {
			do {
				y = tan(M_PI * RandomNumber::get());
				em = sq * y + mean;
			} while (em < 0.0);

			em = floor(em);
			t = 0.9 * (1.0 + (y * y)) * exp(em * almean - gamma(em + 1.0) - g);
		} while (RandomNumber::get() > t);
	}
	return (unsigned int) em;
}

double PoissonRandomNumberGenerator::gamma(double const xx) {
	/**
	 * natural log of gamma function adapted from
	 * Numerical Recipes in Fortran 90, Press et al. 1996
	 */
	double y = xx;
	double ser = 1.000000000190015;
	y += 1.0;
	ser += 76.18009172947146 / y;
	y += 1.0;
	ser += -86.50532032941677 / y;
	y += 1.0;
	ser += 24.01409824083091 / y;
	y += 1.0;
	ser += -1.231739572450155 / y;
	y += 1.0;
	ser += 0.1208650973866179e-2 / y;
	y += 1.0;
	ser += -0.5395239384953e-5 / y;

	double tmp = xx + 5.5;
	tmp = (xx + 0.5) * log(tmp) - tmp;
	return tmp + log(2.5066282746310005 * ser / xx);
}
