#include "ForestType.h"
#include "Log.h"
#include "TreeMarks.h"
#ifdef USE_MPI
#include "MPIUtil.h"
#endif

#include "XercesString.h"
#include "XercesUtilities.h"
#include <xercesc/util/XMLString.hpp>
#include <xercesc/dom/DOMElement.hpp>
XERCES_CPP_NAMESPACE_USE

#include <map>
using std::map;
#include <iostream>
#include <iomanip>
using std::ios;

ForestType::ForestType() 
: initialStandTransformations(TreeMarks::NUMBER_OF_MARKS)
, reproductionTransformations(TreeMarks::NUMBER_OF_MARKS) {

}

ForestType::~ForestType() {

}

#ifdef USE_MPI
void ForestType::broadcast(int const rootId, MPI_Comm const comm) {
	MPIUtil::broadcast(typeName, rootId, comm);
	unsigned int numTypeDatas = getNumSpecies();
	MPIUtil::broadcast(numTypeDatas, rootId, comm);
	Log::log(3, "Number of forest type datas (species) %d", numTypeDatas);
	for (unsigned int index = 0; index < numTypeDatas; index++) {
		ForestTypeData data;
		Log::log(4, "Broadcasting forest type data index %d", index);
		if (MPIUtil::mpiId() == rootId) {
			data = typeData[index];
			data.broadcast(rootId, comm);
		} else {
			data.broadcast(rootId, comm);
			typeData.push_back(data);
		}
	}
	for (unsigned int index = 0; index < TreeMarks::NUMBER_OF_MARKS; ++index) {
		Log::log(4, "Broadcasting transformation for mark %d", index);
		initialStandTransformations[index].broadcast(rootId, comm);
		reproductionTransformations[index].broadcast(rootId, comm);
	}
}
#endif

void ForestType::loadData(DOMElement const * typeDatas) {
	// eUnrecognized denotes an unrecognized tag.
	enum Tags {eUnrecognized, eSpecies, eName, eInitialStandTransformations, eReproductionTransformations};

	map<string, int> tagMap;
	tagMap["species"] = eSpecies;
	tagMap["name"] = eName;
	tagMap["initialStandTransformations"] = eInitialStandTransformations;
	tagMap["reproductionTransformations"] = eReproductionTransformations;

	for (DOMElement * element = XercesUtilities::getFirstChildDOMElement(typeDatas); element != 0; element
			= XercesUtilities::getNextSiblingDOMElement(element)) {
		XercesString tagName = XMLString::replicate(element->getTagName());
		switch (tagMap[tagName.toString()]) {
		case (eSpecies): {
			Log::log(3, "Loading forest type data");
			ForestTypeData data;
			data.setSpeciesName(XercesUtilities::getAttributeValueAsString(element,"name"));
			data.setSpeciesAmount(XercesUtilities::getFirstChildTextContentAsDouble(element));
			// If no reproduction amount is given, the same as for the initial stand will be used
			XercesString name = XMLString::transcode("reproductionAmount");
			if (element->hasAttribute(name.xmlCh())) {
				data.setSpeciesReproductionAmount(XercesUtilities::getAttributeValueAsDouble(element, "reproductionAmount"));
			}
			// If no initial diameter factor is specified, the default is 1.0
			name = XMLString::transcode("initialTreeDiameterFactor");
			if (element->hasAttribute(name.xmlCh())) {
				data.setSpeciesInitialDiameterFactor(XercesUtilities::getAttributeValueAsDouble(element, "initialTreeDiameterFactor"));
			}
			// If no new tree diameter factor is specified, the default is 1.0
			name = XMLString::transcode("newTreeDiameterFactor");
			if (element->hasAttribute(name.xmlCh())) {
				data.setSpeciesNewDiameterFactor(XercesUtilities::getAttributeValueAsDouble(element, "newTreeDiameterFactor"));
			}
			add(data);
			break;
		}
		case (eName): {
			setName(XercesUtilities::getFirstChildTextContentAsString(element));
			break;
		}
		case (eInitialStandTransformations): {
			Log::log(3, "Loading initial stand transformations");
			loadInitialStandTransformations(element);
			break;
		}
		case (eReproductionTransformations): {
			Log::log(3, "Loading reproduction transformations");
			loadReproductionTransformations(element);
			break;
		}
		default: // Handles eUnrecognized.
		{
			throw "Unrecognized XML-element: "+tagName.toString();
		}
		}
	}
}

void ForestType::loadInitialStandTransformations(
		XERCES_CPP_NAMESPACE::DOMElement const * transformation) {
	// eUnrecognized denotes an unrecognized tag.
	enum Tags {eUnrecognized, eTransformation};

	map<string, int> tagMap;
	tagMap["transformation"] = eTransformation;

	for (DOMElement * element = XercesUtilities::getFirstChildDOMElement(transformation); element != 0; element
			= XercesUtilities::getNextSiblingDOMElement(element)) {
		XercesString tagName = XMLString::replicate(element->getTagName());
		switch (tagMap[tagName.toString()]) {
		case (eTransformation): {
			const string markName = XercesUtilities::getAttributeValueAsString(element, "mark");
			const unsigned int markId = TreeMarks::getMarkId(markName);
			initialStandTransformations[markId].loadData(element);
			break;
		}
		default: // Handles eUnrecognized.
		{
			throw "Unrecognized XML-element: "+tagName.toString();
		}
		}
	}
}

void ForestType::loadReproductionTransformations(
		XERCES_CPP_NAMESPACE::DOMElement const * transformation) {
	// eUnrecognized denotes an unrecognized tag.
	enum Tags {eUnrecognized, eTransformation};

	map<string, int> tagMap;
	tagMap["transformation"] = eTransformation;

	for (DOMElement * element = XercesUtilities::getFirstChildDOMElement(transformation); element != 0; element
			= XercesUtilities::getNextSiblingDOMElement(element)) {
		XercesString tagName = XMLString::replicate(element->getTagName());
		switch (tagMap[tagName.toString()]) {
		case (eTransformation): {
			const string markName = XercesUtilities::getAttributeValueAsString(element, "mark");
			const unsigned int markId = TreeMarks::getMarkId(markName);
			reproductionTransformations[markId].loadData(element);
			break;
		}
		default: // Handles eUnrecognized.
		{
			throw "Unrecognized XML-element: "+tagName.toString();
		}
		}
	}
}

ostream & operator<<(ostream & os, ForestType const & type) {

	os << "Forest type name: " << type.getName() << "\n";
	os << "Consists of: ";
	os.setf(ios::fixed);
	os.precision(1);

	double const typeTotalAmount = type.getTotalAmount();
	size_t const numSpeciesInComp = type.getNumSpecies();
	for (size_t speciesIndex=0; speciesIndex<numSpeciesInComp; speciesIndex++) {
		if(speciesIndex>0)
			os << ", ";
		double const percent = type.getSpeciesAmount(speciesIndex) / typeTotalAmount * 100.0;
		os << percent << "% " << type.getSpeciesName(speciesIndex);
	}
	os << "\n";

	return os;
}
