#ifndef FORESTTYPE_H_
#define FORESTTYPE_H_

#include <vector>
using std::vector;
#include <string>
using std::string;
#include <cassert>
#include <ostream>
using std::ostream;
#include "Species.h"

#ifdef USE_MPI
#include <mpi.h>
#endif

#include "ForestTypeData.h"
#include "MarkTransformation.h"

#include <xercesc/util/XercesVersion.hpp>

namespace XERCES_CPP_NAMESPACE {
class DOMElement;
}

class ForestType {
public:
	ForestType();
	~ForestType();

#ifdef USE_MPI
	void broadcast(int const rootId, MPI_Comm const comm);
#endif

	inline string const & getName() const {
		return typeName;
	}
	inline void setName(string name) {
		typeName = name;
	}
	inline void add(ForestTypeData data) {
		typeData.push_back(data);
	}
	inline double getTotalAmount() const {
		double sum = 0.0;
		size_t const numTypeData = getNumSpecies();
		for (size_t index = 0; index < numTypeData; index++) {
			sum += typeData[index].getSpeciesAmount();
		}
		return sum;
	}
	inline double getTotalReproductionAmount() const {
		double sum = 0.0;
		size_t const numTypeData = getNumSpecies();
		for (size_t index = 0; index < numTypeData; index++) {
			sum += typeData[index].getSpeciesReproductionAmount();
		}
		return sum;
	}
	inline size_t getNumSpecies() const {
		return typeData.size();
	}
	inline string const & getSpeciesName(size_t const index) const {
		if (typeData.empty())
			throw "A forest type containing no species was encountered";
		assert(index < getNumSpecies());
		return typeData[index].getSpeciesName();
	}
	inline double getSpeciesAmount(size_t const index) const {
		if (typeData.empty())
			throw "A forest type containing no species was encountered";
		assert(index < getNumSpecies());
		return typeData[index].getSpeciesAmount();
	}
	inline double getSpeciesReproductionAmount(size_t const index) const {
		if (typeData.empty())
			throw "A forest type containing no species was encountered";
		assert(index < getNumSpecies());
		return typeData[index].getSpeciesReproductionAmount();
	}
	inline double getInitialDiameterFactor(size_t const index) const {
		if (typeData.empty())
			throw "A forest type containing no species was encountered";
		assert(index < getNumSpecies());
		return typeData[index].getSpeciesInitialDiameterFactor();
	}

	inline double getInitialTreeDiameterFactor(Species const * const species) const {
		string const & speciesName = species->getName();
		for (size_t i = 0; i < getNumSpecies(); i++) {
			if (getSpeciesName(i) == speciesName) {
				double const diamFactor = getInitialDiameterFactor(i);
				// Log::log(-1,"Using diameter factor %.2f",diamFactor);
				return diamFactor;
			}
		}
		// If the species wasn't found in this forest type there is no value to return.
		throw "The species named "+speciesName+" was not found in forest type named "+typeName;
	}

	inline double getNewDiameterFactor(size_t const index) const {
		if (typeData.empty())
			throw "A forest type containing no species was encountered";
		assert(index < getNumSpecies());
		return typeData[index].getSpeciesNewDiameterFactor();
	}

	inline double getNewTreeDiameterFactor(Species const * const species) const {
		string const & speciesName = species->getName();
		for (size_t i = 0; i < getNumSpecies(); i++) {
			if (getSpeciesName(i) == speciesName) {
				double const diamFactor = getNewDiameterFactor(i);
				// Log::log(-1,"Using diameter factor %.2f",diamFactor);
				return diamFactor;
			}
		}
		// If the species wasn't found in this forest type there is no value to return.
		throw "The species named "+speciesName+" was not found in forest type named "+typeName;
	}

	inline MarkTransformation const & getInitialStandTransformation(unsigned int markId) const {
		return initialStandTransformations[markId];
	}
	
	inline MarkTransformation const & getReproductionTransformation(unsigned int markId) const {
		return reproductionTransformations[markId];
	}

	void loadData(XERCES_CPP_NAMESPACE::DOMElement const * typeDatas);
	friend ostream & operator<<(ostream & os, ForestType const & type);

private:
	void loadInitialStandTransformations(XERCES_CPP_NAMESPACE::DOMElement const * transformation);
	void loadReproductionTransformations(XERCES_CPP_NAMESPACE::DOMElement const * transformation);

	vector<ForestTypeData> typeData;
	string typeName;
	vector<MarkTransformation> initialStandTransformations;
	vector<MarkTransformation> reproductionTransformations;
};

ostream & operator<<(ostream & os, ForestType const & type);

#endif /*FORESTTYPE_H_*/
