#include "ForestTypeContainer.h"
#include "ForestType.h"
#include "Log.h"
#include <cassert>

#ifdef USE_MPI
#include "MPIUtil.h"
#endif

#include "XercesString.h"
#include "XercesUtilities.h"
#include <xercesc/util/XMLString.hpp>
#include <xercesc/dom/DOMElement.hpp>
XERCES_CPP_NAMESPACE_USE
#include <map>
using std::map;

vector<ForestType*> ForestTypeContainer::typeList;

void ForestTypeContainer::Destroy() {
	size_t const numTypes = size();
	for (size_t index=0; index<numTypes; index++) {
		delete typeList[index];
	}
	typeList.clear();
}

#ifdef USE_MPI
void ForestTypeContainer::broadcast(int const rootId, MPI_Comm const comm) {
	unsigned int numTypes = size();
	MPIUtil::broadcast(numTypes, rootId, comm);
	Log::log(1, "Number of forest types %d", numTypes);
	for (unsigned int index=0; index<numTypes; index++) {
		Log::log(2, "Broadcasting forest type index %d", index);
		if (MPIUtil::mpiId()==rootId) {
			typeList[index]->broadcast(rootId, comm);
		} else {
			ForestType * const type = new ForestType();
			type->broadcast(rootId, comm);
			add(type);
		}
	}
}
#endif

ForestType * ForestTypeContainer::get(size_t const id) {
	assert(id < size());
	return typeList[id];
}

unsigned int ForestTypeContainer::getId(string const & name) {
	size_t const numTypes = size();
	for(size_t index=0; index<numTypes; index++) {
		if(typeList[index]->getName() == name)
			return index;
	}
	throw "No forest type named "+name+" was found!";
}

void ForestTypeContainer::add(ForestType * const type) {
	if(type==NULL) {
		throw "Trying to add NULL pointer to forest type to ForestTypeContainer!";
	}
	string const & name = type->getName();
	if(name.size() < 2) {
		Log::log(-1,"Trying to add a forest type with a name less than 2 characters long!");
		throw "Trying to add a forest type with a name less than 2 characters long!";
	}
	size_t const numTypes = size();
	for(size_t index=0; index<numTypes; index++) {
		if(typeList[index]->getName() == name) {
			Log::logError(-1,"Forest type with name \"%s\" already found!",name.c_str());
			throw "Forest type with name " + name + " already found!";
		}
	}
	typeList.push_back(type);
}
void ForestTypeContainer::loadData(DOMElement const * types) {
	// eUnrecognized denotes an unrecognized tag.
	enum Tags {eUnrecognized, eForestType};

	map<string, int> tagMap;
	tagMap["forestType"] = eForestType;

	for (DOMElement * element = XercesUtilities::getFirstChildDOMElement(types); element != 0; element
			= XercesUtilities::getNextSiblingDOMElement(element)) {
		XercesString tagName = XMLString::replicate(element->getTagName());
		switch (tagMap[tagName.toString()]) {
		case (eForestType): {
			Log::log(2, "Loading forest type");
			ForestType * const forestType = new ForestType();
			forestType->loadData(element);
			add(forestType);
			break;
		}
		default: // Handles eUnrecognized.
		{
			throw "Unrecognized XML-element: "+tagName.toString();
		}
		}
	}
}

