#include "MarkDataDistribution.h"

#include "File.h"

#include "XercesString.h"
#include "XercesUtilities.h"
#include <xercesc/util/XMLString.hpp>
#include <xercesc/dom/DOMElement.hpp>
XERCES_CPP_NAMESPACE_USE

#include <iostream>
using std::ios;
using std::ios_base;
#include <cmath>
using std::min;
using std::max;
#include <limits>
using std::numeric_limits;

MarkDataDistribution::MarkDataDistribution() {
}

MarkDataDistribution::MarkDataDistribution(string const & filename) :
	filename(filename) {
}

void MarkDataDistribution::loadData(XERCES_CPP_NAMESPACE::DOMElement const * distribution) {
	filename = XercesUtilities::getFirstChildTextContentAsString(distribution);
}

void MarkDataDistribution::setUp(vector<double> & classes, vector<double> & values) {
	vector<double> points;
	loadEmpiricalData(points);
	unsigned int totalNumberOfPoints = points.size();

	// Find the minimum and maximum mark values
	double lower = numeric_limits<double>::max();
	double upper = 0.0;
	findMinimumAndMaximumMarks(points, lower, upper);

	// Number of classes hardcoded to 100. Increasing the value will result in
	// better approximation, but will also increase memory consumption.
	const int numberOfClasses = 100;
	classes.resize(numberOfClasses);
	values.resize(numberOfClasses);
	vector<unsigned int> frequencies(numberOfClasses - 1);

	setupClasses(lower, upper, classes);
	// Offset the lower and upper class bounds to ensure that all points are
	// included.
	const double tolerance = 0.000000001;
	classes.front() -= tolerance;
	classes.back() += tolerance;

	classifyData(points, classes, frequencies);

	values.front() = 0.0;

	double artificialInc = 0.0;
	int artificialIncIdx = -1;
	for (int i = 0; i < numberOfClasses - 1; i++) {
		if (i <= artificialIncIdx) {
			// Increment artificially
			values[i + 1] = values[i] + artificialInc;
		} else if (frequencies[i] > 0) {
			// Increment as normal
			values[i + 1] = values[i] + static_cast<double>(frequencies[i]) / totalNumberOfPoints;
		} else {
			// The CDF must be monotonically increasing.
			// Find the next class with a frequency greater than 0 and use that value
			// produce artificial points.
			artificialIncIdx = numberOfClasses - 1;
			artificialInc = tolerance;
			for (int j = i + 1; j < numberOfClasses - 1; j++) {
				if (frequencies[j] > 0) {
					artificialIncIdx = j;
					double const temp = static_cast<double>(frequencies[j]) / totalNumberOfPoints;	
					artificialInc = temp / (j - i + 1);
					break;
				}
			}
			values[i + 1] = values[i] + artificialInc;
		}
	}

	values.back() = 1.0;
}

void MarkDataDistribution::loadEmpiricalData(vector<double>& points) const {
	File in(filename, ios_base::in);
	double point;
	while (in.good()) {
		in >> point;
		if (in.good())
			points.push_back(point);
	}
}

void MarkDataDistribution::findMinimumAndMaximumMarks(
		vector<double> const & points, double & lower, double & upper) const {
	const size_t size = points.size();
	for (size_t i = 0; i < size; i++) {
		const double mark = points[i];
		lower = min(lower, mark);
		upper = max(upper, mark);
	}
}

void MarkDataDistribution::setupClasses(double lower, double upper, vector<double> & classBounds) const {
	const unsigned int numberOfClasses = classBounds.size();
	const double increase = (upper - lower) / (numberOfClasses - 1);

	classBounds.front() = lower;
	for (unsigned int i = 1; i < numberOfClasses; ++i) {
		classBounds[i] = classBounds[i-1] + increase;
	}
	classBounds.back() = upper;
}

void MarkDataDistribution::classifyData(
		vector<double> const & points,
		vector<double> const & classBounds,
		vector<unsigned int> & frequencies) const {
	const unsigned int numberOfPoints = points.size();
	const unsigned int numberOfClasses = classBounds.size();
	for (unsigned int i = 0; i < numberOfPoints; ++i) {
		size_t k0 = 0;
		size_t ki = numberOfClasses - 1;
		while (ki - k0 > 1) {
			size_t k = (ki + k0) / 2;
			if (classBounds[k] > points[i]) {
				ki = k;
			} else {
				k0 = k;
			}
		}
		frequencies[k0]++;
	}
}

