#include "MarkNormalDistribution.h"

#include "XercesString.h"
#include "XercesUtilities.h"
#include <xercesc/util/XMLString.hpp>
#include <xercesc/dom/DOMElement.hpp>
XERCES_CPP_NAMESPACE_USE

#include <string>
using std::string;
#include <map>
using std::map;
#include <cmath>
using std::min;
using std::max;
#include <limits>
using std::numeric_limits;

MarkNormalDistribution::MarkNormalDistribution() : mean(0), stddev(0) {
}

MarkNormalDistribution::MarkNormalDistribution(double mean, double stddev) :
	mean(mean), stddev(stddev) {
}

void MarkNormalDistribution::loadData(XERCES_CPP_NAMESPACE::DOMElement const * distribution) {
	// eUnrecognized denotes an unrecognized tag.
	enum Tags {eUnrecognized, eMean, eStandardDeviation};

	map<string, int> tagMap;
	tagMap["mean"] = eMean;
	tagMap["standardDeviation"] = eStandardDeviation;
	bool meanLoaded = false, stddevLoaded = false;

	for (DOMElement * element = XercesUtilities::getFirstChildDOMElement(distribution); element != 0; element
			= XercesUtilities::getNextSiblingDOMElement(element)) {
		XercesString tagName = XMLString::replicate(element->getTagName());
		switch (tagMap[tagName.toString()]) {
		case (eMean): {
			meanLoaded = true;
			mean = XercesUtilities::getFirstChildTextContentAsDouble(element);
			break;
		}
		case (eStandardDeviation): {
			stddevLoaded = true;
			stddev = XercesUtilities::getFirstChildTextContentAsDouble(element);
			break;
		}
		default: { // Handles eUnrecognized.
			throw "Unrecognized XML-element: "+tagName.toString();
		}
		}
	}

	if (!meanLoaded)
		throw "No mean value specified for mark transformation";
	if (!stddevLoaded)
		throw "No standard deviation specified for mark transformation";
	if (mean < 0)
		throw "Mean value for mark transformation must not be negative";
	if (stddev < 0)
		throw "Standard deviation must not be negative";
}

void MarkNormalDistribution::setUp(vector<double> & classes, vector<double> & values) {
	// Number of classes hardcoded to 100. Increasing the value will result in
	// better approximation, but will also increase memory consumption.
	const unsigned int numberOfClasses = 100;
	classes.resize(numberOfClasses);
	values.resize(numberOfClasses);
	// The lower end of the support is the mean value minus three times the
	// standard deviation, or zero, whichever is greater. The upper end of
	// the support is set accordingly.
	const double d = min(3.0*stddev, mean);
	const double lower = mean - d;
	const double upper = mean + d;
	const double increase = (upper - lower) / (numberOfClasses - 1);
	// Precalculate the denominator of the erf argument
	const double denom = 1.0 / (stddev * sqrt(2.0));
	// Since a fraction (the part below the support) of the integral is
	// left out, this is to be subtracted from the cumulative distribution.
	const double lower_integral = erf((lower - mean) * denom);
	// Depending on how much is discarded by the supports, the cumulative
	// distribution will not reach 1. The CDF can hence be scaled by using the
	// CDF value of the upper support.
	const double scale = 1.0 / (0.5 * ( erf((upper - mean) * denom) - lower_integral ));

	classes[0] = lower;
	values[0] = 0;
	for (unsigned int i = 1; i < numberOfClasses; ++i) {
		classes[i] = classes[i-1] + increase;
		values[i] = 0.5 * ( erf((classes[i] - mean) * denom) - lower_integral ) * scale;
	}
}

