#include "MarkWeibullDistribution.h"

#include "XercesString.h"
#include "XercesUtilities.h"
#include <xercesc/util/XMLString.hpp>
#include <xercesc/dom/DOMElement.hpp>
XERCES_CPP_NAMESPACE_USE

#include <string>
using std::string;
#include <map>
using std::map;
#include <cmath>
using std::min;
using std::max;
#include <limits>
using std::numeric_limits;

MarkWeibullDistribution::MarkWeibullDistribution() :
	origin(0), scale(0), shape(0) {
}

MarkWeibullDistribution::MarkWeibullDistribution(double scale, double shape, double origin) :
	origin(origin), scale(scale), shape(shape) {
}

void MarkWeibullDistribution::loadData(XERCES_CPP_NAMESPACE::DOMElement const * distribution) {
	// eUnrecognized denotes an unrecognized tag.
	enum Tags {eUnrecognized, eOrigin, eScale, eShape};

	map<string, int> tagMap;
	tagMap["origin"] = eOrigin;
	tagMap["scale"] = eScale;
	tagMap["shape"] = eShape;

	origin = 0.0;
	bool scaleLoaded = false, shapeLoaded = false;

	for (DOMElement * element = XercesUtilities::getFirstChildDOMElement(distribution); element != 0; element
			= XercesUtilities::getNextSiblingDOMElement(element)) {
		XercesString tagName = XMLString::replicate(element->getTagName());
		switch (tagMap[tagName.toString()]) {
		case (eOrigin): {
			origin = XercesUtilities::getFirstChildTextContentAsDouble(element);
			break;
		}
		case (eScale): {
			scaleLoaded = true;
			scale = XercesUtilities::getFirstChildTextContentAsDouble(element);
			break;
		}
		case (eShape): {
			shapeLoaded = true;
			shape = XercesUtilities::getFirstChildTextContentAsDouble(element);
			break;
		}
		default: { // Handles eUnrecognized.
			throw "Unrecognized XML-element: "+tagName.toString();
		}
		}
	}

	if (!scaleLoaded)
		throw "No scale value specified for mark transformation";
	if (!shapeLoaded)
		throw "No shape specified for mark transformation";
	if (origin < 0.0)
		throw "Origin for mark transformation must not be negative";
	if (scale <= 0.0)
		throw "Scale value for mark transformation must not be negative";
	if (shape <= 0.0)
		throw "Shape value for mark transformation must not be negative";
}

void MarkWeibullDistribution::setUp(vector<double> & classes, vector<double> & values) {
	// Number of classes hardcoded to 100. Increasing the value will result in
	// better approximation, but will also increase memory consumption.
	const unsigned int numberOfClasses = 100;
	classes.resize(numberOfClasses);
	values.resize(numberOfClasses);
	// The support of the Weibull distribution goes from 0 + origin to infinity.
	// We can however approximate the upper end of the support by calculating
	// the inverse of the CDF using an argument reasonably close to 1.
	const double tolerance = 0.00001;
	const double upper = scale * pow(-log(tolerance), 1.0/shape) + origin;
	const double increase = (upper - origin) / (numberOfClasses - 1);

	classes[0] = origin;
	values[0] = 0;
	for (unsigned int i = 1; i < numberOfClasses - 1; ++i) {
		const double x = classes[i-1] + increase;
		classes[i] = x;
		values[i] = 1 - exp( -pow((x - origin)/scale, shape) );
	}
	classes[numberOfClasses - 1] = upper;
	values[numberOfClasses - 1] = 1;
}

