#include "Species.h"
#include "Log.h"

#ifdef USE_MPI
#include "MPIUtil.h"
#endif

#include "XercesString.h"
#include "XercesUtilities.h"
#include <xercesc/util/XMLString.hpp>
#include <xercesc/dom/DOMElement.hpp>
XERCES_CPP_NAMESPACE_USE
#include <map>
using std::map;

Species::Species()
: variables(NUMBER_OF_SPECIES_VARIABLES, 0.0)
{
	variables[SMALL_VOLUME_COEFFICIENT] = 0.6;
	variables[STUMP_MIN_HEIGHT] = 0.0;
	variables[STUMP_DIAMETER_COEFFICIENT] = 0.0;
	variables[STUMP_HEIGHT_COEFFICIENT] = 0.0;
}

Species::~Species() {

}
#ifdef USE_MPI
void Species::broadcast(int const rootId, MPI_Comm const comm) {
	MPIUtil::broadcast(name, rootId, comm);
	MPIUtil::broadcast(variables, rootId, comm);
	competitionDistanceData.broadcast(rootId, comm);
}
#endif

void Species::loadData(DOMElement const * species) {

	// eUnrecognized denotes an unrecognized tag.
	enum Tags {eUnrecognized, eName, eSecondaryModelValue, eSmallVolumeCoefficient, eTaperCurveParameters, eStumpHeightParameters, eCompDistanceModel };

	map<string, int> tagMap;
	tagMap["name"] = eName;
	tagMap["secondaryModelValue"] = eSecondaryModelValue;
	tagMap["smallVolumeCoefficient"] = eSmallVolumeCoefficient;
	tagMap["taperCurveParameters"] = eTaperCurveParameters;
	tagMap["stumpHeightParameters"] = eStumpHeightParameters;
	tagMap["competitionDistanceModel"] = eCompDistanceModel;

	for (DOMElement * element = XercesUtilities::getFirstChildDOMElement(species); element != 0; element
			= XercesUtilities::getNextSiblingDOMElement(element)) {
		XercesString tagName = XMLString::replicate(element->getTagName());
		switch (tagMap[tagName.toString()]) {
		case (eName): {
			name = XercesUtilities::getFirstChildTextContentAsString(element);
			size_t p = name.find_first_not_of(' ');
			if (p > 0 && p != string::npos) {
				// Strip leading whitespace since Tree::operator>> doesn't support it.
				Log::logWarning(-1, "Stripped leading whitespace off species name \"%s\"", name.c_str());
				name.erase(0, p);
			}
			break;
		}
		case (eSecondaryModelValue): {
			variables[SECONDARY_MODEL_VALUE] = XercesUtilities::getFirstChildTextContentAsDouble(element);
			break;
		}
		case (eSmallVolumeCoefficient): {
			variables[SMALL_VOLUME_COEFFICIENT] = XercesUtilities::getFirstChildTextContentAsDouble(element);
			break;
		}
		case (eTaperCurveParameters): {
			loadTaperCurveData(element);
			break;
		}
		case (eStumpHeightParameters): {
			loadStumpHeightData(element);
			break;
		}
		case (eCompDistanceModel): {
			competitionDistanceData.loadData(element);
			break;
		}
		default: // Handles eUnrecognized.
		{
			throw "Unrecognized XML-element: "+tagName.toString();
		}
		}
	}

}

void Species::loadTaperCurveData(XERCES_CPP_NAMESPACE::DOMElement const * parameters) {
	// eUnrecognized denotes an unrecognized tag.
	enum Tags {eUnrecognized, eC1, eC2, eC3, eC5, eC8, eC13, eC21, eC34 };

	map<string, int> tagMap;
	tagMap["coefficient1"] = eC1;
	tagMap["coefficient2"] = eC2;
	tagMap["coefficient3"] = eC3;
	tagMap["coefficient5"] = eC5;
	tagMap["coefficient8"] = eC8;
	tagMap["coefficient13"] = eC13;
	tagMap["coefficient21"] = eC21;
	tagMap["coefficient34"] = eC34;

	for (DOMElement * element = XercesUtilities::getFirstChildDOMElement(parameters); element != 0; element
			= XercesUtilities::getNextSiblingDOMElement(element)) {
		XercesString tagName = XMLString::replicate(element->getTagName());
		switch (tagMap[tagName.toString()]) {
		case (eC1): {
			variables[TAPER_COEFFICIENT1] = XercesUtilities::getFirstChildTextContentAsDouble(element);
			break;
		}
		case (eC2): {
			variables[TAPER_COEFFICIENT2] = XercesUtilities::getFirstChildTextContentAsDouble(element);
			break;
		}
		case (eC3): {
			variables[TAPER_COEFFICIENT3] = XercesUtilities::getFirstChildTextContentAsDouble(element);
			break;
		}
		case (eC5): {
			variables[TAPER_COEFFICIENT5] = XercesUtilities::getFirstChildTextContentAsDouble(element);
			break;
		}
		case (eC8): {
			variables[TAPER_COEFFICIENT8] = XercesUtilities::getFirstChildTextContentAsDouble(element);
			break;
		}
		case (eC13): {
			variables[TAPER_COEFFICIENT13] = XercesUtilities::getFirstChildTextContentAsDouble(element);
			break;
		}
		case (eC21): {
			variables[TAPER_COEFFICIENT21] = XercesUtilities::getFirstChildTextContentAsDouble(element);
			break;
		}
		case (eC34): {
			variables[TAPER_COEFFICIENT34] = XercesUtilities::getFirstChildTextContentAsDouble(element);
			break;
		}
		default: // Handles eUnrecognized.
		{
			throw "Unrecognized XML-element: "+tagName.toString();
		}
		}
	}
}

void Species::loadStumpHeightData(XERCES_CPP_NAMESPACE::DOMElement const * parameters) {
	// eUnrecognized denotes an unrecognized tag.
	enum Tags {eUnrecognized, eMinHeight, eDiameterCoefficient, eHeightCoefficient };

	map<string, int> tagMap;
	tagMap["minimumHeight"] = eMinHeight;
	tagMap["diameterCoefficient"] = eDiameterCoefficient;
	tagMap["heightCoefficient"] = eHeightCoefficient;

	for (DOMElement * element = XercesUtilities::getFirstChildDOMElement(parameters); element != 0; element
			= XercesUtilities::getNextSiblingDOMElement(element)) {
		XercesString tagName = XMLString::replicate(element->getTagName());
		switch (tagMap[tagName.toString()]) {
		case (eMinHeight): {
			variables[STUMP_MIN_HEIGHT] = XercesUtilities::getFirstChildTextContentAsDouble(element);
			break;
		}
		case (eDiameterCoefficient): {
			variables[STUMP_DIAMETER_COEFFICIENT] = XercesUtilities::getFirstChildTextContentAsDouble(element);
			break;
		}
		case (eHeightCoefficient): {
			variables[STUMP_HEIGHT_COEFFICIENT] = XercesUtilities::getFirstChildTextContentAsDouble(element);
			break;
		}
		default: // Handles eUnrecognized.
		{
			throw "Unrecognized XML-element: "+tagName.toString();
		}
		}
	}
}
