#include "SpeciesContainer.h"
#include "Species.h"
#include "Log.h"
#include <cassert>

#ifdef USE_MPI
#include "MPIUtil.h"
#endif

#include "XercesString.h"
#include "XercesUtilities.h"
#include <xercesc/util/XMLString.hpp>
#include <xercesc/dom/DOMElement.hpp>
XERCES_CPP_NAMESPACE_USE
#include <map>
using std::map;
#include <algorithm>
using std::max;

vector<Species*> SpeciesContainer::speciesList;
size_t SpeciesContainer::longestNameLength=0;

void SpeciesContainer::Destroy() {
	size_t const numSpecies = size();
	for (size_t index=0; index<numSpecies; index++) {
		delete speciesList[index];
	}
	speciesList.clear();
}

#ifdef USE_MPI
void SpeciesContainer::broadcast(int const rootId, MPI_Comm const comm) {
	unsigned int numSpecies = size();
	MPIUtil::broadcast(numSpecies, rootId, comm);
	Log::log(1, "Number of species %d", numSpecies);
	for (unsigned int index=0; index<numSpecies; index++) {
		Log::log(2, "Broadcasting Species index %d", index);
		if (MPIUtil::mpiId()==rootId) {
			speciesList[index]->broadcast(rootId, comm);
		} else {
			Species * spec = new Species();
			spec->broadcast(rootId, comm);
			add(spec);
		}
	}
}
#endif

void SpeciesContainer::loadData(DOMElement const * species) {

	// eUnrecognized denotes an unrecognized tag.
	enum Tags {eUnrecognized, eSpecies};

	map<string, int> tagMap;
	tagMap["species"] = eSpecies;

	for (DOMElement * element = XercesUtilities::getFirstChildDOMElement(species); element != 0; element
			= XercesUtilities::getNextSiblingDOMElement(element)) {
		XercesString tagName = XMLString::replicate(element->getTagName());
		switch (tagMap[tagName.toString()]) {
		case (eSpecies): {
			Species * species = new Species();
			species->loadData(element);
			add(species);
			break;
		}
		default: // Handles eUnrecognized.
		{
			throw "Unrecognized XML-element: "+tagName.toString();
		}
		}
	}

}

Species * SpeciesContainer::get(size_t const id) {
	assert(id < size());
	return speciesList[id];
}

void SpeciesContainer::add(Species * const species) {
	if(species==NULL) {
		throw "Trying to add NULL pointer to species to SpeciesContainer!";
	}
	string const & name = species->getName();
	if(name.size() < 2) {
		Log::logError(-1,"Trying to add a species with a name less than 2 characters long!");
		throw "Trying to add a species with a name less than 2 characters long!";
	}
	size_t const numSpecies = size();
	for(size_t index=0; index<numSpecies; index++) {
		if(speciesList[index]->getName() == name) {
			Log::logError(-1,"Species with name \"%s\" already found!",name.c_str());
			throw "Species with name " + name + " already found!";
		}
	}
	speciesList.push_back(species);
	longestNameLength = max(longestNameLength, name.length());
}

unsigned int SpeciesContainer::getId(string const & name) {
	size_t const numSpecies = size();
	for(size_t index=0; index<numSpecies; index++) {
		if(speciesList[index]->getName() == name) {
			return index;
		}
	}
	Log::log(-1,"Species with name \"%s\" not found!",name.c_str());
	throw "Species with name " + name + " not found!";
}
