#include "Compartment.h"
#include "Polygon.h"
#include "Tree.h"
#include "PoissonRandomNumberGenerator.h"
#include "HomogenousPoissonPointGenerator.h"
#include "TreeContainer.h"
#include "CompartmentData.h"
#include "Log.h"
#include "roundToInt.h"
#include "SpeciesContainer.h"
#include "Species.h"
#include "ForestTypeContainer.h"
#include "ForestType.h"
#include "SimulationData.h"
#include "RegenerationModel.h"
#include "RegenerationModelData.h"

#include <iomanip>
using std::ios;
using std::setw;
#include <iostream>
using std::cout;
using std::endl;
#include <algorithm>
using std::max;

#ifdef COMPARISONSEEDING
#include "LimitFinder.h"
#include "UniformRandomNumberGenerator.h"
#endif

#ifdef CALCULATION_INSPECTION
#include "CalculationInspector.h"
#endif

#ifdef USE_MPI
#include "MPIUtil.h"
#endif

Compartment::Compartment(CompartmentData & c)
: compartmentData(c)
, firstTree(NULL)
, lastTree(NULL)
, totalDiameter(0.0)
, totalBasalArea(0.0)
, fullCompartmentVolume(0.0)
, totalNumberOfTrees(0)
, numberOfTrees(0)
, clearCutIteration(0) {
}

Compartment::~Compartment() {
	size_t const numberOfNotAddedTrees = newTrees.size();
	if (numberOfNotAddedTrees>0) {
		Log::logWarning(-1,
				"Warning: a Compartment object containing a vector of new Trees is destroyed, deleting the Tree objects!");
	}
	for(list<Tree*>::const_iterator treeIter=newTrees.begin(); treeIter!=newTrees.end(); ++treeIter) {
		delete *treeIter;
	}
}

void Compartment::addTree(Tree* const tree) {
	tree->setParentCompartment(this);
	// Check if the list is empty
	if (lastTree!=NULL) {
		lastTree->setNextTreeInCompartment(tree);
		tree->setPrevTreeInCompartment(lastTree);
	}
	lastTree=tree;
	if (firstTree==NULL)
		firstTree=tree;
	numberOfTrees++;
}

void Compartment::removeTree(Tree* const tree) {
	Tree* const next = tree->getNextTreeInCompartment();
	Tree* const prev = tree->getPrevTreeInCompartment();
	// Handle the first tree special case
	if (tree==firstTree)
		firstTree=next;
	else
		prev->setNextTreeInCompartment(next);
	//Handle the last tree special case
	if (lastTree==tree)
		lastTree=prev;

	if (next!=NULL)
		next->setPrevTreeInCompartment(prev);

	tree->setPrevTreeInCompartment(NULL);
	tree->setNextTreeInCompartment(NULL);
	tree->setParentCompartment(NULL);
	numberOfTrees--;
}

void Compartment::calculateNewTreePositions(unsigned int const iteration, SimulationData const & simData, TreeContainer & allTrees, TreeContainer & otherCompartmentTrees) {
	const double activeArea = getPolygon()->getActiveArea();
	const double fullArea = getPolygon()->getFullArea();

	unsigned int const compartmentId = compartmentData.getIdNumber();
	Log::log(4, "Calculating new tree positions for compartment %d (area: %.2f / %.2f).", compartmentId, activeArea, fullArea);
	double const treesPerHectare = calculateNumberOfNewTreesPerHectare(simData);
	Log::log(5, "Expected number of new trees per hectare: %.2f", treesPerHectare);

#ifndef COMPARISONSEEDING
	double const expectedNumberOfNewTrees = (double)treesPerHectare * activeArea / 10000.0;
	Log::log(5, "Expecting %.2f points on active area %.2f sqm", expectedNumberOfNewTrees, activeArea);
#else
	double const expectedNumberOfNewTrees = (double)treesPerHectare * fullArea / 10000.0;
	Log::log(5, "Expecting %.2f points on full area %.2f sqm", expectedNumberOfNewTrees, fullArea);
#endif

	unsigned int forestTypeId = getCompartmentData()->getForestTypeId();
	ForestType const * const forestType = ForestTypeContainer::get(forestTypeId);
	double const typeTotalAmount = forestType->getTotalReproductionAmount();
	unsigned int const numSpeciesInComp = forestType->getNumSpecies();
	for (unsigned int speciesIndex=0; speciesIndex<numSpeciesInComp; speciesIndex++) {
		double const expectedNumberOfThisSpecies = expectedNumberOfNewTrees * (forestType->getSpeciesReproductionAmount(speciesIndex) / typeTotalAmount);
		string const & speciesName = forestType->getSpeciesName(speciesIndex);
		unsigned int const speciesId = SpeciesContainer::getId(speciesName);
		unsigned int const numberOfNewTrees = PoissonRandomNumberGenerator::get(expectedNumberOfThisSpecies);
		Log::log(6, "Actual number of new %s trees: %d", speciesName.c_str(), numberOfNewTrees);

		HomogenousPoissonPointGenerator pointGenerator;

		// Generate new tree locations using rejection sampling.
		unsigned int numberOfAcceptedPoints = 0;
		unsigned int requiredPositions = 0;

		// We only have to go through the trees nearby, not the whole forest.
		// Diameter limiting factor here. To avoid checking for updated diameters in
		// the reproduction for each compartment, the limitfinder is updated earlier!
		while (numberOfAcceptedPoints < numberOfNewTrees) {
			requiredPositions++;
			if (requiredPositions>100*numberOfNewTrees) {
				Log::logWarning(6, "Reproduction phase is taking too long. Area probably doesn't have space for more trees.");
				Log::logWarning(6, "Number of attempts: %u", requiredPositions);
				Log::logWarning(6, "Number of trees positioned: %u", numberOfAcceptedPoints);
				Log::logWarning(6, "Giving up...");
				break;
			}
			bool pointIsInsideOldTreeTrunk = false;

			Point point = pointGenerator.generatePointInside(*getPolygon());

#ifdef COMPARISONSEEDING
			UniformRandomNumberGenerator::setSeed(point.x*631262.3 + point.y);
#endif
			double const averageAge = ((double)simData.getIterationLength() + 1.0) / 2.0;
			unsigned int const age = PoissonRandomNumberGenerator::get(averageAge);
			Tree * newTree = new Tree(speciesId, age, point);

			const double fromDiameterInCentimeterToRadiusInMeterCoefficient = 1.0/200.0;
			vector<Tree*> const & trees = allTrees.getTreesCloseTo(newTree);
			for (vector<Tree*>::const_iterator oldTreeIt = trees.begin(); oldTreeIt!=trees.end(); ++oldTreeIt) {
				Tree const * const oldTree = *oldTreeIt;
				const double oldTreeRadius = oldTree->getMark(TreeMarks::DIAMETER)
						* fromDiameterInCentimeterToRadiusInMeterCoefficient;
				const double distanceBetweenTrees = newTree->distanceTo(*oldTree);
				// Check if the new tree is inside the old tree's trunk.
				if (distanceBetweenTrees <= oldTreeRadius) {
					pointIsInsideOldTreeTrunk = true;
					break;
				}
			}

#ifdef COMPARISONSEEDING
#ifdef USE_MPI
			if (otherCompartmentTrees.getForeignTrees().size() > 0) {
				LimitFinder & limitFinder = LimitFinder::Instance();
				limitFinder.setLimitingMode(limitedByDiameterAndCompRad);
				double const limitingValue = limitFinder.getLargestLimitingValue();
				otherCompartmentTrees.setLimitingValue(limitingValue);

				vector<Tree*> const & otherTrees = otherCompartmentTrees.getTreesCloseTo(newTree);
				for (vector<Tree*>::const_iterator oldTreeIt = otherTrees.begin(); oldTreeIt
						!=otherTrees.end(); ++oldTreeIt) {
					Tree const * const oldTree = *oldTreeIt;
					const double oldTreeRadius = oldTree->getMark(TreeMarks::DIAMETER)
							* fromDiameterInCentimeterToRadiusInMeterCoefficient;
					const double distanceBetweenTrees = newTree->distanceTo(*oldTree);
					// Check if the new tree is inside the old tree's trunk.
					if (distanceBetweenTrees <= oldTreeRadius) {
						pointIsInsideOldTreeTrunk = true;
						break;
					}
				}
			}
#endif
#endif

			if (pointIsInsideOldTreeTrunk) {
				// Location discarded.
				delete newTree;
			} else {
				// Location accepted.
#ifdef COMPARISONSEEDING
				if (getPolygon()->activeContains(point)) {
					addNewTree(newTree);
				} else {
					delete newTree;
				}
#else
				addNewTree(newTree);
#endif
				numberOfAcceptedPoints++;
			}
		}
		Log::log(6, "%d %s trees positioned after %d attempts.", numberOfAcceptedPoints, speciesName.c_str(), requiredPositions);
	}
}

void Compartment::calculateInitialStandAge() {
	uint64_t numberOfTrees = 0;
	uint64_t ageSum = 0;
	for (Tree * curTree = firstTree; curTree != NULL; curTree = curTree->getNextTreeInCompartment()) {
		numberOfTrees++;
		ageSum += curTree->getAge();
	}
#ifdef USE_MPI
	MPIUtil::sum(numberOfTrees, compartmentData.getCompartmentCommunicator());
	MPIUtil::sum(ageSum, compartmentData.getCompartmentCommunicator());
#endif
	double age = (double)ageSum / (double)numberOfTrees;
	compartmentData.setInitialStandAge((unsigned int)max(0, roundToInt(age)));
}

double Compartment::calculateNumberOfNewTreesPerHectare(SimulationData const & simData) const {
	Vector variables(RegenerationModel::NUMBER_OF_VARIABLES);

	RegenerationModelData const & data = simData.getRegenerationModelData();
	Vector const & coefficients = data.getCoefficients();
	RegenerationModel::calculateVariables(this,coefficients,variables);

	double const numNewTrees = RegenerationModel::calculateMark(variables,coefficients);
#ifdef CALCULATION_INSPECTION
	CalculationInspector& inspector = CalculationInspector::Instance();
	if (inspector.shouldInspectRegeneration()) {
		inspector.inspectRegenerationModel(data, variables, numNewTrees);
	}
#endif
	return max(0.0,numNewTrees);
}

#ifdef USE_MPI
void Compartment::communicateManagementHistory() {
	if (!getPolygon()->isActive()) {
		return;
	}
	vector<unsigned int> const & templateIds = compartmentData.getManagementTemplateIds();
	for (size_t i = 0, n = templateIds.size(); i < n; i++) {
		unsigned int const templateId = templateIds[i];
		unsigned int iteration = managementIterations[templateId];
		MPIUtil::max(iteration, compartmentData.getCompartmentCommunicator());
		managementIterations[templateId] = iteration;
	}
	MPIUtil::max(clearCutIteration, compartmentData.getCompartmentCommunicator());
}
#endif

