#ifndef COMPARTMENT_H_
#define COMPARTMENT_H_

#include <vector>
using std::vector;
#include <list>
using std::list;
#include <map>
using std::map;
#include <stdint.h>
#include "Polygon.h"
#include "CompartmentData.h"
#include "Tree.h"

class TreeContainer;
class SimulationData;

/**
 * A compartment is a smaller part of a forest, defined by a polygon.
 */
class Compartment {
public:
	Compartment(CompartmentData & c);
	virtual ~Compartment();
	inline const Polygon * getPolygon() const {
		return compartmentData.getCompartmentPolygon();
	}
	inline CompartmentData * getCompartmentData() const {
		return &compartmentData;
	}
	inline Tree* getFirstTree() const {
		return firstTree;
	}
	inline unsigned int size() const {
		return numberOfTrees;
	}
	inline list<Tree*> const & getNewTrees() const {
		return newTrees;
	}
	inline void addNewTree(Tree * const newTree) {
		newTree->setParentCompartment(this);
		newTrees.push_back(newTree);
	}
	inline void addNewTrees() {
		for(list<Tree*>::const_iterator treeIter=newTrees.begin(); treeIter!=newTrees.end(); ++treeIter) {
			addTree(*treeIter);
		}
		newTrees.clear();
	}
	template<class ForwardContainer>
	inline void addAll(ForwardContainer const & trees) {
		for (typename ForwardContainer::const_iterator curTree = trees.begin(); curTree != trees.end(); ++curTree) {
			addTree(*curTree);
		}
	}
	void addTree(Tree* const tree);
	void removeTree(Tree* const tree);

	inline double getTotalDiameter() const {
		return totalDiameter;
	}
	inline void setTotalDiameter(double total) {
		totalDiameter = total;
	}
	inline double getTotalBasalArea() const {
		return totalBasalArea;
	}
	inline void setTotalBasalArea(double total) {
		totalBasalArea = total;
	}
	inline double getTotalVolume() const {
		return fullCompartmentVolume;
	}
	inline void setTotalVolume(double vol) {
		fullCompartmentVolume = vol;
	}
	inline uint64_t getTotalSize() const {
		return totalNumberOfTrees;
	}
	inline void setTotalSize(uint64_t total) {
		totalNumberOfTrees = total;
	}
	inline unsigned int getManagementIteration(unsigned int templateId) const {
		map<unsigned int, unsigned int>::const_iterator it = managementIterations.find(templateId);
		if (it != managementIterations.end()) {
			return it->second;
		} else {
			return 0;
		}
	}
	inline void setManagementIteration(unsigned int templateId, unsigned int iteration) {
		managementIterations[templateId] = iteration;
	}
	inline unsigned int getClearCutIteration() const {
		return clearCutIteration;
	}
	inline void setClearCutIteration(unsigned int iteration) {
		clearCutIteration = iteration;
	}

	void calculateNewTreePositions(unsigned int const iteration, SimulationData const & simData,
			TreeContainer & allTrees, TreeContainer & otherCompartmentTrees);

	void calculateInitialStandAge();

#ifdef USE_MPI
	void communicateManagementHistory();
#endif
private:
	CompartmentData & compartmentData;
	Tree* firstTree;
	Tree* lastTree;
	double totalDiameter;
	double totalBasalArea;
	double fullCompartmentVolume;
	uint64_t totalNumberOfTrees;
	unsigned int numberOfTrees;
	map<unsigned int, unsigned int> managementIterations;
	unsigned int clearCutIteration;

	list<Tree*> newTrees;

	double calculateNumberOfNewTreesPerHectare(SimulationData const & simData) const;
};

#endif /*COMPARTMENT_H_*/
