#include "Polygon.h"
#include "RandomNumber.h"
#include "File.h"
#include "Log.h"

#include <iostream>
using std::ios;
using std::ios_base;
using std::cout;
using std::cerr;
#include <vector>
using std::vector;
using std::make_pair;

#include <sstream>
using std::stringstream;

Polygon::Polygon() :
	id(0), active(NULL) {
}

Polygon::Polygon(string const & filename, unsigned int id) :
	filename(filename), id(id), active(NULL) {
}

Polygon::Polygon(vector<Polygon *> const & polygons) :
	id(0), active(NULL) {
	// Use the first vertex of the first polygon as a junction point
	Point const & junction = polygons.at(0)->vertices.at(0);
	size_t numOfVertices = 0;
	for (size_t i = 0, n = polygons.size(); i < n; ++i) {
		numOfVertices += polygons[i]->vertices.size();
		if (i > 0) {
			// Count the junction vertex for all polygons except the first one.
			numOfVertices++;
		}
	}
	vector<Point> vertices(numOfVertices);
	size_t dest = 0;
	for (size_t i = 0, n = polygons.size(); i < n; ++i) {
		for (size_t j = 0, m = polygons[i]->vertices.size(); j < m; ++j) {
			vertices[dest++] = polygons[i]->vertices[j];
		}
		if (i > 0) {
			// Add a vertex at the junction point to all polygons except the first one.
			vertices[dest++] = junction;
		}
	}
	setVertices(vertices);
	children = polygons;
}

Polygon::~Polygon() {
	if (active != NULL)
		delete active;
	for (size_t i = 0, n = children.size(); i < n; i++)
		delete children[i];
	children.clear();
}

void Polygon::setVertices(vector<Point> const & polyVertices) {
	area = 0.0;

	vertices = polyVertices;
	if (vertices.empty())
		return;

	minCoordinate.x = vertices[0].x;
	minCoordinate.y = vertices[0].y;
	maxCoordinate = minCoordinate;

	// Find max and min coordinate and calculate area
	for (unsigned int i=1; i < vertices.size(); i++) {
		Point const & p = vertices[i];
		if (p.x < minCoordinate.x)
			minCoordinate.x = p.x;
		if (p.y < minCoordinate.y)
			minCoordinate.y = p.y;
		if (p.x > maxCoordinate.x)
			maxCoordinate.x = p.x;
		if (p.y > maxCoordinate.y)
			maxCoordinate.y = p.y;

		area += vertices[i-1].x*vertices[i].y - vertices[i-1].y*vertices[i].x;
	}

	if (area < 0.0)
		area = -area;

	area *= 0.5;

}

#ifdef USE_MPI
void Polygon::broadcast(int const rootId, MPI_Comm const comm) {
	MPIUtil::broadcast(filename, rootId, comm);
	MPIUtil::broadcast(id, rootId, comm);

	vector<double> buffer;
	size_t buffIndex = 0;

	if(MPIUtil::mpiId() == rootId) {
		size_t const numVertices = vertices.size();
		buffer.resize(numVertices*2 + 5);
		for(size_t index=0; index<numVertices; index++) {
			buffer[buffIndex++] = vertices[index].x;
			buffer[buffIndex++] = vertices[index].y;
		}
		buffer[buffIndex++] = minCoordinate.x;
		buffer[buffIndex++] = minCoordinate.y;
		buffer[buffIndex++] = maxCoordinate.x;
		buffer[buffIndex++] = maxCoordinate.y;
		buffer[buffIndex++] = area;
	}
	MPIUtil::broadcast(buffer,rootId,comm);
	if(MPIUtil::mpiId() != rootId) {
		size_t const numVertices = (buffer.size()-5)/2;
		vertices.resize(numVertices);
		for(size_t index=0; index<numVertices; index++) {
			vertices[index].x = buffer[buffIndex++];
			vertices[index].y = buffer[buffIndex++];
		}
		minCoordinate.x = buffer[buffIndex++];
		minCoordinate.y = buffer[buffIndex++];
		maxCoordinate.x = buffer[buffIndex++];
		maxCoordinate.y = buffer[buffIndex++];
		area = buffer[buffIndex++];
	}

	unsigned int numChildren = children.size();
	MPIUtil::broadcast(numChildren, rootId, comm);
	for (unsigned int i = 0; i < numChildren; i++) {
		if (MPIUtil::mpiId() == rootId) {
			children[i]->broadcast(rootId, comm);
		} else {
			Polygon * poly = new Polygon;
			poly->broadcast(rootId, comm);
			children.push_back(poly);
		}
	}
}
#endif

bool Polygon::contains(Point const & point) const {

	// If the polygon is empty, it cannot contain a point...
	if(vertices.size() < 3)
		return false;

	// Check bounding box first to avoid unneccessary work
	if(!boundingBoxContains(point))
		return false;

	// Draw a vertical line upwards from point and count number of intersections
	unsigned int intersections = 0;

	double const x = point.x;
	double const y = point.y;

	unsigned int const numVertices = vertices.size();

	for (unsigned int i=0; i < numVertices-1; i++) {
		Point leftPoint, rightPoint;
		if (vertices[i].x < vertices[i+1].x) {
			leftPoint = vertices[i];
			rightPoint = vertices[i+1];
		} else {
			leftPoint = vertices[i+1];
			rightPoint = vertices[i];
		}

		bool const startToTheLeft = (leftPoint.x <= x);
		bool const endToTheRight = (rightPoint.x > x);

		if (startToTheLeft == endToTheRight && (y-leftPoint.y)*(rightPoint.x-leftPoint.x) < (rightPoint.y-leftPoint.y)
				*(x-leftPoint.x)) {
			intersections++;
		}
	}

	return intersections%2;
}

Point Polygon::getActiveRandomPoint() const {
	if (active != NULL) {
#ifdef COMPARISONSEEDING
		return getRandomPoint();
#else
		return active->getRandomPoint();
#endif
	}
	return getRandomPoint();
}

Point Polygon::getRandomPoint() const {
	while (true) {
		const double x = minCoordinate.x+RandomNumber::get()*(maxCoordinate.x-minCoordinate.x);
		const double y = minCoordinate.y+RandomNumber::get()*(maxCoordinate.y-minCoordinate.y);

		Point point(x, y);
		if (contains(point)) {
			return point;
		}
	}
}

std::ostream& operator<<(std::ostream& s, const Polygon & poly) {
	s << "Polygon (area: "<< poly.area<< ") (Vertices: " << poly.vertices.size() << "): \n";
	s.precision(12);
	s.setf(ios::fixed);
	for (size_t i=0; i < poly.vertices.size(); i++) {
		Point const & p = poly.vertices[i];
		s << p << "\n";
	}
	return s;
}

double Polygon::getDividedArea(DivideType divideType, double splitCoordinate) const {
	vector<Point> newPolygonPoints = getDividedPolygonPoints(divideType, splitCoordinate);
	Polygon p;
	p.setVertices(newPolygonPoints);
	double dividedArea = p.getArea();
	return dividedArea;
}

void Polygon::dividePolygon(DivideType divideType, double splitCoordinate) {
	vector<Point> newPolygonPoints = getDividedPolygonPoints(divideType, splitCoordinate);
	setActivePolygon(newPolygonPoints);
}

vector<Point> Polygon::getDividedPolygonPoints(DivideType divideType, double splitCoordinate) const {
	vector<Point> newPolygonPoints;
	vector<Point> const * verticesPointer = &vertices;
	if (active != NULL) {
		verticesPointer = active->getVertices();
	}

	vector<Point> const & sourcePoints = (*verticesPointer);

	if (sourcePoints.size()==0)
		return vector<Point>();

	int correctSide = (divideType == KEEP_LEFT || divideType == KEEP_ABOVE);
	bool vertical = (divideType == KEEP_BELOW || divideType == KEEP_ABOVE);



	int currentSide = vertical ? (sourcePoints[0].y <= splitCoordinate) : (sourcePoints[0].x <= splitCoordinate);
	if (currentSide == correctSide) {
		newPolygonPoints.push_back(sourcePoints[0]);
	}

	for (unsigned int i=1; i < sourcePoints.size(); i++) {
		Point point = sourcePoints[i];
		int side = vertical ? (point.y <= splitCoordinate) : (point.x <= splitCoordinate);

		if (side != currentSide) {
			// Calculate y
			// using sourcePoints[i-1] == Point on one side
			// and point == point on the other side
			currentSide = side;
			double k, x, y;
			if (vertical) {
				y = splitCoordinate;
				k = (sourcePoints[i-1].x-point.x) /(sourcePoints[i-1].y -point.y);
				x = point.x+k*(splitCoordinate-point.y);
			} else {
				x = splitCoordinate;
				k = (sourcePoints[i-1].y-point.y) /(sourcePoints[i-1].x -point.x);
				y = point.y+k*(splitCoordinate-point.x);
			}
			newPolygonPoints.push_back(Point(x, y));
		}

		if (side == correctSide) {
			if (point != newPolygonPoints[newPolygonPoints.size()-1])
				newPolygonPoints.push_back(point);
		}
	}

	// Close polygon
	if (newPolygonPoints.size() > 1 && newPolygonPoints[0] != newPolygonPoints[newPolygonPoints.size()-1])
		newPolygonPoints.push_back(newPolygonPoints[0]);

	return newPolygonPoints;
}

map<unsigned int, Polygon*> Polygon::loadPolygons(string const & filename) {
	map<unsigned int, Polygon*> polygons;

	Log::log(-1, "Loading polygons from file %s", filename.c_str());

	// Read polygon from file
	File polyFile(filename, ios_base::in);
	vector<Point> currentPolygonVertices;
	currentPolygonVertices.reserve(50);

	while (true) {
		currentPolygonVertices.clear();

		// Read next polygon from file
		int polyId = readNextPolygon(polyFile, currentPolygonVertices);
		if (polyId == -1) {
			break;
		}

		// A polygon can't possibly have fewer than 3 vertices or it has no area...
		size_t const numberOfVertices = currentPolygonVertices.size();
		if (numberOfVertices < 3) {
			stringstream message;
			message << "Polygon: " << polyId << " has too few vertices to be useful. Number of vertices: "
					<< numberOfVertices;
			throw message.str();
		}

		Polygon *p = new Polygon(filename, polyId);
		p->setVertices(currentPolygonVertices);

		polygons.insert(make_pair(polyId, p));

		Log::log(-1, "Loaded polygon %d", polyId);

	}

	return polygons;
}

int Polygon::readNextPolygon(File & polyFile, vector<Point> & polygonVertices) {
	/* First we read the polygon data for the next polygon. Then we check if the data is followed by a polygon with same id. In that case we read the following vertices as cutoutVertices.*/
	int id = -1;
	bool readingCutOut = false;
	Point lastRealPolygonVertex(-1.0, -1.0);
	Point firstCutoutVertex(-1.0, -1.0);
	Point lastPoint(-1.0, -1.0);

	while (true) {
		double x, y;
		char buf[1001];
		polyFile.getline(buf, 1000);

		if (polyFile.eof()) {
			if (id != -1) {
				Log::logError(-1, "End of file encountered while reading polygon %d", id);
			}

			return -1;
		}
		if (strncmp(buf, "END", 3) == 0) {
			if (id != -1) {
				if (readingCutOut) {
					polygonVertices.push_back(firstCutoutVertex);
					polygonVertices.push_back(lastRealPolygonVertex);
				} else {
					lastRealPolygonVertex = lastPoint;
				}

				polyFile.getline(buf, 1000);
				if (polyFile.eof()) {
					return id;
				}

				int nextId;
				if (sscanf(buf, " %d\n", &nextId) && nextId == id) {
					// Continues with same polygon id, treat as cutout
					readingCutOut = true;
					firstCutoutVertex = Point(-1.0, -1.0);
					cerr << "Found cutout for polygon " << id << "\n";
				} else {
					// Continues with another polygon id, seek back so that next read returns this polygon id
					int idRead = polyFile.gcount();
					polyFile.seekg(-idRead, ios_base::cur);
					return id;
				}
			}
		} else if (sscanf(buf, " %lf %lf ", &x, &y) == 2) {
			if (id != -1) {
				if (lastPoint.x == x && lastPoint.y == y) {
				} else {
					lastPoint = Point(x, y);
					if (readingCutOut && firstCutoutVertex.x < 0.0) {
						firstCutoutVertex = lastPoint;
					}
					polygonVertices.push_back(lastPoint);
				}
			} else {
				cerr << "Position '" << buf << "' in polygon file does not belong to any polygon!\n";
			}
		} else if (sscanf(buf, " %d\n", &id) == 1) {
			// Do nothing
		} else if (buf[0] != '\r' && buf[0] != '\n') {
			Log::logError(4, "Invalid line in polygon file: '%s'", buf);
			cerr << "Invalid line in polygon file: '" << buf << "'\n";
		}
	}
}

