#ifndef POLYGON_H_
#define POLYGON_H_

#include <vector>
#include <map>
#include <string>
using std::string;
using std::vector;
using std::map;

#include "Point.h"
#ifdef USE_MPI
#include "MPIUtil.h"
#endif

class File;

enum DivideType { KEEP_LEFT,KEEP_RIGHT,KEEP_ABOVE,KEEP_BELOW};

class Polygon
{
public:
	Polygon();
	Polygon(string const & filename, unsigned int polygonId);
	/*
	 * Create a polygon by combining several ones. The new polygon will have
	 * child polygons.
	 */
	explicit Polygon(vector<Polygon *> const & polygons);

	~Polygon();

	void setVertices(vector<Point> const & polyVertices);
	inline vector<Point> const * getVertices() const {
		return &vertices;
	}
#ifdef USE_MPI
	void broadcast(int const rootId, MPI_Comm const comm);
#endif
	inline string const & getFilename() const {
		return filename;
	}
	inline unsigned int getId() const {
		return id;
	}

	inline Point const & getActiveMinCoordinate() const {
		if (active != NULL)
			return active->getMinCoordinate();
		return getMinCoordinate();
	}
	inline Point const & getActiveMaxCoordinate() const {
		if (active != NULL)
			return active->getMaxCoordinate();
		return getMaxCoordinate();
	}
	inline bool activeContains(Point const & point) const {
		if (active != NULL)
			return active->contains(point);
		return contains(point);
	}
	inline bool fullContains(Point const & point) const {
		return contains(point);
	}
	inline bool hasPoint(Point const & point) const {
		// If the polygon has no points, it doesn't have the point...
		if(vertices.size() < 1)
			return false;
		// Check bounding box first to avoid unneccessary work
		if(!boundingBoxContains(point))
			return false;
		const unsigned int numberOfVertices = this->vertices.size();
		for(unsigned int i=0; i<numberOfVertices; i++) {
			if(this->vertices[i] == point)
				return true;
		}
		return false;
	}
	inline double getActiveArea() const {
		if (active != NULL)
			return active->getArea();
		return getArea();
	}
	inline double getFullArea() const {
		return getArea();
	}
	inline bool isActive() const {
		if (active != NULL)
			return active->isActive();
		if (vertices.size() < 3)
			return false;
		return true;
	}
	Point getActiveRandomPoint() const;

	/*
	 * If the polygon is constructed out of several polygons, this
	 * method will return the sub-polygons. Otherwise it returns an
	 * empty vector.
	 */
	inline vector<Polygon*> const & getChildren() const {
		return children;
	}

	static map<unsigned int, Polygon*> loadPolygons(string const & filename);

	bool mightOverlap(Polygon const * const other) const {
		if (minCoordinate.x >= other->maxCoordinate.x || other->minCoordinate.x >= maxCoordinate.x)
			return false;
		if (minCoordinate.y >= other->maxCoordinate.y || other->minCoordinate.y >= maxCoordinate.y)
			return false;

		return true;
	}
private:
    static int readNextPolygon(File & polyFile, vector<Point> & polygonVertices);

	inline Point const & getMinCoordinate() const {
		return minCoordinate;
	}
	inline Point const & getMaxCoordinate() const {
		return maxCoordinate;
	}
	inline double getArea() const {
		return area;
	}
	inline Point getRelativePosition(Point const & p) const {
		return p-minCoordinate;
	}
    inline void reset() {
    	if (active != NULL)
    		delete active;
    	active = NULL;
    }
    inline void setActivePolygon(vector<Point> & vertices) {
    	if (active != NULL)
    		delete active;
    	active = new Polygon();
    	active->setVertices(vertices);
    }
    inline void setInactive() {
    	vector<Point> p;
    	setActivePolygon(p);
    }
	inline bool boundingBoxContains(Point const & point) const {
		if(minCoordinate <= point && point <= maxCoordinate) {
			return true;
		} else {
			return false;
		}
	}
	friend std::ostream& operator<<(std::ostream& s, const Polygon & poly);
	bool contains(Point const & point) const;
	Point getRandomPoint() const;
    double getDividedArea(DivideType splitType, double splitCoordinate) const;
    void dividePolygon(DivideType splitType, double splitCoordinate);
	void readFromFile(string const filename, int const polyId);
	bool readSetOfVertices(File & polyFile, int const polyId, vector<Point> & vertices);
    vector<Point> getDividedPolygonPoints(DivideType divideType, double splitCoordinate) const;

	string filename;
	unsigned int id;

	vector<Point> vertices;

	Polygon *active;
	Point minCoordinate;
	Point maxCoordinate;
	double area;

	vector<Polygon*> children;

	friend class ForestSplitter;
};

std::ostream& operator<<(std::ostream& s, const Polygon & poly);

#endif /*POLYGON_H_*/
