#include "RectangleContainer.h"
#include "Compartment.h"
#include "Tree.h"
#include "TreeContainer.h"
#include "Log.h"

using std::max;
using std::min;
using std::numeric_limits;
#include <cassert>

RectangleContainer::RectangleContainer(TreeContainer const & parent) :
	itsTreesByLocationUpToDate(false), parentTreeContainer(parent), minPoint(numeric_limits<double>::max(),
			numeric_limits<double>::max()), maxPoint(numeric_limits<double>::min(), numeric_limits<double>::min()),
			limitingValue(10.0), rectanglesUpToDate(false) {
}

RectangleContainer::~RectangleContainer() {
	vector<vector<vector<Tree*> > > ().swap(itsTreesByLocation);
}

vector<Tree*> const & RectangleContainer::getVectorContaining(Tree const * const tree) {
	if (!rectanglesUpToDate) {
		setupRectangles();
	}
	if (!itsTreesByLocationUpToDate) {
		clear();
		placeAllTrees();
	}
#ifndef DISABLE_RECTANGLES
	return getRectangleContainingPoint(tree->getX(), tree->getY());
#else
	//return the whole forest
	return allTrees;
#endif
}

void RectangleContainer::add(Tree* tree) {
	if (!rectanglesUpToDate) {
		setupRectangles();
	}
	if (!itsTreesByLocationUpToDate) {
		clear();
		placeAllTrees();
	}
	placeTree(tree);
}

void RectangleContainer::clear() {
	itsTreesByLocationUpToDate = false;
#ifndef DISABLE_RECTANGLES
	for (size_t i = 0; i < itsTreesByLocation.size(); i++) {
		for (size_t j = 0; j < itsTreesByLocation[i].size(); j++) {
			vector<Tree*> ().swap(itsTreesByLocation[i][j]);
		}
	}
#else
	allTrees.clear();
#endif
}
size_t RectangleContainer::approxSizeInBytes() const {
	size_t const numberOfTrees = parentTreeContainer.size();
#ifdef USE_MPI
	size_t const numberOfForeignTrees = parentTreeContainer.getForeignTrees().size();
	size_t treeBytes = (numberOfTrees + numberOfForeignTrees) * 4 * sizeof(Tree*);
#else
	size_t treeBytes = numberOfTrees * 4 * sizeof(Tree*);
#endif

#ifdef DISABLE_RECTANGLES
	treeBytes /= 4;
#endif
	return treeBytes;
}
size_t RectangleContainer::trueSizeInBytes() const {
	size_t sizeInBytes = 0;

	size_t const rectanglesInX = itsTreesByLocation.size();
	sizeInBytes += rectanglesInX * sizeof(vector<vector<Tree*> > );
	for (size_t i = 0; i < rectanglesInX; i++) {
		size_t const rectanglesInY = itsTreesByLocation[i].size();
		sizeInBytes += rectanglesInY * sizeof(vector<Tree*> );
		for (size_t j = 0; j < rectanglesInY; j++) {
			size_t const numberOfTrees = itsTreesByLocation[i][j].size();
			sizeInBytes += numberOfTrees * sizeof(Tree*);
		}
	}
#ifdef DISABLE_RECTANGLES
	sizeInBytes += allTrees.size() * sizeof(Tree*);
#endif
	return sizeInBytes;
}

void RectangleContainer::setupRectangles() {
#ifndef DISABLE_RECTANGLES
	// Finding the smallest possible bounding box for all our Compartments.
	setBoundingBoxCoordinates();

	assert(minPoint < maxPoint);

	// Dividing the area into a reasonable number of rectangles.
	// Using the limits of the bounding box and the specified limiting value
	// Setting a minimum rectangle size here to avoid too many of them.
	double const limitingValueUsed = max(2.0 * limitingValue, 20.0);

	// Round down to integer number of rectangles.
	size_t numberOfRectanglesInXDirection = (size_t) ((maxPoint.x - minPoint.x) / limitingValueUsed);
	size_t numberOfRectanglesInYDirection = (size_t) ((maxPoint.y - minPoint.y) / limitingValueUsed);
	if (numberOfRectanglesInXDirection == 0)
		numberOfRectanglesInXDirection = 1;
	if (numberOfRectanglesInYDirection == 0)
		numberOfRectanglesInYDirection = 1;

	itsTreesByLocation.clear();
	itsTreesByLocation.resize(numberOfRectanglesInXDirection);

	for (size_t i = 0; i < numberOfRectanglesInXDirection; i++) {
		itsTreesByLocation[i].resize(numberOfRectanglesInYDirection);
	}
#endif
	rectanglesUpToDate = true;
}

void RectangleContainer::placeAllTrees() {
	//	cout << "Placing all trees in rectangles.\n";
	for (TreeContainer::const_iterator treeIter = parentTreeContainer.begin(); treeIter != parentTreeContainer.end(); ++treeIter) {
#ifndef DISABLE_RECTANGLES
		placeTree(*treeIter);
#else
		allTrees.push_back(*treeIter);
#endif
	}
#ifdef USE_MPI
	vector<Tree*> const & foreignTrees = parentTreeContainer.getForeignTrees();
	for (vector<Tree*>::const_iterator treeIter = foreignTrees.begin(); treeIter != foreignTrees.end(); ++treeIter) {
#ifndef DISABLE_RECTANGLES
		placeTree(*treeIter);
#else
		allTrees.push_back(*treeIter);
#endif
	}
#endif
	itsTreesByLocationUpToDate = true;
}

vector<Tree*> & RectangleContainer::getRectangleContainingPoint(double const xCoord, double const yCoord) {
#ifndef DISABLE_RECTANGLES
	size_t const rectX = getRectangleCoordinate(xCoord, minPoint.x, maxPoint.x, itsTreesByLocation.size());
	size_t const rectY = getRectangleCoordinate(yCoord, minPoint.y, maxPoint.y, itsTreesByLocation[0].size());

	return itsTreesByLocation[rectX][rectY];
#else
	return allTrees;
#endif
}

size_t RectangleContainer::getRectangleCoordinate(double const coord, double const minCoord, double const maxCoord,
		size_t const numberOfRectangles) const {
	assert(numberOfRectangles>0);
	// Part of these calculations will "always" return the same and as this function is called a lot,
	// it will benefit a lot from function inlining by the compiler.
	double const offsetCoord = coord - minCoord;
	double const forestSize = maxCoord - minCoord;
	double const rectangleSize = forestSize / numberOfRectangles;
	size_t rectangleCoord;
	// Check for trees outside the process borders.
	if (offsetCoord <= 0.0)
		rectangleCoord = 0;
	else
		rectangleCoord = (size_t) (offsetCoord / rectangleSize);
	if (rectangleCoord >= numberOfRectangles)
		rectangleCoord = numberOfRectangles - 1;
	return rectangleCoord;
}
void RectangleContainer::setBoundingBoxCoordinates() {
	minPoint.x = numeric_limits<double>::max();
	minPoint.y = numeric_limits<double>::max();
	maxPoint.x = numeric_limits<double>::min();
	maxPoint.y = numeric_limits<double>::min();
	bool hasPolygonCoordinates = false;
	for (TreeContainer::const_comp_iterator compIter = parentTreeContainer.compBegin(); compIter != parentTreeContainer.compEnd(); ++compIter) {
		Polygon const * const polygon = (*compIter)->getPolygon();
		if (polygon->isActive()) {
			Point const minPolygonCoord = polygon->getActiveMinCoordinate();
			Point const maxPolygonCoord = polygon->getActiveMaxCoordinate();
			minPoint.x = min(minPolygonCoord.x, minPoint.x);
			minPoint.y = min(minPolygonCoord.y, minPoint.y);
			maxPoint.x = max(maxPolygonCoord.x, maxPoint.x);
			maxPoint.y = max(maxPolygonCoord.y, maxPoint.y);
			hasPolygonCoordinates = true;
		}
	}
	if (!hasPolygonCoordinates) {
		for (TreeContainer::const_iterator treeIter = parentTreeContainer.begin(), endIter = parentTreeContainer.end(); treeIter != endIter; ++treeIter) {
			Point const position = (*treeIter)->getPosition();
			minPoint.x = min(position.x, minPoint.x);
			minPoint.y = min(position.y, minPoint.y);
			maxPoint.x = max(position.x, maxPoint.x);
			maxPoint.y = max(position.y, maxPoint.y);
		}
	}
#ifdef USE_MPI
	vector<Tree*> const & foreignTrees = parentTreeContainer.getForeignTrees();
	for (vector<Tree*>::const_iterator treeIter = foreignTrees.begin(); treeIter != foreignTrees.end(); ++treeIter) {
		double const xCoord = (*treeIter)->getX();
		double const yCoord = (*treeIter)->getY();
		minPoint.x = min(xCoord, minPoint.x);
		minPoint.y = min(yCoord, minPoint.y);
		maxPoint.x = max(xCoord, maxPoint.x);
		maxPoint.y = max(yCoord, maxPoint.y);
	}
#endif
}

void RectangleContainer::placeTree(Tree * const curTree) {
	double const xCoord = curTree->getX();
	double const yCoord = curTree->getY();

	// The coordinates passed here will of course often be outside the actual area.
	vector<Tree*> & rectangle1 = getRectangleContainingPoint(xCoord + limitingValue, yCoord + limitingValue);
	vector<Tree*> & rectangle2 = getRectangleContainingPoint(xCoord + limitingValue, yCoord - limitingValue);
	vector<Tree*> & rectangle3 = getRectangleContainingPoint(xCoord - limitingValue, yCoord + limitingValue);
	vector<Tree*> & rectangle4 = getRectangleContainingPoint(xCoord - limitingValue, yCoord - limitingValue);

	rectangle1.push_back(curTree);
	if (!(&rectangle1 == &rectangle2))
		rectangle2.push_back(curTree);
	if (!(&rectangle1 == &rectangle3 || &rectangle2 == &rectangle3))
		rectangle3.push_back(curTree);
	if (!(&rectangle1 == &rectangle4 || &rectangle2 == &rectangle4 || &rectangle3 == &rectangle4))
		rectangle4.push_back(curTree);
}
