#include "TreeContainer.h"
#include "Compartment.h"
#include "File.h"
#include "PaddedField.h"
#include "parameters.h"
#include "CompartmentData.h"
#include "Polygon.h"
#include "Point.h"
#include "StandardNormalRandomNumberGenerator.h"
#ifdef COMPARISONSEEDING
#include "UniformRandomNumberGenerator.h"
#endif

#include <algorithm>
using std::remove;
#include <ios>
using std::ios_base;
using std::ios;

TreeContainer::TreeContainer(bool owns) :
	rectangleContainer(*this),
	ownsObjects(owns) {
}

TreeContainer::~TreeContainer() {
	// When the treecontainer is destroyed it should take care of all trees it had.
	clear();
}

TreeContainer::CompartmentMap TreeContainer::getCompartmentMap() const {
	CompartmentMap compMap;	
	for (const_comp_iterator compIter = compBegin(), compEndIter = compEnd();
			compIter != compEndIter; ++compIter) {
		unsigned int const compartmentId = (*compIter)->getCompartmentData()->getIdNumber();
		compMap[compartmentId] = *compIter;
	}
	return compMap;
}

void TreeContainer::clear() {
	if (ownsObjects) {
		const size_t numberOfTrees = itsTrees.size();
		const size_t numberOfCompartments = compartments.size();
		for (size_t curTreeIndex = 0; curTreeIndex<numberOfTrees; curTreeIndex++) {
			delete itsTrees[curTreeIndex];
		}
		for (size_t curCompartmentIndex = 0; curCompartmentIndex<numberOfCompartments; curCompartmentIndex++) {
			delete compartments[curCompartmentIndex];
		}
#ifdef USE_MPI
		const size_t numberOfForeignTrees = foreignTrees.size();
		for (size_t curForeignTreeIndex = 0; curForeignTreeIndex<numberOfForeignTrees; curForeignTreeIndex++) {
			delete foreignTrees[curForeignTreeIndex];
		}
#endif
	}
	vector<Tree*>().swap(itsTrees);
	vector<Compartment*>().swap(compartments);
#ifdef USE_MPI
	vector<Tree*>().swap(foreignTrees);
#endif
	clearRectangles();
}
void TreeContainer::add(Tree* tree) {
	itsTrees.push_back(tree);
	rectangleContainer.add(tree);
}
void TreeContainer::innerAdd(Tree* tree) {
	itsTrees.push_back(tree);
	// Adding a tree will make the contents of the rectangles outdated.
	rectangleContainer.setTreesOutDated();
}

void TreeContainer::addCompartment(Compartment* compartment) {
	compartments.push_back(compartment);
	rectangleContainer.setRectanglesOutDated();
	Tree* tree = compartment->getFirstTree();
	itsTrees.reserve(itsTrees.size() + compartment->size());
	while (tree!=NULL) {
		innerAdd(tree);
		tree=tree->getNextTreeInCompartment();
	}
}

void TreeContainer::prepareForErase(iterator treePosition) {
	// erasing a tree must remove it from its compartment
	(*treePosition)->getParentCompartment()->removeTree(*treePosition);
	// erasing a tree must free the allocated memory
	if (ownsObjects)
		delete *treePosition;
	// erasing a tree will make all pointers to it bad so they can't be used.
	rectangleContainer.setTreesOutDated();
	// setting the deleted pointer to point to NULL
	*treePosition = NULL;
}
void TreeContainer::prepareForErase(Tree const * const toErase) {
	for(iterator curTree = itsTrees.begin(); curTree!=itsTrees.end(); curTree++) {
		if(*curTree == toErase) {
			prepareForErase(curTree);
			return;
		}
	}
}
void TreeContainer::massErase() {
	// Rebuild the vector, excluding all nullpointers.
	Tree * const toDelete = NULL;
	itsTrees.erase(remove(itsTrees.begin(),itsTrees.end(),toDelete),itsTrees.end());
}

TreeContainer::iterator TreeContainer::erase(iterator treePosition) {
	prepareForErase(treePosition);
	// erase the pointer from the vector.
	return itsTrees.erase(treePosition);
}

TreeContainer::iterator TreeContainer::erase(Tree const * const toErase) {
	for(iterator curTree = itsTrees.begin(); curTree!=itsTrees.end(); curTree++) {
		if(*curTree == toErase) {
			return erase(curTree);
		}
	}
	return itsTrees.end();
}

void TreeContainer::eraseDeadAndMarked() {
	for(iterator treeIter=itsTrees.begin(); treeIter!=itsTrees.end(); ++treeIter) {
		Tree const * const curTree = *treeIter;
		if( curTree->isDead() || curTree->isMarked() ) {
			prepareForErase(treeIter);
		}
	}
	massErase();
}

vector<Tree*> const & TreeContainer::getTreesCloseTo(Tree const * const tree) {
	return rectangleContainer.getVectorContaining(tree);
}

#ifdef INCLUDE_ENVIRONMENTAL_EFFECTS
void TreeContainer::assignRandomNumbers(bool includeSets[Tree::NUMBER_OF_RND_NUM_SETS]) {
	iterator i = begin(), iend = end();
	for (; i != iend; ++i) {
		for (unsigned int set = 0; set < Tree::NUMBER_OF_RND_NUM_SETS; ++set) {
			if (includeSets[set]) {
				for (unsigned int mark = 0; mark < TreeMarks::NUMBER_OF_MARKS; ++mark) {
#ifdef COMPARISONSEEDING
					UniformRandomNumberGenerator::setSeed((*i)->getInitialHeterogeneity());
#endif
					(*i)->setRandomNumber(set, mark, StandardNormalRandomNumberGenerator::get());
				}
			}
		}
	}
}
#endif

#ifdef USE_MPI
void TreeContainer::clearForeignTrees() {
	rectangleContainer.setTreesOutDated();
	if (ownsObjects) {
		const size_t numberOfForeignTrees = foreignTrees.size();
		for (size_t index=0; index<numberOfForeignTrees; index++) {
			delete foreignTrees[index];
		}
	}
	vector<Tree*>().swap(foreignTrees);
}
void TreeContainer::addForeignTrees(vector<Tree*> const & foreign) {
	rectangleContainer.setTreesOutDated();
	rectangleContainer.setRectanglesOutDated();
	foreignTrees.insert(foreignTrees.end(), foreign.begin(), foreign.end());
}
vector<Tree*> const TreeContainer::getCompartmentForeignTrees(Compartment const * compartment) const {
	vector<Tree*> currentCompartmentForeignTrees;
	const unsigned int numberOfForeignTrees = foreignTrees.size();
	Polygon const * const compartmentPolygon = compartment->getPolygon();
	for (size_t curForeignTreeIndex = 0; curForeignTreeIndex<numberOfForeignTrees; curForeignTreeIndex++) {
		Tree * const curTree = foreignTrees[curForeignTreeIndex];
		if (compartmentPolygon->fullContains(curTree->getPosition())) {
			currentCompartmentForeignTrees.push_back(curTree);
		}
	}
	return currentCompartmentForeignTrees;
}
#endif

