#ifndef TREECONTAINER_H_
#define TREECONTAINER_H_

#include <vector>
using std::vector;
#include <string>
using std::string;
#include <map>
using std::map;
#include <iterator>
using std::input_iterator_tag;


#include "RectangleContainer.h"
#include "Tree.h"

class Compartment;
class CompartmentData;

/**
 * This is an advanced container for trees.
 * Provides methods for adding and removing trees, for iteration of the container
 * and for finding the neighbours of trees.
 * Internally handles all Trees by storing them both by location and by Compartment.
 */
class TreeContainer {
public:

	TreeContainer(bool owns=true);
	~TreeContainer();

	// Returns iterators that work similarly to the STL iterators.
	// For trees
	typedef vector<Tree*>::iterator iterator;
	typedef vector<Tree*>::const_iterator const_iterator;
	inline iterator begin() { return itsTrees.begin(); }
	inline iterator end() { return itsTrees.end(); }
	inline const_iterator begin() const { return itsTrees.begin(); }
	inline const_iterator end() const { return itsTrees.end(); }
	inline Tree* operator[](unsigned int const index) const { return itsTrees[index]; }
	inline Tree* & operator[](unsigned int const index) { return itsTrees[index]; }
	inline size_t size() const { return itsTrees.size(); }

	// For compartments
	typedef vector<Compartment*>::const_iterator const_comp_iterator;
	inline const_comp_iterator compBegin() const { return compartments.begin(); }
	inline const_comp_iterator compEnd() const { return compartments.end(); }
	inline size_t compSize() const { return compartments.size(); }

	typedef map<unsigned int, Compartment*> CompartmentMap;
	CompartmentMap getCompartmentMap() const;

	/**
	 * Deletes all trees, all compartments and forces the size of the vectors to zero.
	 */
	void clear();
	/**
	 * Adds the specified tree to the TreeContainer AND the RectangleContainer.
	 */
	void add(Tree* tree);
	/**
	 * Adds the given Compartment and all its trees to this container.
	 */
	void addCompartment(Compartment * compartment);
	/**
	 * Adds all trees in the specified container to this container.
	 * Assumes that they belong to a compartment.
	 */
	template<class ForwardContainer>
	inline void addAll(ForwardContainer const & trees){
		itsTrees.reserve(itsTrees.size() + trees.size());
		for (typename ForwardContainer::const_iterator curTree = trees.begin(); curTree != trees.end(); ++curTree) {
			innerAdd(*curTree);
		}
	}
	/**
	 * Returns a vector containing trees which are relevant
	 * for the calculation of competitionindexes .
	 */
	vector<Tree*> const & getTreesCloseTo(Tree const * const tree);
#ifdef INCLUDE_ENVIRONMENTAL_EFFECTS
	/**
	 * Assign a normally distributed random number for each mark of every tree
	 * in the container.
	 */
	void assignRandomNumbers(bool includeSets[Tree::NUMBER_OF_RND_NUM_SETS]);
#endif
	/**
	 * Sets the value that limits the size of the rectangles in the RectangleContainer
	 */
	inline void setLimitingValue(double const val) {
		rectangleContainer.setLimitingValue(val);
	}
	void prepareForErase(iterator treePosition);
	void prepareForErase(Tree const * const toErase);
	void massErase();
	/**
	 * Erases the tree at the specified position.
	 * Returns an iterator to the tree following the tree that was erased.
	 */
	iterator erase(iterator treePosition);
	iterator erase(Tree const * const toErase);
	/*
	 * Erases all dead trees or trees marked for removal. More efficient
	 * that erasing element by element.
	 */
	void eraseDeadAndMarked();

	inline void clearRectangles() {
		rectangleContainer.clear();
	}
	inline size_t approxRectangleSizeInBytes() const {
		return rectangleContainer.approxSizeInBytes();
	}
	inline size_t trueRectangleSizeInBytes() const {
		return rectangleContainer.trueSizeInBytes();
	}
	inline bool isOwner() const {
		return ownsObjects;
	}

#ifdef USE_MPI
	void clearForeignTrees();
	inline vector<Tree*> const & getForeignTrees() const {
		return foreignTrees;
	}
	void addForeignTrees(vector<Tree*> const & foreign);
	vector<Tree*> const getCompartmentForeignTrees(Compartment const * compartment) const;
#endif

private:
	vector<Tree*> itsTrees;
	vector<Compartment*> compartments;
#ifdef USE_MPI
	vector<Tree*> foreignTrees;
#endif

	RectangleContainer rectangleContainer;
	bool const ownsObjects;

	/**
	 * Adds the tree to the TreeContainer, NOT the RectangleContainer.
	 */
	void innerAdd(Tree* tree);

};

#endif /*TREECONTAINER_H_*/
