/*
 * ForestInitializer.cpp
 *
 *  Created on: 2008-nov-05
 *      Author: joschori
 */
#include "ForestInitializer.h"
#include "TreeContainer.h"
#include "TreePersistence.h"
#include "parameters.h"
#include "PaddedField.h"
#include "File.h"
#include "Tree.h"
#include "Log.h"
#include "TreeSender.h"
#include "Compartment.h"
#include "InitialCompartmentGenerator.h"
#include "LimitFinder.h"
#include "Timer.h"
#include "UniformRandomNumberGenerator.h"

#ifdef USE_MPI
#ifdef INCLUDE_ENVIRONMENTAL_EFFECTS
#include "TreeSorter.h"
#endif
#endif

#include <iostream>
using std::ios_base;
#include <sstream>
using std::istringstream;
#include <iomanip>
using std::skipws;

bool ForestInitializer::treesLoaded = false;
bool ForestInitializer::positionLoaded = false;
bool ForestInitializer::speciesLoaded = false;
bool ForestInitializer::heterogeneityLoaded = false;
bool ForestInitializer::ageLoaded = false;
bool ForestInitializer::marksLoaded[TreeMarks::NUMBER_OF_MARKS];

void ForestInitializer::load(string const & filename, TreeContainer & trees) {
	for (size_t index = 0; index < TreeMarks::NUMBER_OF_MARKS; index++) {
		marksLoaded[index] = false;
	}

	if (filename.size() > 0) {
		checkFile(filename);
		TreePersistence::read(filename, trees);
	}
}

void ForestInitializer::generate(InitialCompartmentGenerator const & compartmentGenerator, int const replication,
		TreeContainer & trees) {
	LimitFinder & limitFinder = LimitFinder::Instance();

	if (!treesLoaded || !positionLoaded || !speciesLoaded || !heterogeneityLoaded || !ageLoaded) {
		for (TreeContainer::const_comp_iterator compIter = trees.compBegin(); compIter != trees.compEnd(); ++compIter) {
			Compartment * const compartment = *compIter;
			Polygon const * const polygon = compartment->getPolygon();
			if (polygon->isActive()) {
#ifdef COMPARISONSEEDING
				size_t const index = distance(trees.compBegin(),compIter);
				UniformRandomNumberGenerator::setSeed(2145.0*(index+1)+replication*trees.compSize());
#endif
				compartmentGenerator.generateTreeData(compartment);
				// Add the newly generated trees to the treecontainer and compartments.
				trees.addAll(compartment->getNewTrees());
				compartment->addNewTrees();
			}
		}
	}
	if (ageLoaded) {
		for (TreeContainer::const_comp_iterator compIter = trees.compBegin(); compIter != trees.compEnd(); ++compIter) {
			Compartment * const compartment = *compIter;
			CompartmentData * const compData = compartment->getCompartmentData();
			Polygon const * const polygon = compartment->getPolygon();
			if (polygon->isActive()) {
				compartment->calculateInitialStandAge();
				Log::log(2, "Stand age of compartment %u is %u years", compData->getIdNumber(), compData->getInitialStandAge());
			}
		}
	}
#ifdef INCLUDE_ENVIRONMENTAL_EFFECTS
	if (compartmentGenerator.hasEnvironmentalEffects()) {
		// Generate and assign random numbers for environmental effects
		bool includeSets[Tree::NUMBER_OF_RND_NUM_SETS] = {true, false};
		trees.assignRandomNumbers(includeSets);
#ifdef USE_MPI
		// Sort the trees so that correlated random numbers can be calculated in
		// correct order.
		Log::log(3, "Sorting trees for local Cholesky decomposition.");
		Timer sortingTime;
		TreeSorter::sortForLocalCholesky(trees);
		Log::log(3, "Tree sorting took %.2fs.", sortingTime.getElapsed());
#endif
	}
#endif
#ifdef USE_MPI
	vector<Tree*> foreignTrees = TreeSender::transferForeignTrees(trees);
	trees.addForeignTrees(foreignTrees);
#endif
	limitFinder.update(trees);
	limitFinder.setLimitingMode(limitedByCompRad);
	trees.setLimitingValue(limitFinder.getLargestLimitingValue());
	if (!ForestInitializer::allMarksLoaded()) {
		compartmentGenerator.generateTreeMarks(trees);
	}
}

void ForestInitializer::checkFile(string const & filename) {

	// Read the first line from the specified tree data file.
	File in(filename, ios_base::in);
	string loadedLine;
	getline(in, loadedLine);

	// The number of expected delimiters depends on what the << and >> operators in Tree work on.
	size_t const expectedNumberOfDelimiters = 4 + TreeMarks::NUMBER_OF_MARKS;

	// Count the number of delimiters in the loaded string.
	size_t numberOfDelimiters = 0;
	for (size_t pos = 0; pos < loadedLine.size(); pos++) {
		if (loadedLine[pos] == delimiterChar) {
			numberOfDelimiters++;
		}
	}

	// Throw error if the file contains too few columns.
	if (numberOfDelimiters != expectedNumberOfDelimiters) {
		stringstream messagestream;
		messagestream << "Loaded tree data file contains only ";
		messagestream << numberOfDelimiters;
		messagestream << " column delimiters, expected ";
		messagestream << expectedNumberOfDelimiters;
		messagestream << ".";
		throw messagestream.str();
	}

	// File generally OK.
	treesLoaded = true;

	// Check which columns contain data
	size_t start = 0;
	vector<bool> columnOk(expectedNumberOfDelimiters + 1, false);
	vector<bool>::iterator colIter;
	for (colIter = columnOk.begin(); colIter != columnOk.end(); ++colIter) {
		size_t next = loadedLine.find(delimiterChar, start);
		istringstream stream;
		stream.str(loadedLine.substr(start, next - start));
		string innerstring;
		stream >> skipws >> innerstring;
		Log::log(-1, "Parsed column (start %u and end %u) contained \"%s\".", start, next, innerstring.c_str());
		if (innerstring.size() > 0) {
			Log::log(-1, "Column accepted.");
			*colIter = true;
		}
		start = (next != string::npos) ? next + 1 : loadedLine.size() - 1;
	}

	// Mark which data is loaded (i.e. not generated later)
	// Column ids here depend on the order in the Tree class!

	// X and Y
	positionLoaded = columnOk[0] && columnOk[1];
	// Marks
	for (size_t mark = 0; mark < TreeMarks::NUMBER_OF_MARKS; mark++) {
		marksLoaded[mark] = columnOk[2 + mark];
	}
	// Age
	ageLoaded = columnOk[2 + TreeMarks::NUMBER_OF_MARKS];
	// Heterogeneity
	heterogeneityLoaded = columnOk[3 + TreeMarks::NUMBER_OF_MARKS];
	// Species
	speciesLoaded = columnOk[4 + TreeMarks::NUMBER_OF_MARKS];
}
