#include "InitialCompartmentGenerator.h"
#include "loadCoefficients.h"
#include "TreeContainer.h"
#include "Point.h"
#include "HomogenousPoissonPointGenerator.h"
#include "InhomogenousPoissonPointGenerator.h"
#include "HomogenousClusteredPointGenerator.h"
#include "InhomogenousClusteredPointGenerator.h"
#include "HomogenousRegularPointGenerator.h"
#include "InhomogenousRegularPointGenerator.h"
#include "NullPointGenerator.h"
#include "StandardNormalRandomNumberGenerator.h"
#include "PoissonRandomNumberGenerator.h"
#include "Model.h"
#include "SecondaryModel.h"
#include "Species.h"
#ifdef INCLUDE_ENVIRONMENTAL_EFFECTS
#include "LocalCorrelation.h"
#include "Matrix.h"
#endif
#include "Polygon.h"
#include "Compartment.h"
#include "Tree.h"
#include "Log.h"
#include "UniformRandomNumberGenerator.h"
#ifdef USE_MPI
#include "CompartmentCommunicator.h"
#endif
#include "CompartmentData.h"
#include "SimulationData.h"
#include "LimitFinder.h"
#include "ForestTypeContainer.h"
#include "ForestType.h"
#include "SpeciesContainer.h"
#include "ForestInitializer.h"
#include "RandomNumber.h"
#include "classify.h"
#include "MarkTransformer.h"
#include "roundToInt.h"

#include <xercesc/util/XMLString.hpp>
#include <xercesc/dom/DOMElement.hpp>
XERCES_CPP_NAMESPACE_USE

#include "XercesString.h"
#include "XercesUtilities.h"

#ifdef CALCULATION_INSPECTION
#include "CalculationInspector.h"
#endif

#include <map>
using std::map;
#include <vector>
using std::vector;
#include <string>
using std::string;

#include "File.h"
#include <ios>
using std::ios_base;
using std::ios;
#include <iostream>
using std::cout;
using std::max;

InitialCompartmentGenerator::InitialCompartmentGenerator(SimulationData const & sim) :
	simulationData(sim), pointGenerator(new NullPointGenerator), initializationDatas(TreeMarks::NUMBER_OF_MARKS),
			averageInitialAge(1) {
}

InitialCompartmentGenerator::~InitialCompartmentGenerator() {
}
#ifdef USE_MPI
void InitialCompartmentGenerator::broadcast(int const rootId, MPI_Comm const comm) {
	PGObjectType pgType = pointGenerator->getObjectType();
	int pgTypeAsInt = (int) pgType;
	MPIUtil::broadcast(pgTypeAsInt, rootId, comm);
	if (MPIUtil::mpiId() != rootId) {
		PGObjectType receivePGType = (PGObjectType) pgTypeAsInt;
		switch (receivePGType) {
		case eNULLPG:
			pointGenerator.reset(new NullPointGenerator);
			break;
		case eHOMOCLUSPG:
			pointGenerator.reset(new HomogenousClusteredPointGenerator);
			break;
		case eHOMOPOISPG:
			pointGenerator.reset(new HomogenousPoissonPointGenerator);
			break;
		case eHOMOREGUPG:
			pointGenerator.reset(new HomogenousRegularPointGenerator);
			break;
		case eINHOCLUSPG:
			pointGenerator.reset(new InhomogenousClusteredPointGenerator);
			break;
		case eINHOPOISPG:
			pointGenerator.reset(new InhomogenousPoissonPointGenerator);
			break;
		case eINHOREGUPG:
			pointGenerator.reset(new InhomogenousRegularPointGenerator);
			break;
		}
	}
	pointGenerator->broadcast(rootId, comm);

	for (size_t index = 0; index < TreeMarks::NUMBER_OF_MARKS; index++) {
		initializationDatas[index].broadcast(rootId, comm);
	}

	MPIUtil::broadcast(averageInitialAge, rootId, comm);
}
#endif

void InitialCompartmentGenerator::generateTreeData(Compartment * compartment) const {
	CompartmentData * const compData = compartment->getCompartmentData();
	Log::log(2, "Generating tree data for compartment id: %u", compData->getIdNumber());

	double const numFactor = compData->getNumberOfInitialTreesFactor();
	double const ageFactor = compData->getAgeOfInitialTreesFactor();
	unsigned int forestTypeId = compData->getForestTypeId();
	ForestType const * const forestType = ForestTypeContainer::get(forestTypeId);
	double const typeTotalAmount = forestType->getTotalAmount();
	unsigned int const numSpeciesInComp = forestType->getNumSpecies();

	if (ForestInitializer::treesExist()) {
		// Generate data for trees which exist!
		// Loop through all trees, generate data.

		// Set up classbounds specifying the percentage of each species in the compartment.
		vector<double> speciesBounds(numSpeciesInComp + 1);
		speciesBounds[0] = 0.0;
		for (unsigned int spIndex = 0; spIndex < numSpeciesInComp; spIndex++) {
			speciesBounds[spIndex + 1] = speciesBounds[spIndex] + (forestType->getSpeciesAmount(spIndex)
					/ typeTotalAmount);
		}
		speciesBounds[numSpeciesInComp] = 1.0000001;

		for (Tree * curTree = compartment->getFirstTree(); curTree != NULL; curTree
				= curTree->getNextTreeInCompartment()) {
			if (!ForestInitializer::positionIsLoaded()) {
				// TODO: generate positions for already existing tree objects.
				// This is very difficult to achieve with all pointgenerators.

			}
			if (!ForestInitializer::speciesIsLoaded()) {
				// Generate species for the existing trees.
				// Set a correct percentage of all trees to each species
				unsigned int spIndex;
				if (numSpeciesInComp > 1) {
#ifdef COMPARISONSEEDING
					UniformRandomNumberGenerator::setSeeds(curTree, 733.332);
#endif
					spIndex = classify(speciesBounds, RandomNumber::get());
				} else {
					spIndex = 0;
				}
				unsigned int const speciesId = SpeciesContainer::getId(forestType->getSpeciesName(spIndex));
				curTree->setSpeciesId(speciesId);
			}
			if (!ForestInitializer::ageIsLoaded()) {
				// Generate age for the existing tree.
#ifdef COMPARISONSEEDING
				UniformRandomNumberGenerator::setSeed(curTree->getX()*631262.3 + curTree->getY()+6.234);
#endif
				unsigned int const age = PoissonRandomNumberGenerator::get(ageFactor * (double) averageInitialAge);
				curTree->setAge(age);
			}
			if (!ForestInitializer::heterogeneityIsLoaded()) {
				// Generate initial heterogeneity for existing tree.
#ifdef COMPARISONSEEDING
				UniformRandomNumberGenerator::setSeeds(curTree, 71);
#endif
				curTree->setInitialHeterogeneity(StandardNormalRandomNumberGenerator::get());
			}
		}

	} else {
		// Generate completely new trees.

		for (unsigned int speciesIndex = 0; speciesIndex < numSpeciesInComp; speciesIndex++) {
			Log::log(2, "Adding trees of species %s", forestType->getSpeciesName(speciesIndex).c_str());
			double const percent = numFactor * forestType->getSpeciesAmount(speciesIndex) / typeTotalAmount;
			unsigned int const speciesId = SpeciesContainer::getId(forestType->getSpeciesName(speciesIndex));

			vector<Point> treeLocations;
			pointGenerator->generatePointsInside(*compartment->getPolygon(), percent, treeLocations);
			for (vector<Point>::const_iterator point = treeLocations.begin(); point != treeLocations.end(); ++point) {
#ifdef COMPARISONSEEDING
				// Only add trees that go into the active part of the compartment
				if (!compartment->getPolygon()->activeContains(*point))
				continue;
				UniformRandomNumberGenerator::setSeed(point->x*631262.3 + point->y+6.234);
#endif
				unsigned int const age = PoissonRandomNumberGenerator::get(ageFactor * (double) averageInitialAge);
				Tree* newTree = new Tree(speciesId, age, *point);
				compartment->addNewTree(newTree);
#ifdef COMPARISONSEEDING
				UniformRandomNumberGenerator::setSeeds(newTree, 71);
#endif
				newTree->setInitialHeterogeneity(StandardNormalRandomNumberGenerator::get());
			}
		}
	}
	
	if (!ForestInitializer::treesExist() || !ForestInitializer::ageIsLoaded()) {
		unsigned int standAge = (unsigned int)max(0, roundToInt(ageFactor * (double)averageInitialAge));
		compData->setInitialStandAge(standAge);
		Log::log(2, "Stand age of compartment %u is %u years", compData->getIdNumber(), standAge);
	}
}

void InitialCompartmentGenerator::generateTreeMarks(TreeContainer & allTrees) const {
	size_t const numTrees = allTrees.size();
	// The limitFinder should be set to limitedByCompRad before this function is called.
	LimitFinder & limitFinder = LimitFinder::Instance();
	double const competitionRadius = limitFinder.getLargestLimitingValue();
#ifdef INCLUDE_ENVIRONMENTAL_EFFECTS
	allTrees.setLimitingValue(competitionRadius * simulationData.getCorrelationDistanceFactor());
#else
	allTrees.setLimitingValue(competitionRadius);
#endif

#ifdef INCLUDE_ENVIRONMENTAL_EFFECTS
	vector<Vector> environmentalEffects(TreeMarks::NUMBER_OF_MARKS);
	for (unsigned int index = 0; index < TreeMarks::NUMBER_OF_MARKS; index++) {
		Log::log(2, "Will allocate %dx%d = %d bytes for environmental effects vector", numTrees, sizeof(double),
			numTrees*sizeof(double));
		environmentalEffects[index] = Vector(numTrees);
		if (initializationDatas[index].hasEnvironmentalEffects()) {
			LocalCorrelation correlation;
			correlation.setCompetitionRadius(competitionRadius);
			correlation.setDistanceFactor(simulationData.getCorrelationDistanceFactor());
			correlation.setCoefficients(initializationDatas[index].getEnvironmentalEffectCoefficients());
			correlation.setRandomNumberSetIndex(0);
			correlation.setMarkIndex(index);
			for (unsigned int curTreeIndex = 0; curTreeIndex < numTrees; ++curTreeIndex) {
				Tree const * const curTree = allTrees[curTreeIndex];
				double const x = correlation.calculateCorrelatedRandomNumber(curTree, allTrees.getTreesCloseTo(curTree));
				environmentalEffects[index][curTreeIndex] = x;
			}
			unsigned int numCalculations = correlation.getNumberOfCalculations();
			double avgMatrixSize = correlation.getAverageMatrixSize();
			unsigned int maxMatrixSize = correlation.getMaximumMatrixSize();
			Log::log(2, "Performed %u local covariance calculations with average matrix size %3.2fx%3.2f and maximum matrix size %ux%u",
					numCalculations, avgMatrixSize, avgMatrixSize, maxMatrixSize, maxMatrixSize);
		}
	}

	allTrees.setLimitingValue(competitionRadius);
#endif

	Vector modelVariables(Model::NUMBER_OF_MODEL_VARIABLES);
	Vector secondaryVariables(SecondaryModel::numberOfVariables());

#ifdef CALCULATION_INSPECTION
	CalculationInspector& inspector = CalculationInspector::Instance();
#endif

	// Calculate tree marks.
	for (unsigned int index = 0; index < TreeMarks::NUMBER_OF_MARKS; index++) {
		if (ForestInitializer::markIsLoaded(index))
			continue;

		for (size_t curTreeIndex = 0; curTreeIndex < numTrees; curTreeIndex++) {
			Tree * const curTree = allTrees[curTreeIndex];

			SecondaryModel::calculateSecondaryVariables(
					curTree,
					curTree->getParentCompartment()->getCompartmentData(),
					simulationData.getSizeClassHandler(),
					simulationData.getAgeClassHandler(),
					simulationData.getHeightClassHandler(),
					secondaryVariables);

			modelVariables.clear();
#ifdef COMPARISONSEEDING
			UniformRandomNumberGenerator::setSeeds(curTree, 71);
			// pretend to draw initial heterogeneity here to test comparisonseeding
			// hope the compiler doesn't remove "unnecessary code"
			if(index==0)
			double dummy = StandardNormalRandomNumberGenerator::get();

#endif
			modelVariables[Model::TAU] = curTree->getInitialHeterogeneity();
			modelVariables[Model::XI] = 0.0;
#ifdef INCLUDE_ENVIRONMENTAL_EFFECTS
			modelVariables[Model::XI] = environmentalEffects[index][curTreeIndex];
#endif
			// Calculate competition indexes for this mark
			Model::calculateModelVariables(curTree, allTrees.getTreesCloseTo(curTree), competitionRadius,
					modelVariables);

			double initialMark = Model::calculateMark(modelVariables, secondaryVariables,
					initializationDatas[index].getModelTerms());
#ifdef CALCULATION_INSPECTION
			if (inspector.shouldInspectInitialStand()) {
				inspector.inspectModel("Initial Stand", *curTree, curTreeIndex, index, initializationDatas[index], modelVariables, secondaryVariables, initialMark);
			}
#endif
			curTree->setMark(index, initialMark);
		}

		// Transform tree marks
		const string markName = TreeMarks::getMarkName(index);

		// Calculate an initial CDF for all trees
		MarkTransformer transformer(index);
		transformer.calculateInitialCDF(allTrees);

		// We need to loop over the compartments to provide forest type-specific target distributions
		TreeContainer::const_comp_iterator compIter = allTrees.compBegin(), compEnd = allTrees.compEnd();
		for (; compIter != compEnd; ++compIter) {
			Compartment * const compartment = *compIter;
			if (compartment->getPolygon()->isActive()) {
				CompartmentData const * const compData = compartment->getCompartmentData();

				unsigned int forestTypeId = compData->getForestTypeId();
				ForestType const * const forestType = ForestTypeContainer::get(forestTypeId);
				MarkTransformation const & transformation = forestType->getInitialStandTransformation(index);
				if (transformation.isLoaded()) {
					Log::log(3, "Transforming initial mark %s in compartment %d", markName.c_str(), compData->getIdNumber());
					transformer.setTargetTransformation(transformation);

					TreeContainer compartmentTrees(false);
					compartmentTrees.addCompartment(compartment);
					transformer.transform(compartmentTrees);

#ifdef CALCULATION_INSPECTION
					if (inspector.shouldInspectInitialStand()) {
						size_t const numCompTrees = compartmentTrees.size();
						for (size_t curTreeIndex = 0; curTreeIndex < numCompTrees; curTreeIndex++) {
							Tree * const curTree = compartmentTrees[curTreeIndex];
							inspector.print("Transformed initial mark '%s' for tree [%d] compartment %d: %f", markName.c_str(), curTreeIndex, compData->getIdNumber(), curTree->getMark(index));
						}
					}
#endif
				} else {
					Log::log(3, "Skipping transformation of mark %s. No distribution defined.", markName.c_str());
				}

				if (index == TreeMarks::DIAMETER) {
					// Multiply the diameter by the initial tree diameter factor
					for (Tree * curTree = compartment->getFirstTree(); curTree != NULL; curTree = curTree->getNextTreeInCompartment()) {
						Species const * const species = SpeciesContainer::get(curTree->getSpeciesId());
						double const diamFactor = forestType->getInitialTreeDiameterFactor(species);
						curTree->setMark(index, curTree->getMark(index) * diamFactor);	
					}
				}
			}
		}
	}

}

#ifdef INCLUDE_ENVIRONMENTAL_EFFECTS
bool InitialCompartmentGenerator::hasEnvironmentalEffects() const {
	for (unsigned int index = 0; index < TreeMarks::NUMBER_OF_MARKS; ++index) {
		if (initializationDatas[index].hasEnvironmentalEffects())
			return true;
	}
	return false;
}
#endif

void InitialCompartmentGenerator::loadData(DOMElement const * initialCompartmentGenerator) {
	// eUnrecognized denotes an unrecognized tag.
	enum Tags {
		eUnrecognized, ePointProcess, eInitializationParameters, eInitialAge
	};

	map<string, int> tagMap;
	tagMap["pointProcess"] = ePointProcess;
	tagMap["initializationParameters"] = eInitializationParameters;
	tagMap["averageInitialAge"] = eInitialAge;

	for (DOMElement * element = XercesUtilities::getFirstChildDOMElement(initialCompartmentGenerator); element != 0; element
			= XercesUtilities::getNextSiblingDOMElement(element)) {
		XercesString tagName = XMLString::replicate(element->getTagName());
		switch (tagMap[tagName.toString()]) {
		case (ePointProcess): {
			loadPointProcess(element);
			break;
		}
		case (eInitializationParameters): {
			enum Tags2 {
				eUnrecognized, eMark
			};
			map<string, int> tagMap;
			tagMap["mark"] = eMark;

			for (DOMElement * mark = XercesUtilities::getFirstChildDOMElement(element); mark != 0; mark
					= XercesUtilities::getNextSiblingDOMElement(mark)) {
				XercesString tagName = XMLString::replicate(mark->getTagName());
				switch (tagMap[tagName.toString()]) {
				case (eMark): {
					string name = XercesUtilities::getAttributeValueAsString(mark, "name");
					unsigned int markId = TreeMarks::getMarkId(name);
					initializationDatas[markId].loadData(mark);
					break;
				}
				default: // Handles eUnrecognized.
				{
					throw "Unrecognized XML-element: " + tagName.toString()
							+ ", expected 'mark name=\"...\"' at this level.";
				}
				}
			}
			break;
		}
		case (eInitialAge): {
			averageInitialAge = XercesUtilities::getFirstChildTextContentAsInt(element);
			break;
		}
		default: // Handles eUnrecognized.
		{
			throw "Unrecognized XML-element: " + tagName.toString();
		}
		}
	}
}

void InitialCompartmentGenerator::loadPointProcess(DOMElement const * pointProcess) {
	// eUnrecognized denotes an unrecognized tag.
	enum Tags {
		eUnrecognized,
		eHomogeneousPoisson,
		eInhomogeneousPoisson,
		eHomogeneousClustered,
		eInhomogeneousClustered,
		eHomogeneousRegular,
		eInhomogeneousRegular
	};

	map<string, int> tagMap;
	tagMap["homogeneousPoisson"] = eHomogeneousPoisson;
	tagMap["inhomogeneousPoisson"] = eInhomogeneousPoisson;
	tagMap["homogeneousClustered"] = eHomogeneousClustered;
	tagMap["inhomogeneousClustered"] = eInhomogeneousClustered;
	tagMap["homogeneousRegular"] = eHomogeneousRegular;
	tagMap["inhomogeneousRegular"] = eInhomogeneousRegular;

	for (DOMElement * element = XercesUtilities::getFirstChildDOMElement(pointProcess); element != 0; element
			= XercesUtilities::getNextSiblingDOMElement(element)) {
		XercesString tagName = XMLString::replicate(element->getTagName());
		switch (tagMap[tagName.toString()]) {
		case (eHomogeneousPoisson): {
			// Log::log(4,"Homogeneous Poisson point process:");
			pointGenerator.reset(new HomogenousPoissonPointGenerator);
			break;
		}
		case (eInhomogeneousPoisson): {
			// Log::log(4,"Inhomogeneous Poisson point process:");
			pointGenerator.reset(new InhomogenousPoissonPointGenerator);
			break;
		}
		case (eHomogeneousClustered): {
			// Log::log(4,"Homogeneous clustered point process:");
			pointGenerator.reset(new HomogenousClusteredPointGenerator);
			break;
		}
		case (eInhomogeneousClustered): {
			// Log::log(4,"Inhomogeneous clustered point process:");
			pointGenerator.reset(new InhomogenousClusteredPointGenerator);
			break;
		}
		case (eHomogeneousRegular): {
			// Log::log(4,"Homogeneous regular point process:");
			pointGenerator.reset(new HomogenousRegularPointGenerator);
			break;
		}
		case (eInhomogeneousRegular): {
			// Log::log(4,"Inhomogeneous regular point process:");
			pointGenerator.reset(new InhomogenousRegularPointGenerator);
			break;
		}
		default: // Handles eUnrecognized.
		{
			throw "Unrecognized XML-element: " + tagName.toString();
		}
		}
		pointGenerator->loadData(element);
	}
}
